/*
 * Copyright (c) 1998-2004 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */
package jgroup.test.jini.reggie;

import java.io.IOException;
import java.rmi.Remote;
import java.rmi.RemoteException;

import jgroup.core.ConfigManager;
import jgroup.util.Network;
import net.jini.config.Configuration;
import net.jini.config.ConfigurationProvider;
import net.jini.core.discovery.LookupLocator;
import net.jini.core.entry.Entry;
import net.jini.core.lookup.ServiceID;
import net.jini.discovery.LookupDiscovery;
import net.jini.discovery.LookupDiscoveryManager;
import net.jini.export.Exporter;
import net.jini.jeri.BasicILFactory;
import net.jini.jeri.BasicJeriExporter;
import net.jini.jeri.tcp.TcpServerEndpoint;
import net.jini.lookup.JoinManager;
import net.jini.lookup.ServiceIDListener;
import net.jini.lookup.entry.Name;

import org.apache.log4j.Logger;

import com.sun.jini.config.Config;
import com.sun.jini.start.LifeCycle;


/**
 * Simple server to test Reggie.
 * 
 * @author Rohnny Moland
 */
public class ReggieHelloServer 
  implements ServiceIDListener, ReggieHello
{  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Logger
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Obtain logger for this class */
  private static final Logger log = Logger.getLogger(ReggieHelloServer.class);

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Message returned by sayHello */ 
  private String answer;
  
  /** The join manager instance */
  private JoinManager joinManager;
  
  /** The exporter instance */
  private Exporter exporter;
  
  /** The exported proxy instance */
  private Remote proxy;
  
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructor and main
  ////////////////////////////////////////////////////////////////////////////////////////////

 
  /**
   * Constructor compatible with the jini service starter.
   *
   * @param configArgs <code>String</code> array whose elements are
   *                   the arguments to use when creating the server.
   * @param lifeCycle  instance of <code>LifeCycle</code> that, if
   *                   non-<code>null</code>, will cause this object's
   *                   <code>unregister</code> method to be invoked during
   *                   shutdown to notify the service starter framework that
   *                   the reference to this service's implementation can be
   *                   'released' for garbage collection. A value of
   *                   <code>null</code> for this argument is allowed.
   *
   * @throws Exception If there was a problem initializing the service.
   */
  public ReggieHelloServer(String[] configArgs, LifeCycle lifeCycle) 
    throws Exception
  { 
    ConfigManager.init();
    
    final Configuration config = ConfigurationProvider.getInstance(configArgs, getClass()
        .getClassLoader());
 
    answer = new String("Hello from " + Network.getLocalHostName());
   
    
    /** Exports the remote object and joins it into reggie */ 
    exporter = (Exporter) 
      Config.getNonNullEntry(config, "jgroup.test.jini.reggie", 
                      "serverExporter", Exporter.class, 
                      new BasicJeriExporter(TcpServerEndpoint.getInstance(0), 
                                            new BasicILFactory()));
    proxy = exporter.export(this);

    Entry[] entries = (Entry[]) 
      Config.getNonNullEntry(config, "jgroup.test.jini.reggie", "entries",
                             Entry[].class, new Entry[] {new Name("ReggieHelloServer")});

    String[] groups = (String[]) 
      Config.getNonNullEntry(config, "jgroup.test.jini.reggie", "initialLookupGroups", 
          String[].class, new String[] {System.getProperty("reggiehello.lookup.group", "jgroup.sf.net")});
    
    try {
      LookupDiscovery ld = new LookupDiscovery(groups);
      joinManager = new JoinManager(proxy, entries, this, ld, null);
    } catch (Exception e) {
      log.error(e.getMessage());
      System.exit(1);
    }    
  } 
  
  /**
   * Helper method used to retrieve the LookupDiscoveryManager
   * from a started service.
   */
  private LookupDiscoveryManager getDiscoveryManager()
  {
    return (LookupDiscoveryManager) joinManager.getDiscoveryManager();
  }
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from ReggieHello
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  External method
   */
  public String sayHello() throws java.rmi.RemoteException
  {    
    log.debug("sayHello: returning " + answer);
    return answer;
  }

  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from ServiceIDListener
  ////////////////////////////////////////////////////////////////////////////////////////////

  public void serviceIDNotify(ServiceID serviceID)
  {
    log.debug("Service id: " + serviceID);
  }

  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from JoinAdmin
  ////////////////////////////////////////////////////////////////////////////////////////////

  /* (non-Javadoc)
   * @see net.jini.admin.JoinAdmin#getLookupAttributes()
   */
  public Entry[] getLookupAttributes() throws RemoteException
  {
    return joinManager.getAttributes();
  }


  /* (non-Javadoc)
   * @see net.jini.admin.JoinAdmin#addLookupAttributes(net.jini.core.entry.Entry[])
   */
  public void addLookupAttributes(Entry[] attrSets) throws RemoteException
  {
    joinManager.addAttributes(attrSets);
  }


  /* (non-Javadoc)
   * @see net.jini.admin.JoinAdmin#modifyLookupAttributes(net.jini.core.entry.Entry[], net.jini.core.entry.Entry[])
   */
  public void modifyLookupAttributes(Entry[] attrSetTemplates, Entry[] attrSets) throws RemoteException
  {
    joinManager.modifyAttributes(attrSetTemplates, attrSets);
  }


  /* (non-Javadoc)
   * @see net.jini.admin.JoinAdmin#getLookupGroups()
   */
  public String[] getLookupGroups() throws RemoteException
  {
    return getDiscoveryManager().getGroups();
  }


  /* (non-Javadoc)
   * @see net.jini.admin.JoinAdmin#addLookupGroups(java.lang.String[])
   */
  public void addLookupGroups(String[] groups) throws RemoteException
  {
    try {
      getDiscoveryManager().addGroups(groups);      
    } catch (IOException e) {
      log.error(e.getMessage());
    }
  }


  /* (non-Javadoc)
   * @see net.jini.admin.JoinAdmin#removeLookupGroups(java.lang.String[])
   */
  public void removeLookupGroups(String[] groups) throws RemoteException
  {
    getDiscoveryManager().removeGroups(groups);
  }


  /* (non-Javadoc)
   * @see net.jini.admin.JoinAdmin#setLookupGroups(java.lang.String[])
   */
  public void setLookupGroups(String[] groups) throws RemoteException
  { 
    try {
      getDiscoveryManager().setGroups(groups);
    } catch (IOException e) {
      log.error(e.getMessage());
    }
  }


  /* (non-Javadoc)
   * @see net.jini.admin.JoinAdmin#getLookupLocators()
   */
  public LookupLocator[] getLookupLocators() throws RemoteException
  {
    return getDiscoveryManager().getLocators();
  }


  /* (non-Javadoc)
   * @see net.jini.admin.JoinAdmin#addLookupLocators(net.jini.core.discovery.LookupLocator[])
   */
  public void addLookupLocators(LookupLocator[] locators) throws RemoteException
  {
    getDiscoveryManager().addLocators(locators);
  }


  /* (non-Javadoc)
   * @see net.jini.admin.JoinAdmin#removeLookupLocators(net.jini.core.discovery.LookupLocator[])
   */
  public void removeLookupLocators(LookupLocator[] locators) throws RemoteException
  {
    getDiscoveryManager().removeLocators(locators);
  }


  /* (non-Javadoc)
   * @see net.jini.admin.JoinAdmin#setLookupLocators(net.jini.core.discovery.LookupLocator[])
   */
  public void setLookupLocators(LookupLocator[] locators) throws RemoteException
  {
    getDiscoveryManager().setLocators(locators);
  }

  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from DestroyAdmin
  ////////////////////////////////////////////////////////////////////////////////////////////

  /* (non-Javadoc)
   * @see com.sun.jini.admin.DestroyAdmin#destroy()
   */
  public void destroy() throws RemoteException
  {
    exporter.unexport(true);
  }

  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from Administrable
  ////////////////////////////////////////////////////////////////////////////////////////////

  /* (non-Javadoc)
   * @see net.jini.admin.Administrable#getAdmin()
   */
  public Object getAdmin() throws RemoteException
  {
    return proxy;
  }

} 
