/*
 * Copyright (c) 1998-2006 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */
package jgroup.test.jini.txn;

import java.io.Serializable;
import java.rmi.Remote;
import java.rmi.RemoteException;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.Map;
import java.util.Properties;

import javax.jdo.JDOHelper;
import javax.jdo.PersistenceManager;
import javax.jdo.PersistenceManagerFactory;
import javax.jdo.Query;

import jgroup.core.ConfigManager;
import net.jini.config.Configuration;
import net.jini.config.ConfigurationProvider;
import net.jini.core.entry.Entry;
import net.jini.core.lookup.ServiceID;
import net.jini.core.transaction.UnknownTransactionException;
import net.jini.core.transaction.server.TransactionConstants;
import net.jini.core.transaction.server.TransactionManager;
import net.jini.core.transaction.server.TransactionParticipant;
import net.jini.discovery.LookupDiscovery;
import net.jini.export.Exporter;
import net.jini.jeri.BasicILFactory;
import net.jini.jeri.BasicJeriExporter;
import net.jini.jeri.tcp.TcpServerEndpoint;
import net.jini.lookup.JoinManager;
import net.jini.lookup.ServiceIDListener;
import net.jini.lookup.entry.Name;

import org.apache.log4j.Logger;

import com.sun.jini.config.Config;
import com.sun.jini.start.LifeCycle;

/**
 * Bank server without replication support used as participant in a
 * transaction.
 * 
 * @author Rohnny Moland
 */
public class NonReplicatedBankServer 
  implements Bank, TransactionParticipant, ServiceIDListener
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Logger
  ////////////////////////////////////////////////////////////////////////////////////////////

  private static final Logger log = Logger.getLogger(NonReplicatedBankServer.class);


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Maps used to store pending transactions and the bank accounts. */
  private Map<Long, CreditDebit> pendingTxns;
  private Map<Integer, Integer> accounts;

  /** Name of the bank service */
  private String name; 

  /** Database specific fields */
  private PersistenceManagerFactory pmf;
  private PersistenceManager pm;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructor and main
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Constructor compatible with the jini service starter.
   *
   * @param configArgs <code>String</code> array whose elements are
   *                   the arguments to use when creating the server.
   * @param lifeCycle  instance of <code>LifeCycle</code> that, if
   *                   non-<code>null</code>, will cause this object's
   *                   <code>unregister</code> method to be invoked during
   *                   shutdown to notify the service starter framework that
   *                   the reference to this service's implementation can be
   *                   'released' for garbage collection. A value of
   *                   <code>null</code> for this argument is allowed.
   *
   * @throws Exception If there was a problem initializing the service.
   */
  public NonReplicatedBankServer(String[] configArgs, LifeCycle lifeCycle) 
    throws Exception
  { 
    ConfigManager.init();

    pendingTxns = Collections.synchronizedMap(new HashMap<Long, CreditDebit>());
    accounts = Collections.synchronizedMap(new HashMap<Integer, Integer>());
    
    name = System.getProperty("bank.name", "NOKAS/RobberyBank");
  
    Entry[] entries = new Entry[] { new Name(name) };
    final Configuration config = ConfigurationProvider.getInstance(configArgs, getClass()
        .getClassLoader());
    Exporter exporter = (Exporter) Config.getNonNullEntry(config, "jgroup.test.jini.txn",
        "serverExporter", Exporter.class, 
         new BasicJeriExporter(TcpServerEndpoint.getInstance(0), new BasicILFactory()));

    Remote proxy = exporter.export(this);
    String[] groups = new String[] {System.getProperty("bank.lookup.group", "jgroup.sf.net")};
    LookupDiscovery ld = new LookupDiscovery(groups);
    new JoinManager(proxy, entries, this, ld, null);
    
    initDB();  
   }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from Bank
  ////////////////////////////////////////////////////////////////////////////////////////////
   
  /* (non-Javadoc)
   * @see jgroup.test.jini.txn.Bank#getBalance(long)
   */
  public Integer getBalance(long accNumber) 
    throws RemoteException
  {
    Integer balance = accounts.get(new Long(accNumber));
    if (balance == null) balance = new Integer(0); 
    return balance;
  }


  /* (non-Javadoc)
   * @see jgroup.test.jini.txn.Bank#withdraw(long, int, net.jini.core.transaction.Transaction)
   */
  public void withdraw(long accountNumber, int amount, long transactionID)
    throws RemoteException
  {
//    pendingTxns.put(new Long(transactionID), new CreditDebit(accountNumber, 0, amount));
    log.debug("Withdraw=" + amount + ",Account=" + accountNumber + ",tid=" + transactionID);
  }


  /* (non-Javadoc)
   * @see jgroup.test.jini.txn.Bank#deposit(long, int, net.jini.core.transaction.Transaction)
   */
  public void deposit(long accountNumber, int amount, long transactionID)
    throws RemoteException
  {
//    pendingTxns.put(new Long(transactionID), new CreditDebit(accountNumber, amount, 0));
    log.debug("Deposit=" + amount + ",Account=" + accountNumber + ",tid=" + transactionID);
  }


  /* (non-Javadoc)
   * @see jgroup.test.jini.txn.Bank#getName()
   */
  public String getName()
    throws RemoteException
  {
    return name;
  }


  /**
   * Do the credit/debit operation.
   * 
   * @param accNumber Account number to credit/debit
   * @param credit Amount to credit
   * @param debit Amount to debit
   */
  private void doCreditDebit(long accNumber, int credit, int debit)
  { 
//    log.debug("credit = " + credit + " debit = " + debit);
    Account account = getAccount(pm, accNumber);
//    log.debug("Balance BEFORE commit for acc.nr " + accNumber + " is " + account.getBalance());
    int newBalance = account.getBalance();
    newBalance += credit;
    newBalance -= debit;
    updateAccount(pm, accNumber, newBalance);    
//    log.debug("Balance AFTER commit for acc.nr " + accNumber + " is " + account.getBalance());
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from TransactionParticipant
  ////////////////////////////////////////////////////////////////////////////////////////////

  /* (non-Javadoc)
   * @see net.jini.core.transaction.server.TransactionParticipant#prepare(net.jini.core.transaction.server.TransactionManager, long)
   */
  public int prepare(TransactionManager mgr, long id) 
    throws UnknownTransactionException, RemoteException
  {
    log.debug(name + " reached PREPARE state, txn id:" + id);
    return TransactionConstants.PREPARED;
  }


  /* (non-Javadoc)
   * @see net.jini.core.transaction.server.TransactionParticipant#commit(net.jini.core.transaction.server.TransactionManager, long)
   */
  public void commit(TransactionManager mgr, long id) 
    throws UnknownTransactionException, RemoteException
  {
    log.debug(name + " reached COMMIT state, txn id: " + id);
 
//    CreditDebit temp = pendingTxns.get(new Long(id));
//    if (temp != null)
//      doCreditDebit(temp.accNumber, temp.credit, temp.debit);
//    
//    pendingTxns.remove(new Long(id));
//    log.debug("commit(): New size of pendingTxns: " + pendingTxns.size());
  }


  /* (non-Javadoc)
   * @see net.jini.core.transaction.server.TransactionParticipant#abort(net.jini.core.transaction.server.TransactionManager, long)
   */
  public void abort(TransactionManager mgr, long id) 
    throws UnknownTransactionException, RemoteException
  {
    log.debug(name + " reached ABORT state for txn id: " + id);
    
//    CreditDebit temp = pendingTxns.get(new Long(id));
//    log.debug(temp.toString());
//    pendingTxns.remove(new Long(id));
  }


  /* (non-Javadoc)
   * @see net.jini.core.transaction.server.TransactionParticipant#prepareAndCommit(net.jini.core.transaction.server.TransactionManager, long)
   */
  public int prepareAndCommit(TransactionManager mgr, long id) 
    throws UnknownTransactionException, RemoteException
  {
    int result = prepare(mgr, id);
    if (result == TransactionConstants.PREPARED) {
      commit(mgr, id);
      result = TransactionConstants.COMMITTED;
    }
    return result;    
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Database methods
  ////////////////////////////////////////////////////////////////////////////////////////////
  
  /**
   * Initialize database. Get properties from objectdb.xml.
   */
  private void initDB() 
  { 
    String syspmf = System.getProperty("javax.jdo.PersistenceManagerFactoryClass", 
                                        "com.objectdb.jdo.PMF");
    String connectionURL = System.getProperty("bank.connectionURL", "/tmp/bank.odb");
    
    log.debug("syspmf: " + syspmf);
    log.debug("connectionURL: " + connectionURL);
    
    Properties properties = new Properties();
    properties.setProperty(
        "javax.jdo.PersistenceManagerFactoryClass", syspmf);
    properties.setProperty(
        "javax.jdo.option.ConnectionURL", connectionURL); 

    pmf = JDOHelper.getPersistenceManagerFactory(properties, 
                                                    JDOHelper.class.getClassLoader());          
  
    /** If db does not contain data, insert some default accounts and balances. */
    Account account = null;
    account = getAccount(pm, 111111111);
    if (account == null) {
      for (int i=1; i < 10; i++) {
        insertAccount(pm, 111111111*i, 100000);
      }
    }
  }


  /**
   * Insert an account object into database.
   * 
   * @param pm The persistence manager to use
   * @param number The account number
   * @param balance The balance of the account
   */
  private void insertAccount(PersistenceManager pm, long number, int balance) 
  {
    Account account = new Account(number, balance);
    try {
      pm = pmf.getPersistenceManager();
      pm.currentTransaction().begin();
      pm.makePersistent(account);
      pm.currentTransaction().commit();
    } catch (Exception e) {
      pm.currentTransaction().rollback();
      log.error(e.getMessage());
    } finally {
      pm.close();
    }
  }


  /**
   * Update an account object.
   * 
   * @param pm The persistence manager to use
   * @param number The account number to update
   * @param balance The new balance of the account
   */
  private boolean updateAccount(PersistenceManager pm, long number, int balance)
  {
    Account account = null;
    boolean update = true;
    
    try {
      pm = pmf.getPersistenceManager();
      pm.currentTransaction().begin();
      Query query = pm.newQuery(Account.class, "number == " + number);
      
      Collection result = (Collection) query.execute();
      
      account = (Account) result.iterator().next();
      
      if (account == null) {
        update = false;
        log.debug("Did not find a matching instance");
        pm.currentTransaction().rollback();  
      }
      
      account.setBalance(balance);
      pm.currentTransaction().commit(); 
    } catch (Exception e) {
      pm.currentTransaction().rollback();
      log.error(e.getMessage());
    } finally {  
        pm.close();        
      }
      return update;
    }


  /**
   * Returns an account object with given account number.
   * 
   * @param pm The persistence manager to use
   * @param number Account number of account to return
   * @return The account object
   */
  private Account getAccount(PersistenceManager pm, long number)
  {
    Account account = null;
    
    try {
      pm = pmf.getPersistenceManager();
  
      Query query = pm.newQuery(Account.class, "number == " + number);
  
      Collection result = (Collection) query.execute();
      
      account = (Account) result.iterator().next();
      
      if (account == null) {
        log.debug("Did not find a matching instance");
        pm.currentTransaction().rollback();  
      }
    } catch (Exception e) {
      log.error(e.getMessage());
    } finally {
      pm.close();
    }
    return account;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Inner classes
  ////////////////////////////////////////////////////////////////////////////////////////////  
  
  /**
   * Class that holds a temporary credit/debit operation, before doing a commit/abort operation.
   */
  public class CreditDebit implements Serializable
  {
    private static final long serialVersionUID = -923925461553488080L;
    int credit;
    int debit;
    long accNumber;
    
    CreditDebit(long accNumber, int credit, int debit) 
    {
     this.accNumber = accNumber;
     this.credit = credit;
      this.debit = debit;
    }
    
    public String toString() 
    {
      String temp = "CreditDebit { \n";
      temp += "accNumber=" + accNumber + "\n";
      temp += "credit=" + credit + "\n";
      temp += "debit=" + debit + "\n";
      temp += "} \n";
      return temp;
    }
  }

  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from ServiceIDListener
  ////////////////////////////////////////////////////////////////////////////////////////////  

  /* (non-Javadoc)
   * @see net.jini.lookup.ServiceIDListener#serviceIDNotify(net.jini.core.lookup.ServiceID)
   */
  public void serviceIDNotify(ServiceID serviceID) 
  {
    log.debug("Service ID: " + serviceID);
  }
  
} // END NonReplicatedBankServer
 