/*
 * Copyright (c) 1998-2006 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.test.jini.txn;

import java.io.BufferedReader;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.rmi.RMISecurityManager;
import java.util.HashMap;
import java.util.Map;
import java.util.Random;

import jgroup.core.ConfigManager;
import jgroup.jini.LookupManager;
import net.jini.core.entry.Entry;
import net.jini.core.lease.Lease;
import net.jini.core.lookup.ServiceTemplate;
import net.jini.core.transaction.server.TransactionManager;
import net.jini.core.transaction.server.TransactionParticipant;
import net.jini.lookup.entry.Name;

import org.apache.log4j.Logger;

import com.sun.jini.lookup.entry.BasicServiceType;

/**
 * Client for running Jini transactions, using either replicated or non-replicated
 * transaction components.<p>
 * 
 * This client can perform transactions using three different models:
 * <ol>
 * <li> Using replicated bank services and a replicated transaction manager.</li>
 * <li> Using non-replicated bank services and a non-replicated transaction manager. In 
 *   this model, classic (transient) Jini transactions are performed without replication 
 *   support at all.</li> 
 * <li> Using non-replicated bank services and a transaction manager with replication 
 *   support. In this mode, the state of the transaction manager is transfered to the
 *   transaction manager replicas, when performing transactions.</li>
 * </ol>
 * The different models are set by passing two system properties to TxnClient:
 * <ul>
 * <li> tm.replication </li>
 * <li> banks.replication </li>
 * </ul>
 * 
 * @author Rohnny Moland
 */
public class TxnClient
{
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Logger
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Obtain logger for this class */
  private static final Logger log = Logger.getLogger(TxnClient.class);


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Map holding transaction scenarios */
  private Map<Integer,TxnData> txnScenarios = new HashMap<Integer,TxnData>();

  /** Bank service proxies */
  private Bank bank1;
  private Bank bank2;
  
  /** The local TransactionManager service proxy */
  private TransactionManager mgr;
  
  /** Boolean determining if the current transaction has committed. */
  private volatile boolean committed;
  
  /** Boolean telling if client should lookup bank services with replication support */
  private boolean banksReplication;
  
  /** Boolean telling if client should lookup bank services with passive replication support */
  private boolean pBanksReplication;
  
  /** Boolean telling if client should lookup transaction manager with replication support */
  private boolean mgrReplication;

  /** Boolean telling if client should lookup transaction manager with passive replication support */
  private boolean pMgrReplication;
  
  /** Holds the transaction number performed */
  private int txnNumber = 1;
  
  /** measurement instance fields */
  private long start;
  private long end;
  private long totalstart;
  private long totalend;

  /** Default path to the txn simulation file */
  private static final String defaultSimulationPath = "target/resources/simulation/txn-simulation";

  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructor and main
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Instatiates a TxnClient object
   */
  public static void main(String args[])
    throws Exception
  {
    new TxnClient();
  } 
  
  /**
   * Constructor invoked by main
   */    
  public TxnClient()
    throws Exception 
  {
    System.setSecurityManager(new RMISecurityManager());
    ConfigManager.init();
    
    /** get the system properties */
    banksReplication = Boolean.parseBoolean(System.getProperty("banks.replication", "true"));
    pBanksReplication = Boolean.parseBoolean(System.getProperty("banks.passive", "false"));
    mgrReplication = Boolean.parseBoolean(System.getProperty("tm.replication", "true"));
    pMgrReplication = Boolean.parseBoolean(System.getProperty("tm.passive", "false"));
    int txns = Integer.getInteger("num.of.txns", 9).intValue();    
    String simFile = System.getProperty("txn.simulation.file", defaultSimulationPath);  
    
    log.debug("banks.replication: " + banksReplication);
    log.debug("banks.passive: " + pBanksReplication);
    log.debug("tm.replication: " + mgrReplication);
    log.debug("tm.passive: " + pMgrReplication);
    log.debug("txns: " + txns);
    log.debug("simFile: " + simFile);

    /** Lookup the services */
    lookup();    
    
    /** load transaction scenarios */
    loadTxnScenarios(simFile);    
    
    /** Perform some random transactions */
    performTxnScenarios(txns);
    
    /** exit from client, to be able to run post experiment */
    System.exit(0);
  }
  
  
  /**
   * Lookup group TM and transaction participants.
   */
  private void lookup() 
  {
    try {
      log.debug("try to find bank services...");
      if (banksReplication) {
        if (pBanksReplication) {
          bank1 = (Bank) LookupManager.lookup("Bank/TheLoanBox");
          log.debug("Try to lookup Bank/LargeFees");
          bank2 = (Bank) LookupManager.lookup("Bank/LargeFees");           
        } else {
          bank1 = (Bank) LookupManager.lookup("NOKAS/RobberyBank");
          bank2 = (Bank) LookupManager.lookup("Bank/YadaSavings");           
        }
     } else {
        bank1 = (Bank) LookupManager.lookup(new Name("NOKAS/RobberyBank"));
        bank2 = (Bank) LookupManager.lookup(new Name("Bank/YadaSavings"));
      }
      log.debug("bank1 name: " + bank1.getName());
      log.debug("bank2 name: " + bank2.getName());

      log.debug("try to find a TM...");

      if (mgrReplication) {
        if (pMgrReplication) {
          mgr = (TransactionManager) LookupManager.lookup("Jgroup/PassiveGroupMahalo");
        } else {
          mgr = (TransactionManager) LookupManager.lookup("Jgroup/ActiveGroupMahalo");
        }
      } else {
        Class[] classes = new Class[] { TransactionManager.class };
        Entry[] attr = new Entry[] {new BasicServiceType("Transaction Manager")};
        ServiceTemplate template = new ServiceTemplate(null, classes, attr);
        mgr = (TransactionManager) LookupManager.lookup(template);
      }
    } catch (IOException e) {
      log.error("FAiled to look service", e);
    }       
  }
  
   
  /**
   * Performs a transaction. Could be:
   * 1. Replicated transaction manager and replicated bank services.
   * 2. Non-replicated transaction manager and non-replicated bank services.
   * 3. Replicated transaction manager and non-replicated bank services.
   */
  private void performTransaction(int txnnr)
  {
    TxnData txnData = txnScenarios.get(new Integer(txnnr));

    TransactionManager.Created tcs = null;
    try {
      tcs = mgr.create(Lease.FOREVER);
    } catch (Exception e) {
      log.error("Unable to create transaction", e);
    }
    long transactionID = tcs.id;
    log.debug("-- Created transaction number: " + txnNumber++ + " with txn id=" + transactionID);
    
    try {
      bank1.withdraw(txnData.srcAccount, txnData.credit, transactionID);
      bank2.deposit(txnData.destAccount, txnData.debit, transactionID);
      mgr.join(transactionID, (TransactionParticipant) bank1, 0);
      mgr.join(transactionID, (TransactionParticipant) bank2, 0);      
    } catch (Exception e) {
      log.error("Unable to join in participants", e);
    }
    
    try {
      mgr.commit(transactionID);
      log.debug("Transaction " + transactionID + " committed");
    } catch (Exception e) {
      log.error("Unable to commit transaction", e);
    }
  }
  
  
  /**
   * Loads the txn scenarios.
   */  
  private void loadTxnScenarios(String simulationFile)
  {
    try {
      BufferedReader in = new BufferedReader(new FileReader(simulationFile));
      String txnScenario;
      log.debug("txnnr. srcAccount destAccount credit debit");
      log.debug("--------------------------------------------------");      
      for (int i=0; (txnScenario = in.readLine()) != null; i++) {  
        String[] txnInfo = txnScenario.split("\\s+");
        TxnData txnData = new TxnData(Long.parseLong(txnInfo[1]), Long.parseLong(txnInfo[2]), 
                             Integer.parseInt(txnInfo[3]), Integer.parseInt(txnInfo[4]) );        
        txnScenarios.put(new Integer(Integer.parseInt(txnInfo[0])), txnData);
        log.debug(txnInfo[0] + "  " + txnInfo[1] + "  " + txnInfo[2] + "  " + txnInfo[3] + "  " + txnInfo[4]);
      }
      in.close();
    } catch (FileNotFoundException e) {
      log.error("Could not find the specified file: " + simulationFile, e);
    } catch (IOException e) {
      log.error(e.getMessage());
    }
  }


  /**
   * Perform some random transactions.
   * 
   * @param number The number of transactions to perform.
   * @throws InterruptedException 
   */
  private void performTxnScenarios(int number)
  {
    Random rnd = new Random(); 
    for (int i=0; i<number; i++) {
      
      /** Ensure that a client can only perform one transaction at once */
      performTransaction(1+rnd.nextInt(8));

//      while (!committed) {
//      }
//      committed = false;
    }
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Inner classes
  ////////////////////////////////////////////////////////////////////////////////////////////  
  
  /**
   * This class holds the txn scenario datas.
   */
  private class TxnData 
  {    
    private long srcAccount;
    private long destAccount;
    private int credit;
    private int debit;
     
    public TxnData(long srcAccount, long destAccount, int credit, int debit)
    {
      this.srcAccount = srcAccount;
      this.destAccount = destAccount;
      this.credit = credit;
      this.debit = debit;  
    }
  }

} // END TxnClient
