/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.test.performance;

import java.io.PrintWriter;

import jgroup.util.log.Eventlogger;

/**
 *  This class is used by both Jgroup and standard RMI test to perform
 *  repeated invocations and take mean times.  A single test consists of
 *  NUM_CALLS invocations of a method; the time needed to perform an
 *  invocation is the average computed from the total time.  Each test
 *  is repeated NUM_ROUNDS times.
 *
 *  @author Alberto Montresor
 *  @author Hein Meling
 *  @since Jgroup 1.1
 */
class ClientTest
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constants
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Max number of */
  private static final int PRE_NUM_ROUNDS      = 5000;
  private static final int POST_NUM_ROUNDS     = 5000;
  private static final int NUM_ROUNDS          = 100000;
  private static final int NUM_CALLS_PER_ROUND = 1;
  private static final int MIN_SIZE = 1000;
  private static final int MAX_SIZE = 1000;
  private static final int INC_SIZE = 1000;


  /**
   *  Anycast/unicast invocations, with byte arrays of variable size
   *  as arguments and return values.
   */
  private static void pre_test2(SpeedTest test)
    throws Exception
  {
    /* Create a buffer of size MIN_SIZE */
    byte[] buffer = new byte[MIN_SIZE];
    /* Run some iterations to initialize the experiment */
    for (int i=0; i < PRE_NUM_ROUNDS; i++) {
      // Perform test 
      for (int j=1; j <= NUM_CALLS_PER_ROUND; j++) {
        test.test(buffer);
      }
    }
  }

  /**
   *  Post experiment run to allow other clients to also complete the
   *  main part of the experiment under identical load conditions.
   */
  private static void post_test2(SpeedTest test)
    throws Exception
  {
    /* Create a buffer of size MIN_SIZE */
    byte[] buffer = new byte[MIN_SIZE];
    /* Run some iterations to finalize the experiment */
    for (int i=0; i < POST_NUM_ROUNDS; i++) {
      // Perform test 
      for (int j=1; j <= NUM_CALLS_PER_ROUND; j++) {
        test.test(buffer);
      }
    }
  }

  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Anycast/unicast invocations, without arguments/return values
   */
  static void test1(SpeedTest test, PrintWriter out)
    throws Exception
  {
    /* Min, max and average time; initialization */
    long min = Long.MAX_VALUE, max = 0, avg = 0, start = 0, time = 0;

    System.out.println("\n\n-- Test1: test() -----------------------------------------------------");
    out.println("\n\n-- Test1: test() -----------------------------------------------------");
    System.out.println("ROUND NO ; NUM OF CALLS ; TOTAL TIME ; AVG TIME PR CALL");
    out.println("ROUND NO ; NUM OF CALLS ; TOTAL TIME ; AVG TIME PR CALL");

    for (int i=0; i < NUM_ROUNDS; i++) {

      // Perform test and measure times
      start = System.currentTimeMillis();
      for (int j=1; j <= NUM_CALLS_PER_ROUND; j++) {
        System.out.print("+");
        if (j % 20 == 0) {
          System.out.print("+");
        }
        test.test();
      }
      System.out.println("");

      // Compute statics
      time = System.currentTimeMillis() - start;
      if (time < min)
        min = time;
      if (time > max)
        max = time;
      avg += time;

      // Print statistics
      System.out.println(i + " ; " + NUM_CALLS_PER_ROUND + " ; " + time + " ; " + ((float) time / NUM_CALLS_PER_ROUND));
      out.println(i + " ; " + NUM_CALLS_PER_ROUND + " ; " + time + " ; " + ((float) time / NUM_CALLS_PER_ROUND));
    }
    System.out.println("MIN TIME ; MAX TIME ; AVG TIME");
    System.out.print(((float) min/NUM_CALLS_PER_ROUND) + " ; ");
    System.out.print(((float) max/NUM_CALLS_PER_ROUND) + " ; ");
    System.out.println(((float) avg/(NUM_CALLS_PER_ROUND*NUM_ROUNDS)));
    out.println("MIN TIME ; MAX TIME ; AVG TIME");
    out.print(((float) min/NUM_CALLS_PER_ROUND) + " ; ");
    out.print(((float) max/NUM_CALLS_PER_ROUND) + " ; ");
    out.println(((float) avg/(NUM_CALLS_PER_ROUND*NUM_ROUNDS)));
    System.out.println("----------------------------------------------------------------------");
    out.println("----------------------------------------------------------------------");
  }

  /**
   *  Anycast/unicast invocations, with byte arrays of variable size
   *  as arguments and return values.
   */
  static void test2(SpeedTest test, PrintWriter out)
    throws Exception
  {
    pre_test2(test);
    out.println("\n-- byte[] test(byte[]) ----------------------------------------");
    /* Start and delay time values */
    long start = 0, delay = 0;
    /* Create a buffer */
    byte[] buffer;
    for (int k = MIN_SIZE; k <= MAX_SIZE; k += INC_SIZE) {
      /* Create a buffer of size k */
      buffer = new byte[k];
      for (int i=0; i < NUM_ROUNDS; i++) {
        // Perform test and measure times
//        start = System.currentTimeMillis();
        //Debug.logMem("start");
        Eventlogger.logEvent("start");
        for (int j=1; j <= NUM_CALLS_PER_ROUND; j++) {
          test.test(buffer);
        }
        //Debug.logMem("stop");
        Eventlogger.logEvent("stop");
//        delay = System.currentTimeMillis() - start;
//        float callTime = ((float) delay / NUM_CALLS_PER_ROUND);
//        out.println("k=" + k + "; " + callTime);
      }
      //Debug.flushMem();
      Eventlogger.flushMem();
    }
    out.println("----------------------------------------------------------------------");
    post_test2(test);
  }


  /**
   *  Anycast/unicast invocations, with string of variable size
   *  as arguments and return values.
   */
  static void test3(SpeedTest test, PrintWriter out)
    throws Exception
  {
    /* Min, max and average time; initialization */
    long min = Long.MAX_VALUE, max = 0, avg = 0, start = 0, time = 0;
  
    System.out.println("\n\n-- Test3: String test(String) ----------------------------------------");
    out.println("\n\n-- Test3: String test(String) ----------------------------------------");

//    for (int k = MIN_STR_SIZE; k <= MAX_STR_SIZE; k += INC_STR_SIZE) {
    int k = 50;
      /* Create a string of size k */
      StringBuilder buf = new StringBuilder();
      for (int i = 0; i < k; i++) {
        buf.append(i);
      }
      String str = buf.toString();

      /* Reset the min, max and avg values for the next array size */
      min = Long.MAX_VALUE;  max = 0;  avg = 0;

      for (int i=0; i < NUM_ROUNDS; i++) {

        // Perform test and measure times
        start = System.currentTimeMillis();
        for (int j=1; j <= NUM_CALLS_PER_ROUND; j++) {
          String reply = test.test(str);
//          System.out.println(reply.substring(0, 30));
        }

        // Compute statics
        time = System.currentTimeMillis() - start;
        if (time < min)
          min = time;
        if (time > max)
          max = time;
        avg += time;
  
        // Print statistics
        System.out.println(i + " ; " + NUM_CALLS_PER_ROUND + " ; " + time + " ; " + ((float) time / NUM_CALLS_PER_ROUND));
//        out.println(i + " ; " + NUM_CALLS + " ; " + time + " ; " + ((float) time / NUM_CALLS));
        out.println(((float) time / NUM_CALLS_PER_ROUND));
      }
      System.out.println("SIZE ; MIN TIME ; MAX TIME ; AVG TIME");
      System.out.print(k + " ; ");
      System.out.print(((float) min/NUM_CALLS_PER_ROUND) + " ; ");
      System.out.print(((float) max/NUM_CALLS_PER_ROUND) + " ; ");
      System.out.println(((float) avg/(NUM_CALLS_PER_ROUND*NUM_ROUNDS)));
/*
      out.println("SIZE ; MIN TIME ; MAX TIME ; AVG TIME");
      out.print(k + " ; ");
      out.print(((float) min/NUM_CALLS) + " ; ");
      out.print(((float) max/NUM_CALLS) + " ; ");
      out.println(((float) avg/(NUM_CALLS*NUM_ROUNDS)));
      */
//    }
    System.out.println("----------------------------------------------------------------------");
    out.println("----------------------------------------------------------------------");
  }

  /**
   * Multicast invocations, without arguments/return values
   */
  static void mtest1(SpeedTest test, PrintWriter out)
    throws Exception
  {
    /* Min, max and average time; initialization */
    long min = Long.MAX_VALUE, max = 0, avg = 0, start = 0, time = 0;

    System.out.println("\n\n-- Test3: mtest() ----------------------------------------------------");
    out.println("\n\n-- Test3: mtest() -----------------------------------------------------");
    System.out.println("ROUND NO ; NUM OF CALLS ; TOTAL TIME ; AVG TIME PR CALL");
    out.println("ROUND NO ; NUM OF CALLS ; TOTAL TIME ; AVG TIME PR CALL");

    for (int i=0; i < NUM_ROUNDS; i++) {

      // Perform test and measure times
      start = System.currentTimeMillis();
      for (int j=1; j <= NUM_CALLS_PER_ROUND; j++) {
        System.out.print("+");
        if (j % 20 == 0) {
          System.out.print("+");
        }
        test.mtest();
      }
      System.out.println("");

      // Compute statics
      time = System.currentTimeMillis() - start;
      if (time < min)
        min = time;
      if (time > max)
        max = time;
      avg += time;

      // Print statistics
      System.out.println(i + " ; " + NUM_CALLS_PER_ROUND + " ; " + time + " ; " + ((float) time / NUM_CALLS_PER_ROUND));
      out.println(i + " ; " + NUM_CALLS_PER_ROUND + " ; " + time + " ; " + ((float) time / NUM_CALLS_PER_ROUND));
    }
    System.out.println("MIN TIME ; MAX TIME ; AVG TIME");
    System.out.print(((float) min/NUM_CALLS_PER_ROUND) + " ; ");
    System.out.print(((float) max/NUM_CALLS_PER_ROUND) + " ; ");
    System.out.println(((float) avg/(NUM_CALLS_PER_ROUND*NUM_ROUNDS)));
    out.println("MIN TIME ; MAX TIME ; AVG TIME");
    out.print(((float) min/NUM_CALLS_PER_ROUND) + " ; ");
    out.print(((float) max/NUM_CALLS_PER_ROUND) + " ; ");
    out.println(((float) avg/(NUM_CALLS_PER_ROUND*NUM_ROUNDS)));
    System.out.println("----------------------------------------------------------------------");
    out.println("----------------------------------------------------------------------");
  }

  /**
   *  Multicast invocations, with byte arrays of variable size
   *  as arguments and return values.
   */
  static void mtest2(SpeedTest test, PrintWriter out)
    throws Exception
  {
//    pre_test2(test);
    System.out.println("\n-- byte[] mtest(byte[]) ----------------------------------------");
    /* Start and stop time values */
    long start = 0, stop = 0;
    /* Create a buffer */
    byte[] buffer;
    for (int k = MIN_SIZE; k <= MAX_SIZE; k += INC_SIZE) {
      /* Create a buffer of size k */
      buffer = new byte[k];
      for (int i=0; i < NUM_ROUNDS; i++) {
        // Perform test and measure times
        start = System.currentTimeMillis();
        for (int j=1; j <= NUM_CALLS_PER_ROUND; j++) {
//          Thread.sleep(50);
          test.mtest(buffer);
        }
        stop = System.currentTimeMillis() - start;
        float callTime = ((float) stop / NUM_CALLS_PER_ROUND);
        System.out.println("i=" + i*NUM_CALLS_PER_ROUND + ", k=" + k + "; " + callTime);
      }
    }
    System.out.println("----------------------------------------------------------------------");
//    post_test2(test);
  }


  /**
   *  Multicast invocations, with byte arrays of variable size
   *  as arguments and return values.
   */
  static void xmtest2(SpeedTest test, PrintWriter out)
    throws Exception
  {
    /* Min, max and average time; initialization */
    long min = Long.MAX_VALUE, max = 0, avg = 0, start = 0, time = 0;

    System.out.println("\n\n-- Test4: byte[] mtest(byte[]) ---------------------------------------");
    System.out.println("SIZE ; MIN TIME ; MAX TIME ; AVG TIME");
    out.println("\n\n-- Test4: byte[] mtest(byte[]) ----------------------------------------");
    out.println("SIZE ; MIN TIME ; MAX TIME ; AVG TIME");

    for (int k=0; k <= MAX_SIZE; k += INC_SIZE) {
//    for (int k=INC_SIZE; k <= MAX_SIZE; k += INC_SIZE) {
System.out.print(" k=" + k);
      byte[] buffer = new byte[k];
      min = Long.MAX_VALUE;  max = 0;  avg = 0;

      for (int i=0; i < NUM_ROUNDS; i++) {

        // Perform test and measure times
        start = System.currentTimeMillis();
        for (int j=1; j <= NUM_CALLS_PER_ROUND; j++) {
          System.out.println(j);
          if (j % 20 == 0) {
            System.out.println("+");
          }
          Thread.sleep(500);
          test.mtest(buffer);
        }
        System.out.println("");

        // Compute statics
        time = System.currentTimeMillis() - start;
        if (time < min)
          min = time;
        if (time > max)
          max = time;
        avg += time;

        // Print statistics
        System.out.println(i + " ; " + NUM_CALLS_PER_ROUND + " ; " + time + " ; " + ((float) time / NUM_CALLS_PER_ROUND));
        out.println(i + " ; " + NUM_CALLS_PER_ROUND + " ; " + time + " ; " + ((float) time / NUM_CALLS_PER_ROUND));
      }
      System.out.print(k + " ; ");
      System.out.print(((float) min/NUM_CALLS_PER_ROUND) + " ; ");
      System.out.print(((float) max/NUM_CALLS_PER_ROUND) + " ; ");
      System.out.println(((float) avg/(NUM_CALLS_PER_ROUND*NUM_ROUNDS)));
      out.print(k + " ; ");
      out.print(((float) min/NUM_CALLS_PER_ROUND) + " ; ");
      out.print(((float) max/NUM_CALLS_PER_ROUND) + " ; ");
      out.println(((float) avg/(NUM_CALLS_PER_ROUND*NUM_ROUNDS)));
    }
    System.out.println("----------------------------------------------------------------------");
    out.println("----------------------------------------------------------------------");
  }

  /**
   *  Multicast invocations, with byte arrays of variable size
   *  as arguments and return values.
   */
  static void mtest3(SpeedTest test, PrintWriter out)
    throws Exception
  {
    System.out.println("\n\n-- Test5: byte[] mtest(byte[]) ---------------------------------------");
    System.out.println("JIT Compiling and optimization phase:");
    long start = 0;

    byte[] buffer = new byte[10000];
    int calls = NUM_ROUNDS*NUM_CALLS_PER_ROUND;
    long[] timeVals = new long[calls];
    for (int j=0; j < calls; j++) {
      start = System.nanoTime();
      test.mtest(buffer);
      timeVals[j] = System.nanoTime() - start;
    }
    for (int j=0; j < calls; j++) {
      System.out.println(timeVals[j]);
    }

    System.out.println("\n\nStandard measurements:");
    for (int k=0; k <= MAX_SIZE; k += INC_SIZE) {
      System.out.println("-- k=" + k + " ------------------------------------------------");
      buffer = new byte[k];
      for (int j=0; j < calls; j++) {
        start = System.nanoTime();
        test.mtest(buffer);
        timeVals[j] = System.nanoTime() - start;
      }
      for (int j=0; j < calls; j++) {
        System.out.println(timeVals[j]);
      }
    }
  }

  /**
   *  Multicast invocations, with byte arrays of variable size
   *  as arguments and return values.
   */
  static void mtest4(SpeedTest test, PrintWriter out)
    throws Exception
  {
    System.out.println("\n\n-- Test6: void vmtest(byte[]) ---------------------------------------");
    System.out.println("JIT Compiling and optimization phase:");

    byte[] buffer = new byte[10000];
    test.vmtest(buffer);
    System.out.println("Completed void vmtest(byte[])");
  }

} // END ClientTest
