/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.test.performance.upgrade;

import jgroup.test.performance.SpeedTest;


/**
 *  Test client thread that is able to work in a pool of threads for coordination.
 *  The class itself manages a pool of its intances. The ClientThreadPool instances are 
 *  created and started (@see java.lang.Thread.start()), they are added to the
 *  pool. The threads wait until allStart() method is invoked and then all of them
 *  start at once executing method run1(). 
 *  The run cycle may be repeated by calling allStart().
 *  Additionally, method getResponseTime() may be invoked to check
 *  the result of the run1() computations.
 *
 *  Note1: Many threads running in one jvm can only generate bigger traffic if and 
 *  only if the jvm runs on a multiprosessor machine. Another possibility is to 
 *  extend the coordination to "remote" threads.
 *
 * Note2: this class should be split into three for better reuse: one implementing
 * the coordination scheme between the threads in pool, one implementing a pool 
 * container with typical membership operations like join, leave and one class or
 * interface for the implementation of the actual logic.
 * Thus we could have many thread pools, each managing of different, potentially
 * heterogeneous "worker" Thread instances.
 *
 * 
 *  @author       Marcin Solarski
 *  @since        Jgroup 2.1
 */
class ClientThreadPool extends Thread {
    
  public ClientThreadPool(int method, int calibReqCallNo, int rps, int times, SpeedTest server ) throws Exception{

    if( ! initiated ) throw new RuntimeException("Not initialized properly");   
    test = new UpgradeClientTest(server);
    synchronized (ClientThreadPool.class) {id = instances ++;  }
    AvRespTime[id][0] = test.calibrate(method, calibReqCallNo);
    startCycleTime[0] = System.currentTimeMillis();
    this.rps = rps;
    this.times = times;
    setName("TestClient " + id);
  }

  
  /**
   *  initializes the pool. The operation has to be invoked before any thread
   *  is created.
   *  @param threads number of threads to be coordinated in the pool
   *  @param cycles number of cycles of synchronization 
   */

  public static void init(int threads, int cycles) {
    threads_no = threads;
    cycles_no = cycles;
    //init the table for results, position 0 is reserved for calibration
          AvRespTime= new float[threads][cycles+1];
    startCycleTime = new long[cycles+2];
    initiated = true;
  }

  /**
   * the method implements synchronization of starting executing method run1()
   * at once by all the threads in the pool.
   */
  public void run() {
    try {
      while ( true ) { 
        results_ready = false;
        synchronized (poolSemaphore) {
           active_instances++;
           if(active_instances < threads_no) {
             //System.out.println(getName() + " waiting for TestTread.class, active instances: " + active_instances);
             poolSemaphore.wait();
             //System.out.println(getName() + " after waiting for TestTread.class");

           } else {

             active_instances = 0;
             synchronized (startSemaphore) {
               synchronized (start_request) {
                 waiting_for_start = true; 
                 start_request.notify();
                 //System.out.println(getName() + " (last thread) waiting for startSemaphore...." + waiting_for_start);
               }
               startSemaphore.wait();
               waiting_for_start = false; 
         
             }
             //System.out.println(getName() + " notifies all threads waiting for poolSemaphore");
             poolSemaphore.notifyAll();
             startCycleTime[cur_cycle+1] = System.currentTimeMillis();
           }
           if( exit ) return;
        }

        //do the work
        run1();
        //System.out.println(getName() + " result.notify for cycle " + cur_cycle);
        synchronized (result) {cur_cycle++; results_ready = true; result.notifyAll();}
        //notify that the result is ready  gerRespoTime() 
      }
    } catch(InterruptedException e) { 
      System.out.println(getName() + " interrupted");
    }
  }

  /**
   * the method implement the actual work to do by each of the threads in the pool
   */
  synchronized void run1() {
    //System.out.println("Starting " + getName());
    AvRespTime[id][cur_cycle+1] = test.generate_traffic(rps, times);  
  }

  private int id;
  private volatile  Object  result = new Object();
  private volatile int cur_cycle = 0;
  private volatile boolean  results_ready = false;


  private volatile static  int   instances=0, active_instances=0;
  private volatile static Object startSemaphore = new Object();
  private volatile static Object poolSemaphore = new Object(); 
  private volatile static Object start_request = new Object(); 
  private volatile static boolean  waiting_for_start = false;
  private static boolean  exit = false;
  private static boolean  first_started = true;

  private volatile static int threads_no = 0;
  private volatile static int cycles_no = -1;
  private volatile static boolean initiated = false;

  //Test specific 
  private int rps, times;
  private UpgradeClientTest test;
  private static float AvRespTime[][];
  private static long startCycleTime[];

  /**
   * the method returns the number of threads in the pool.
   */
  public static synchronized int getRunningInstances() {
    return instances;
  }


  /**
   * the method makes all threads to stop after the current execution
   * of method run1() is over.
   */
  public static synchronized void allStop() {
    exit = true;
    synchronized (startSemaphore) {startSemaphore.notify();}

  }

  /**
   * waits until the results of the given cycle are ready
   */
  public  void CycleDone(int cycle) {
    synchronized (result) {
      if(cur_cycle  >= cycle) return ;
             while (cur_cycle < cycle -1 ) {
         try {
           result.wait();
               } catch(InterruptedException e) {}
      }
      if( results_ready) {
        //System.out.println(getName() + " returning from CycleDone " + cur_cycle);
        return;
      } else {
        try {
          //System.out.println(getName() + " has to wait in CycleDone in the last cycle because the result are not ready " + cur_cycle);
          result.wait();
        } catch(InterruptedException e) {}
      }
    }
  }



  /**
   * the method triggers all threads in the pool to start executing
   * method run1()
   */
  public static  void allStart() {
    //System.out.println("Starting all threads");
    if (! first_started) {
      try {
        synchronized (poolSemaphore ) {
          poolSemaphore.wait();
        }
      } catch(InterruptedException e) {}
      first_started = false;
    }

    while( true ) {
      try {
        synchronized( start_request ) { 
          if( waiting_for_start ) { 
            //System.out.println("waiting for start is true"); 
            waiting_for_start = false;
            break;
          } else {
            //System.out.println("waiting for start is false"); 
            start_request.wait();
          }
        }
      } catch(InterruptedException e) {}
    }
    
    //System.out.println("Sending startSemaphore.notify()");
    synchronized( startSemaphore ) { 
      startSemaphore.notify(); 
    }
  }

  /**
   * returns the times at which a given cycle started
   *  @param cycle number of the cycle
   *  @return time of the cycle start
   */
  public static long getStartCycleTime(int cycle) {
     if(cycle<0 || cycle>cycles_no) return -1;
     return startCycleTime[cycle];
  }
    
  /**
   * returns the computations of method run1(). In this case the avrage response
   * time of calls to the server.
   */
  public static float getRespTime(int thread, int cycle) {
    //if(thread>0 && thread <threads_no && cycle>=0 && cycle < cycles_no)
      return AvRespTime[thread][cycle];
    //else return 0;
  }

} // END ClientThreadPool