/*
 * Copyright (c) 1998-2004 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package com.sun.jini.mahalo;

import java.rmi.RemoteException;
import java.util.Iterator;
import java.util.Map;
import java.util.Vector;

import jgroup.core.GroupManager;
import jgroup.core.MemberId;
import jgroup.core.MembershipListener;
import jgroup.core.MembershipService;
import jgroup.core.View;
import jgroup.core.protocols.Anycast;
import jgroup.core.protocols.Atomic;
import jgroup.jini.txn.InternalGroupTransactionManager;
import jgroup.jini.txn.InternalPassiveGroupTransactionManager;
import net.jini.core.lease.LeaseDeniedException;
import net.jini.core.transaction.CannotAbortException;
import net.jini.core.transaction.CannotCommitException;
import net.jini.core.transaction.CannotJoinException;
import net.jini.core.transaction.TimeoutExpiredException;
import net.jini.core.transaction.UnknownTransactionException;
import net.jini.core.transaction.server.CrashCountException;
import net.jini.core.transaction.server.TransactionManager;
import net.jini.core.transaction.server.TransactionParticipant;

import org.apache.log4j.Logger;

import com.sun.jini.start.LifeCycle;

/**
 * This class is to be used together with the
 * {@link com.sun.jini.start.ServiceStarter} framework to start
 * a group enabled <i>transient</i> implementation of Mahalo.
 * 
 * @author Rohnny Moland
 * @author Heine Kolltveit
 */
public class PassiveGroupMahalo 
//  extends GroupTxnManagerImpl 
  implements TransactionManager, MembershipListener, InternalPassiveGroupTransactionManager,
    InternalGroupTransactionManager
{
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Logger
  ////////////////////////////////////////////////////////////////////////////////////////////

  private final static Logger log = Logger.getLogger(PassiveGroupMahalo.class);
 
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////
  
  private static final long serialVersionUID = -6405403802047754780L;

  /** Service proxies */
  private static InternalPassiveGroupTransactionManager ipgtm;
  private InternalGroupTransactionManager igtm;
  //  private SemiActiveService sas;
  private MembershipService pgms;
  
  /** The <code>MemberId</code> of this replica */
  private MemberId me;
  
  /** Boolean telling if this replica is a leader of the group */
  private boolean leader;

  /** The mahalo object */
  private GroupTxnManagerImpl                  tm;

  /** Measurement fields */
  private long prepareStart;
  private long prepareEnd;
  private long commitStart;
  private long commitEnd;
  
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructors
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Constructs a transaction manager with replication support.
   *
   * @param configArgs <code>String</code> array whose elements are
   *                   the arguments to use when creating the server.
   * @param lifeCycle  instance of <code>LifeCycle</code> that, if
   *                   non-<code>null</code>, will cause this object's
   *                   <code>unregister</code> method to be invoked during
   *                   shutdown to notify the service starter framework that
   *                   the reference to this service's implementation can be
   *                   'released' for garbage collection. A value of
   *                   <code>null</code> for this argument is allowed.
   *
   * @throws Exception If there was a problem initializing the service.
   */
  public PassiveGroupMahalo(String[] configArgs, LifeCycle lifeCycle) 
    throws Exception
  {
    if (log.isDebugEnabled())
      log.debug("Trying to start pgtm...");

    /** Obtains the group manager */
    GroupManager gm = GroupManager.getGroupManager(this);
    
    /** Obtains the service proxies */    
    ipgtm = (InternalPassiveGroupTransactionManager) gm
        .getService(InternalPassiveGroupTransactionManager.class);
    igtm = (InternalGroupTransactionManager) gm
        .getService(InternalGroupTransactionManager.class);
    
//    sas = (SemiActiveService) gm.getService(SemiActiveService.class);
    pgms = (MembershipService) gm.getService(MembershipService.class);
    me = pgms.getMyIdentifier();
    
    tm = new GroupTxnManagerImpl(configArgs, lifeCycle, false);
    
    if (log.isDebugEnabled()) 
      log.debug("Started PGTM!");
  }

  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from MembershipListener
  ////////////////////////////////////////////////////////////////////////////////////////////
  
  /* (non-Javadoc)
   * @see jgroup.core.MembershipListener#viewChange(jgroup.core.View)
   */
  public void viewChange(View view)
  {
    if (log.isDebugEnabled()) {
      log.debug("View id: " + view);
    }  
    
    leader =  view.isLeader(me);
    if (log.isDebugEnabled() && view.size() > 1 && leader) {
      log.debug("I am the leader replica.");
    }
  }
  
  /* (non-Javadoc)
   * @see jgroup.core.MembershipListener#prepareChange()
   */
  public void prepareChange()
  {
  }

  /* (non-Javadoc)
   * @see jgroup.core.MembershipListener#hasLeft()
   */
  public void hasLeft()
  {
  }
 
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from TransactionManager
  ////////////////////////////////////////////////////////////////////////////////////////////

  /* (non-Javadoc)
   * @see net.jini.core.transaction.server.TransactionManager#create(long)
   */
  @Anycast
  public TransactionManager.Created create(long lease) 
    throws LeaseDeniedException, RemoteException
  {
    if (log.isDebugEnabled())
      log.debug("Invoking tm.create on mahalo..");
    
    TransactionManager.Created created = tm.create(lease);
    
    if (log.isDebugEnabled()) {
      log.debug("PassiveGroupTransientMahalo.create() created  txn with id: " + created.id);
      //printTxns(txns);
    }

    GroupTxnManagerTransaction tmt = (GroupTxnManagerTransaction) tm.txns.get(new Long(created.id));    
    try {
      igtm.internalCreate(tmt, created.id);
    } catch (RemoteException e) {
      log.debug(e.getMessage());
    }
    return created;
  }

  /* (non-Javadoc)
   * @see net.jini.core.transaction.server.TransactionManager#
   * join(long, net.jini.core.transaction.server.TransactionParticipant, long)
   */
  /** FIXME: This should be "passively" replicated as well. */
  @Atomic
  public void join(long id, TransactionParticipant part, long crashCount) 
    throws UnknownTransactionException, CannotJoinException, CrashCountException, RemoteException
  {
    if (log.isDebugEnabled()) {      
      log.debug("PassiveGroupTransientMahalo.join() triggered with txn id: " + id);
    }
 
    tm.join(id, part, crashCount);
    if (log.isDebugEnabled()) {
      //printTxns(txns);
      log.debug("PassiveGroupTransientMahalo.join() done with txn id: " + id);
    }
  }


  /* (non-Javadoc)
   * @see net.jini.core.transaction.server.TransactionManager#getState(long)
   */
  @Anycast
  public int getState(long id) throws UnknownTransactionException
  {
    int state = tm.getState(id);
    if (log.isDebugEnabled())
      log.debug("PassiveGroupTransientMahalo.getState() for txn id: " + id + " is " + state);
    return state;
  }


  /* (non-Javadoc)
   * @see net.jini.core.transaction.server.TransactionManager#commit(long)
   */
  @Anycast
  public void commit(long id) 
    throws UnknownTransactionException, CannotCommitException, RemoteException
  {
    if (log.isDebugEnabled())
      log.debug("PassiveGroupTransientMahalo.commit() invoked with txn id: " + id);
    
    tm.commit(id);  
    
    if (log.isDebugEnabled())
      log.debug("PassiveGroupTransientMahalo.commit() finished with txn id: " + id);
  }


  /* (non-Javadoc)
   * @see net.jini.core.transaction.server.TransactionManager#commit(long, long)
   */
  @Anycast
  public void commit(long id, long waitFor) 
    throws UnknownTransactionException, CannotCommitException, 
      TimeoutExpiredException, RemoteException
  {
    if (log.isDebugEnabled())
      log.debug("PassiveGroupTransientMahalo.commit() invoked with txn id: " + id);
    
    tm.commit(id, waitFor);  
    
    if (log.isDebugEnabled())
      log.debug("PassiveGroupTransientMahalo.commit() finished with txn id: " + id);
  }


  /* (non-Javadoc)
   * @see net.jini.core.transaction.server.TransactionManager#abort(long)
   */
  @Anycast
  public void abort(long id) 
    throws UnknownTransactionException, CannotAbortException
  {
    tm.abort(id);
    if (log.isDebugEnabled())
      log.debug("Transaction " + id + " aborted.");
  }

  
  /* (non-Javadoc)
   * @see net.jini.core.transaction.server.TransactionManager#abort(long, long)
   */
  @Anycast
  public void abort(long id, long waitFor) throws UnknownTransactionException,
      CannotAbortException, TimeoutExpiredException
  {
    if (log.isDebugEnabled())
      log.debug("Transaction " + id + " aborted.");
    tm.abort(id, waitFor);
  }


  /* (non-Javadoc)
   * @see jgroup.core.MergingListener#getState(jgroup.core.MemberId[])
   */
  public Object getState(MemberId[] dests) 
  {
//    if (Debug.DEBUG)
//      log.debug("PassiveGroupTransientMahalo.getState() triggered.");
    return tm.txns;
  }

  /* (non-Javadoc)
   * @see jgroup.core.MergingListener#putState(java.lang.Object, jgroup.core.MemberId[])
   */
  public void putState(Object state, MemberId[] sources) 
  { 
    if (log.isDebugEnabled())
      log.debug("PassiveGroupTransientMahalo.putState() invoked.");

    if (state != null) { 
      Map remoteTxns = (Map) state;
  
      synchronized (remoteTxns) 
      {
        for(Iterator iter=remoteTxns.keySet().iterator();iter.hasNext();) {
          Long remoteTransactionID = (Long) iter.next();
          GroupTxnManagerTransaction remoteTxn = (GroupTxnManagerTransaction) remoteTxns
              .get(remoteTransactionID);
          GroupTxnManagerTransaction localTxn = (GroupTxnManagerTransaction) tm.txns
              .get(remoteTransactionID);
          
          if (localTxn == null) {
            remoteTxn.setLogManager(tm.getLogManager());
            remoteTxn.setTaskManager(tm.getTaskManager());
            remoteTxn.setWakeupManager(tm.getWakeupManager());
            remoteTxn.setSettler(tm);
            remoteTxn.setLocks();
            tm.txns.put(remoteTransactionID, remoteTxn);
            
            if (log.isDebugEnabled())
              log.debug("Transaction " + remoteTransactionID + "added");
            
          /*} else if (remoteTxn.getState() != localTxn.getState()) {
            if (remoteTxn.getState() == ABORTED && localTxn.getState(COMMITTED))
              log.error("Remote transaction status is ABORTED and local is COMMITTED!! Unrecoverable error!");*/
          } else {
            if (log.isDebugEnabled())
              log.debug("Transaction " + remoteTransactionID + "existed locally");
          }
          
        }
      }
    }
  }

  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from InternalPassiveGroupTransactionManager
  ////////////////////////////////////////////////////////////////////////////////////////////

  /* (non-Javadoc)
   * @see jgroup.test.jini.txn.InternalPassiveGroupTransactionManager#allPartsPrepared(Object, long)
   */
  public void transPrepared(Object txnMgr, long tid) 
    throws RemoteException
  {
    //printTxns(txns);
    prepareStart = System.nanoTime();
    if (log.isDebugEnabled())
      log.debug("backupTransaction() - txn id: " + tid);
    
    GroupTxnManagerTransaction mgrTransaction = (GroupTxnManagerTransaction) txnMgr;
    
    /** Must set locks here because its used by getState(..) */
    mgrTransaction.setLocks();
    
    if (tm.txns.containsKey(tid)) {
//      if (Debug.DEBUG)
//        log.debug("txn " + tid + " already exists");
      //Is the state of the transaction outdated?;
      GroupTxnManagerTransaction current = (GroupTxnManagerTransaction) tm.txns.get(new Long(tid)); 
      if (current.getState() != mgrTransaction.getState())
        //Outdated: Remove the old. The new is added below as a new transaction below 
        tm.txns.remove(tid);
      else {
        prepareEnd = System.nanoTime();
        return;
      }
    }

    mgrTransaction.setLogManager(tm.getLogManager());
    mgrTransaction.setTaskManager(tm.getTaskManager());
    mgrTransaction.setWakeupManager(tm.getWakeupManager());
    mgrTransaction.setSettler(tm);    
    tm.txns.put(new Long(tid), mgrTransaction);
    prepareEnd = System.nanoTime();
    //printTxns(txns);
  }
  
  
  /* (non-Javadoc)
   * @see jgroup.jini.txn.InternalPassiveGroupTransactionManager#allPartsCommitted(long)
   */
  public void transCommitted(long id) 
    throws UnknownTransactionException, RemoteException
  {
    commitStart = System.nanoTime();
    //if (Debug.DEBUG)
      log.debug("entering finalCommit() - txn id: " + id);
    //printTxns(txns);
    
    Long txnid = new Long(id); 
    
    if (tm.txns.containsKey(txnid))
      tm.txns.remove(txnid);
    else
      throw new UnknownTransactionException("No transaction with id " + id + " found");
    
    if (log.isDebugEnabled())
      log.debug("leaving finalCommit() - txn id: " + id);
    //printTxns(txns);
    commitEnd = System.nanoTime();
    log.warn("execute: " + ((prepareEnd - prepareStart)/1000000) + "\t" + ((commitEnd - commitStart)/1000000));
  }

  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Instance methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Prints transactions debug/state info.
   */
  public void printTxns(Map<Long,GroupTxnManagerTransaction> txns) 
  {
    Vector<GroupTxnManagerTransaction> keys = new Vector<GroupTxnManagerTransaction>(txns
        .values());
    int n = keys.size();
    log.debug("Current Transactions:\n");
    for (int i = 0; i < n; i++) {
      GroupTxnManagerTransaction tmt = (GroupTxnManagerTransaction) keys.get(i);
      if (tmt != null) {
        long id = tmt.getID();
        int state = tmt.getState();
        String stateString = "Not valid!";
        /*
         *    ACTIVE = 1
         *    VOTING = 2
         *    PREPARED = 3
         *    NOTCHANGED = 4
         *    COMMITTED = 5
         *    ABORTED = 6
         */
        log.debug("Txn-id: " + id);
        switch (state) {
          case 1: stateString = "active";
                  break;
          case 2: stateString = "voting";
                  break;
          case 3: stateString = "prepared";
                  break;
          case 4: stateString = "notchanged";
                  break;
          case 5: stateString = "committed";
                  break;
          case 6: stateString = "aborted";
                  break;
        }
        log.debug(" State = " + stateString + "\n");
      }
    }
    log.debug("");
  }
  
  public String toString()
  {
    return "PassiveGroupTransientMahalo";
  }
  
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from InternalPassiveGroupTransactionManager
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Get an InternalPassiveGroupTransactionManager.
   * 
   * @return The InternalPassiveGroupTransactionManager
   */
  protected static InternalPassiveGroupTransactionManager getIPGTM()
  {
  	return ipgtm;
  }

  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from InternalGroupTransactionManager
  ////////////////////////////////////////////////////////////////////////////////////////////

  /* (non-Javadoc)
   * @see jgroup.jini.txn.InternalGroupTransactionManager#internalCreate(java.lang.Object, long)
   */
  public void internalCreate(Object txnMgr, long id) 
    throws RemoteException
  {
    if (log.isDebugEnabled())
      log.debug("internalCreate() - txn id: " + id);

    if (leader)
      return;
    
    /** Insert and initialize the transaction on follower replicas */
    GroupTxnManagerTransaction mgrTransaction = (GroupTxnManagerTransaction) txnMgr;

    mgrTransaction.setLogManager(tm.getLogManager());
    mgrTransaction.setTaskManager(tm.getTaskManager());
    mgrTransaction.setWakeupManager(tm.getWakeupManager());
    mgrTransaction.setSettler(tm);
    mgrTransaction.setLocks();
    tm.txns.put(new Long(id), mgrTransaction);
  }

  /* (non-Javadoc)
   * @see jgroup.jini.txn.InternalGroupTransactionManager#internalCommit(long)
   */
  public void internalCommit(long id)
    throws UnknownTransactionException, CannotCommitException, RemoteException
  {
    throw new UnsupportedOperationException("Not implemented.");
  }
  
} // END PassiveGroupMahalo
