/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.arm;

import java.util.Collections;
import java.util.Iterator;
import java.util.Map;
import java.util.TreeMap;

import jgroup.core.ConfigurationException;
import jgroup.core.MemberId;
import jgroup.core.View;
import jgroup.core.arm.DistributionScheme;
import jgroup.core.arm.GroupExistsException;
import jgroup.core.arm.RedundancyException;
import jgroup.core.arm.UnknownGroupException;
import jgroup.relacs.config.AppConfig;
import jgroup.relacs.config.Domain;
import jgroup.relacs.config.DomainSet;
import jgroup.relacs.config.Host;
import jgroup.relacs.config.HostSet;

import org.apache.log4j.Logger;

/**
 *  ReplicaCount class is used to compute replica allocation on
 *  domains/hosts in the distributed system.
 *
 *  FIXME HEIN: This class needs to implement the DomainListener and
 *  HostListener.  Be aware the the host set below is a clone (shallow
 *  copy) of the original.  This means that we should work on the
 *  original when adding new host, but also add the new host to the host
 *  set use to represent the various replica counts.  That is, adding a
 *  new host, the replica count for it should be set to zero.
 *
 *  @author Hein Meling
 *  @since Jgroup 1.2
 */
final class ReplicaCount
  implements DistributionScheme
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Logger
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Obtain logger for this class */
  private static final Logger log = Logger.getLogger(ReplicaCount.class);


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constants
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Zero value for storing in a collection. */
  private static final Integer ZERO = new Integer(0);

  /**
   *  Key for the <code>Host</code> content map; to access the number of
   *  replicas executing on the associated host.
   */
  private static final String REPLICA_COUNT = "ReplicaCount";

  /**
   *  Key for the <code>Domain</code> content map; to obtain access to
   *  the TreeMap of sorted hosts.  Sorting is based on the number of
   *  replicas executing on each host.
   */
  private static final String REPLICA_COUNT_MAP = "ReplicaCountMap";


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** The set of domains over which we want to assign replicas. */
  private DomainSet domains;

  /** Table for managing groups and applications */
  private GroupTable groupTable;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructor
  ////////////////////////////////////////////////////////////////////////////////////////////

  ReplicaCount(GroupTable groupTable, DomainSet domains)
    throws ConfigurationException
  {
    this.groupTable = groupTable;
    this.domains = domains;

    for (Domain domain : domains) {
      HostSet hosts = (HostSet) domain.getHostSet().clone();
      Map<Integer,HostSet> countMap =
        Collections.synchronizedMap(new TreeMap<Integer,HostSet>());
      countMap.put(ZERO, hosts);
      domain.put(REPLICA_COUNT_MAP, countMap);
      for (Iterator j = hosts.iterator(); j.hasNext(); ) {
        Host host = (Host) j.next();
        host.put(REPLICA_COUNT, ZERO);
      }
    }
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from the DistributionScheme interface
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Assign the host location for the specified application.  The
   *  <code>AppConfig</code> object also contains the required redundancy.
   *  <P>
   *  We first try to place replicas in different domains.  But if some
   *  domains are unavailable, or the requested redundancy is greater
   *  than the number of domains, this is not possible.  In this case we
   *  place replicas in different domains were possible and then in the
   *  same domain as existing replicas.  We never place replicas on the
   *  same host.
   *  <P>
   *  FIXME HEIN: This method should prehaps be synchronized on the
   *  'domains' object in order to avoid concurrent changes to the set
   *  of domains while iterating.  For instance if a domain is removed
   *  after we determined that the size was sufficient to satisfy the
   *  given redundancy, the method would behave incorrectly.
   *
   *  @param app
   *    An <code>AppConfig</code> object specifying the class parameters
   *    and redundancy requirements.
   *  @return 
   *    A <code>HostSet</code> specifying the locations where the replicas
   *    should be created.
   *  @exception RedundancyException 
   *    Is raised if an the specified redundancy could not be satisfied.
   *  @exception GroupExistsException 
   *    Raised if the specified application group already exists (has been
   *    assigned to some hostset).
   */
  public HostSet assignReplicas(AppConfig app)
    throws RedundancyException, GroupExistsException
  {
    if (app == null)
      throw new NullPointerException("No application object specified for assignReplicas()");
    if (groupTable.contains(app))
      throw new GroupExistsException("Cannot assign replicas", app.getGroupId());

    /* Get the number of domains in the distributed system. */
    int numDomains = domains.size();
    if (numDomains == 0)
      throw new RedundancyException("No domains available to assign replicas.");

    /* Create an empty HostSet with capacity := redundancy */
    int redundancy = app.getInitialRedundancy();
    HostSet assignedHosts = new HostSet(redundancy);

    /*
     * We exit the outer loop once the redundancy has been satisfied or
     * the number of domains to avoid is equal to the number of
     * available domains.  In the latter case, we must rollback and
     * throw an exception (see below.)
     */
    for (DomainSet avoid = new DomainSet();
         redundancy > 0 && avoid.size() < numDomains; ) {

      /* Iterate over domains */
      for (Iterator i = domains.iterator(); redundancy > 0 && i.hasNext(); ) {
        Domain domain = (Domain) i.next();
        if (!avoid.containsDomain(domain)) {
          Host host = getHost(domain, assignedHosts);
          /*
           * If getHost returns null the domain is marked as avoid. This
           * is because it is either unavailable or all hosts in it has
           * been used to place replicas of this type.
           */
          if (host == null)
            avoid.addDomain(domain);
          else {
            /* We found an available host; adding it to the host set */
            assignedHosts.addHost(host);
            redundancy--;
          }
        }
      }
    }

    /*
     * If the number of hosts obtained is less than the requested
     * redundancy we rollback any updates to the system tables and throw
     * an exception.
     */
    if (assignedHosts.size() < redundancy) {
      removeReplicas(assignedHosts);
      throw new RedundancyException("Could not satisfy requested redundancy for " + app.getClassName());
    }

    /*
     * If the group table already has assigned this application,
     * a GroupExistsException is thrown.
     */
    groupTable.add(app, assignedHosts);

    return assignedHosts;
  }


  public HostSet collocateReplicas(AppConfig app, AppConfig collocateWithApp)
    throws GroupExistsException, RedundancyException, UnknownGroupException
  {
    if (groupTable.contains(app))
      throw new GroupExistsException("Cannot assign replicas", app.getGroupId());

    if (app.getInitialRedundancy() != collocateWithApp.getInitialRedundancy() ||
        app.getMinimalRedundancy() != collocateWithApp.getMinimalRedundancy()) {
      RedundancyException e = new RedundancyException("Cannot collocate replicas.");
      log.error("Cannot collocate replicas with different redundancy requirements:\n"
        + app + "\n" + collocateWithApp, e);
      throw e;
    }
    HostSet assignedHosts = groupTable.getAssignedHosts(collocateWithApp);
    if (assignedHosts.isEmpty()) {
      throw new UnknownGroupException("Application not previously assigned: "
        + collocateWithApp.getClassName(), collocateWithApp.getGroupId());
    }
    return assignedHosts;
  }


  /**
   *  Remove replica assignment data by updating the content map of
   *  hosts and their corresponding domain content map.  In addition
   *
   *  @param app
   *    The application replicas to be removed.
   *  @throws UnknownGroupException
   *    Raised if the specified application group is unknown.
   */
  public HostSet removeReplicas(AppConfig app)
    throws UnknownGroupException
  {
    if (!groupTable.contains(app))
      throw new UnknownGroupException(app.getGroupId());
    HostSet hosts = groupTable.getAssignedHosts(app);
    removeReplicas(hosts);
    groupTable.remove(app);
    return hosts;
  }


  /**
   *  Reassign the application replica running on the specified host.
   *
   *  @param app
   *    The application group to which the replica belongs.
   *  @param host
   *    The host from which to remove the replica.
   *  @return 
   *    The host on which a new replica should be created.
   *  @exception UnknownGroupException
   *    The given application group is unknown.
   *  @exception RedundancyException
   *    Raised if the system could not reassign the replica to any host
   *    in the system.
   */
  public Host reassignReplica(AppConfig app, Host host)
    throws UnknownGroupException, RedundancyException
  {
    if (!groupTable.contains(app))
      throw new UnknownGroupException(app.getGroupId());
    HostSet assignedHosts = groupTable.getAssignedHosts(app);

    /* First try to get a new host within the same domain. */
    Host newHost = getHost(host.getDomain(), assignedHosts);
    if (newHost == null) {

      /* Iterate over domains to try get an alternative location. */
      for (Iterator iter = domains.iterator(); newHost == null && iter.hasNext(); )
        newHost = getHost((Domain) iter.next(), assignedHosts);

      if (newHost == null)
        throw new RedundancyException("Could not reassign the given replica " + app + ", away from " + host);
    }

    /*
     * Update the host set for the given application to reflect the
     * reassignment.  Note that we may already have removed the host
     * from the assignedHosts set during the viewChange() method
     * invocation.
     */
    groupTable.removeHost(app, host);
    groupTable.addHost(app, newHost);

    return newHost;
  }


  /* (non-javadoc)
   * @see jgroup.core.arm.DistributionScheme#viewChangeEvent(jgroup.core.View)
   */
  public void viewChangeEvent(View view)
  {
    // Update the GroupTable according to this view change event
    groupTable.viewChangeEvent(view);
    /*
     * Cancel the renew timer for the application given by the provided
     * view.  This is required to prevent the renew mechanism from
     * triggering, in cases when a new view has been installed, but the
     * members has recently been initialized, and are thus out of sync
     * with the renew timer at the RM replicas.
     */
    int groupSize = view.size();
    if (groupSize > 0) {
      ReplicaPingEvent pingEvent =
        ReplicaPingEvent.getNewPingEvent(view.getGid(), groupSize);
      if (pingEvent != null)
        pingEvent.handle(this);
    }
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Private methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Remove replica assignment data by updating the content map of
   *  hosts and their corresponding domain content map.
   *
   *  @param hosts
   *    A <code>HostSet</code> containing the hosts that should have
   *    their content map updated due to this removal.
   *
   *  @throws IllegalStateException
   *    Raised if the domain of one of the hosts in the set was not
   *    initialized correctly, or if the countMap does not contain the
   *    specified host in the expected position.  This exception should
   *    never occur unless invoked at an incorrect time.
   */
  private void removeReplicas(final HostSet hosts)
  {
    for (Iterator i = hosts.iterator(); i.hasNext(); )
      removeReplica((Host) i.next());
  }


  /**
   *  Remove a replica assignment from the specified host,
   *  by updating the content map of the host and its
   *  corresponding domain content map.
   *
   *  @param host
   *    The <code>Host</code> that should have its content map
   *    updated due to this removal.
   *
   *  @throws IllegalStateException
   *    Raised if the domain of the host was not initialized correctly,
   *    or if the countMap does not contain the specified host in the
   *    expected position.  This exception should never occur unless
   *    this method is invoked at an incorrect time.
   */
  @SuppressWarnings("unchecked")
  public static void removeReplica(Host host)
  {
    Integer count = (Integer) host.get(REPLICA_COUNT);
    Integer oldCount = new Integer(count.intValue() - 1);
    host.put(REPLICA_COUNT, oldCount);
    Domain domain = host.getDomain();
    domain.decReplicaCount();

    Map<Integer,HostSet> countMap = (Map) domain.get(REPLICA_COUNT_MAP);
    if (countMap == null)
      throw new IllegalStateException("Domain has not been initialized: " + domain);

    /*
     * Get the host set with the current replica count from which this
     * host should be removed.
     */
    HostSet hostSet = countMap.get(count);
    if (hostSet == null) {
      //FIXME HACK: to avoid NPE problem.
      for (HostSet hosts : countMap.values()) {
        if (hosts.containsHost(host)) {
          hostSet = hosts;
          break;
        }
      }
    }
    if (!hostSet.removeHost(host))
      throw new IllegalStateException("Host not in host set for the current replica count.");

    if (hostSet.isEmpty())
      countMap.remove(count);

    /*
     * Get the host set with the old replica count to which this host
     * should be added.
     */
    hostSet = countMap.get(oldCount);
    if (hostSet == null) {
      /*
       * There is no other hosts with matching old replica count value,
       * therefore we create a new (empty) host set and place it in the
       * sorted countMap, into which the given host will be put.
       */
      hostSet = new HostSet();
      countMap.put(oldCount, hostSet);
    }

    /*
     * Add the given host to the set of hosts with the old replica count
     * value.
     */
    hostSet.addHost(host);
  }


  /**
   *  Update the internal tables to include a replica placed on
   *  the given host.
   */
  public static void updateHost(Host host)
  {
    Domain domain = host.getDomain();
    /*
     * We clone the domain host set since we need to remove a
     * host from it before using it as the avoidHostSet.
     */
    HostSet avoidHostSet = (HostSet) domain.getHostSet().clone();
    /*
     * Remove the host to place, from the avoidHostSet to force
     * the getHost() method to update the table correctly, and
     * return the same host; if not, we are in trouble.
     */
    avoidHostSet.removeHost(host);
    Host newHost = getHost(domain, avoidHostSet);
    if (newHost == null || !newHost.equals(host)) {
      log.warn(host.getHostName() + " could not be added to the replica assignment; "
        + "the avoid set contains these hosts: " + avoidHostSet);
    } else {
      if (log.isDebugEnabled())
        log.debug("Updated host to include one more replica: " + host);
    }
  }

  /**
   *  Returns a host from the given domain that is not in the avoid set.
   *  The host returned is a host with the lowest replica count in this
   *  domain.
   *
   *  @param domain
   *    Domain from which to get a host.
   *  @param avoid
   *    A set of hosts in this domain to avoid.  That is, a new replica
   *    should not be placed on the same host.  It is the responsibility
   *    of the caller to determine which hosts are already used for this
   *    invocation series.  To start an invocation series the avoid set
   *    may be <code>null</code> or just an empty <code>HostSet</code>.
   *  @return 
   *    An availble host in the given domain, and not in the avoid set.  
   *    If no more hosts are available in this domain, <code>null</code>
   *    is returned.
   *  @throws IllegalStateException
   *    Raised if the specified domain has not been correctly initialized.
   */
  @SuppressWarnings("unchecked")
  private static Host getHost(Domain domain, HostSet avoid)
  {
    if (avoid == null)
      avoid = new HostSet(0);

    Map<Integer,HostSet> countMap = (Map) domain.get(REPLICA_COUNT_MAP);
    if (countMap == null)
      throw new IllegalStateException("Domain has not been initialized: " + domain);

    /*
     * Get the host set in this domain with the lowest replica count.
     */
    synchronized (countMap) {
      for (Iterator cntIter = countMap.keySet().iterator(); cntIter.hasNext(); ) {

        /*
         * Since the countMap is a sorted map, the keys will be iterated
         * in sorted order from low to high.
         */
        Integer count = (Integer) cntIter.next();

        /* Obtain the host set associated with this count value; the lowest possible. */
        HostSet hosts = (HostSet) countMap.get(count);
        /* If this count value contains no hosts, we try the next one. */
        if (hosts == null)
          continue;

        synchronized (hosts) {
          for (Iterator iter = hosts.iterator(); iter.hasNext(); ) {
            Host host = (Host) iter.next();
            if (!avoid.containsHost(host) && host.isAvailable()) {
              /* Found an available host with a low replica count. */
              foundHost(host, hosts, countMap, count);
              return host;
            }
          }
        }
      }
    }

    /* We could not find any unused and available hosts. */
    return null;
  }


  private static void foundHost(Host host, HostSet hosts,
      Map<Integer,HostSet> countMap, Integer count)
  {
    /*
     * Remove this host from its present host set.  If the host
     * set is empty after this removal, we must remove the entry
     * from the sorted countMap which is stored in the domain
     * content map.
     */
    hosts.removeHost(host);
    if (hosts.isEmpty())
      countMap.remove(count);

    /* New replica count value; for the selected host */
    Integer newCount = new Integer(count.intValue() + 1);

    if (countMap.containsKey(newCount)) {

      /*
       * Add this host to an existing set of hosts with the new
       * replica count value.
       */
      ((HostSet) countMap.get(newCount)).addHost(host);

    } else {

      /*
       * There is no other hosts with matching replica count
       * value, we add our host to a new host set, that we put in
       * the sorted countMap.
       */
      hosts = new HostSet();
      hosts.addHost(host);
      countMap.put(newCount, hosts);

    }

    /* Update the count value associated with this host and its domain. */
    host.put(REPLICA_COUNT, newCount);
    Domain domain = host.getDomain();
    domain.incReplicaCount();
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Object methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  public String toString()
  {
    StringBuilder b = new StringBuilder();
    for (Iterator iter = groupTable.iterator(); iter.hasNext();) {
      GroupData gd = (GroupData) iter.next();
      HostSet hosts = gd.getAssignedHosts();
      for (Iterator it = hosts.iterator(); it.hasNext();) {
        Host host = (Host) it.next();
        Integer count = (Integer) host.get(REPLICA_COUNT);
        b.append(host);
        b.append(" has assigned ");
        b.append(count);
        b.append(" replicas\n");
      }
    }
    return groupTable.toString();
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // MergingListener methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  public Object getState(MemberId[] dests)
  {
    return domains;
  }


  public void putState(Object status, MemberId[] sources)
  {
    if (log.isDebugEnabled()) {
      log.debug("RC-putstate");
      for (int i = 0; i < sources.length; i++) {
        log.debug("sources: " + sources[i]);
      }
    }
    DomainSet rDomains = (DomainSet) status;
    for (Domain rDomain : rDomains) {
      Domain lDomain = domains.getDomain(rDomain);
      if (rDomain.getReplicaCount() > lDomain.getReplicaCount()) {
        Map<Integer,HostSet> oldcountMap = (Map) lDomain.get(REPLICA_COUNT_MAP);
        Map<Integer,HostSet> countMap = (Map) rDomain.get(REPLICA_COUNT_MAP);
        if (log.isDebugEnabled()) {
          log.debug(lDomain.getAddress().getHostAddress()  + " (local): " + lDomain.getReplicaCount());
          log.debug(rDomain.getAddress().getHostAddress() + " (remote): " + rDomain.getReplicaCount());
          log.debug("old countmap: " + oldcountMap);
          log.debug("new countmap: " + countMap);
        }
        lDomain.put(REPLICA_COUNT_MAP, countMap);
      }
    }
  }

} // End of ReplicaCount
