/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.arm;

import static jgroup.util.log.ReplicaEvent.Type.Initialized;

import java.rmi.RemoteException;
import java.util.HashMap;
import java.util.Map;

import jgroup.arm.GroupTable.MergeRecord;
import jgroup.core.ConfigManager;
import jgroup.core.GroupManager;
import jgroup.core.JgroupException;
import jgroup.core.MemberId;
import jgroup.core.MembershipListener;
import jgroup.core.MergingListener;
import jgroup.core.View;
import jgroup.core.arm.ARMEvent;
import jgroup.core.arm.DistributionScheme;
import jgroup.core.arm.GroupExistsException;
import jgroup.core.arm.RedundancyException;
import jgroup.core.arm.ReplicationManager;
import jgroup.core.arm.UnknownGroupException;
import jgroup.core.protocols.Multicast;
import jgroup.relacs.config.AppConfig;
import jgroup.relacs.config.DistributedSystemConfig;
import jgroup.relacs.config.Host;
import jgroup.relacs.config.HostSet;
import jgroup.relacs.registry.RegistryImpl;
import jgroup.upgrade.UpgradeManagerImpl;
import jgroup.util.Abort;
import jgroup.util.Util;
import jgroup.util.log.Eventlogger;
import jgroup.util.log.ReplicaEvent;

import org.apache.log4j.Logger;

/**
 *  This implementation of the replication manager interfaces make use
 *  of a variant of the semi-active replication scheme, in which only
 *  the <i>leader</i> RM replica receive and process the requests,
 *  while the <i>follower</i> replicas receive state update messages
 *  through the <code>StateListener</code> interface.  This permits
 *  that only the leader perform creation and removal of replicas at
 *  hosts in the distributed system, ensuring that only one group is
 *  created in response to a <code>createGroup()</code> request. <p>
 *
 *  This is done to prevent problems of non-determinism due to
 *  multithreading in the <code>Correlator</code>. <p>
 *
 *  The RM leader replica communicate with <code>ExecDaemon</code>
 *  (factory) objects at the hosts that are supposed to server replicas
 *  of any type.  The operations that can be performed on the
 *  <code>ExecDaemon</code>s are typically idempotent. <p>
 *
 *  Other application-level layers (or even servers) may also
 *  communicate arbitrary events to the replication manager.  For instance,
 *  the <code>RecoveryLayer</code> will provide <code>ViewChangeEvent</code>
 *  notifications to the RM leader when the group assoicated with a
 *  <code>RecoveryLayer</code> instance has installed a new view.
 *  Furthermore, the <code>RecoveryLayer</code> also implements a lease
 *  based liveness timer in case multiple replicas fail simultaneously. <p>
 *
 *  The events that can be communicated to the replication manager must
 *  implement the <code>ARMEvent</code> interface, and must handle its
 *  own event when received at the replication manager.
 *
 *  The replication manager supports several different implementations of
 *  the <code>RecoveryStrategy</code> interface, allowing different
 *  applications to provide specific recovery needs.  For instance, one
 *  such recovery strategy is the <code>KeepMinimalInPartition</code>,
 *  which does just that.
 *
 *  The replication manager also provide self-recovery, by reusing the
 *  same infrastructure used by the application-level recovery.  However,
 *  there is no recovery in case all replication manager replicas has been
 *  exhausted.
 *
 *  @author Hein Meling
 *  @since Jgroup 1.2
 */
public class ReplicaManagerImpl
  implements ReplicationManager, MembershipListener, MergingListener
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Logger
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Obtain logger for this class */
  private static final Logger log = Logger.getLogger(ReplicaManagerImpl.class);


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constants
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  The default period to await view installations after replicas
   *  has been (re)assigned.
   */
  private static final int DEFAULT_VIEW_MONITOR_PERIOD = 5000;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Private Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /*
   * Note that these fields cannot be initialized here, since the
   * constructor must ensure to parse the configuration file before
   * creating an instance of the distribution scheme object.
   */

  /** The distribution scheme (interface) */
  private final DistributionScheme distScheme;

  /** The local instance of the dependable registry. */
  private RegistryImpl dregistry;

  /** Mapping: groupid --> ManagementCallback object */
  private Map<Integer,ManagementCallback> callbackMap;

  /** This application object for the RM group */
  private final AppConfig thisApp;

  /** My member identifier */
  private MemberId myId;

  /** True if I'm the leader (if single member, it is always leader) */
  private boolean iamLeader = true;

  /** The table of GroupData objects */
  private final GroupTable grouptable;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructors
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Bootstrap or create local <code>ReplicaManagerImpl</code>
   *  instance.  Bootstrapping involves some initialization of some
   *  replication manager components that is required for creating
   *  remote instances of the repliation manager.  Creating a local
   *  instance of the <code>ReplicaManagerImpl</code> involves the same
   *  initialization procedure and in addition creating a local
   *  dependable registry instance and obtaining a group manager and
   *  related service references.
   *
   *  @param bootstrap
   *    True indicates that this is <code>ReplicaManagerImpl</code> is
   *    to be used for bootstrapping only, while false indicate that it
   *    will be a local replication manager instance.
   *
   *  @exception JgroupException
   *    Raised if the replication manager could not be created for
   *    numerous reasons.
   *  @exception RemoteException
   *    Raised if there was a problem with the dependable registry.
   */
  private ReplicaManagerImpl(boolean bootstrap)
    throws JgroupException, RemoteException
  {
    /* Initialize the application name for the Replication Mananger */
    ConfigManager.init();
    /* Update reachability status for all hosts in the distributed system */
    boolean someSuspected = DistributedSystemConfig.updateAll();
    /* Read configuration parameters, and construct tables */
    thisApp = AppConfig.getApplication(this);
    AppConfig dregApp = AppConfig.getApplication(RegistryImpl.class);
    dregApp.setRedundancy(thisApp.getMinimalRedundancy(), thisApp.getInitialRedundancy());
    int viewMonitorPeriod = thisApp.getIntParam("ViewMonitorPeriod", DEFAULT_VIEW_MONITOR_PERIOD);
    grouptable = new GroupTable(viewMonitorPeriod, thisApp.getGroupId());
    DistributedSystemConfig dsc = ConfigManager.getDistributedSystem();
    distScheme = new ReplicaCount(grouptable, dsc.getDomainSet());
    /*
     * Export the distribution scheme interface for the purpose of
     * recovery strategy initialization in the AppConfig class.
     */
    AppConfig.setDistributionScheme(distScheme);
    /*
     * Ensure that the RM application properties are set appropriately for both
     * the RM and DR applications.  That is, check the system properties for minimal
     * and initial redundancy configuration parameters and initialize accordingly.
     * Otherwise, the applications.xml defaults are used.
     */
    thisApp.setProperties(null);
    dregApp.setProperties(null);

    boolean createLocalRegistry = thisApp.getBooleanParam("CreateLocalRegistry");
    if (bootstrap) {
      if (!createLocalRegistry) {
        /* Create the DependableRegistry group on hosts in the distributed system */
        createGroup(dregApp);
        // Wait for the dregistry to get up and running.
        Util.sleep(2000);
      }
      createGroup(thisApp);
      log.info("Replication Manager(s) started and ready.");
    } else {
      if (createLocalRegistry) {
        /* Create a local DependableRegistry in the same JVM as the ReplicaManager */
        dregistry = new RegistryImpl();
      } else {
        // Since we have installed dregistry external to the ReplicaManager, initialize state appropriately
        distScheme.assignReplicas(dregApp);
        // Since the dregistry does not support RecoveryLayer, we remove it from the grouptable
        grouptable.remove(dregApp);
      }
      //FIXME this initialization does not give correct host assignment when
      // creating new RM replicas that is exposed to network partitions
      /* Initialize the replica assignments for the RM group (initialize state) */
      distScheme.assignReplicas(thisApp);

      /* Obtain a group manager for this application */
      GroupManager gm = GroupManager.getGroupManager(this);
      myId = gm.getLocalMember();
      if (thisApp.getBooleanParam("UpgradeManager"))
        new UpgradeManagerImpl(dregistry);
      if (Eventlogger.ENABLED)
        Eventlogger.logEventFlush(new ReplicaEvent(Initialized, gm.getGroupId()));
      log.info("Replication Manager ready...");
    }
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // ReplicationManager interface methods (event notifications for the Replication Manager)
  ////////////////////////////////////////////////////////////////////////////////////////////

  @Multicast public synchronized void notifyEvent(ARMEvent event)
    throws RemoteException, Exception
  {
    event.handle(distScheme);
    int groupId = event.getGroupId();
    if (callbackMap != null && groupId >= 0) {
      ManagementCallback callback = callbackMap.get(groupId);
      if (callback != null) {
        try {
          callback.notifyEvent(event);
        } catch (RemoteException e) {
          log.warn("Management callback client is unavailable");
          callbackMap.remove(groupId);
        }
      }
    }
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // ReplicationManager interface methods (methods for group management)
  ////////////////////////////////////////////////////////////////////////////////////////////

  @Multicast public synchronized int createGroup(AppConfig app)
    throws RemoteException, RedundancyException, GroupExistsException
  {
    /*
     * Assign replicas to a set of available hosts in the distributed
     * system.  Throws an exception if the requested redundancy could
     * not be satisfied, or if the group was already installed through
     * this method.
     */
    HostSet assignedHosts = distScheme.assignReplicas(app);
    if (!iamLeader)
      return app.getGroupId();
//    if (log.isDebugEnabled())
//      log.debug(app.getClassData().getShortName() + " assigned to " + assignedHosts);

    /* Cycle through the host list and create replicas using the ExecDaemon */
    for (Host host : assignedHosts) {
      if (!host.createReplica(app))
        host.suspect();
    }
//    if (log.isDebugEnabled())
//      log.debug("Created group: " + app.getGroupId());
    return app.getGroupId();
  }


  @Multicast public synchronized void removeGroup(AppConfig app)
    throws RemoteException, UnknownGroupException
  {
    HostSet hosts = distScheme.removeReplicas(app);
    if (!iamLeader)
      return;

    /* Cycle through the host list and remove replicas using the ExecDaemon */
    for (Host host : hosts) {
      if (!host.removeReplica(app))
        host.suspect();
    }
  }


  /* (non-Javadoc)
   * @see jgroup.core.arm.ReplicationManager#updateGroup(jgroup.relacs.config.AppConfig)
   */
  @Multicast public synchronized void updateGroup(AppConfig app)
  {
    // method to facilitate runtime changes to the replication policy
  }


  /* (non-Javadoc)
   * @see jgroup.core.arm.ReplicationManager#subscribe(jgroup.core.arm.ReplicationManager.ManagementCallback, int)
   */
  @Multicast public synchronized void subscribe(ManagementCallback callback, int gid)
    throws RemoteException, UnknownGroupException
  {
    if (!grouptable.contains(gid)) {
      throw new UnknownGroupException(gid);
    }
    if (callback != null) {
      if (callbackMap == null)
        callbackMap = new HashMap<Integer,ManagementCallback>(5);
      callbackMap.put(Integer.valueOf(gid), callback);
    }
  }


  /* (non-Javadoc)
   * @see jgroup.core.arm.ReplicationManager#unsubscribe(int)
   */
  @Multicast public synchronized void unsubscribe(int gid)
    throws RemoteException, UnknownGroupException
  {
    if (!grouptable.contains(gid) || callbackMap == null) {
      throw new UnknownGroupException(gid);
    }
    callbackMap.remove(Integer.valueOf(gid));
    if (callbackMap.isEmpty())
      callbackMap = null;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // MembershipListener interface methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * View change for the replication manager group; set the leader
   * replica, and ensure that the grouptable is correctly updated
   * with respect to this view change from the RM group.
   */
  public void viewChange(View view)
  {
    /*
     * Note that in the first view, myId will be null.
     * However, this is ok, since the isLeader() returns true
     * whenever there is only a single member in the view.
     */
    iamLeader = view.isLeader(myId);
    grouptable.viewChange(view, iamLeader);
    /* This is needed for activating self-recovery */
    grouptable.viewChangeEvent(view);
  }


  public void hasLeft()
  {
  }


  public void prepareChange()
  {
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // MergingListener interface methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /* (non-Javadoc)
   * @see jgroup.core.MergingListener#getState(jgroup.core.MemberId[])
   */
  public Object getState(MemberId[] dests)
  {
//    if (log.isDebugEnabled())
//      log.debug("RM.getState");
    return grouptable.getState(dests, myId);
  }


  /* (non-Javadoc)
   * @see jgroup.core.MergingListener#putState(java.lang.Object, jgroup.core.MemberId[])
   */
  public void putState(Object state, MemberId[] sources)
  {
//    if (log.isDebugEnabled())
//      log.debug("RM.putState");
    grouptable.putState((MergeRecord) state);
    /*
     * For each application in the stored grouptable, check if we need
     * to schedule the view monitor after the state of the replication
     * manager has been updated.
     */
    for (GroupData gd : grouptable) {
      gd.scheduleViewMonitor();
    }
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Main method
  ////////////////////////////////////////////////////////////////////////////////////////////

  public static void main(String[] argv)
    throws Exception
  {
    /* By default value, we don't bootstrap */
    boolean bootstrap = false;

    /* Parse the command line arguments */
    String usageString = "replicamanager [--bootstrap]";
    for (int i = 0; i < argv.length; i++) {
      if (argv[i].equals("--bootstrap")) {
        bootstrap = true;
      } else {
        System.err.println("Unknown option " + argv[i]);
        Abort.usage(usageString);
      }
    }

    try {
      new ReplicaManagerImpl(bootstrap);
    } catch (Exception e) {
      Abort.exit("Failed to start the replication manager", e, 1);
    }
    if (bootstrap) {
      System.exit(0);
    }
  }

} // END ReplicaManagerImpl
