/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.core;

import java.lang.annotation.Documented;
import java.lang.annotation.ElementType;
import java.lang.annotation.Retention;
import java.lang.annotation.RetentionPolicy;
import java.lang.annotation.Target;

/**
 *  In order to be notified of membership events like view changes, an
 *  object must implement this interface.
 *
 *  @author Alberto Montresor
 *  @since Jgroup 0.1
 */
public interface MembershipListener
{

  /**
   * Annotation marker used to indicate that the <code>viewChange()</code>
   * method should allow invocations in which the two consecutive views have the
   * same set of members.  <i>The default is to suppress duplicate views.</i>
   *
   * @author Hein Meling
   * @since Jgroup 3.0
   */
  @Documented
  @Retention(RetentionPolicy.RUNTIME)
  @Target(ElementType.METHOD)
  public @interface AllowDuplicateViews { }

  /**
   *  Upcall invoked on members implementing the <code>MembershipListener</code>
   *  interface, when a view change occurs.
   *
   *  @param view
   *    The new view that has been installed.
   *  @see jgroup.core.MembershipService
   */
  public void viewChange(View view);


  /**
   *  Upcall that is invoked by the group membership service to notify
   *  that the current view is not valid any more, and the group
   *  membership service is trying to reach an agreement on a new view.
   *
   *  @see jgroup.core.MembershipService
   */
  public void prepareChange();


  /**
   *  Upcall that is invoked by Jgroup to acknowledge the fact that this
   *  object has left the group.  This method is only invoked after the
   *  <code>leave</code> method has been invoked on the group membership
   *  service.
   * 
   *  @see jgroup.core.MembershipService
   */
  public void hasLeft();

} // END MembershipListener
