/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.core;

import jgroup.core.Layer.FinalizeLayer;

/**
 *  The facilities provided by the group membership service included in
 *  Jgroup may be accessed using the <code>MembershipService</code>
 *  interface.  Methods in this interfaces enable objects to become
 *  members of a group and subsequentially leave it. <p>
 *
 *  Instances of <code>MembershipService</code> may be obtained through
 *  the <code>GroupManager</code> class.
 *
 *  @author Alberto Montresor
 *  @author Hein Meling
 *  @since Jgroup 0.1 (Revised Jgroup 2.2)
 *  @see jgroup.core.GroupManager
 */
public interface MembershipService
  extends Layer, FinalizeLayer
{

  /**
   *  Requests to join a group identified by <code>groupId</code>.  <p>
   *
   *  @param groupId
   *    Non-negative integer identifying the group.
   *  @exception IllegalArgumentException
   *    Raised if <code>groupId</code> is negative.
   *  @exception JgroupException
   *    Raised if the membership service receives a request to join a
   *    group to which it is already a member.
   */
  public void join(int groupId)
    throws JgroupException;


  /**
   *  Requests to join the default group associated with the Jgroup daemon. <p>
   *
   *  Each Jgroup daemon has a default group.  If a Jgroup daemon does not need
   *  to be shared by multiple groups, the application servers may simply join
   *  the default group without having to specify group identifiers.
   *
   *  @exception JgroupException
   *    Raised if the membership service receives a request to join a
   *    group to which it is already a member.
   */
  public void join()
    throws JgroupException;


  /**
   *  Request to leave a group.  Note that an object member may receive
   *  event notifications such as view installations for a group even
   *  after having requested to leave the group itself.  However, the
   *  member will eventually leave the group.  When this happens, its
   *  <code>hasLeft</code> method is invoked.  After this invocation,
   *  the member will not receive any further event notification related
   *  to the group.
   *
   *  @exception JgroupException
   *    Rasied if the server is not member of any group, or is already
   *    leaving its current group.
   */
  public void leave()
    throws JgroupException;


  /**
   *  Returns true if the <code>GroupManager</code> is member of some
   *  group.  Otherwise, false is returned.
   */
  public boolean isMember();


  /**
   *  Returns true if the <code>GroupManager</code> is in the process of
   *  joining some group.  Otherwise, false is returned.
   */
  public boolean isJoining();


  /**
   *  Returns true if the <code>GroupManager</code> is member of some group
   *  or is in the process of joining some group.  Otherwise, false is returned.
   */
  public boolean isMemberOrJoining();


  /**
   * Returns true if the local member is the group leader.
   */
  public boolean isLeader();


  /**
   *  Returns the member identifier of the local member.
   */
  public MemberId getMyIdentifier();


  /**
   *  Returns the group identifier.
   */
  public int getGid();


  /**
   *  Returns a <code>MemberTable</code> helper object that will be
   *  notified of changes in the membership.  This is used to provide
   *  application developers with additional information about members
   *  of their current view.
   */
  public MemberTable getMemberTable();


  /**
   *  Return the number of members of this group.
   */
  public int members();


  /**
   *  Position index of this member in the array containing
   *  the members of the current view.
   */
  public int getViewIndex();


  /**
   *  Position index of this member in the array containing 
   *  the local members of the current view.
   */
  public int getMemberIndex();

} // END MembershipService
