/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.core.arm;

import java.rmi.Remote;
import java.rmi.RemoteException;
import java.util.Set;

import jgroup.relacs.config.ClassData;


/**
 *  The <code>ExecService</code> interface describes the methods of an
 *  execution service that can be remotely invoked. Such an execution
 *  service can be used to start other java applications (replicas) on
 *  the remote machine on which the execution service is running.
 *
 *  @author  Hein Meling
 *  @since   Jgroup 1.2
 */
public interface ExecService
  extends Remote
{

  /////////////////////////////////////////////////////////////////////////////
  // Constants
  /////////////////////////////////////////////////////////////////////////////

  /**
   *  Service name used to register each execution daemon in its local
   *  registry, or bootstrap registry.
   */
  public static final String EXEC_DAEMON_NAME = "ExecutionDaemon";

  /**
   *  Keyword used to communicate to a replica that is should do a
   *  graceful shutdown.
   */
  public static final String SHUTDOWN_REPLICA = "SHUTDOWN";


  /////////////////////////////////////////////////////////////////////////////
  // Methods
  /////////////////////////////////////////////////////////////////////////////

  /**
   *  Instantiate and start the given class (replica), within the same
   *  JVM as the execution service.  Be aware that starting a new
   *  replica in this way, may cause either the replica(s) or the
   *  execution service to crash due to the failure of only a single
   *  replica.  Also, the use of System.exit() from such as replica will
   *  cause the JVM to exit.  Thus, this approach should be used with
   *  causion.
   *
   *  @param classData 
   *    The class data object representing the replica to start.  The
   *    specified class must implement the remove method.
   *  @return
   *    False if the replica already exists on the associated execution
   *    service; true is returned if the replica was created.
   * 
   *  @throws RemoteException 
   *    If remote operation failed.
   *  @throws ExecException 
   *    If the specified class (replica) could not be instantiated or
   *    if the class (replica) does not implement the remove method.
   */
  public boolean createExecReplica(ClassData classData)
    throws RemoteException, ExecException;

  /**
   *  Instantiate and start the given class (replica) in a JVM process
   *  separate from the execution service JVM process.
   *
   *  @param classData 
   *    The class data object representing the replica to start.
   *  @return
   *    False if the replica already exists on the associated execution
   *    service; true is returned if the replica was created.
   * 
   *  @throws RemoteException 
   *    If remote operation failed.
   *  @throws ExecException 
   *    If the specified class (replica) could not be instantiated.
   */
  public boolean createReplica(ClassData classData)
    throws RemoteException, ExecException;

  /**
   *  Remove the specified class (replica) from the execution service
   *  associated with the remote reference.
   *
   *  @param classData 
   *    The class data object representing the replica to start.
   *  @return 
   *    False if the specified class (replica) could not be found on
   *    the associated execution service; true if the replica was
   *    successfully removed.
   * 
   *  @throws RemoteException 
   *    If remote operation failed.
   */
  public boolean removeReplica(ClassData classData)
    throws RemoteException;

  /**
   *  Return the set of replicas currently running on this execution service.
   *
   *  @return 
   *    The set of class data objects representing replicas that is currently
   *    running within this <code>ExecService</code> instance.
   *  @throws RemoteException 
   *    If remote operation failed.
   */
  public Set<ClassData> queryReplicas()
    throws RemoteException;

  /**
   *  Method used to check if the execution service is still running.
   *
   *  @throws RemoteException 
   *    If remote operation failed.  This will indicate that the execution
   *    service is no longer running.
   */
  public void ping()
    throws RemoteException;

  /**
   *  Request the execution service on this particular host to shutdown
   *  in approximately <code>delay</code> milliseconds.  If delay is negative
   *  all local replicas are removed, and shutdown will be immediate and no
   *  shutdown hooks are executed.  If the delay is positive, shutdown hooks
   *  will be executed during the shutdown sequence.
   *
   *  @param delay
   *    The delay to wait before initiating shutdown.
   *  @throws RemoteException
   *    If remote operation failed.  This will indicate that the execution
   *    service is no longer running.
   */
  public void shutdown(int delay)
    throws RemoteException;

} // END ExecService
