/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */ 

package jgroup.core.arm;

import java.rmi.Remote;
import java.rmi.RemoteException;

import jgroup.core.ConfigurationException;
import jgroup.core.ExternalGMIListener;
import jgroup.relacs.config.AppConfig;


/**
 *  The <code>ReplicationManager</code> interface describes the methods
 *  that a management client can invoke on the replication manager to
 *  create and remove object groups within the distributed system.
 *
 *  Note that all methods in this interface has no declared invocation
 *  semantics.  The invocation semantics depends on the implementation
 *  of this interface, and the <code>ReplicaManagerImpl</code> class
 *  provides one such implementation in which all methods are defined
 *  to have <code>@Multicast</code> invocation semantics.  However, in
 *  this implementation of this interface only the <i>leader</i> replica
 *  actually perform the create and remove replica operations on the 
 *  execution service, while the <i>follower</i> replicas just keeps
 *  track of the state by updating its own group table.
 *
 *  The above semantics require the replica distribution scheme to
 *  perform deterministic replica placements.
 *
 *  The <code>ReplicationManager</code> interface also describe the
 *  methods that ARM components, such as the <code>RecoveryLayer</code>
 *  and <code>ExecDaemon</code> can invoke on the replication manager
 *  to notify it of certain events.
 *
 *  @author Hein Meling
 *  @since Jgroup 1.2
 */
public interface ReplicationManager
  extends ExternalGMIListener
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Event notification method (notifications for the Replication Manager)
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  The <code>notifyEvent</code> method is used by components external
   *  to the replication manager to notify it of important events, such
   *  as the presence of a new host (<code>ExecDaemon</code>) or view
   *  change events from the <code>RecoveryLayer</code>. <p>
   *
   *  The method takes arbitrary event types as long as they implement
   *  the interface <code>ARMEvent</code>.
   *
   *  @param event
   *    An event notification for the replication manager.
   * @exception RemoteException
   *    Raised if the remote operation failed.
   */
  public void notifyEvent(ARMEvent event)
    throws RemoteException, Exception;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods for group management
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Create a group for the specified application, with the
   *  applications required redundancy.
   *
   *  @param app
   *    The application details required to create a group for it.
   *  @return 
   *    The group identifier.
   *  @exception RemoteException
   *    Raised if there was communication problems; only when invoked remotely.
   *  @exception ConfigurationException
   *    Raised if the specified application recovery strategy could not be initialized.
   *  @exception RedundancyException
   *    Raised if the specified redundancy could not be satisfied.
   *  @exception GroupExistsException
   *    Raised if the specified group already exists.
   */
  public int createGroup(AppConfig app)
    throws RemoteException, RedundancyException, GroupExistsException;


  /**
   *  Remove the group for the given application.
   *
   *  @param app
   *    The application group to remove.
   *  @exception RemoteException
   *    Raised if there was communication problems; only when invoked remotely.
   *  @exception UnknownGroupException
   *    Raised if the specified application has no group in the system.
   */
  public void removeGroup(AppConfig app)
    throws RemoteException, UnknownGroupException;


  /**
   *  Update the replication policy for the group for the given application.
   *
   *  @param app
   *    The application group to remove.
   *  @exception RemoteException
   *    Raised if there was communication problems; only when invoked remotely.
   *  @exception UnknownGroupException
   *    Raised if the specified application has no group in the system.
   */
  public void updateGroup(AppConfig app)
    throws RemoteException, UnknownGroupException;


  /**
   *  Register to receive notifications for the given group.
   *
   *  @param callback
   *    The callback server (typically a management client)
   *  @param group
   *    The group for which notifications are desired
   *  @exception RemoteException
   *    Raised if there was communication problems; only when invoked remotely.
   *  @exception UnknownGroupException
   *    Raised if the specified application has no group in the system.
   */
  public void subscribe(ManagementCallback callback, int group)
    throws RemoteException, UnknownGroupException;


  /**
   *  Unregister to receive notifications for the given group.
   *
   *  @param group
   *    The group for which notifications are no longer desired
   *  @exception RemoteException
   *    Raised if there was communication problems; only when invoked remotely.
   *  @exception UnknownGroupException
   *    Raised if the specified application has no group in the system.
   */
  public void unsubscribe(int group)
    throws RemoteException, UnknownGroupException;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Callback interface for group management
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Callback interface for monitoring installed applications.
   */
  public interface ManagementCallback
    extends Remote
  {

    /**
     *  Notify the management application that a new view has been
     *  installed.
     */
    public void notifyEvent(ARMEvent event)
      throws RemoteException;

  }

} // END ReplicationManager
