/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.core.registry;

import java.rmi.RemoteException;
import java.rmi.server.RMIClientSocketFactory;
import java.rmi.server.RMIServerSocketFactory;

import jgroup.core.ConfigManager;
import jgroup.core.ConfigurationException;
import jgroup.core.JgroupException;


/**
 *  Class <code>RegistryFactory</code> is used to obtain a group proxy
 *  for the dependable registry service, or to create a dependable
 *  registry replica on the local machine. <p>
 *
 *  Every instance (replica) of the dependable registry service is
 *  associated with a standard RMI registry, which is used to retrieve a
 *  group reference for the replicated remote registries.
 *
 *  @see jgroup.relacs.registry.RelacsRegistryLocator
 *  @author Alberto Montresor
 *  @author Hein Meling
 *  @since Jgroup 0.8
 */
public final class RegistryFactory
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Static section
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** The registry locator object needed to obtain a registry instance */
  private static RegistryLocator locator;


  /**
   *  Set the registry locator object.
   */
  public static void setLocator(RegistryLocator reglocator)
  {
    if (reglocator == null)
      throw new IllegalStateException("Cannot set registry locator to null");
    locator = reglocator;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructor
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Private constructor to prevent public construction.
   */
  private RegistryFactory() {}


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Get registry methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Returns a proxy for the dependable registry service.
   *
   *  @return 
   *    A proxy for the dependable registry service.
   *
   *  @exception RemoteException 
   *    Raised if the proxy could not be obtained.
   *  @exception ConfigurationException
   *    Raised if there was a problem parsing the distributed system
   *    configuration file.
   */
  public static DependableRegistry getRegistry()
    throws RemoteException, JgroupException
  {
    ConfigManager.init();
    if (locator == null)
      throw new IllegalStateException("Registry locator not yet initialized");
    return locator.getRegistry();
  }

  /**
   * @return
   *   Returns an interface clients can use to obtain
   *   a remote reference from registry. 
   */
  public static LookupRegistry getLookupRegistry()
    throws RemoteException, JgroupException
  {
    return locator.getLookupRegistry();
  }

  /**
   *  Returns a proxy for the dependable registry service.
   *
   *  Communication with this remote registry will use the supplied
   *  <code>RMIClientSocketFactory</code> to create socket connections
   *  to the dependable registry on the distributed system and port.
   *
   *  @param csf
   *    Client-side socket factory used to make connections to the
   *    dependable registry. If <code>csf</code> is null, then the
   *    default client-side socket factory will be used.
   *
   *  @return 
   *    A proxy for the dependable registry service.
   *
   *  @exception RemoteException 
   *    Raised if the proxy could not be obtained.
   *  @exception ConfigurationException
   *    Raised if there was a problem parsing the distributed system
   *    configuration file.
   */
  public static DependableRegistry getRegistry(RMIClientSocketFactory csf)
    throws RemoteException, JgroupException
  {
    ConfigManager.init();
    if (locator == null)
      throw new IllegalStateException("Registry locator not yet initialized");
    return locator.getRegistry(csf);
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Create registry methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Creates and exports a dependable registry replica on the local
   *  host listening for incoming requests on an the port specified in
   *  the system configuration file. <p>
   *
   *  @return 
   *    The newly created dependable registry replica.
   *
   *  @exception RemoteException
   *    Raised if the registry could not be exported.
   *  @exception JgroupException
   *    Raised if a registry instance could not be created.
   *  @exception ConfigurationException
   *    Raised if there was a problem parsing the distributed system
   *    configuration file.
   */
  public static DependableRegistry createRegistry()
    throws RemoteException, JgroupException
  {
    ConfigManager.init();
    if (locator == null)
      throw new IllegalStateException("Registry locator not yet initialized");
    return locator.createRegistry();
  }


  /**
   *  Creates and exports a dependable registry replica on the local
   *  host listening for incoming requests on an the port specified in
   *  the system configuration file. <p>
   *
   *  The new instance of the dependable registry listens to incoming
   *  requests using a <code>ServerSocket</code> created from the
   *  supplied <code>RMIServerSocketFactory</code>.  A client that
   *  receives a reference to this registry will then use a
   *  <code>Socket</code> created from the supplied
   *  <code>RMIClientSocketFactory</code>.
   *
   *  @param csf
   *    Client-side <code>Socket</code> factory used to make connections
   *    to the registry.
   *  @param ssf
   *    Server-side <code>ServerSocket</code> factory used to accept
   *    connections to the registry.
   *
   *  @return 
   *    The newly created dependable registry replica.
   *
   *  @exception RemoteException
   *    Raised if the registry could not be exported.
   *  @exception JgroupException
   *    Raised if a registry instance could not be created.
   *  @exception ConfigurationException
   *    Raised if there was a problem parsing the distributed system
   *    configuration file.
   */
  public static DependableRegistry createRegistry(RMIClientSocketFactory csf, RMIServerSocketFactory ssf)
    throws RemoteException, JgroupException
  {
    ConfigManager.init();
    if (locator == null)
      throw new IllegalStateException("Registry locator not yet initialized");
    return locator.createRegistry(csf, ssf);
  }

} // END RegistryFactory
