/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.core.registry;

import java.rmi.AccessException;
import java.rmi.NotBoundException;
import java.rmi.RemoteException;

import jgroup.core.IID;
import jgroup.core.JgroupException;
import jgroup.core.Layer;

/**
 *  This interface is used by servers to bind and unbind their reference
 *  with a specified name from the dependable registry.  That is the name
 *  to server reference bindings setup through this interface will be
 *  stored in the dependable registry.  This interface is accessed by
 *  server objects (and layers) through the group manager.
 *
 *  @author Hein Meling
 *  @since Jgroup 2.2
 */
public interface RegistryService
  extends Layer
{

  /**
   *  The default lease time used by the dependable registry (milliseconds)
   *  when clearing out expired leases.  This assumes that the
   *  <code>LeaseLayer</code> is configured in the application layer stack;
   *  otherwise lease expiration checks in the dependable registry is disabled.
   *  Changes to the lease time should be configured in the
   *  <code>applications.xml</code> configuration file.
   *
   *  @see jgroup.relacs.gm.LeaseService#DEFAULT_REFRESH_RATE
   */
  public static final int DEFAULT_LEASE_TIME = 60000;

  /**
   *  Bind the server associated with the <code>RegistryLayer</code>
   *  to the object group associated with the name obtained from the
   *  <code>application.xml</code> file in the dependable registry
   *  running in the same distributed system as this group manager.
   *
   *  @return 
   *    A binding identifier <code>IID</code> that can be used to
   *    remove this server binding from the registry.
   *
   *  @exception RemoteException
   *    Raised when the remote operation failed.
   *  @exception AccessException
   *    Raised when this operation is not permitted.
   *  @exception JgroupException
   *    Raised in the event of a group communication failure.
   */
  public IID bind()
    throws RemoteException, JgroupException;

  /**
   *  Bind the given server to the object group associated with the
   *  name obtained from the <code>application.xml</code> file in the
   *  dependable registry running in the same distributed system as
   *  this group manager.
   *
   *  @param server
   *    The server in the object group to associated with the given name.
   *  @return 
   *    A binding identifier <code>IID</code> that can be used to
   *    remove this server binding from the registry.
   *
   *  @exception RemoteException
   *    Raised when the remote operation failed.
   *  @exception AccessException
   *    Raised when this operation is not permitted.
   *  @exception JgroupException
   *    Raised in the event of a group communication failure.
   */
  public IID bind(Object server)
    throws RemoteException, JgroupException;

  /**
   *  Bind the given server to the object group associated with the
   *  name in the dependable registry running in the same distributed
   *  system as this group manager.
   *
   *  @param name
   *    The name associated with the object group.
   *  @param server
   *    The server in the object group to associated with the given name.
   *  @return 
   *    A binding identifier <code>IID</code> that can be used to
   *    remove this server binding from the registry.
   *
   *  @exception RemoteException
   *    Raised when the remote operation failed.
   *  @exception AccessException
   *    Raised when this operation is not permitted.
   *  @exception JgroupException
   *    Raised in the event of a group communication failure.
   */
  public IID bind(String name, Object server)
    throws RemoteException, JgroupException;

  /**
   *  Bind the given server to the object group associated with the
   *  specified name in the specified dependable registry.
   *
   *  @param name
   *    The name associated with the object group.
   *  @param server
   *    The server in the object group to associated with the given name.
   *  @param registry
   *    The dependable registry in which to bind the given server.
   *  @return 
   *    A binding identifier <code>IID</code> that can be used to
   *    remove this server binding from the registry.
   *
   *  @exception RemoteException
   *    Raised when the remote operation failed.
   *  @exception AccessException
   *    Raised when this operation is not permitted.
   *  @exception JgroupException
   *    Raised in the event of a group communication failure.
   */
  public IID bind(String name, Object server, DependableRegistry registry)
    throws RemoteException, JgroupException;

  /**
   *  Unbind the reference associated with the <code>RegistryLayer</code>
   *  from the dependable registry running in the same distributed system
   *  as this group manager.
   *
   *  @exception RemoteException
   *    Raised when the remote operation failed.
   *  @exception NotBoundException
   *    Raised the given binding identifier was not in the dependable registry.
   *  @exception AccessException
   *    Raised when this operation is not permitted.
   *  @exception JgroupException
   *    Raised in the event of a group communication failure.
   */
  public void unbind()
    throws RemoteException, NotBoundException, JgroupException;

  /**
   *  Unbind the reference associated with the specified binding
   *  identifier from the dependable registry running in the same
   *  distributed system as this group manager.
   *
   *  @param bindId
   *    The binding identifier to be removed from the dependable registry.
   * 
   *  @exception RemoteException
   *    Raised when the remote operation failed.
   *  @exception NotBoundException
   *    Raised the given binding identifier was not in the dependable registry.
   *  @exception AccessException
   *    Raised when this operation is not permitted.
   *  @exception JgroupException
   *    Raised in the event of a group communication failure.
   */
  public void unbind(IID bindId)
    throws RemoteException, NotBoundException, JgroupException;

  /**
   *  Unbind the reference associated with the specified binding
   *  identifier from the specified dependable registry.
   *
   *  @param bindId
   *    The binding identifier to be removed from the dependable registry.
   *  @param registry
   *    The dependable registry from which to unbind the given binding.
   *
   *  @exception RemoteException
   *    Raised when the remote operation failed.
   *  @exception NotBoundException
   *    Raised the given binding identifier was not in the dependable registry.
   *  @exception AccessException
   *    Raised when this operation is not permitted.
   *  @exception JgroupException
   *    Raised in the event of a group communication failure.
   */
  public void unbind(IID bindId, DependableRegistry registry)
    throws RemoteException, NotBoundException;

  /**
   *  Refresh the lease for the local server.
   * 
   *  @exception RemoteException
   *    Raised when the remote operation failed.
   *  @exception JgroupException
   *    Raised in the event of a group communication failure.
   */
  public void refreshLease()
    throws RemoteException, JgroupException;

  /**
   *  Returns the local invocation identifier bound to the dependable registry.
   *  This can be used to <code>unbind()</code> the associated server from
   *  the dependable registry.
   *
   *  @return
   *    The local invocation identifier bound to the depedable registry.
   *  @throws NullPointerException
   *    Raised if the replica is not yet bound to the dependable registry.
   */
  public IID getIID()
    throws NullPointerException;

} // END RegistryService
