/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.relacs.config;

import java.io.Externalizable;
import java.io.IOException;
import java.io.ObjectInput;
import java.io.ObjectOutput;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.NoSuchElementException;
import java.util.StringTokenizer;

import jgroup.core.ConfigurationException;
import jgroup.core.MemberId;
import jgroup.core.View;
import jgroup.core.arm.DistributionScheme;
import jgroup.core.arm.RecoveryStrategy;
import jgroup.relacs.types.ViewImpl;

import org.apache.log4j.Logger;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;


/**
 *  Class for holding application information, including replication
 *  style and recovery strategy for the dependable applications.
 *
 *  @author Hein Meling
 *  @since Jgroup 1.2
 */
public class AppConfig
  implements Externalizable, ConfigurationObject
{

  private static final long serialVersionUID = 1335688016655019665L;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Logger
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Obtain logger for this class */
  private static final Logger log = Logger.getLogger(AppConfig.class);


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Tags
  ////////////////////////////////////////////////////////////////////////////////////////////

  // Attributes for the App tag
  private static final String NAME_ATTRIB                = "name";
  private static final String GROUP_ATTRIB               = "group";

  private static final String CLASS_TAG_NAME             = "Class";
  private static final String EXEC_JVM_ATTRIB            = "inExecJVM";
  private static final String ARGS_ATTRIB                = "args";

  private static final String PARAM_TAG_NAME             = "Param";
  private static final String VALUE_ATTRIB               = "value";

  private static final String LAYERSTACK_TAG_NAME        = "LayerStack";
  private static final String ORDER_ATTRIB               = "order";

  private static final String SERVICE_TAG_NAME           = "Service";
  private static final String AUTO_ATTRIB                = "auto";

  private static final String RECOVERYSTRATEGY_TAG_NAME  = "RecoveryStrategy";
  private static final String REDUNDANCY_TAG_NAME        = "Redundancy";
  private static final String MINIMAL_ATTRIB             = "minimal";
  private static final String INITIAL_ATTRIB             = "initial";


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constants
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Lower and upper bound on the redundancy level allowed.  Used for
   *  checking if the specified redundancy levels are valid.
   */
  private final static int REDUNDANCY_LOWER_BOUND = 1;
  private final static int REDUNDANCY_UPPER_BOUND = 20;

  /**
   *  Initial capacity is 37, since the expected number of application
   *  entries is less than 28.  If this succeeds we avoid having to
   *  rehash the HashMap.  @See java.util.HashMap for details.
   */
  private static final int APPLICATION_HASH_SIZE = 37;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Static collections (for inter-object or class access)
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Application registry name -> AppConfig object */
  private static Map<String,AppConfig> appMap =
    new HashMap<String,AppConfig>(APPLICATION_HASH_SIZE);

  /** Application class name -> AppConfig object */
  private static Map<String,AppConfig> appClazzMap =
    new HashMap<String,AppConfig>(APPLICATION_HASH_SIZE);

  /** Application group identifier -> AppConfig object */
  private static Map<Integer,AppConfig> appGroupMap =
    new HashMap<Integer,AppConfig>(APPLICATION_HASH_SIZE);


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Transient and static fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Recovery strategy instance */
  private transient RecoveryStrategy recoveryStrategy = null;

  /** The distribution scheme of the local replication manager */
  private static DistributionScheme dScheme = null;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** The name used in the registry for this application */
  private String regName;

  /** Group identifier */
  private int groupId;

  /** ClassData (className, classArgs) */
  private ClassData classData;

  /** The service set or layer stack order for this application */
  private List<String> serviceSet;

  /** Map containing parameters associated with this application */
  private Map<String,String> paramMap;

  /**
   *  Boolean value used to indicate that if the application
   *  should be started within the same JVM as the execution
   *  daemon.  If the value is false the application will be
   *  started in a separate JVM.
   */
  private boolean inExecJVM;

  /** Recovery strategy class */
  private Class<?> strategyClass;

  /** Initial redundancy for this application */
  private int initialRedundancy = -1;

  /** Minimal redundancy for this application */
  private int minimalRedundancy = -1;

  /** The current view of this application */
  private View view;

  /** The set of hosts on which this application is known to be located */
  private HostSet viewHosts = new HostSet();


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructors
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Creates a new <code>AppConfig</code> instance, from the given class
   *  name, arguments and group identifier.  The <code>AppConfig</code>
   *  object <i>must</i> also be initialized with the required
   *  redundancy and replication style parameters through the
   *  corresponding set methods; otherwise an
   *  <code>IllegalStateException</code> will be thrown upon using the
   *  <code>AppConfig</code> object.
   *
   *  @param className
   *     The classname for this application object.
   *  @param args
   *     The arguments for the application object.
   *  @param groupId
   *     The group identifier for this application object.
   *  @exception ClassNotFoundException 
   *     Raised if the given class could not be found.
   */
  public AppConfig(String regName, String className, String args, int groupId, boolean inExecJVM)
//    throws ClassNotFoundException
  {
    this.regName = regName;
    this.groupId = groupId;
    this.view = new ViewImpl(groupId);
    StringTokenizer argTokens = new StringTokenizer(args);
    String[] classArgs = new String[argTokens.countTokens()];
    for (int i = 0; argTokens.hasMoreTokens(); i++)
      classArgs[i] = argTokens.nextToken();
    classData = new ClassData(className, classArgs);
    this.inExecJVM = inExecJVM;
  }


  /**
   *  Augment the system with an additional application.
   *
   *  @param regname
   *     the registry name for this application
   *  @param classname
   *     the class name
   *  @param args
   *     argument string
   *  @param group
   *     the group identifier for the application
   *  @param inExecJVM
   *     must be true if the application should be started in the execution daemon JVM.
   *     if set to false, the application will be started in a separate process.
   *  @return 
   *    application info object associated with the <classname, args> pair 
   *  @throws ConfigurationException
   *    if the specified class name is already registered
   */
  public static AppConfig addApplication(String regname, String classname, String args,
                                  int group, boolean inExecJVM)
    throws ConfigurationException
  {
//    if (appMap.containsKey(regname))
//      throw new ConfigurationException("Registry name already defined: " + regname);
    /*
     * We don't allow multiple application definitions for the same class.
     */
    if (appClazzMap.containsKey(classname))
      throw new ConfigurationException("Class name already defined: " + classname);
    AppConfig app = new AppConfig(regname, classname, args, group, inExecJVM);
//    try {
//      app = new AppConfig(regname, classname, args, group, inExecJVM);
//    } catch (ClassNotFoundException e) {
//      throw new ConfigurationException("Could not find the specified application class: "
//        + e.getMessage());
//    }
    appMap.put(regname, app);
    appClazzMap.put(classname, app);
    appGroupMap.put(new Integer(group), app);
    return app;
  }


  /**
   * Add the given set of applications to the static maps.
   *
   * @param apps
   * @throws ConfigurationException
   */
  public static void addApplications(AppConfig[] apps)
    throws ConfigurationException
  {
    for (int i = 0; i < apps.length; i++) {
      AppConfig app = apps[i];
      String classname = app.getClassName();
      if (appClazzMap.containsKey(classname))
        throw new ConfigurationException("Class name already defined: " + classname);
      appMap.put(app.getRegistryName(), app);
      appClazzMap.put(classname, app);
      appGroupMap.put(new Integer(app.getGroupId()), app);
    }
  }


  /**
   *  Returns an array of all applications registered.
   */
  public static AppConfig[] getApplications()
  {
    return (AppConfig[]) appMap.values().toArray(new AppConfig[appMap.size()]);
  }


  /**
   *  Returns an <CODE>AppConfig</CODE> object associated with the
   *  specified name.  The name may be either the registry name or the
   *  class name of the application.
   *
   *  @exception NullPointerException
   *    Raised if the specified name has no corresponding class in the
   *    applications configuration.
   */
  public static AppConfig getApplication(String name)
  {
    AppConfig app = (AppConfig) appMap.get(name);
    if (app == null) {
      app = (AppConfig) appClazzMap.get(name);
      if (app == null)
        throw new NullPointerException("Unknown application: " + name);
    }
    return app;
  }


  /**
   *  Returns an <CODE>AppConfig</CODE> object associated with the
   *  specified object instance.
   *
   *  @exception NullPointerException
   *    Raised if the specified object instance name has no
   *    corresponding class in the applications configuration.
   */
  public static AppConfig getApplication(Object appObj)
  {
    if (appObj == null)
      throw new NullPointerException("No application object specified");
    AppConfig app = (AppConfig) appClazzMap.get(appObj.getClass().getName());
    if (app == null)
      throw new NullPointerException("Unknown application: " + appObj.getClass().getName());
    return app;
  }


  /**
   *  Returns an <CODE>AppConfig</CODE> object associated with the
   *  specified class.
   *
   *  @exception NullPointerException
   *    Raised if the specified class has no corresponding class
   *    in the applications configuration.
   */
  public static AppConfig getApplication(Class cl)
  {
    AppConfig app = (AppConfig) appClazzMap.get(cl.getName());
    if (app == null)
      throw new NullPointerException("Unknown application: " + cl.getName());
    return app;
  }


  /**
   *  Returns an <CODE>AppConfig</CODE> object associated with the
   *  specified group identifier.
   *
   *  @exception NullPointerException
   *    Raised if the specified group identifier has no corresponding application
   *    in the applications configuration.
   */
  public static AppConfig getApplication(int gid)
  {
    AppConfig app = (AppConfig) appGroupMap.get(new Integer(gid));
    if (app == null)
      throw new NullPointerException("Unknown application for group: " + gid);
    return app;
  }

  /**
   *  Method to store the distribution scheme of the local
   *  replication manager.  This is required for recovery strategy
   *  construction and initialization.
   */
  public static void setDistributionScheme(DistributionScheme distScheme)
  {
    dScheme = distScheme;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from ConfigurationObject
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Parses the specified element.
   */
  public void parse(Element elm)
    throws ConfigurationException
  {
    /* 
     *  Read attributes of the App tag
     */
    String regname = ConfigParser.getAttrib(elm, NAME_ATTRIB, true);
    int group = ConfigParser.getIntAttrib(elm, GROUP_ATTRIB, true);
    AppConfig app = null;

    /*
     * Nodes within the application tag should be either Class,
     * LayerStack, ReplicationStyle or RecoveryStrategy tags.
     */
    NodeList tagList = elm.getChildNodes();

    /* Parse the list of tags */
    for (int i = 1; i <= tagList.getLength(); i++) {
      Node node = tagList.item(i);
      if (node != null && !node.getNodeName().startsWith("#")) {
        elm = (Element)node;

        if (ConfigParser.checkTag(elm, CLASS_TAG_NAME)) {
          String classname = ConfigParser.getAttrib(elm, NAME_ATTRIB, true);
          boolean inExecJVM = ConfigParser.getBooleanAttrib(elm, EXEC_JVM_ATTRIB, false);
          String args = ConfigParser.getAttrib(elm, ARGS_ATTRIB, false);
          /* Add the application data to the system */
          app = AppConfig.addApplication(regname, classname, args, group, inExecJVM);

        } else if (ConfigParser.checkTag(elm, PARAM_TAG_NAME)) {
          String pName = ConfigParser.getAttrib(elm, NAME_ATTRIB, true);
          String pValue = ConfigParser.getAttrib(elm, VALUE_ATTRIB, true);
          app.addParam(pName, pValue);

        } else if (ConfigParser.checkTag(elm, LAYERSTACK_TAG_NAME)) {
          if (!parseServiceSet(app, elm)) {
            /*
             * The service set was not defined using the Service tag,
             * check if there is an order atttribute instead.
             */
            String stackOrder = ConfigParser.getAttrib(elm, ORDER_ATTRIB, true);
            app.setServiceSet(stackOrder);
          }

        } else if (ConfigParser.checkTag(elm, RECOVERYSTRATEGY_TAG_NAME)) {
          parseRecoveryStrategy(app, elm);
        }
      }
    }
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Private methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  private boolean parseServiceSet(AppConfig app, Element elm)
    throws ConfigurationException
  {
    NodeList serviceList = elm.getChildNodes();
    if (serviceList.getLength() == 0) {
      return false;
    }

    /* Parse the list of tags */
    for (int i = 1; i <= serviceList.getLength(); i++) {
      Node node = serviceList.item(i);
      if (node != null && !node.getNodeName().startsWith("#")) {
        elm = (Element)node;
        if (ConfigParser.checkTag(elm, SERVICE_TAG_NAME)) {
          String serviceName = ConfigParser.getAttrib(elm, NAME_ATTRIB, true);
          boolean auto = ConfigParser.getBooleanAttrib(elm, AUTO_ATTRIB, false);
          if (auto)
            app.addParam(serviceName + "." + AUTO_ATTRIB, Boolean.toString(auto));
          app.addService(serviceName);
          parseServiceParams(serviceName, app, elm);
        }
      }
    }
    return true;
  }


  private void parseServiceParams(String serviceName, AppConfig app, Element elm)
    throws ConfigurationException
  {
    NodeList paramList = elm.getChildNodes();

    /* Parse the list of tags */
    for (int i = 1; i <= paramList.getLength(); i++) {
      Node node = paramList.item(i);
      if (node != null && !node.getNodeName().startsWith("#")) {
        elm = (Element)node;
        if (ConfigParser.checkTag(elm, PARAM_TAG_NAME)) {
          String pName = serviceName + "." + ConfigParser.getAttrib(elm, NAME_ATTRIB, true);
          String pValue = ConfigParser.getAttrib(elm, VALUE_ATTRIB, true);
          app.addParam(pName, pValue);
        }
      }
    }
  }


  private void parseRecoveryStrategy(AppConfig app, Element elm)
    throws ConfigurationException
  {
    int initial = 0, minimal = 0;

    String strategy = ConfigParser.getAttrib(elm, NAME_ATTRIB, true);

    NodeList redundancyList = elm.getChildNodes();

    /* Parse the list of tags */
    for (int i=1; i <= redundancyList.getLength(); i++) {
      Node node = redundancyList.item(i);
      if (node != null && !node.getNodeName().startsWith("#")) {
        elm = (Element)node;
        if (ConfigParser.checkTag(elm, REDUNDANCY_TAG_NAME)) {
          initial = ConfigParser.getIntAttrib(elm, INITIAL_ATTRIB, true);
          minimal = ConfigParser.getIntAttrib(elm, MINIMAL_ATTRIB, true);
        }
      }
    }
    app.setRecoveryStrategy(strategy);
    /*
     * Check if the redundancy parameters should be overridden by the
     * ones defined as system properties.
     */
    String minimalStr = System.getProperty("minimal.redundancy");
    String initialStr = System.getProperty("initial.redundancy");
    if (minimalStr != null && minimalStr.length() > 0 && !minimalStr.equals("default")) {
      minimal = Integer.parseInt(minimalStr);
    }
    if (initialStr != null && initialStr.length() > 0 && !initialStr.equals("default")) {
      initial = Integer.parseInt(initialStr);
    }
    app.setRedundancy(minimal, initial);
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Application Info interface methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Returns the name associated with this application in the
   *  (dependable) registry.
   */
  public String getRegistryName()
  {
    return regName;
  }


  /**
   *  Returns a class data object for the application.
   *
   *  @return 
   *    the class data object for the application.
   */
  public ClassData getClassData()
  {
    return classData;
  }


  /**
   * Set the properties that should apply for this application.
   * If no properties map is provided, the properties is initialized
   * according to this applications properties.
   */
  public void setProperties(Map<String,String> properties)
  {
    if (properties == null) {
      /*
       * If the given properties map is null, we initialize a new
       * map with the properties of this application.
       */
      properties = new HashMap<String,String>(3);
    }
    if (!properties.containsKey("minimal.redundancy"))
      properties.put("minimal.redundancy", Integer.toString(getMinimalRedundancy()));
    if (!properties.containsKey("initial.redundancy"))
      properties.put("initial.redundancy", Integer.toString(getInitialRedundancy()));
    classData.setProperties(properties);
  }


  public int getGroupId()
  {
    return groupId;
  }


  /**
   *  Returns the class name as a string.
   *
   *  @return 
   *    Returns the class name as a string.
   */
  public String getClassName()
  {
    return classData.getName();
  }


  /**
   *  Returns the class arguments as a string array.
   *
   *  @return 
   *    Returns the class arguments as a string array.
   */
  public String[] getClassArgs()
  {
    return classData.getArgs();
  }


  /**
   *  Returns true if this application should be started within the
   *  same JVM as the execution daemon.  False indicates that the
   *  application should be started external the the execution daemon,
   *  that is in a separate JVM.
   */
  public boolean startInExecJVM()
  {
    return inExecJVM;
  }


  /**
   *  Returns the value associated with the parameter name provided
   *  to this method.
   *
   *  @param name
   *    The name of the parameter to retreive from the parameter map.
   *  @return
   *    The parameter value associated with the given parameter name.
   *  @throws NoSuchElementException
   *    Raised if the given parameter name was not found in the internal
   *    parameter map, or if there is no parameters associated with this
   *    application.
   */
  public String getParam(String name)
    throws NoSuchElementException
  {
    if (paramMap == null) {
      throw new NoSuchElementException("The application has no associated parameters");
    }
    String strval = (String) paramMap.get(name);
    if (strval == null) {
      throw new NoSuchElementException("No parameter named '" + name
        + "' found to be associated with application: " + regName);
    }
    return strval;
  }


  /**
   *  Returns the value associated with the parameter name provided
   *  to this method; if this application has no such parameter defined
   *  the provided default value will be returned.
   *
   *  @param name
   *    The name of the parameter to retreive from the parameter map.
   *  @param defaultVal
   *    The default value to return if the given parameter is not
   *    defined in the parameter map.
   *  @return
   *    The parameter value associated with the given parameter name.
   */
  public String getParam(String name, String defaultVal)
  {
    if (paramMap == null) {
      return defaultVal;
    }
    String strval = (String) paramMap.get(name);
    if (strval == null) {
      return defaultVal;
    }
    return strval;
  }


  /**
   *  Returns the integer value associated with the parameter name
   *  provided to this method.
   *
   *  @param name
   *    The name of the parameter to retreive from the parameter map.
   *  @return
   *    The <code>int</code> parameter value associated with the given
   *    parameter name.
   *  @throws NoSuchElementException
   *    Raised if the given parameter name was not found in the internal
   *    parameter map.
   */
  public int getIntParam(String name)
    throws NoSuchElementException
  {
    return Integer.valueOf(getParam(name)).intValue();
  }


  /**
   *  Returns the integer value associated with the parameter name
   *  provided to this method.  If this application has no associated
   *  value for this parameter, the provided default value is returned.
   *
   *  @param name
   *    The name of the parameter to retreive from the parameter map.
   *  @param defaultValue
   *    The default value of the parameter to be return if no value
   *    is available in the parameter map.
   *  @return
   *    The <code>int</code> parameter value associated with the given
   *    parameter name.  If the parameter have no specified value for
   *    the associated application, we return the provided default value.
   */
  public int getIntParam(String name, int defaultValue)
  {
    if (paramMap == null) {
      return defaultValue;
    }
    String strval = (String) paramMap.get(name);
    if (strval == null) {
      return defaultValue;
    }
    return Integer.valueOf(strval).intValue();
  }


  /**
   *  Returns the boolean value associated with the parameter name
   *  provided to this method.  If this application has no associated
   *  value for this parameter, false is returned.
   *
   *  @param name
   *    The name of the parameter to retreive from the parameter map.
   *  @return
   *    The <code>boolean</code> parameter value associated with the given
   *    parameter name.  If the parameter have no specified value for
   *    the associated application, false is returned.
   */
  public boolean getBooleanParam(String name)
  {
    if (paramMap == null) {
      return false;
    }
    String strval = (String) paramMap.get(name);
    if (strval == null) {
      return false;
    }
    return Boolean.valueOf(strval).booleanValue();
  }


  /**
   *  Retrieve the set of services used by this application.
   *
   *  @return 
   *    A <code>String[]</code> of services used by this application,
   *    ordered from the lowest layer at index position zero.
   *  @exception IllegalStateException
   *    Raised if the service set has not yet been initialized.
   */
  public String[] getServiceSet()
  {
    if (serviceSet == null) {
      throw new IllegalStateException("Service set not yet initialized.");
    }
    return (String[]) serviceSet.toArray(new String[serviceSet.size()]);
  }


  /**
   *  Returns true if this application contains the specified service,
   *  otherwise false is returned.
   *
   *  @return 
   *    True if this application contains the specified service.
   *  @exception IllegalStateException
   *    Raised if the service set has not yet been initialized.
   */
  public boolean hasService(String service)
  {
    if (serviceSet == null) {
      throw new IllegalStateException("Service set not yet initialized.");
    }
    return serviceSet.contains(service);
  }


  public RecoveryStrategy getRecoveryStrategy()
  {
    if (strategyClass == null)
      throw new IllegalStateException("RecoveryStrategy not set");
    if (recoveryStrategy == null) {
      try {
        initRecoveryStrategy();
      } catch (ConfigurationException e) {
        throw new IllegalStateException(e.getMessage());
      }
    }
    return recoveryStrategy;
  }


  /**
   *  Set the minimal and initial redundancy for this application.
   *
   *  @param initial  
   *    initial redundancy for this application
   *  @param minimal
   *    minimal redundancy for this application
   *  @exception ConfigurationException
   *    if the redundancy specifications are out of bound.
   */
  public void setRedundancy(int minimal, int initial)
    throws ConfigurationException
  {
    if (initial < minimal)
      throw new ConfigurationException("Initial redundancy is less than the minimal redundancy.");

    if (initial >= REDUNDANCY_LOWER_BOUND && initial <= REDUNDANCY_UPPER_BOUND)
      initialRedundancy = initial;
    else if (initial > REDUNDANCY_UPPER_BOUND)
      throw new ConfigurationException("Initial redundancy is too high.");
    else
      throw new ConfigurationException("Initial redundancy is too low.");

    if (minimal >= REDUNDANCY_LOWER_BOUND && minimal <= REDUNDANCY_UPPER_BOUND)
      minimalRedundancy = minimal;
    else if (minimal > REDUNDANCY_UPPER_BOUND)
      throw new ConfigurationException("Minimal redundancy is too high.");
    else
      throw new ConfigurationException("Minimal redundancy is too low.");
  }

  public int getInitialRedundancy()
  {
    if (initialRedundancy == -1) 
      throw new IllegalStateException("Inital redundancy is not set: " + this);
    return initialRedundancy;
  }


  public int getMinimalRedundancy()
  {
    if (minimalRedundancy == -1) 
      throw new IllegalStateException("Minimal redundancy is not set");
    return minimalRedundancy;
  }


  public int getCurrentGroupSize()
  {
    return view.size();
  }

  /**
   *  Returns true if this application needs recovery,
   *  otherwise false is returned.
   */
  public boolean needsRecovery()
  {
    RecoveryStrategy rs = getRecoveryStrategy();
    return rs.needsRecovery();
  }

  ////////////////////////////////////////////////////////////////////////////////////////////
  // MembershipListener interface methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Update this application group's view and set of view hosts.
   */
  public synchronized void viewChange(View view)
  {
    if (!this.view.equals(view)) {
      this.view = view;
      //FIXME before updating the viewHosts set, we should update the status of the replicas on these hosts
      // to something other than NORMAL/JOINING...
      viewHosts.removeAllHosts();
      MemberId[] members = view.getMembers();
      for (int i = 0; i < members.length; i++) {
        Host host = DistributedSystemConfig.getHost(members[i]);
        // Update the replica status on the host to normal
        host.viewChange(this);
        viewHosts.addHost(host);
      }
    }
  }

  /**
   * Returns the current view of this application, which may be
   * an empty view in case the application has not installed
   * any views yet.
   */
  public View getView()
  {
    return view;
  }

  /**
   * Returns the set of hosts on which replicas of this application
   * is known to be located.
   */
  public HostSet getViewHosts()
  {
    return viewHosts;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Initialization methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Sets the recovery strategy for this application.
   *
   *  @param strategy
   *    class name denoting the recovery strategy
   *  @exception ConfigurationException
   *    if the recovery strategy is not supported.
   */
  public void setRecoveryStrategy(String strategy)
    throws ConfigurationException
  {
    /*
     * Check if the given strategy use full package naming; if not, we
     * augment it with the default recovery strategy location.
     */
    strategy = (strategy.indexOf(".") != -1) ? strategy : "jgroup.arm.recovery." + strategy;
    try {
      strategyClass = Class.forName(strategy);
    } catch (ClassNotFoundException e) {
      throw new ConfigurationException("Unsupported recovery strategy specified: " + strategy);
    }
  }


  /**
   *  Set the service set or stack order for this application.  The
   *  layer stack order defines the layers that the application requires
   *  to be used, and the group manager will construct the required set
   *  of layers in the given order.
   *
   *  @param stackOrder
   *    The colon separate string of services that this application
   *    requires.
   *  @exception ConfigurationException
   *    Thrown if the string contains an unspecified service.
   */
  public void setServiceSet(String stackOrder)
    throws ConfigurationException
  {
    String[] stack = stackOrder.split(":");
    for (int i = 0; i < stack.length; i++) {
      addService(stack[i]);
    }
  }


  /**
   *  Add the given service to the protocol stack for this application.
   *  This method must be invoked in the order in which the layers should
   *  be organized in the stack.
   *
   *  @param serviceName
   *    The service name to add to the application.
   */
  public void addService(String serviceName)
    throws ConfigurationException
  {
    if (serviceSet == null) {
      serviceSet = new ArrayList<String>();
    }
    serviceSet.add(serviceName);
  }


  /**
   *  Add the given name-value pair to the internal parameter map.
   *
   *  @param name
   *    The name of the parameter to add to the map.
   *  @param value
   *    The value to associate with the parameter.
   *
   *  @throws ConfigurationException
   *    Raised if the given parameter name already exists within the
   *    internal parameter map.
   */
  public void addParam(String name, String value)
    throws ConfigurationException
  {
    if (paramMap == null) {
      paramMap = new HashMap<String,String>();
    } else {
      if (paramMap.containsKey(name)) {
        throw new ConfigurationException("Multiply defined parameter name: " + name);
      }
    }
    paramMap.put(name, value);
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Private methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Initialize the recovery strategy for this application.  The
   *  recovery strategy should not be instantiated on other host than
   *  the ones running the replication manager replicas.  That is, since
   *  we expect the recovery strategy to make use of the local
   *  replication manager in order to perform recovery.  Creating an
   *  instance on all hosts parsing the <code>application.xml</code>
   *  would thus result in an instance of the replication manager on
   *  each host as well.  Therefore, we only init the recovery strategy
   *  upon invocations of the getRecoveryStrategy() method, typically
   *  only invoked from within the replication manager (Correlator).
   *
   *  @exception ConfigurationException
   *    If the recovery strategy associated with this application
   *    could not be initialized.
   */
  private void initRecoveryStrategy()
    throws ConfigurationException
  {
    if (dScheme == null) {
      throw new IllegalStateException("Cannot initialize recovery strategy "
          + "without providing a distribution scheme.");
    }
    try {
      /*
       * Check if the recovery strategy used by this application has
       * already been initialized; if so, we just return now.
       */
      if (recoveryStrategy != null)
        return;

      /*
       * A no-argument constructor must be provided by a class
       * implementing a recovery strategy.  This pattern cannot be
       * enforced by the RecoveryStratgy interface, thus we must check
       * this at runtime instead.
       */
      Constructor<?> c = strategyClass.getConstructor( new Class[] {} );
      recoveryStrategy = (RecoveryStrategy) c.newInstance( new Object[] {} );
      /* Initialize the recovery strategy. */
      recoveryStrategy.initialize(dScheme, this);

    } catch (NoSuchMethodException e) {
      throw new ConfigurationException(strategyClass + " does not implement a non-argument constructor.");
    } catch (IllegalArgumentException e) {
      throw new ConfigurationException(strategyClass + " does not implement a non-argument constructor.");
    } catch (IllegalAccessException e) {
      throw new ConfigurationException(strategyClass + " does not provide an accessible constructor.");
    } catch (InstantiationException e) {
      throw new ConfigurationException(strategyClass + " represents an abstract class.");
    } catch (InvocationTargetException e) {
      throw new ConfigurationException(strategyClass + " has thrown an exception.", e.getTargetException());
    }
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from Object
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Returns a hash code value for this application info object.
   */
  public int hashCode()
  {
    return regName.hashCode() ^ classData.hashCode() ^ groupId;
  }


  /**
   *  Returns true if the provided object is equal to this object.
   */
  public boolean equals(Object obj)
  {
    if (!(obj instanceof AppConfig)) {
      return false;
    }
    AppConfig app = (AppConfig) obj;
    return (regName.equals(app.regName) &&
            classData.equals(app.classData) &&
            groupId == app.groupId);
  }


  /**
   *  Returns a string representation of this object
   */
  public String toString() 
  {
    return toString(false);
  }


  /**
   * Returns a string representation of this object.  If the 
   * <code>full</code> parameter is false, only a short version
   * is returned, otherwise a full version is returned.
   */
  public String toString(boolean full)
  {
    StringBuilder buf = new StringBuilder();
    if (full) {
      buf.append("[AppConfig: ");
      buf.append(regName);
      buf.append(", gid=");
      buf.append(groupId);
      buf.append(": ");
    }
    buf.append(classData.getShortName());
    buf.append(": ");
    if (full) {
      String[] srSet = getServiceSet();
      for (int i = 0; i < srSet.length; i++) {
        buf.append((i != 0) ? ":" : "");
        buf.append(srSet[i]);
      }
      buf.append(";  ");
      if (strategyClass != null)
        buf.append(strategyClass.getName());
      buf.append(", ");
    }
    buf.append("{");
    buf.append(initialRedundancy);
    buf.append(", ");
    buf.append(minimalRedundancy);
    buf.append("}, viewHosts=");
    buf.append(viewHosts);
    if (full)
      buf.append("]");
    return buf.toString();
  }

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from Externalizable
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Default constructor for externalization and configuration parser construction.
   */
  public AppConfig() { }


  /**
   *  @see java.io.Externalizable#readExternal(ObjectInput)
   */
  @SuppressWarnings("unchecked")
  public void readExternal(ObjectInput in)
    throws IOException, ClassNotFoundException
  {
    regName = in.readUTF();
    groupId = in.readInt();
    classData = (ClassData) in.readObject();
    serviceSet = (List) in.readObject();
    paramMap = (Map) in.readObject();
    inExecJVM = in.readBoolean();
    strategyClass = (Class) in.readObject();
    initialRedundancy = in.readInt();
    minimalRedundancy = in.readInt();
    view = new ViewImpl();
    view.readExternal(in);
    viewChange(view);
    if (log.isDebugEnabled())
      log.debug("Received AppConfig object: " + toString());
    /*
     * Replace the application in the local maps with the application
     * being unmarshalled, so as to update the various parameters at
     * other hosts.
     */
    appMap.put(regName, this);
    appClazzMap.put(classData.getName(), this);
    appGroupMap.put(new Integer(groupId), this);
  }


  /**
   *  @see java.io.Externalizable#writeExternal(ObjectOutput)
   */
  public void writeExternal(ObjectOutput out)
    throws IOException
  {
    out.writeUTF(regName);
    out.writeInt(groupId);
    out.writeObject(classData);
    out.writeObject(serviceSet);
    out.writeObject(paramMap);
    out.writeBoolean(inExecJVM);
    out.writeObject(strategyClass);
    out.writeInt(initialRedundancy);
    out.writeInt(minimalRedundancy);
    view.writeExternal(out);
    if (log.isDebugEnabled())
      log.debug("Sending AppConfig object: " + toString());
  }

} // END AppConfig
