/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.relacs.config;

import java.io.Externalizable;
import java.io.IOException;
import java.io.ObjectInput;
import java.io.ObjectOutput;
import java.net.InetAddress;
import java.net.UnknownHostException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.NoSuchElementException;

import jgroup.core.MemberId;
import jgroup.util.Abort;
import jgroup.util.Network;

/**
 *  Collection class for storing information about a set of hosts.
 *
 *  @author Hein Meling
 *  @since Jgroup 1.2
 */
public final class HostSet
  implements Cloneable, Externalizable, Iterable<Host>
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constants
  ////////////////////////////////////////////////////////////////////////////////////////////

  private static final long serialVersionUID = 2517425024237000760L;

  /**
   *  Default initial capacity.  If we manage to stay below this value
   *  we avoid to rehash the set.  @See java.util.HashMap for details.
   */
  private static final int HOST_HASH_SIZE = 50;

  /**
   *  Default load factor.
   */
  private static final float LOAD_FACTOR = (float) 0.75;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * The unsyncronized map of <code>Host</code> objects.
   * This is needed for cloning purposes.
   */
  private Map<InetAddress, Host> unsyncHosts;

  /** Synchronized map of <code>Host</code> objects (InetAddress -> Host object)*/
  private Map<InetAddress, Host> hosts;

  /**
   *  Dynamic array of host listeners.  Note that this field cannot be
   *  declared as just the List interface due to this class implementing
   *  the Cloneable interface.
   */
  private ArrayList<HostListener> hostListeners = new ArrayList<HostListener>();


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructors
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Constructs an empty <code>HostSet</code> with default capacity and
   *  load factor.  Also the default constructor for externalization.
   */
  public HostSet()
  {
    this(HOST_HASH_SIZE, LOAD_FACTOR);
  }


  /**
   *  Constructs an empty <code>HostSet</code> with default load factor,
   *  with space for <code>maxEntries</code> elements.
   */
  public HostSet(int maxEntries)
  {
    this(maxEntries, LOAD_FACTOR);
  }


  /**
   *  Constructs an empty <code>HostSet</code> with space for
   *  <code>maxEntries</code> elements, using the specified load factor.
   */
  public HostSet(int maxEntries, float loadFactor)
  {
    int initialCapacity = Math.round(maxEntries / loadFactor) + 1;
    unsyncHosts = new HashMap<InetAddress, Host>(initialCapacity, loadFactor);
    hosts = Collections.synchronizedMap(unsyncHosts);
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Listener interface
  ////////////////////////////////////////////////////////////////////////////////////////////

  public void addHostListener(HostListener listener)
  {
    hostListeners.add(listener);
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // HostSet interface methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Returns the <code>Host</code> object associated with the specified
   *  <code>InetAddress</code>.
   *
   *  @param inetAddress
   *    The <code>InetAddress</code> of the <code>Host</code> object to get.
   *  @return 
   *    The <code>Host</code> object associated with the given
   *    <code>InetAddress</code>.  <code>null</code> is returned if the given
   *    IP address host is not in this host set.
   */
  public Host getHost(InetAddress inetAddress)
  {
    return (Host) hosts.get(inetAddress);
  }


  /**
   *  Returns the <code>Host</code> object associated with the specified
   *  <code>MemberId</code>.
   *
   *  @param member
   *    The <code>MemberId</code> of the <code>Host</code> object to get.
   *  @return 
   *    The <code>Host</code> object associated with the given
   *    <code>MemberId</code>.  <code>null</code> is returned if the given
   *    member host is not in this host set.
   */
  public Host getHost(MemberId member)
  {
    return (Host) hosts.get(member.getEndPoint().getAddress());
  }


  /**
   *  Add a new host to this <code>HostSet</code>.
   *
   *  @param hostName
   *    The host name.
   *  @param domain
   *    The <code>Domain</code> associated with this host.
   *  @param port
   *    The port number for this host.
   *  @return 
   *    True if the <code>HostSet</code> did not already contain this host.
   *  @throws UnknownHostException
   *    Is raised if the host was not recognized.
   */
  public boolean addHost(String hostName, Domain domain, int port)
    throws UnknownHostException
  {
    return addHost(new Host(hostName, domain, port));
  }


  /**
   *  Add a new host to this <code>HostSet</code>.
   *
   *  @param host
   *    A host object.
   *  @return 
   *    True if the <code>HostSet</code> did not already contain this host.
   */
  public boolean addHost(Host host)
  {
    Host oldHost = (Host) hosts.put(host.getAddress(), host);
    boolean added = !host.equals(oldHost);
    if (added)
      for (int i = 0, size = hostListeners.size(); i < size; i++)
        ((HostListener) hostListeners.get(i)).addHost(host);
    return added;
  }


  /**
   *  Add the given host to the this host set.  The given host
   *  is represented as an integer, and must already exist as a
   *  <code>Host</code> object in the <code>DistributedSystemConfig</code>
   *  object.
   *
   *  @param hostIntAdr
   *    An <code>int</code> representation of the host to add to this
   *    host set.
   *  @return
   *    True if the <code>HostSet</code> did not already contain this host.
   *  @throws UnknownHostException
   *    If the <code>int</code> representation of the host does not
   *    correspond to a valid host address, or if it could not
   *    be found.
   */
  public boolean addHost(int hostIntAdr)
    throws UnknownHostException
  {
    InetAddress inetAdr = Network.translate(hostIntAdr);
    Host host = DistributedSystemConfig.getHost(inetAdr);
    return addHost(host);
  }


  /**
   *  Add all hosts in the given <code>HostSet</code> to this
   *  <code>HostSet</code>.
   *
   *  @param hostSet
   *    The set of hosts to add to this host set.
   */
  public void addHosts(HostSet hostSet)
  {
    for (Iterator iter = hostSet.iterator(); iter.hasNext();) {
      Host host = (Host) iter.next();
      addHost(host);
    }
  }


  /**
   *  Remove a host (member) from this <code>HostSet</code>.
   *
   *  @param member
   *    A member object representing a host object.
   *  @return 
   *    True if the <code>Host</code> was removed from the
   *    <code>HostSet</code>, that is if it was present in the
   *    <code>HostSet</code>; false otherwise.
   */
  public boolean removeHost(MemberId member)
  {
    Host h = getHost(member);
    if (h != null)
      return removeHost(h);
    else
      return false;
  }


  /**
   *  Remove a host from this <code>HostSet</code>.
   *
   *  @param host
   *    A host object.
   *  @return 
   *    True if the <code>Host</code> was removed from the
   *    <code>HostSet</code>, that is if it was present in the
   *    <code>HostSet</code>; false otherwise.
   */
  public boolean removeHost(Host host)
  {
    Host oldHost = (Host) hosts.remove(host.getAddress());
    boolean removed = (oldHost != null);
    if (removed)
      for (int i = 0, size = hostListeners.size(); i < size; i++)
        ((HostListener) hostListeners.get(i)).removeHost(host);
    return removed;
  }


  /**
   *  Remove all hosts in the given <code>HostSet</code> from this
   *  <code>HostSet</code>.
   *
   *  @param hostSet
   *    The set of hosts to remove from this host set.
   */
  public void removeHosts(HostSet hostSet)
  {
    if (hostSet == this) {
      removeAllHosts();
      return;
    }
    for (Iterator iter = hostSet.iterator(); iter.hasNext();) {
      Host host = (Host) iter.next();
      removeHost(host);
    }
  }


  /**
   *  Remove all hosts in this <code>HostSet</code>.
   */
  public void removeAllHosts()
  {
    hosts.clear();
    hostListeners.clear();
  }


  /**
   *  Check if the given host set is fully contained in
   *  this <code>HostSet</code>.
   *
   *  @param hostSet
   *    A host set.
   *  @return 
   *    True if the given <code>HostSet</code> is contained in
   *    this host set.
   */
  public boolean containsAll(HostSet hostSet)
  {
    for (Iterator iter = hostSet.iterator(); iter.hasNext();) {
      Host host = (Host) iter.next();
      if (!hosts.containsKey(host.getAddress())) {
        return false;
      }
    }
    return true;
  }


  /**
   *  Check if the host is in this <code>HostSet</code>.
   *
   *  @param host
   *    A host object.
   *  @return 
   *    True if the <code>HostSet</code> contained the specified host.
   */
  public boolean containsHost(Host host)
  {
    return hosts.containsKey(host.getAddress());
  }


  /**
   *  Check if the host represented by the given InetAddress is in this
   *  <code>HostSet</code>.
   *
   *  @param inetAddress
   *    The <code>InetAddress</code> representation of the host to
   *    test for presence in the <code>HostSet</code>.
   *  @return 
   *    True if the <code>HostSet</code> contained the specified host.
   */
  public boolean containsHost(InetAddress inetAddress)
  {
    return hosts.containsKey(inetAddress);
  }


  /**
   *  Returns the number of hosts stored in this <code>HostSet</code>.
   */
  public int size()
  {
    return hosts.size();
  }


  /**
   *  Returns true if this <code>HostSet</code> is empty.
   */
  public boolean isEmpty()
  {
    return hosts.isEmpty();
  }


  /**
   *  Returns an iterator over the values of the <code>HostSet</code>.
   *  The values returned by the <code>next</code> method call of the
   *  iterator should be cast to <code>Host</code>.
   *
   *  @return 
   *    An <code>Iterator</code> for the <code>HostSet</code>.
   */
  public Iterator<Host> iterator()
  {
    return hosts.values().iterator();
  }


  /**
   *  Returns the set of hosts that are considered available; i.e. running
   *  an <code>ExecService</code> instance.
   */
  public HostSet getAvailHosts()
  {
    //FIXME since this method may block due to machines being unavail;
    // it should be wrapped in a thread with a timeout mechanism. TimerTask stuff.
    HostSet availableHost = (HostSet) this.clone();
    for (Iterator iter = availableHost.iterator(); iter.hasNext();) {
      Host host = (Host) iter.next();
      if (!host.isAvailable()) {
        iter.remove();
      }
    }
    return availableHost;
  }


  public Host[] toArray()
  {
    return (Host[]) hosts.values().toArray(new Host[hosts.size()]);
  }


  /**
   *  Returns an <code>int[]</code> array representing the content of
   *  this <code>HostSet</code>.
   */
  public int[] toIntArray()
  {
    int i = 0;
    int[] inetAdr = new int[hosts.size()];
    synchronized (hosts) {
      for (Iterator assignedIter = hosts.values().iterator(); assignedIter.hasNext();) {
        Host host = (Host) assignedIter.next();
        inetAdr[i++] = Network.translate(host.getAddress());
      }
    }
    return inetAdr;
  }


  /**
   *  Returns the first <code>Host</code> returned by the iterators next
   *  method.  There is no guarantees that this method returns the same
   *  host for multiple invocations, nor does it promise to return a
   *  different host on the next invocation.  This method is useful to
   *  obtain just one host in the <code>HostSet</code>, when it does not
   *  matter which host to get.
   *
   *  @return 
   *    The <code>Host</code> object found.
   *  @throws NoSuchElementException
   *    Raised if there are no hosts in this <code>HostSet</code>.
   */
  public Host getFirst()
  {
    Iterator<Host> iter = iterator();
    if (!iter.hasNext())
      throw new NoSuchElementException("No hosts in the HostSet");
    return iter.next();
  }


  /**
   *  Removes the first <code>Host</code> returned by the iterators next
   *  method.  There is no guarantees that this method returns the same
   *  host for multiple invocations.  This method is useful to
   *  obtain just one host in the <code>HostSet</code>, when it does not
   *  matter which host to get.
   *
   *  @return 
   *    The <code>Host</code> object found.
   *  @throws NoSuchElementException
   *    Raised if there are no hosts in this <code>HostSet</code>.
   */
  public synchronized Host removeFirst()
  {
    Iterator<Host> iter = iterator();
    if (!iter.hasNext())
      throw new NoSuchElementException("No hosts in the HostSet");
    Host host = iter.next();
    iter.remove();
    return host;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Cloneable interface
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Returns a shallow copy of this <code>HostSet</code> instance.
   *  (The elements themselves are not copied.)
   *
   *  @return 
   *    a clone of this <code>HostSet</code> instance.
   */
  @SuppressWarnings("unchecked")
  public Object clone()
  {
    try {
      HostSet newHostSet = (HostSet) super.clone();
      newHostSet.unsyncHosts = (HashMap) ((HashMap) unsyncHosts).clone();
      newHostSet.hosts = Collections.synchronizedMap(newHostSet.unsyncHosts);
      newHostSet.hostListeners = (ArrayList) hostListeners.clone();
      return newHostSet;
    } catch(CloneNotSupportedException e) {
      /* Cannot happen since HashMap supports clone. */
      throw new Abort("Internal error", e);
    }
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Override methods in Object
  ////////////////////////////////////////////////////////////////////////////////////////////

  public boolean equals(Object o)
  {
    if (o != null && o instanceof HostSet) {
      HostSet hostSet = (HostSet) o;
      boolean eq = (size() == hostSet.size());
      // iterator over this host set and compare with the provided host set
      for (Iterator iter = iterator(); eq && iter.hasNext();) {
        Host host = (Host) iter.next();
        eq = hostSet.containsHost(host);
      }
      return eq;
    } else {
      return false;
    }
  }


  /**
   *  Return a hash code for this host set.
   */
  public int hashCode()
  {
    return hosts.hashCode();
  }


  /**
   *  Returns a string representation of this object
   */
  public String toString()
  {
    StringBuilder buf = new StringBuilder();
    buf.append("[HostSet: ");
    boolean first = true;
    synchronized (hosts) {
      for (Iterator i = iterator(); i.hasNext(); ) {
        if (first)
          first = false;
        else
          buf.append(", ");
        buf.append(((Host)i.next()).toString());
      }
    }
    buf.append("]");
    return buf.toString();
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  //  Marshaling/unmarshaling methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /* (non-Javadoc)
   * @see java.io.Externalizable#writeExternal(java.io.ObjectOutput)
   */
  public void writeExternal(ObjectOutput out)
    throws IOException
  {
//    int[] hosts = toIntArray();
//    out.writeInt(hosts.length);
//    for (int i = 0; i < hosts.length; i++) {
//      out.writeInt(hosts[i]);
//    }

    out.writeInt(size());
    for (Iterator iter = this.hosts.values().iterator(); iter.hasNext();) {
      Host host = (Host) iter.next();
      out.writeInt(host.getIntAddress());
      out.writeObject(host.getContentMap());
    }
  }


  /* (non-Javadoc)
   * @see java.io.Externalizable#readExternal(java.io.ObjectInput)
   */
  public void readExternal(ObjectInput in)
    throws IOException, ClassNotFoundException
  {
    int len = in.readInt();
    for (int k = 0; k < len; k++) {
      int hostadr = in.readInt();
      try {
        addHost(hostadr);
      } catch (UnknownHostException e) {
        e.printStackTrace();
      }
      Host host = getHost(Network.translate(hostadr));
      host.setContentMap((Map) in.readObject());
    }
  }

} // END HostSet
