/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.relacs.config;

import jgroup.core.ConfigurationException;

/**
 *  The <code>TransportConfig</code> class contains the configuration
 *  parameters needed by the underlying transport protocol used in
 *  Relacs.
 *
 *  @author Alberto Montresor
 *  @author Hein Meling
 *  @author Marcin Solarski
 *  @since Jgroup 0.1
 */
final public class TransportConfig
  implements ConfigurationObject
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Tags
  ////////////////////////////////////////////////////////////////////////////////////////////

  private static final String dsPayloadAttrib             = "payload";
  private static final String dsMulticastTTLAttrib        = "multicastTTL";
  private static final String dsAlfaNAttrib               = "alfaN";
  private static final String dsAlfaDAttrib               = "alfaD";
  private static final String dsRoutingTimeoutAttrib      = "routingTimeout";
  private static final String dsCongestionTimeoutAttrib   = "congestionTimeout";
  private static final String dsStartingWindowAttrib      = "startingWindow";
  private static final String dsMaxWindowAttrib           = "maxWindow";
  private static final String dsMinWindowAttrib           = "minWindow";
  private static final String dsMaxTTLAttrib              = "maxTTL";
  private static final String dsTTLWarningAttrib          = "TTLWarning";
  private static final String dsMaxPathLengthAttrib       = "maxPathLength";
  private static final String dsPathLengthWarningAttrib   = "pathLengthWarning";
  private static final String dsLocalPingTimeoutAttrib    = "localPingTimeout";
  private static final String dsdaemonLookupRetryNoAttrib    = "daemonLookupRetryNo";
  private static final String dsdaemonLookupRetryDelayAttrib = "daemonLookupRetryDelay";


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constants
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Default payload */
  private static final int PAYLOAD = 1024;

  /** Default multicast TTL */
  private static final int MULTICAST_TTL = 2;

  /** Default alfan */
  private static final int ALFAN = 7;

  /** Default alfad */
  private static final int ALFAD = 8;

  /** Default routing timeout */
  private static final int ROUTING_TIMEOUT = 1000;

  /** Default routing timeout */
  private static final int CONGESTION_TIMEOUT = 1000;

  /** Default initial size of the window */
  private static final int STARTING_WINDOW = 100;

  /** Default max size of the window */
  private static final int MAX_WINDOW = 10;

  /** Default min size of the window */
  private static final int MIN_WINDOW = 1;

  /** Default max TTL (reachability information) */
  private static final int MAX_TTL = 6;

  /** Default TTL warning level (reachability information) */
  private static final int TTL_WARNING = MAX_TTL / 2;

  /** Default delta infinite (routing constants) */
  private static final int MAX_PATH_LENGTH = 6;

  /** Default delta warning level (routing constants) */
  private static final int PATH_LENGTH_WARNING = MAX_PATH_LENGTH / 2;

  /** Default local ping timeout */
  private static final int LOCAL_PING_TIMEOUT = 2000;

 /** The number of retries to find another local daemon. */
  private static final int RETRIES = 5;

  /** The default  delay (in milliseconds) between retries. */
  private static final int RETRY_DELAY = 2000;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Payload length of a mss message */
  private int payload;

  /** multicast TTL  */
  private int multicastTTL;

  /** Failure detector parameter */
  private int alfan;

  /** Failure detector parameter */
  private int alfad; 

  /** Routing timeout */
  private int routingTimeout;

  /** Congestion timeout */
  private int congestionTimeout;

  /** Initial size of the window */
  private int startingWindow;

  /** Maximum size of the window */
  private int maxWindow;

  /** Maximum size of the window */
  private int minWindow;

  /** Maximum TTL (reachability information) */
  private int maxTTL;

  /** TTL warning level (reachability information) */
  private int ttlWarning;

  /** Delta infinite (routing constants) */
  private int maxPathLength;

  /** Delta warning level (routing constants) */
  private int pathLengthWarning;

  /** Local ping timeout */
  private int localPingTimeout;
  
  /** Number of retries */
  private int retryNo;

  /** Delay between the retries */
  private int retryDelay;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructors
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Constructs a <code> TransportConfig </code> object using the
   *  default values.
   */
  public TransportConfig()
  {
    this(PAYLOAD, MULTICAST_TTL, ALFAN, ALFAD, ROUTING_TIMEOUT, CONGESTION_TIMEOUT,
      STARTING_WINDOW, MAX_WINDOW, MIN_WINDOW, MAX_TTL, TTL_WARNING,
      MAX_PATH_LENGTH, PATH_LENGTH_WARNING, LOCAL_PING_TIMEOUT, RETRIES, RETRY_DELAY);
  }


  /**
   *  Constructs a <code> TransportConfig </code> object from the given
   *  parameters.
   */
  public TransportConfig(int payload, int multicastTTL, int alfan, int alfad, int congestionTimeout,
    int routingTimeout, int startingWindow, int maxWindow, int minWindow, int maxTTL,
    int ttlWarning, int maxPathLength, int pathLengthWarning, int localPingTimeout, int retryNo, int retryDelay)
  {
    this.payload           = payload;
    this.multicastTTL      = multicastTTL;
    this.alfan             = alfan;
    this.alfad             = alfad;
    this.congestionTimeout = congestionTimeout;
    this.routingTimeout    = routingTimeout;
    this.startingWindow    = startingWindow;
    this.maxWindow         = maxWindow;
    this.minWindow         = minWindow;
    this.maxTTL            = maxTTL;
    this.ttlWarning        = ttlWarning;
    this.maxPathLength     = maxPathLength;
    this.pathLengthWarning = pathLengthWarning;
    this.localPingTimeout  = localPingTimeout;
    this.retryNo = retryNo;
    this.retryDelay = retryDelay;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from ConfigurationObject
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Parses the specified element.
   */
  public void parse(org.w3c.dom.Element elm)
    throws ConfigurationException
  {
    payload            = ConfigParser.getIntAttrib(elm, dsPayloadAttrib, false);
    multicastTTL       = ConfigParser.getIntAttrib(elm, dsMulticastTTLAttrib, false);
    alfan              = ConfigParser.getIntAttrib(elm, dsAlfaNAttrib, false);
    alfad              = ConfigParser.getIntAttrib(elm, dsAlfaDAttrib, false);
    routingTimeout     = ConfigParser.getIntAttrib(elm, dsRoutingTimeoutAttrib, false);
    congestionTimeout  = ConfigParser.getIntAttrib(elm, dsCongestionTimeoutAttrib, false);
    startingWindow     = ConfigParser.getIntAttrib(elm, dsStartingWindowAttrib, false);
    maxWindow          = ConfigParser.getIntAttrib(elm, dsMaxWindowAttrib, false);
    minWindow          = ConfigParser.getIntAttrib(elm, dsMinWindowAttrib, false);
    maxTTL             = ConfigParser.getIntAttrib(elm, dsMaxTTLAttrib, false);
    ttlWarning         = ConfigParser.getIntAttrib(elm, dsTTLWarningAttrib, false);
    maxPathLength      = ConfigParser.getIntAttrib(elm, dsMaxPathLengthAttrib, false);
    pathLengthWarning  = ConfigParser.getIntAttrib(elm, dsPathLengthWarningAttrib, false);
    localPingTimeout   = ConfigParser.getIntAttrib(elm, dsLocalPingTimeoutAttrib, false);
    retryNo   = ConfigParser.getIntAttrib(elm, dsdaemonLookupRetryNoAttrib, false);
    retryDelay   = ConfigParser.getIntAttrib(elm, dsdaemonLookupRetryDelayAttrib, false);
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Getter methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Get the size of the payload field of application packets.
   */
  public int getPayload()
  {
    return payload;
  }


  /**
   *  Get the value of the multicast TTL used to send the packets over the 
   *  multicast sockets. The TTL determines the administrative
   *  scope of the ip multicast used for group member communication. It can have
   *  value in range from 1 (scope of LAN) to 16 (world-wide scope). Use this 
   *  param with care. The default value is set to 2 which allows for deploying
   *  group memebers in adjacent networks (connected to the same router).
   *
   *  @since Jgroup 2.1
   */
  public int getMulticastTTL()
  {
    return multicastTTL;
  }


  /**
   *  Returns the <code>alfan</code> parameter, which is used for
   *  computing the average round-trip time (RTT), and it determines the
   *  weight to give to the old values.  The formula used is given in
   *  {@link jgroup.relacs.mss.MssHost#updateTimeout updateTimeout}.
   *
   *  @see jgroup.relacs.mss.MssHost#updateTimeout
   */
  public int getAlfan()
  {
    return alfan;
  }


  /**
   *  Returns the <code>alfan</code> parameter, which is used for
   *  computing the average round-trip time (RTT), and it determines the
   *  weight to give to the old values.  The formula used is given in
   *  {@link jgroup.relacs.mss.MssHost#updateTimeout updateTimeout}.
   *
   *  @see jgroup.relacs.mss.MssHost#updateTimeout
   */
  public int getAlfad()
  {
    return alfad;
  }


  /**
   *  Get the time interval between two routing messages sent by a
   *  daemon.
   */
  public int getRoutingTimeout()
  {
    return routingTimeout;
  }


  /**
   *  Set the time interval between two routing messages sent by a
   *  daemon.
   */
  public void setRoutingTimeout(int rt)
  {
    routingTimeout = rt;
  }


  /**
   *  Get the time interval between two routing messages sent by a
   *  daemon.
   */
  public int getCongestionTimeout()
  {
    return congestionTimeout;
  }


  /**
   *  Get the timeout value used to by members to suspect the daemon.
   *  This value is related to the ping frequency of the members.
   *  It is a factor two longer than the local ping frequency.
   *
   *  @see #getLocalPingRate
   */
  public int getDaemonSuspectTimeout()
  {
    return localPingTimeout * 2;
  }


  /**
   *  Get the time interval between two ping messages sent by the
   *  daemon to its associated members.
   */
  public int getLocalPingRate()
  {
    return localPingTimeout;
  }


  /**
   *  Get the size of the starting window
   */
  public int getStartingWindow()
  {
    return startingWindow;
  }


  /**
   *  Get the maximum size of the window
   */
  public int getMaxWindow()
  {
    return maxWindow;
  }


  /**
   *  Get the minimum size of the window
   */
  public int getMinWindow()
  {
    return minWindow;
  }


  /**
   *  Returns the reachability threshold used to evaluate the
   *  reachability of hosts within a cluster for the purpose of
   *  selecting a cluster leader.  Currently, the threshold is set to
   *  half the <code>maxTTL</code> value specified in the XML
   *  configuration.  Note that this value should never exceed the
   *  <code>maxTTL</code> value.
   */
  public int getReachabilityThreshold()
  {
    return (maxTTL / 2) + 1;
  }


  /**
   *  Get the maximum TTL value acceptable before considering an
   *  endpoint as unreachable.
   */
  public int getMaxTTL()
  {
    return maxTTL;
  }


  /**
   *  Get the TTL warning level.
   */
  public int getTTLWarning()
  {
    return ttlWarning;
  }


  /**
   *  Get the max path length.
   */
  public int getMaxPathLength()
  {
    return maxPathLength;
  }


  /**
   *  Get the path length warning level.
   */
  public int getPathWarning()
  {
    return pathLengthWarning;
  }

  /**
   *  Get the number of retries of checking the reachability of other 
   *  Jgroup Daemons.
   *
   *  An example working value for a LAN environment with Intel Pentium II 333Mhz 
   *  amounts to 6.
   *
   *  @since Jgroup 2.1
   */
  public int getRetryNo() 
  {
    return retryNo;
  }
  
  /**
   *  Get the delay [in ms] between the Jgroup daemon reachibility retries. 
   *  This parameter allows for tuning the speed of coming to a conclusion
   *  on suspecting other Jgroup daemons to be down. The optimal value of this 
   *  parameter strongly depends on the CPU power of the hosts where the 
   *  Jgroup daemons (and group memebers) run as well as the network 
   *  characteristics of the links between them (max return trip time). 
   *  An example working value for a LAN environment with Intel Pentium II 333Mhz 
   *  amounts to 300 ms.
   *
   *  @since Jgroup 2.1
   */  	
   public int getRetryDelay() 
   {
     return retryDelay;	
   }

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Override methods in Object
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Returns a string representation of this object
   */
  public String toString() {
    StringBuilder buf = new StringBuilder();
    buf.append("[p=");
    buf.append(payload);
    buf.append(", mTTL=");
    buf.append(multicastTTL);
    buf.append(", an=");
    buf.append(alfan);
    buf.append(", ad=");
    buf.append(alfad);
    buf.append(", rt=");
    buf.append(routingTimeout);
    buf.append(", ct=");
    buf.append(congestionTimeout);
    buf.append(", sw=");
    buf.append(startingWindow);
    buf.append(", maxw=");
    buf.append(maxWindow);
    buf.append(", minw=");
    buf.append(minWindow);
    buf.append(", maxttl=");
    buf.append(maxTTL);
    buf.append(", ttlw=");
    buf.append(ttlWarning);
    buf.append(", maxpathl=");
    buf.append(maxPathLength);
    buf.append(", pathw=");
    buf.append(pathLengthWarning);
    buf.append(", ping=");
    buf.append(localPingTimeout);
    buf.append("]");
    return buf.toString();
  }

} // END TransportConfig
