/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.relacs.daemon;

import java.io.IOException;
import java.rmi.Remote;
import java.rmi.RemoteException;
import java.rmi.server.ExportException;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;

import jgroup.core.ConfigManager;
import jgroup.core.ConfigurationException;
import jgroup.core.EndPoint;
import jgroup.core.JgroupException;
import jgroup.core.MemberId;
import jgroup.core.View;
import jgroup.core.registry.BootstrapRegistry;
import jgroup.relacs.config.DistributedSystemConfig;
import jgroup.relacs.config.TransportConfig;
import jgroup.relacs.events.DeliveryAck;
import jgroup.relacs.events.Event;
import jgroup.relacs.events.EventTags;
import jgroup.relacs.events.InstallAck;
import jgroup.relacs.events.InstallEvent;
import jgroup.relacs.events.JoinRequest;
import jgroup.relacs.events.LeaveRequest;
import jgroup.relacs.events.MulticastRequest;
import jgroup.relacs.events.PrepareAck;
import jgroup.relacs.mss.HostTable;
import jgroup.relacs.mss.Mss;
import jgroup.relacs.mss.MssConstants;
import jgroup.relacs.mss.MssHost;
import jgroup.relacs.mss.MssUser;
import jgroup.relacs.types.GroupId;
import jgroup.relacs.types.MemberIdImpl;
import jgroup.relacs.types.ViewId;
import jgroup.relacs.types.ViewImpl;
import jgroup.util.InMessage;
import jgroup.util.Network;
import jgroup.util.ThreadMonitor;
import jgroup.util.Util;
import net.jini.export.Exporter;
import net.jini.jeri.BasicILFactory;
import net.jini.jeri.BasicJeriExporter;
import net.jini.jeri.tcp.TcpServerEndpoint;

import org.apache.log4j.Logger;
import org.apache.log4j.MDC;
import org.apache.log4j.NDC;

/**
 *  The <code>Daemon</code> is the access interface for the group
 *  membership service and reliable multicast service.
 *
 *  @author Alberto Montresor
 *  @author Hein Meling
 *  @author Marcin Solarski
 *  @since Jgroup 0.1
 */
public final class Daemon
  implements DaemonService, MssUser, Tag, EventTags, MssConstants
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Logger
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Obtain logger for this class */
  private static final Logger log = Logger.getLogger(Daemon.class);


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Static fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Number of members created so far */
  private static int members = 0;

  /** Compose the daemon registry name including the daemon port of the local host */
  private static final String daemonRegName =
    DAEMON_NAME + DistributedSystemConfig.getLocalHost().getPort();


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Static methods for creating and obtaining a daemon reference
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Create a daemon object on the local host.
   *
   *  @param conf
   *    Transport configuration parameters
   *  @return 
   *    A daemon instance.
   *
   *  @exception IOException
   *    Raised if a daemon could not be created due to a problem with
   *    port initialization.  Another application (or daemon) may be
   *    using our socket port.
   *  @exception ConfigurationException
   *    Raised if a daemon could not be created due to a
   *    misconfiguration.
   */
  public static Daemon createDaemon(TransportConfig conf)
    throws ConfigurationException, IOException
  {
    if (log.isDebugEnabled())
      log.debug("Attempting to create a daemon on the localhost");
    /* Obtain distributed system configuration information */
    DistributedSystemConfig dsc = ConfigManager.getDistributedSystem();
    /* Try to create a daemon locally. */
    Daemon daemon = new Daemon(dsc, conf);
    if (log.isDebugEnabled())
      log.debug("Successfully created a daemon on localhost: " + daemon);
    return daemon;
  }


  public static void bindLocalDaemon(DaemonService daemonService)
    throws JgroupException
  {
    try {
      /*
       * The daemon was successfully created within the local JVM; we export it
       * and publish its reference in case another MembershipLayer attemps to
       * initialize a new daemon on the same or a different host/JVM.  This allows
       * the MembershipLayer to use the daemon created in this JVM, through RMI.
       */
      Exporter exporter =
        new BasicJeriExporter(TcpServerEndpoint.getInstance(0), new BasicILFactory());
      Remote daemonProxy = exporter.export(daemonService);
      if (log.isDebugEnabled())
        log.debug("DaemonService has been exported");
      BootstrapRegistry.rebind(daemonRegName, daemonProxy);
      if (log.isDebugEnabled())
        log.debug("DaemonService has been bound to the bootstrap registry");
    } catch (ExportException ee) {
      if (log.isDebugEnabled())
        log.warn("Failed to export DaemonService", ee);
      throw new JgroupException("Failed to export the Daemon", ee);
    } catch (IOException ioe) {
      if (log.isDebugEnabled())
        log.warn("Failed to bind the DaemonService", ioe);
      throw new JgroupException("Failed to bind the Daemon", ioe);
    }
  }


  /**
   *  Obtain a remote reference for a local daemon. <p>
   *
   *  This method can be invoked if we are unable to bind a daemon
   *  instance to the port specifed in the distributed system
   *  configuration.  This may occur on two occasions; (i) there is
   *  another daemon running on the local host, or (ii) some other
   *  application is using the specified port.  The former case may also
   *  occur before the other daemon has managed to register its
   *  reference in the bootstrap registry, thus we need to retry a few
   *  times before concluding that it is some other application.
   *
   *  @return
   *    The remote <code>DaemonService</code> object
   *
   *  @exception JgroupException
   *    Raised if no daemon could be found (e.g., not stored in the registry).
   */
  public static DaemonService lookupLocalDaemon()
    throws JgroupException
  {
    Exception ex = null;
    for (int retryCount = retries; retryCount > 0; retryCount--) {
      try {
        /* Check if there is another daemon available for this host. */
        DaemonService daemonService = (DaemonService) BootstrapRegistry.lookup(daemonRegName);
        if (log.isDebugEnabled())
          log.debug("Successfully obtained RMI-based daemon reference for localhost");
        ex = null;
        return daemonService;

      } catch (Exception e) {
        try {
          BootstrapRegistry.refreshRegistryStub();
        } catch (RemoteException e1) {
          log.warn("Problem connecting to the bootstrap registry.", e1);
        }
        ex = e;
        log.warn("Retry (" + (retries-retryCount) + "/" + retries
          + ") to fetch Daemon named " + daemonRegName + " failed");
        /*
         * This catches both NotBoundException and RemoteException. <p>
         *
         * This basically means that we failed to lookup the daemon in
         * the registry of the localhost, and we simply retry a few
         * times, giving it a little delay between each retry.  We sleep
         * for retryDelay milliseconds, to give other JVMs the
         * opportunity to register themselves in the bootstrap registry.
         */
        try {
          Thread.sleep(retryDelay);
        } catch (InterruptedException ie) {
          /* Do nothing */
        }

      }
    }
    String mesg = "Cannot obtain a reference for a remote daemon (on the localhost)";
    if (log.isDebugEnabled())
      log.error(mesg);
    throw new JgroupException(mesg, ex);
  }

  /**
   *  Lookup a local daemon without retrying.
   */
  public static DaemonService quickLookupLocalDaemon()
  {
    /* Check if there is another daemon available for this host. */
    try {
      DaemonService ds = (DaemonService) BootstrapRegistry.lookup(daemonRegName);
      if (log.isDebugEnabled())
        log.debug("Quick lookup successfully obtained RMI-based daemon reference for localhost");
      return ds;
    } catch (Exception e) {
      return null;
    }
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** The number of retries to find another local daemon. */
  private static int retries;

  /** The delay (in milliseconds) between retries. */
  private static int retryDelay;

  /** Table containing the groups handled by this daemon */
  private final GroupTable grouptable = new GroupTable();

  /** The set of members associated with this Jgroup daemon */
  private final Set<MemberId> localMembers = new HashSet<MemberId>(5);   

  /** Reference to this host */
  private final MssHost me;

  /** MSS layer used for communication */
  private final Mss mss;

  /** Number of suspect events */
  private int nsuspects;

  /** Reachable set of last suspect event */
  private EndPoint[] trset;

  /** New hosts in reachable set of last suspect event */
  private EndPoint[] nrset;

  /** New hosts in unreachable set of last suspect event */
  private EndPoint[] nuset;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructors
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Private daemon constructor.
   */
  private Daemon(DistributedSystemConfig ds, TransportConfig conf)
    throws ConfigurationException, IOException
  {
    //initialize retries and retryDelay
    retries = conf.getRetryNo();
    retryDelay = conf.getRetryDelay(); 

    // Initialize mss
    mss = new Mss(this, ds, conf);

    // Obtain information
    HostTable hosttable = mss.getHostTable();
    me = HostTable.getLocalHost();

    // Initialize suspect information
    nsuspects = 0;
    trset = new EndPoint[] { me.getEndPoint() };
    nrset = new EndPoint[0];
    nuset = new EndPoint[0];
    
    // Initialize symmetry set information
    for (int i=0; i < hosttable.size(); i++) {
      MssHost scan = hosttable.get(i);
      EndPoint[] symset = new EndPoint[] { scan.getEndPoint() };
      grouptable.setSymset(scan.getEndPoint(), symset);
    }
    
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Communication with members (DaemonService interface)
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Returns a new member id for a new member. The new member id is obtained
   *  by concatenating the endpoint of this daemon, the incarnation number
   *  used to distinguish different incarnation of the same endpoint, and
   *  a sequence number used to distinguish different members.
   */
  public MemberId getMemberId(EndPoint localEndPoint)
    throws RemoteException
  {
    MemberId member = new MemberIdImpl(me.getEndPoint(), me.getIncarnationId(), localEndPoint, ++members);
    localMembers.add(member);
    if (log.isDebugEnabled()) {
      log.debug("Current set of local members: ");
      for (Iterator iter = localMembers.iterator(); iter.hasNext();) {
        log.debug(iter.next());
      }
    }
    return member;
  }


  /**
   *  Returns the number of members associated with this Jgroup daemon.
   */
  public int members()
    throws RemoteException
  {
    if (log.isDebugEnabled()) {
      log.debug("I'm the daemon, and " + members + " members are associated with me");
      ThreadMonitor.dumpAllThreads();
    }
    return members;
  }


  /**
   *  Adds an event generated by a member to the local queue of events.
   */
  public void addEvent(Event event)
    throws RemoteException
  {
    if (log.isDebugEnabled()) {
      MDC.put("group", "[Group: " + event.getGid() + "]");
      log.debug("Received member event " + event);
    }
    mss.lsend(event);
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Implementation of methods inherited from the MssUser interface
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Handles a message received from a remote host.
   *  
   *  @param tag the tag identifying the type of message
   *  @param stream the message, represented by a <CODE>InMessage</CODE> 
   *    object
   *  @param src the sender of the message
   */
  public void remoteReceive(byte tag, InMessage stream, EndPoint src)
  {
    if (src.isLocal()) {
      if (log.isDebugEnabled())
        log.debug("Local message received; discarded at the daemon layer");
      return;
    }
    if (log.isDebugEnabled()) {
      NDC.push(Network.getMachineName(src.getAddress().getHostName()));
    }
    try {
      /* 
       *  Search the group identifier in the group table; if not present,
       *  this daemon is not interested in the group and discards the 
       *  message.
       */
      int gid = GroupId.unmarshal(stream);
      Group group = grouptable.getGroup(gid);
      if (log.isDebugEnabled()) {
        String tagString = Tag.R_TAGS[0];
        if (tag > 0 && tag <= R_RESULT)
          tagString = Tag.R_TAGS[tag];
        else if (tag >= R_NOGID && tag <= R_JOINACK)
          tagString = Tag.R_ExTAGS[tag-R_NOGID];
        if (group == null)
          MDC.put("group", "[Group: " + gid + "]");
        else
          MDC.put("group", group.toString());
        log.debug("remoteReceive: tag=" + tagString + ", gid=" + gid + ", src=" + src);
      }
      if (group == null) {
        if (log.isDebugEnabled())
          log.debug("Not interested in group: " + gid + "; sending R_LEAVE to: " + src);
        try {
          MsgGroup tosend = MsgGroup.marshal(gid, new java.util.ArrayList(0), 1);
          mss.send(R_LEAVE, tosend, src);
        } catch (IOException e) {
          log.error("Unable to marshal group message", e);
        }
      } else {
        /*
         *  Based on the tag, invokes the appropriate handling method. 
         */
        switch(tag) {
        case R_MCAST:
          group.handleMsgMcast  (MsgMcast.unmarshal(stream),  src);
          break;
        case R_ACK:
          group.handleMsgAck    (MsgAck.unmarshal(stream),    src);
          break;
        case R_PROP:
          group.handleMsgProp   (MsgProp.unmarshal(stream),   src);
          break;
        case R_ESTIM:
          group.handleMsgEstim  (MsgEstim.unmarshal(stream),  src);
          break;
        case R_FORWARD:
          group.handleMsgDlvrd  (MsgMcast.unmarshal(stream),  src);
          break;
        case R_SYNCH:
          group.handleMsgSynch  (MsgSynch.unmarshal(stream),  src);
          break;
        case R_SYMM:
          group.handleMsgSymm   (MsgSymm.unmarshal(stream),   src);
          break;
        case R_VIEW:
          group.handleMsgView   (MsgView.unmarshal(stream),   src);
          break;
        case R_RESULT:
          group.handleMsgResult (MsgResult.unmarshal(stream), src);
          break;
        case R_JOIN:
        {
          boolean memberAdded = group.handleMsgJoin(MsgGroup.unmarshal(stream), src);
          if (memberAdded && Util.in(trset, src)) {
            group.handleRemoteSuspect(nsuspects, trset, nrset, nuset);
          }
          break;
        }
        case R_JOINACK:
        {
          boolean memberAdded = group.handleMsgJoinAck(MsgGroup.unmarshal(stream), src);
          if (memberAdded && Util.in(trset, src)) {
            group.handleRemoteSuspect(nsuspects, trset, nrset, nuset);
          }
          break;
        }
        case R_LEAVE:
          group.leave(src);  
          break;
        default:
          log.warn("Ignoring unknown message tag: " + tag);
        } // END switch
        if (group.isEmpty()) {
          // The group has no local members anymore
          grouptable.removeGroup(gid);
          group.sendLeave();
        }
      }
    } catch (IOException e) {
      log.error("Unable to handle mreceived message", e);
    } catch (ClassNotFoundException e) {
      log.error("Internal error: Cannot unmarshal mreceived message", e);
    } finally {
      NDC.pop();
    }
  }


  /**
   *  Handles a message received by a local member.
   *  
   *  @param event represents the received message
   */
  public void localReceive(Event event)
  {
    int gid = event.getGid();
    Group group = grouptable.getGroup(gid);
    if (event.getTag() == JOIN_REQUEST && group != null && group.isEmpty()) {
       grouptable.removeGroup(gid);
       group = null;
     }
    if (group != null) {
      if (log.isDebugEnabled()) {
        NDC.push("local");
        MDC.put("group", group.toString());
        log.debug("localReceive: " + event); 
      }
      /*
       *  If this event is associated to an existing group,
       *  the group object will handle it.
       */
      switch (event.getTag()) {
      case MULTICAST_REQUEST:
        group.handleMulticastRequest  ((MulticastRequest) event);
        break;
      case DELIVERY_ACK:
        group.handleLocalDlvrAck      ((DeliveryAck) event);
        break;
      case INSTALL_ACK:
        group.handleLocalViewAck      ((InstallAck) event);
        break;
      case PREPARE_ACK:
        group.handleLocalPrepareAck   ((PrepareAck) event);
        break;
      case JOIN_REQUEST:
        group.handleLocalJoin         ((JoinRequest) event, this);
        break;
      case LEAVE_REQUEST:
        group.handleLeaveRequest      ((LeaveRequest) event);
        break;
      }
    } else if (event.getTag() == JOIN_REQUEST) {
      if (log.isDebugEnabled()) {
        NDC.push("local");
        MDC.put("group", "[Group: " + gid + "]");
      }
      /*
       *  If this event is associated to a non-existing group,
       *  and it is a join event, create a new group object.
       */
      createGroup((JoinRequest) event);
    } else {
      log.warn("Illegal local event (unknown group " + gid + ")");
    }
    if (log.isDebugEnabled())
      NDC.pop();
  }


  /**
   *  Handles variations in the mss-reachability
   *
   *  @param trset   set of all (total) reachable hosts
   *  @param nrset   set of reachable hosts that were unreachable
   *  @param nuset   set of unreachable hosts that were reachable
   *  @param newinc  set of hosts who has a new incarnation identifier
   */
  public void remoteSuspect(EndPoint[] trset, EndPoint[] nrset, EndPoint[] nuset, EndPoint[] newinc)
  {
    /* Increment the number of generated suspects */
    nsuspects++;

    if (log.isDebugEnabled()) {
      log.debug("nsuspects=" + nsuspects);
      StringBuilder buf = new StringBuilder("trset(");
      buf.append(trset.length);
      buf.append("): ");
      for (int i = 0; i < trset.length; i++) {
        buf.append(trset[i]);
        buf.append(" ");
      }
      log.debug(buf);

      buf = new StringBuilder("nrset(");
      buf.append(nrset.length);
      buf.append("): ");
      for (int i = 0; i < nrset.length; i++) {
        buf.append(nrset[i]);
        buf.append(" ");
      }
      log.debug(buf);

      buf = new StringBuilder("nuset(");
      buf.append(nuset.length);
      buf.append("): ");
      for (int i = 0; i < nuset.length; i++) {
        buf.append(nuset[i]);
        buf.append(" ");
      }
      log.debug(buf);
    }

    /*
     * Sets symmetry set to the previous trset value;
     * then stores trset, nrset, nuset
     */
    for(int i=0; i < nrset.length; i++)
      grouptable.setSymset(nrset[i], this.trset);  
    this.trset = trset;
    this.nrset = nrset;
    this.nuset = nuset;

    /* Send JOIN messages to the newly reachable members (nrset) */
    if (nrset.length > 0) {
      for (Iterator iter = grouptable.iterator(); iter.hasNext(); ) {
        Group group = (Group) iter.next();
        MDC.put("group", group.toString());
        try {
          MsgGroup tosend = MsgGroup.marshal(group.getGid(), group.getHosts(), nrset.length);
          mss.msend(R_JOIN, tosend, nrset);
        } catch (IOException e) {
          log.error("Unable to marshal group message", e);
        }
        if (log.isDebugEnabled())
          log.debug("Sent JOIN message for group (" + group.getGid() + ") to new reachable hosts");
      }
    }

    /* Updates version numbers in case of new incarnations */
    if (newinc != null) {
      for (int i=0; i < newinc.length; i++) {
        /* Iterator over the groups with new incarnations. */
        for (Iterator iter = grouptable.getGroups(newinc[i]); iter.hasNext(); ) {
          Group group = (Group) iter.next();
          HostData hd = group.getHost(newinc[i]);
          if (hd != null) 
            hd.newIncarnationId();
          if (log.isDebugEnabled()) {
            MDC.put("group", group.toString());
            log.debug("Newinc: " + newinc[i]);
          }
        }
      }
    }
    activateGroups(nrset);
    activateGroups(nuset);
  }

  /**
   *  Invoked to notify the objects implementing this interface that
   *  a remote object is not responding to remote method invocations,
   *  and thus should be crashed.
   *  
   *  @param memberId a callback object identifying the suspected object;
   *    this object has been stored in the Mss through the addPingListener
   *    invocation on the Mss.
   */
  public void localSuspect(MemberId memberId)
  {
    /*
     *  Iterates over the groups contained in the group table to 
     *  notify that the given member id is crashed.
     */
    for (Iterator iter = grouptable.iterator(); iter.hasNext(); ) {
      Group group = (Group) iter.next();
      MDC.put("group", group.toString());
      group.handleLocalSuspect(memberId);
    }
  }

  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // General functions
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Constructs the datastructures required for a group and broadcast
   *  an R_JOIN event through the cluster leaders.  This leads to the
   *  installation of a single member view, before the member
   *  potentially receives a JOINACK.
   *
   *  @param request
   *     A <code>JoinRequest</code> object.
   */
  private void createGroup(JoinRequest request)
  {
    int gid = request.getGid();
    if (log.isDebugEnabled()) {
      MDC.put("group", "[Group: " + gid +"]");
      log.debug("Creating new (local) group: " + gid);
    }
    EndPoint endpoint = me.getEndPoint();
    long vid = ViewId.create(endpoint, 0);
    MemberId joiningMember = request.getMemberId();
    View view = new ViewImpl(gid, vid, joiningMember);
    MemberData md = new MemberData(joiningMember, request.getDispatcher());
    md.schedulePingTimer(this);
    HostData hd = new HostData(me);
    Group group = new Group(grouptable, mss, gid, view, hd, md);
    if (log.isDebugEnabled())
      MDC.put("group", group.toString());
    grouptable.insertGroup(group);
    grouptable.addEndpointToGroup(endpoint, group);
    InstallEvent event = new InstallEvent(gid, view, 0, 0);
    boolean success = DaemonInteraction.addEvent(md, event);
    if (!success) {
      /*
       * If have no success to deliver the install event to the member,
       * we assume it is crashed. Since this is the only member in the
       * group, and it is crashed, we remove it and we return from the
       * method without sending messages outside.
       */
      grouptable.removeGroup(gid);
      if (log.isDebugEnabled())
        log.debug("Failed to create group " + gid + "; new member crashed");
      return;
    }
    try {
      MsgGroup tosend = MsgGroup.marshal(gid, group.getHosts(), ALL);
      /*
       * Sending the join request to everyone, as we want to advertise
       * the existence of this member in the specified group.
       */
      mss.msend(R_JOIN, tosend);
    } catch (IOException e) {
      log.error("Unable to marshal group message", e);
    }
    if (log.isDebugEnabled())
      log.debug("Successfully created (local) group: " + gid);
  }


  /**
   *  Handle suspects
   */
  private void activateGroups(EndPoint[] hosts)
  {
    for (int i=0; i < hosts.length; i++) {
      for (Iterator iter = grouptable.getGroups(hosts[i]); iter.hasNext(); ) {
        Group group = (Group) iter.next();
        MDC.put("group", group.toString());
        group.handleRemoteSuspect(nsuspects, trset, nrset, nuset);
      }
    }
  }

} // END Daemon
