/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.relacs.daemon;

import java.io.IOException;
import java.rmi.RemoteException;
import java.util.Iterator;

import jgroup.core.ConfigManager;
import jgroup.core.JgroupException;
import jgroup.core.MemberId;
import jgroup.core.registry.BootstrapRegistry;
import jgroup.relacs.config.DistributedSystemConfig;
import jgroup.relacs.config.Domain;
import jgroup.relacs.config.Host;
import jgroup.relacs.config.HostSet;
import jgroup.relacs.config.TransportConfig;
import jgroup.relacs.events.Event;
import jgroup.relacs.gm.RemoteDispatcher;
import jgroup.util.MsgFactory;
import jgroup.util.ThreadMonitor;

import org.apache.log4j.LogManager;
import org.apache.log4j.Logger;
import org.apache.log4j.MDC;


/**
 *  The <code>DaemonInteraction</code> class facilitate a common
 *  interaction point between the daemon and group managers.  Its main
 *  purpose is to provide common error handling.
 *
 *  @author Hein Meling
 *  @since Jgroup 1.2
 */
public class DaemonInteraction
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Logger
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Obtain logger for this class */
  private static final Logger log = Logger.getLogger(DaemonInteraction.class);


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Static fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** 
   * The remote daemon object reference (same host, so also local, but
   * going through RMI).  
   */
  private static DaemonService daemonService;

  /** Transport configuration information */
  private static TransportConfig tconf;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Static methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Initialize the Jgroup daemon, according to the specified number of
   *  daemons that should exist in this domain.
   *
   *  @throws JgroupException
   */
  public static void initDaemon()
    throws JgroupException
  {
    /* Obtains configuration information */
    tconf = (TransportConfig) ConfigManager.getConfig(TransportConfig.class);

    /* Obtain distributed system configuration information */
    DistributedSystemConfig dsc = ConfigManager.getDistributedSystem();

    /* Message factory initialization */
    MsgFactory.initMsgFactory(tconf.getPayload(), dsc.size(), dsc.numOfDomains());

    Domain localDomain = DistributedSystemConfig.getLocalDomain();
    if (localDomain.allDaemons()) {
      /* All the hosts in this domain should have a Jgroup daemon. */
      init();

    } else if (localDomain.hasNoDaemons()) {
      /*
       * The local domain should not have any daemons; we need to query
       * the remote domains for any Jgroup daemons.
       */
       throw new UnsupportedOperationException("The local domain must have at least one Jgroup daemon");

    } else {
      /* Check if the local JVM already has a daemon */
      if (checkLocalDaemon() < 0) {
        try {
          /* Try to obtain a daemon running in another JVM on the localhost */
          daemonService = Daemon.lookupLocalDaemon();
          if (log.isDebugEnabled()) {
            log.debug("Daemon found in bootstrap registry");
          }
        } catch (JgroupException e) {
          /* There is no daemon running on the local host. */
          checkRemoteDaemons(localDomain);
        }
      }
    }
  }


  /**
   *  Returns the number of members known to the local daemon.  If a negative
   *  value is returned, the local daemon references is either <code>null</code>
   *  or no longer usable.
   */
  private static int checkLocalDaemon()
  {
    if (daemonService == null)
      daemonService = Daemon.quickLookupLocalDaemon();
    if (daemonService != null) {
      try {
        /* Check if the currently available daemon reference is valid. */
        return daemonService.members();
      } catch (RemoteException e) {
        /* The locally known daemon reference is not valid. */
      }
    }
    return -1;
  }


  /**
   *  Check all hosts in the provided domain for available daemons.
   *  Once we have determined the hosts that are running daemons, we
   *  check if there is enough daemons in this domain, according to the
   *  specified number of daemons that should reside in this domain.
   *  If there is enough, we attach ourselves to the daemon with the
   *  least associated members.  If there is not enough daemons, we
   *  attempt to create one locally.
   *
   *  @param domain
   *    The domain in which to look for Jgroup daemons.
   *  @throws JgroupException
   *    Raised if a local Jgroup daemon could not be created.
   */
  private static void checkRemoteDaemons(Domain domain)
    throws JgroupException
  {
    HostSet hosts = domain.getHostSet();
    int jdaemons = domain.numOfDaemons(), minMembers = Integer.MAX_VALUE;
    for (Iterator iter = hosts.iterator(); jdaemons > 0 && iter.hasNext();) {
      Host host = (Host) iter.next();
      String dRegName = DaemonService.DAEMON_NAME + host.getPort();
      try {
        DaemonService jdService = (DaemonService) host.lookup(dRegName);

        /* Number of members associated with this particular daemon */
        int members = jdService.members();
        /* If we get here, we have found another daemon */
        jdaemons--;
        /*
         * Until we know if we should create a local daemon, we assume
         * that the daemon recently found to be hosting the least amount
         * of members will be selected as our daemon.
         */
        if (members < minMembers) {
          daemonService = jdService;
          minMembers = members;
        }
        if (log.isDebugEnabled()) {
          log.debug("Found daemon on " + host);
          log.debug("Members: " + members);
        }

      } catch (Exception e) {
        /*
         * This will catch NotBoundException, AccessException
         * and RemoteException.
         */
      }
    }
    if (jdaemons > 0) {
      if (log.isDebugEnabled()) {
        log.debug("remaining jdaemons to create: " + jdaemons);
        log.debug("domain.jdaemons: " + domain.numOfDaemons());
      }
      /*
       * Since there is not enough daemons in this domain yet,
       * we create one locally.
       */
      init();

    } else {
      /*
       * Since there is already enough Jgroup daemons in this domain
       * we simply get the one with the least members assigned to it.
       */
      if (log.isDebugEnabled()) {
        log.debug("Using: " + daemonService);
      }

    }
  }


  /**
   *  Create or attach to a Jgroup daemon on the local host.
   */
  private static void init()
    throws JgroupException
  {
    try {
      BootstrapRegistry.refreshRegistryStub();
    } catch (RemoteException e) {
      throw new JgroupException("Could not initialize the bootstrap registry");
    }

    /* Check if the local JVM already has a daemon */
    if (checkLocalDaemon() < 0) {
      try {
        /* Try to create a daemon in the local JVM. */
        daemonService = Daemon.createDaemon(tconf);
        if (log.isDebugEnabled())
          log.debug("Daemon created");
        Daemon.bindLocalDaemon(daemonService);
      } catch (IOException ioe) {
        if (log.isDebugEnabled())
          log.warn("Failed to create local Daemon", ioe);
        /*
         * Try to obtain an RMI-based daemon running in another JVM,
         * since we failed to create a daemon in the local JVM.
         */
        daemonService = Daemon.lookupLocalDaemon();
        if (log.isDebugEnabled())
          log.debug("Daemon found in bootstrap registry");
      }
    }
  }


  /**
   *  Returns a <code>MemberId</code> for a new member.  The new
   *  identifier is obtained by concatenating the endpoint of this
   *  daemon, the incarnation number used to distinguish different
   *  incarnation of the same endpoint, and a sequence number used to
   *  distinguish different members running on the same host (different
   *  JVMs).
   *
   *  @exception JgroupException
   *    Raised if we could not obtain a member identifier due to
   *    communication problems with the daemon.
   */
  public static MemberId getMemberId()
    throws JgroupException
  {
    try {
      return daemonService.getMemberId(DistributedSystemConfig.getLocalHost());
    } catch (RemoteException e) {
      throw new JgroupException("Unable to obtain my member identifier", e);
    }
  }


  /**
   *  Adds an event generated by a member to the daemon's event queue.
   * 
   *  @return
   *    True if the event was successfully added; false is returned
   *    if the daemon has crashed, and not recovered.
   */
  public static boolean addEvent(Event event)
  {
    try {
      if (log.isDebugEnabled()) {
        MDC.put("group", "[Group: " + event.getGid() + "]");
        log.debug("Member->Daemon: " + event);
      }
      daemonService.addEvent(event);
      if (log.isDebugEnabled())
        log.debug("addEvent returned successfully (M->D)");
      return true;
    } catch (RemoteException e) {
      /*
       * If the system property 'jgroup.daemon.recover.locally' is set to
       * true, we try to recover from a crashed daemon by performing the
       * initDaemon() method.  Otherwise, we simply perform a system halt()
       * to ensure that this JVM shuts itself down immediately.  This is
       * useful to avoid that a member generates "noise" while it is in an
       * unstable state, and may not recover successfully anyway.  This
       * allows remaining members of the group to quickly detect this
       * member's failure, and inform the ARM framework of the crash whom
       * will take the opportune countermeasures.
       */
      log.warn("Member failed to send event to Daemon", e);
      return handleDaemonFailure();
    }
  }


  /**
   *  Adds an event generated by the daemon to given (local) member's
   *  event queue.
   *
   *  @param member
   *    The recipient member (Dispatcher) of this event.
   *  @param event
   *    The event to be added to the given member's event queue.
   *  @return
   *    True if communication is ok; false otherwise.
   */
  public static boolean addEvent(MemberData member, Event event)
  {
    try {
      if (log.isDebugEnabled()) {
        log.debug("Daemon->Member(" + member + "): " + event);
      }
      RemoteDispatcher dispatcher = member.getRemoteDispatcher();
      synchronized (dispatcher) {
        dispatcher.addEvent(event);
      }
      if (log.isDebugEnabled())
        log.debug("addEvent returned successfully (D->M)");
      return true;
    } catch (RemoteException e) {
      /*
       * Return false to notify the daemon that communication
       * with the member has failed.  If the member is local
       * (i.e., resides in the same virtual machine), this can
       * never happend.  Otherwise, the daemon will take the
       * opportune actions to install a new view.  Exception
       * is printed in the debug log to enable to check for
       * remote exceptions not related to failed communication.
       */
      log.warn("Daemon failed to send event to " + member, e);
      return false;
    } 
  }


  /**
   *  Handle daemon-member interaction failures. <p>
   *
   *  If the system property 'jgroup.daemon.recover.locally' is true,
   *  and there is a confirmed daemon crash, local recovery is performed.
   *  Otherwise, this member will commit suicide.
   *
   *  @return
   *    True is returned if a new daemon was successfully initialized.
   */
  public static boolean handleDaemonFailure()
  {
    if (DaemonInteraction.suspectDaemon()) {
      /*
       * The daemon was correctly suspected.  Check if we should
       * try to recover locally, or leave it to ARM.  Default is
       * to not try to recover the daemon locally.
       */
      if (Boolean.getBoolean("jgroup.daemon.recover.locally")) {
        /*
         * Try to recover a Jgroup daemon on the localhost.
         */
        if (log.isDebugEnabled())
          log.debug("Trying to recover daemon locally");
        try {
          initDaemon();
          return true;
        } catch (JgroupException e) {
          log.warn("Local daemon recovery failed.", e);
          return false;
        }
      } else {
        /*
         * By default, we do not try to recover the daemon locally;
         * instead we ensure to crash this JVM so that ARM can handle
         * this crash.
         */
        log.error("Crashing this member dispatcher since the daemon crashed");
        // Ensure that log files are closed
        LogManager.shutdown();
        Runtime.getRuntime().halt(13);
        return false;
      }
    } else {
      /*
       * The daemon is still alive, but was suspected;
       * appears not to be pinging me anymore.
       */
      log.error("DUMPING THREADS");
      for (int i = 0; i < 10; i++) {
        ThreadMonitor.dumpAllThreads();
        try {
          Thread.sleep(200);
        } catch (InterruptedException e) { }
      }
      log.error("COMMITTING SUICIDE");
      Runtime.getRuntime().halt(13);
      return false;
    }
  }


  /**
   *  Method used to indicate that the daemon has not responded within
   *  some predefined timeout.  Here we try to recover by installing
   *  a new daemon locally.
   *
   *  @return true if the daemon is not responding, false otherwise.
   */
  public static boolean suspectDaemon()
  {
    assert daemonService != null : "DaemonService should not be null";
    try {
      int members = daemonService.members();
      if (log.isDebugEnabled())
        log.debug("Daemon is alive with members: "+ members);
      return false;
    } catch (RemoteException e1) {
      if (log.isDebugEnabled()) {
        log.debug("Daemon correctly suspected", e1);
      }
      return true;
    }
  }

} // END DaemonInteraction
