/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.relacs.daemon;

import java.util.HashSet;
import java.util.Set;
import java.util.Timer;
import java.util.TimerTask;

import jgroup.core.ConfigManager;
import jgroup.core.MemberId;
import jgroup.relacs.config.TransportConfig;
import jgroup.relacs.gm.PingListener;
import jgroup.relacs.gm.RemoteDispatcher;
import jgroup.relacs.mss.MssUser;

import org.apache.log4j.Logger;


/**
 *  The <code>MemberData</code> class is a data structure containing
 *  information about a member on a per-group basis. In particular,
 *  the current status of the member with respect to group membership
 *  is stored in this data structure.
 *
 *  @author Alberto Montresor
 *  @since  Jgroup 0.1
 */
final class MemberData
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Logger
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Obtain logger for this class */
  private static final Logger log = Logger.getLogger(MemberData.class); 
  
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constants
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** 
   *  Members having this status have joined the group and there neither
   *  crashed nor leaving the group
   */
  private static final int S_OK = 1;
  
  /** 
   *  Members having this status have requested to leave the group.
   *  They are in the process of being removed.
   */
  private static final int S_LEAVING = 2;
  
  /** 
   *  Members having this status have crashed; are in the process of
   *  of being removed
   */
  private static final int S_CRASHED = 3;

  /**
   *  String representation of member status.
   */
  private static final String S_STAT[] = {
    "INVALID_STATUS", "OK", "LEAVING", "CRASHED"
  };


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Identifier */
  private MemberId memberId;
  
  /** Reference to the dispatcher service of this member */
  private RemoteDispatcher dispatcher;

  /** Current status of the member */
  private int status;

  /** Indicates if the member has acknowledged the most recent PrepareEvent */
  private boolean prepared;

  /** The set of acks that the daemon is awaiting from this member. */
  private Set<Integer> acks = new HashSet<Integer>(5);

  // Local Member Ping Management

  /** Ping rate; how often local members are ping'ed */
  private int pingRate;

  /** The ping timer object associated with this member */
  private Timer pingTimer;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructors
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Creates a new <CODE>MemberData</CODE> object associated a member
   * identifier to a remote dispatcher and a ping listener.<p>
   *
   * @param id the identifier of the member associated to this
   *    data structure 
   * @param dispatcher the dispatcher of the member
   */
  MemberData(MemberId id, RemoteDispatcher dispatcher)
  {
    this.memberId = id;
    this.dispatcher = dispatcher;
    this.status = S_OK;
    this.prepared = false;
    TransportConfig tconf =
      (TransportConfig) ConfigManager.getConfig(TransportConfig.class);
    pingRate = tconf.getLocalPingRate();
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Returns the member identifier associated to this member.
   */
  MemberId getMemberId()
  {
    return memberId;
  }

  
  /**
   *  Returns true if this member is leaving the group. 
   */
  boolean isLeaving()
  {
    return status == S_LEAVING;
  }
  
  /**
   *  Returns true if this member has crashed and left the group. 
   */
  boolean isCrashed()
  {
    return status == S_CRASHED;
  }

  /**
   *  Returns true if this member has acknowledged the PrepareEvent.
   */
  boolean isPrepared()
  {
    return prepared;
  }

  /**
   *  Set this member to the leaving status.
   */
  void setLeaving()
  {
    this.status = S_LEAVING;
  }

  /**
   *  Set the leaving status of this member.
   * 
   *  @param status the new status of the member
   */
  void setCrashed()
  {
    if (log.isDebugEnabled())
      log.debug("Marking " + memberId + " as CRASHED");
    this.status = S_CRASHED;
  }

  /**
   *  Set the prepared status to true for this member.
   */
  void setPrepared()
  {
    prepared = true;
  }
  
  /**
   *  Set the prepared status to false for this member.
   */
  void resetPrepared()
  {
    prepared = false;
  }

  /**
   *  Returns a reference to the remote dispatcher that deals with events
   *  associated to this member. The remote dispatcher can be actually
   *  remote, if the member is in a JVM different from the local one;
   *  or can be in the local JVM.
   */
  RemoteDispatcher getRemoteDispatcher()
  {
    return dispatcher;
  }
  
  /**
   *  Returns true if the daemon is awaiting one or more ACKs from the member.
   */
  boolean hasOutstandingAcks()
  {
    return (acks.size() != 0);
  }

  /**
   *  Returns true if the daemon is awaiting for the given ACK for this member.
   */
  boolean awaitingAck(int mid)
  {
    return acks.contains(mid);
  }

  /**
   *  The daemon is awaiting an ACK for the given message from this member.
   */
  void setAck(int mid)
  {
    acks.add(mid);
  }

  /**
   *  The daemon is no longer awaiting an ACK for the given message from this member.
   */
  void removeAck(int mid)
  {
    acks.remove(mid);
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods for object polling
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Schedule ping timer to ping the dispatcher object periodically.
   *
   * @param mssuser the daemon object assoicated with this member;
   *   to be notified if this member is suspected to have failed.
   */
  public void schedulePingTimer(final MssUser mssuser)
  {
    if (pingTimer != null) {
      /*  
       * Ignored; it can be caused by the same member
       * joining more than one group.
       */
      if (log.isDebugEnabled())
        log.debug("Ping timer already scheduled for: " + memberId);
      return;
    }
    TimerTask pingTT = new TimerTask() {
      public void run() {
        try {
          /* Performs pinging */
          if (log.isDebugEnabled())
            log.debug("Pinging local member: " + memberId);
          synchronized (dispatcher) {
            ((PingListener) dispatcher).ping();
          }
          if (log.isDebugEnabled())
            log.debug("Ping local member completed: " + memberId);
        } catch (Exception e) {
          mssuser.localSuspect(memberId);
          cancelPingTimer();
          log.warn("Mss unable to connect to replica", e);
        }
      }
    };
    pingTimer = new Timer("PingTimer-" + memberId.getLocalId(), true);
    pingTimer.schedule(pingTT, pingRate, pingRate);
    if (log.isDebugEnabled())
      log.debug("Ping timer scheduled for: " + memberId);
  }

  /**
   * Cancel ping timer for this member.
   */
  public void cancelPingTimer()
  {
    if (pingTimer != null) {
      pingTimer.cancel();
      if (log.isDebugEnabled())
        log.debug("Ping timer canceled: " + memberId);
    }
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from Object
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Returns a string representation of this object.
   */
  public String toString()
  {
    StringBuilder buf = new StringBuilder();
    buf.append("[MemberData: ");
    buf.append(memberId);
    buf.append(", ");
    buf.append(S_STAT[status]);
    if (hasOutstandingAcks()) {
      buf.append(", outstanding=");
      buf.append(acks);
    }
    buf.append("]");
    return buf.toString();
  }

} // END MemberData
