/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.relacs.daemon;


import java.io.IOException;
import java.io.ObjectInput;
import java.util.Collection;
import java.util.Iterator;

import jgroup.core.EndPoint;
import jgroup.relacs.mss.MssConstants;
import jgroup.relacs.types.EndPointImpl;
import jgroup.relacs.types.GroupId;
import jgroup.relacs.types.GroupIndex;
import jgroup.util.MsgFactory;
import jgroup.util.OutMessage;


/**
 *  The <code>MsgGroup</code> class
 *
 *  @author Alberto Montresor
 *  @author Hein Meling
 *  @since Jgroup 0.1
 */
final class MsgGroup
  implements MssConstants, DaemonMsg
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Position constants
  ////////////////////////////////////////////////////////////////////////////////////////////

  private static final int START       = MSS_HEADER_SIZE;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Unique message instance
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * As these messages are never stored by the daemon, a single message
   * instance is sufficient.  In this way, we save the costs of
   * allocating and garbage collecting messages.
   */
  private static MsgGroup msg = new MsgGroup();

  /** The marshalled message data (also static for same reason as above) */
  private static OutMessage outmsg;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** List of host identifiers associated with the group */
  EndPoint[] hosts;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Returns a <code>MsgGroup</code> object that contains the decoding
   *  of the m-received input stream.
   *
   *  @param msgstream   the message input stream to decode
   */
  static MsgGroup unmarshal(ObjectInput msgstream)
    throws IOException, ClassNotFoundException
  {
    int len  = GroupIndex.unmarshal(msgstream);
    msg.hosts = new EndPoint[len];
    for (int i=0; i<len; i++) {
      msg.hosts[i] = new EndPointImpl();
      msg.hosts[i].readExternal(msgstream);
    }

    return msg;
  }


  /**
   *  Returns a output stream containing the encoding of a
   *  <code>MsgGroup</code> message.
   *
   *  @param     gid       group identifier
   */
  static MsgGroup marshal(int gid, Collection list, int nreceivers)
    throws IOException
  {
    // Allocates data buffer
    int size = START + GroupId.SIZE + GroupIndex.SIZE + list.size() * EndPointImpl.SIZE;
    outmsg = MsgFactory.get(size, nreceivers);
    outmsg.seek(START);

    // Writes group id and number of hosts
    GroupId.marshal(outmsg, gid);
    GroupIndex.marshal(outmsg, list.size());

    // Writes hosts
    Iterator iterator = list.iterator();
    while (iterator.hasNext()) {
      HostData scan = (HostData) iterator.next();
      scan.getEndPoint().writeExternal(outmsg);
    }
    return msg;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from DaemonMsg
  ////////////////////////////////////////////////////////////////////////////////////////////

  public int size()
  {
    /* Note; it is ok to use the outmsg size since MsgGroup will never be
     * forwarded or resent (it is not queued). */
    return outmsg.size();
  }

  public OutMessage getOutMessage()
  {
    return outmsg;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from Object
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Returns a string representation of this object
   */
  public String toString()
  {
    StringBuilder buffer = new StringBuilder();
    buffer.append("[MsgGroup: hosts={");
    for (int i=0; i < hosts.length; i++) {
      buffer.append(hosts[i]);
      if (i < hosts.length-1)
        buffer.append(", ");
    }
    buffer.append("}]");
    return buffer.toString();
  }

} // END MsgGroup
