/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.relacs.daemon;

import java.io.IOException;
import java.io.ObjectInput;
import java.io.OutputStream;

import jgroup.core.MemberId;
import jgroup.core.multicast.AckListener;
import jgroup.relacs.events.MulticastRequest;
import jgroup.relacs.mss.MssConstants;
import jgroup.relacs.types.Flag;
import jgroup.relacs.types.GroupId;
import jgroup.relacs.types.GroupIndex;
import jgroup.relacs.types.MemberIdImpl;
import jgroup.relacs.types.MessageId;
import jgroup.relacs.types.MessageLen;
import jgroup.relacs.types.ViewId;
import jgroup.util.InMessage;
import jgroup.util.MsgFactory;
import jgroup.util.OutMessage;

/**
 * The <code>MsgMcast</code> class
 *
 * @author  Alberto Montresor
 * @since   Jgroup 0.1
 */
public final class MsgMcast
  implements MssConstants, DaemonMsg
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Position constants
  ////////////////////////////////////////////////////////////////////////////////////////////

  private static final int START_FIXED = MSS_HEADER_SIZE;

  private static final int FIXED_SIZE  = GroupId.SIZE + Flag.SIZE*2 + MemberIdImpl.SIZE + MessageLen.SIZE;
  private static final int START_VAR   = START_FIXED + FIXED_SIZE;

  private static final int VAR_SIZE    = GroupIndex.SIZE + MessageId.SIZE + ViewId.SIZE;
  private static final int START_DATA  = START_VAR + VAR_SIZE;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Serialized fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** True if this message carries a single object */
  boolean isObject;

  /** True if an ack is required */
  boolean ackr;
  
  /** Identifier of the sender */
  MemberId sender;
  
  /** Position of the sender in view identified by <code>vid</code> */
  int hpos;
  
  /** View identifier */
  long vid;
  
  /** Message identifier */
  int mid;
  
  /** Ack listener */
  AckListener ackl;

  /** Message to be sent */
  private OutMessage outmsg;
  
  /** Message received from the Mss */
  private InMessage inmsg;
  
  /** Set of acks associated with this message */
  // FIXME: Should be moved in an external datastructure
  private boolean[] ackset;

  /** Number of false values in ackset */
  private int nfalse;

  /** MsgMcast message size (excluding the mss header) */
  private int size;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructors and additional methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /*
   *  Constructs a <code>MsgMcast</code> data structure starting from a
   *  m-received message.
   */
  private MsgMcast(InMessage inmsg)
    throws IOException, ClassNotFoundException
  {
    isObject = Flag.unmarshal(inmsg);
    ackr = Flag.unmarshal(inmsg);
    sender = new MemberIdImpl();
    sender.readExternal(inmsg);
    size = MessageLen.unmarshal(inmsg);
    hpos = GroupIndex.unmarshal(inmsg);
    vid  = ViewId.unmarshal(inmsg);
    mid  = MessageId.unmarshal(inmsg);

    this.inmsg = inmsg;
    inmsg.seek(START_DATA);
    outmsg = new OutMessage(inmsg);
  }


  /**
   *  Creates an uncompleted <code>MsgMcast</code> object; other fields
   *  will be set when the information will be available.
   */
  MsgMcast(MulticastRequest mcastRequest)
    throws IOException
  {
    this.outmsg = mcastRequest.getPayload();
    /*
     * Computes the total size of this MsgMcast message, including that
     * application level data of the mcast.  This excludes the size
     * of the mss header fields.  Note that we can use the outmsg.size()
     * method to compute the application data size, since only app data
     * has been written to the outmsg stream.
     */
    size = outmsg.size() + FIXED_SIZE + VAR_SIZE;

    // Store info for local message
    this.isObject = mcastRequest.isObject();
    this.ackl = mcastRequest.getAckListener();
    this.ackr = (ackl != null);
    this.sender = mcastRequest.getSender();

    // Writes fixed part of the header
    outmsg.seek(START_FIXED);
    GroupId.marshal(outmsg, mcastRequest.getGid());
    Flag.marshal(outmsg, isObject);
    Flag.marshal(outmsg, ackr);
    sender.writeExternal(outmsg);
    MessageLen.marshal(outmsg, size);
  }


  /**
   *  Complete the message structure with the data related to the view
   *  in which the message is sent.
   *
   *  @param     hpos    host index in the view.
   *  @param     vid       view identifier.
   *  @param     mid     message identifier.
   */
  void complete(int hpos, long vid, int mid)
    throws IOException
  {
    // Writes variable part of the header
    outmsg.seek(START_VAR);
    GroupIndex.marshal(outmsg, hpos);
    ViewId.marshal(outmsg, vid);
    MessageId.marshal(outmsg, mid);

    // Store info for local message
    this.vid  = vid;
    this.hpos = hpos;
    this.mid  = mid;

    inmsg = new InMessage(outmsg);
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Static methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *
   */
  public static OutputStream createOutputStream()
  {
    OutMessage stream = MsgFactory.get();
    stream.seek(START_DATA);
    return stream;
  }


  /**
   *  Returns a <code>MsgMcast</code> object that contains the decoding
   *  of the m-received input stream.
   *
   *  @param inmsg    the message input stream to decode
   */
  static MsgMcast unmarshal(ObjectInput inmsg)
    throws IOException, ClassNotFoundException
  {
    return new MsgMcast((InMessage) inmsg);
  }


  /**
   *
   */
  InMessage getInMessage()
  {
    /*
     * This must be synchronized on the inmsg object so as to avoid
     * concurrent access to the inmsg object, since most methods in 
     * the InMessage class modify its state (the position field in
     * particular), including methods such as seek(), read(), insert()
     * and so on.  Such concurrent access may occur when two competing
     * threads (Ehandler and Dispatcher) want to access the inmsg object.
     * For example, one thread may be reading the inmsg object for
     * delivery to an upper-layer (see IntGroupHandler.deliverStream()),
     * while another thread wants to send the inmsg to another member of
     * the group, and thus needs to seek to the START_DATA position;
     * see MsgMcast.getInMessage().
     */
    synchronized (inmsg) {
      inmsg.seek(START_DATA);
      return inmsg;
    }
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from DaemonMsg
  ////////////////////////////////////////////////////////////////////////////////////////////

  public int size()
  {
    return size;
  }

  public OutMessage getOutMessage()
  {
    return outmsg;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Ack management
  ////////////////////////////////////////////////////////////////////////////////////////////

  void setAckArray(int len)
  {
    ackset = new boolean[len];
    nfalse = len;
  }
  
  void setAck(int pos)
  {
    if (!ackset[pos])
      nfalse--;
    ackset[pos] = true;
  }

  boolean isComplete()
  {
    return (nfalse == 0);
  }


  /**
   *  Returns a string representation of this object
   */
  public String toString()
  {
    StringBuilder b = new StringBuilder();
    b.append("[MsgMcast: ");
    b.append("mid=");
    b.append(mid);
    b.append(", sender=");
    b.append(sender);
    b.append(", vid=");
    b.append(vid);
    if (inmsg != null) {
      b.append(", ");
      b.append(inmsg);
    }
    b.append("]");
    return b.toString();
  }

} // END MsgMcast
