/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */


package jgroup.relacs.daemon;

import java.io.IOException;
import java.io.ObjectInput;

import jgroup.core.MemberId;
import jgroup.relacs.mss.MssConstants;
import jgroup.relacs.types.Flag;
import jgroup.relacs.types.GroupId;
import jgroup.relacs.types.GroupIndex;
import jgroup.relacs.types.MemberIdImpl;
import jgroup.relacs.types.MessageId;
import jgroup.relacs.types.ViewId;
import jgroup.util.MsgFactory;
import jgroup.util.OutMessage;

/**
 *
 * @author  Alberto Montresor
 * @since   Jgroup 0.8
 */
final class MsgResult
  implements MssConstants, DaemonMsg
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Position constants
  ////////////////////////////////////////////////////////////////////////////////////////////

  private static final int START  = MSS_HEADER_SIZE;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Group identifier */
  int gid;

  /** View identifier */
  long vid;

  /** Message identifier */
  int mid;
  
  /**
   * Identifier of the sender of the message to which this message is a
   * response.
   */
  MemberId sender;
  
  /** Number of messages delivered by all local members */
  int dlvr;
  
  /** Host index in the view */
  int hpos;
  
  /** Member index in the view */
  int vpos;
  
  /** Result value */
  Object result;         

  /** The marshalled message stream for this object */
  private OutMessage outmsg;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructors
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Default constructors used to create an empty ResultMessage in the
   *  decoding process.
   */
  MsgResult()
  {
  }
  
  /**
   *  Default constructors used to create an empty ResultMessage in the
   *  decoding process.
   */
  MsgResult(int gid, long vid, int mid, MemberId sender, int dlvr, int hpos, int vpos, 
    Object result)
  {
    this.gid = gid;
    this.vid = vid;
    this.mid = mid;
    this.sender = sender;
    this.dlvr = dlvr;
    this.hpos = hpos;
    this.vpos = vpos;
    this.result = result;
  }
  
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Returns an output stream containing the encoding of this message.
   */
  MsgResult marshal(int nreceivers)
    throws IOException
  {
    outmsg = marshal(gid, vid, mid, sender, dlvr, hpos, vpos, result, nreceivers);
    return this;
  }
  
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Static methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Returns a <code>MsgResult</code> object that contains the decoding
   *  of an m-received input stream.
   *
   *  @param stream    the message input stream to decode
   */
  static MsgResult unmarshal(ObjectInput stream)
    throws IOException, ClassNotFoundException
  {
    /*
     * Currently, this message type cannot be a singleton since the
     * SendBuffer class stores instances of MsgResult in its ackbuffer.
     */
    MsgResult msg = new MsgResult();

    msg.vid    = ViewId.unmarshal(stream);
    msg.mid    = MessageId.unmarshal(stream);
    msg.sender = new MemberIdImpl();
    msg.sender.readExternal(stream);
    msg.dlvr   = MessageId.unmarshal(stream);
    msg.hpos   = GroupIndex.unmarshal(stream);
    msg.vpos   = GroupIndex.unmarshal(stream);
    // check if stream contains a result object.
    if (Flag.unmarshal(stream)) {
      msg.result = stream.readObject();
    }
    return msg;
  }

  /**
   *  Returns a output stream containing the encoding of a
   *  <code>MsgResult</code> message.
   *
   *  @param     gid     group identifier
   *  @param     vid     current view identifier
   *  @param     mid     identifier of the original message to which this message is a response
   *  @param     sender  identifier of the sender of the original message to which this message is
   *    a response
   *  @param     hpos    host index in the view
   *  @param     mpos    member index in the view
   *  @param     result  result value
   */
  private static OutMessage marshal(int gid, long vid, int mid, MemberId sender, int dlvr, int hpos,
                            int vpos, Object result, int nreceivers)
    throws IOException
  {
    OutMessage stream;

    // Allocates data buffer
    if (result == null) {
      int size = START + GroupId.SIZE + ViewId.SIZE + MessageId.SIZE * 2 + 
                 GroupIndex.SIZE*2 + MemberIdImpl.SIZE + Flag.SIZE;
      stream = MsgFactory.get(size, nreceivers);
    } else {
      stream = MsgFactory.get();
    }
    stream.seek(START);

    // Writes single data
    GroupId.marshal(stream, gid);
    ViewId.marshal(stream, vid);
    MessageId.marshal(stream, mid);
    sender.writeExternal(stream);
    MessageId.marshal(stream, dlvr);
    GroupIndex.marshal(stream, hpos);
    GroupIndex.marshal(stream, vpos);
    boolean hasResult = (result != null);
    Flag.marshal(stream, hasResult);
    if (hasResult) {
      stream.writeObject(result);
    }
    return stream;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from DaemonMsg
  ////////////////////////////////////////////////////////////////////////////////////////////

  public int size()
  {
    /* Note; it is ok to use the outmsg size since MsgResult will never be
     * forwarded or resent (it is queued, but only to deliver later, I think). */
    return outmsg.size();
  }

  public OutMessage getOutMessage()
  {
    return outmsg;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from Object
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Returns a string representation of this object
   */
  public String toString()
  {
    StringBuilder buf = new StringBuilder();
    buf.append("[MsgResult: gid=");
    buf.append(gid);
    buf.append(", vid=");
    buf.append(vid);
    buf.append(", mid=");
    buf.append(mid);
    buf.append(", sender=");
    buf.append(sender);
    buf.append(", dlvr=");
    buf.append(dlvr);
    buf.append(", hpos=");
    buf.append(hpos);
    buf.append(", vpos=");
    buf.append(vpos);
    buf.append(", result=");
    buf.append(result);
    return buf.toString();
  }

} // END MsgResult
