/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.relacs.events;

import java.io.IOException;
import java.io.ObjectInput;
import java.io.ObjectOutput;

import jgroup.core.MemberId;
import jgroup.relacs.types.GroupIndex;
import jgroup.relacs.types.MemberIdImpl;
import jgroup.relacs.types.MessageId;

/**
 *  The <code>DeliveryAck</code> class represents a local message
 *  exchanged between members and the daemon; it is used by the member
 *  to notify the daemon that a message has been successfully delivered.
 *
 *  @author Alberto Montresor
 *  @since  Jgroup 1.2
 */
public final class DeliveryAck
  extends Event
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  private static final long serialVersionUID = -6414737364782334061L;

  /** Host position of the member in the current view */
  private int hostIndex;

  /** Local member position of the member in the current view */
  private int memberIndex;

  /** */
  private int viewIndex;

  /** Message identifier */
  private int mid;

  /** Identifier of the sender of the original message */
  private MemberId sender;

  /** Result value associated to the message acknowledgement */
  private Object result;

  /** Sender of this ack message */
  private MemberId ackSender;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructors
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Default constructor for externalization.
   */
  public DeliveryAck()
  {
  }


  /**
   *  Creates a <code>DeliveryAck</code> with the specified data.
   *  
   *  @param gid  group identifier
   *  @param hostIndex host position in the current view
   *  @param memberIndex member position in the current view
   *  @param viewIndex
   *  @param mid  message identifier
   *  @param sender sender identifier
   *  @param result result associated to the message acknowledgement
   */
  public DeliveryAck(int gid, int hostIndex, int memberIndex, int viewIndex, 
    int mid, MemberId sender, Object result, MemberId ackSender)
  {
    super(DELIVERY_ACK, gid);
    this.hostIndex  = hostIndex;
    this.memberIndex  = memberIndex;
    this.viewIndex = viewIndex;
    this.mid = mid;
    this.sender = sender;
    this.result = result;
    this.ackSender = ackSender;
  }

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Getter / setter methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Returns the position index of the local host in the array of the
   *  hosts contained in the current view.
   */
  public int getHostIndex()
  {
    return hostIndex;
  }

  /**
   *  Returns the position index of this member in the array of
   *  the local members contained in the current view.
   */
  public int getMemberIndex()
  {
    return memberIndex;
  }

  /**
   *  Returns the position index of this member in the array of
   *  the members contained in the current view.
   */
  public int getViewIndex()
  {
    return viewIndex;
  }

  /**
   *  Returns the message identifier.
   */
  public int getMessageId()
  {
    return mid;
  }

  /**
   *  Returns the sender identifier of the original message sender.
   */
  public MemberId getSender()
  {
    return sender;
  }

  /**
   *  Returns the result object associated to the message acknowledgement.
   */
  public Object getResult()
  {
    return result;
  }

  /**
   *  Returns the sender of this ack message.
   */
  public MemberId getAckSender()
  {
    return ackSender;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from Externalizable
  ////////////////////////////////////////////////////////////////////////////////////////////
  
  /**
   *  Restores the content of this object from the marshalled data contained
   *  in the specified input stream.
   * 
   *  @param in the stream to be read
   */
  public void readExternal(ObjectInput in)
    throws IOException, ClassNotFoundException
  {
    super.readExternal(in);
    hostIndex = GroupIndex.unmarshal(in);
    memberIndex = GroupIndex.unmarshal(in);
    viewIndex = GroupIndex.unmarshal(in);
    mid = MessageId.unmarshal(in);
    sender = new MemberIdImpl();
    sender.readExternal(in);
    result = in.readObject();
    ackSender = new MemberIdImpl();
    ackSender.readExternal(in);
  }

  /**
   *  Marshals the content of this object to the specified output stream.
   * 
   *  @param out the stream to be written
   */
  public void writeExternal(ObjectOutput out)
    throws IOException
  {
    super.writeExternal(out);
    GroupIndex.marshal(out, hostIndex);
    GroupIndex.marshal(out, memberIndex);
    GroupIndex.marshal(out, viewIndex);
    MessageId.marshal(out, mid);
    sender.writeExternal(out);
    out.writeObject(result);
    ackSender.writeExternal(out);
  }
  
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from Object
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Returns a string representation of the object.
   */
  public String toString()
  {
    StringBuilder buf = new StringBuilder();
    buf.append(super.toString());
    buf.append("[DeliveryAck: hostIndex=");
    buf.append(hostIndex);
    buf.append(", memberIndex=");
    buf.append(memberIndex);
    buf.append(", viewIndex=");
    buf.append(viewIndex);
    buf.append(", mid=");
    buf.append(mid);
    buf.append(", sender=");
    buf.append(sender);
    buf.append(", ackSender=");
    buf.append(ackSender);
    buf.append("]");
    return buf.toString();
  }

} // END DeliveryAck
