/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.relacs.events;

import java.io.IOException;
import java.io.ObjectInput;
import java.io.ObjectOutput;

import jgroup.core.MemberId;
import jgroup.relacs.gm.RemoteDispatcher;
import jgroup.relacs.types.MemberIdImpl;

/**
 *  The <code>JoinRequest</code> class represents a local message
 *  exchange between members and the daemon; it is used by the member to
 *  notify the daemon of a join request from the specified member.
 *
 *  @author Alberto Montresor
 *  @since Jgroup 1.2
 */
public final class JoinRequest
  extends Event
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  private static final long serialVersionUID = -749635009327118327L;

  /** Local member identifier */
  private MemberId memberId;

  /** Reference to the member */
  private RemoteDispatcher dispatcher;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructors
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Default constructor for externalization.
   */
  public JoinRequest()
  {
  }

  /**
   *  Creates a new <code>JoinRequest</code> event for the specified
   *  <code>memberId</code> to join group <code>gid</code>.
   */
  public JoinRequest(int gid, MemberId memberId, RemoteDispatcher dispatcher)
  {
    super(JOIN_REQUEST, gid);
    this.memberId = memberId;
    this.dispatcher = dispatcher;
  }

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods
  ////////////////////////////////////////////////////////////////////////////////////////////
  
  /**
   * 
   */
  public MemberId getMemberId()
  {
    return memberId;
  }

  /**
   * 
   */
  public RemoteDispatcher getDispatcher()
  {
    return dispatcher;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  //  Methods from Externalizable
  ////////////////////////////////////////////////////////////////////////////////////////////
  
  /**
   *  Restores the content of this object from the marshalled data contained
   *  in the specified input stream.
   * 
   *  @param in the stream to be read
   */
  public void readExternal(ObjectInput in)
    throws IOException, ClassNotFoundException
  {
    super.readExternal(in);
    memberId = new MemberIdImpl();
    memberId.readExternal(in);
    dispatcher = (RemoteDispatcher) in.readObject();
  }
  
  /**
   *  Marshals the content of this object to the specified output stream.
   * 
   *  @param out the stream to be written
   */
  public void writeExternal(ObjectOutput out)
    throws IOException
  {
    super.writeExternal(out);
    memberId.writeExternal(out);
    out.writeObject(dispatcher);
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from Object
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Returns a string representation of the object.
   */
  public String toString()
  {
    StringBuilder buf = new StringBuilder();
    buf.append(super.toString());
    buf.append("[JoinRequest:");
    buf.append(" memberId=");
    buf.append(memberId);
    buf.append(" remoteDispatcher=");
    buf.append(dispatcher);
    buf.append("]");
    return buf.toString();
  }

} // END JoinRequest
