/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.relacs.events;

import java.io.IOException;
import java.io.ObjectInput;
import java.io.ObjectOutput;

import jgroup.core.MemberId;
import jgroup.core.multicast.AckListener;
import jgroup.core.multicast.ChainIdentifier;
import jgroup.relacs.types.Flag;
import jgroup.relacs.types.MemberIdImpl;
import jgroup.util.OutMessage;

/**
 *  This class represents a local message exchanged between members and 
 *  the daemon.  It is used by members to multicast a message.
 *
 *  @author Alberto Montresor
 *  @since  Jgroup 1.2
 */
public final class MulticastRequest
  extends Event
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  private static final long serialVersionUID = 6936202150321190893L;

  /** Message codified as message payload */
  private OutMessage payload;
  
  /** True if the payload contains a single object */
  private boolean isObject;
  
  /** Identifier of the sender */
  private MemberId sender;
  
  /** Ack listener to be notified */
  private AckListener ackListener;

  /** The chain identifier for this multicast message (may be null) */
  private ChainIdentifier chainId;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructors
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Default constructor for externalization.
   */
  public MulticastRequest() { }


  /**
   *  Creates a <CODE>MulticastRequest</CODE> event.
   * 
   *  @param gid group identifier
   *  @param isObject true if the message contains a single message
   *  @param payload the payload of the message
   *  @param sender sender identifier
   *  @param ackListener  ack listener
   *  @param chainId the chain identifier
   */
  public MulticastRequest(int gid, boolean isObject, OutMessage payload,
      MemberId sender, AckListener ackListener, ChainIdentifier chId)
  {
    super(MULTICAST_REQUEST, gid);
    this.payload = payload;
    this.isObject = isObject;
    this.sender = sender;
    this.ackListener = ackListener;
    this.chainId = chId;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Accessor methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * 
   */
  public OutMessage getPayload()
  {
    return payload;
  }
  
  /**
   * 
   */
  public boolean isObject()
  {
    return isObject;
  }

  /**
   * 
   */
  public MemberId getSender()
  {
    return sender;
  }

  /**
   * 
   */
  public AckListener getAckListener()
  {
    return ackListener;
  }

  /**
   * 
   */
  public ChainIdentifier getChainId()
  {
    return chainId;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Object methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  public String toString()
  {
    StringBuilder buf = new StringBuilder();
    buf.append("[");
    buf.append(super.toString());
    buf.append(", sender=");
    buf.append(sender);
    buf.append(", isObj=");
    buf.append(isObject);
    buf.append(", ackListener=");
    buf.append(ackListener);
    buf.append(", payload=");
    buf.append(payload);
    if (chainId != null) {
      buf.append(", chainId=");
      buf.append(chainId);
    }
    return buf.toString();    
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from Externalizable
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Restores the content of this object from the marshalled data contained
   *  in the specified input stream.
   * 
   *  @param in the stream to be read
   */
  public void readExternal(ObjectInput in)
    throws IOException, ClassNotFoundException
  {
    super.readExternal(in);
    sender = new MemberIdImpl();
    sender.readExternal(in);
    isObject = Flag.unmarshal(in);
    ackListener = (AckListener) in.readObject();
    payload = (OutMessage) in.readObject();
    chainId = (ChainIdentifier) in.readObject();
  }
  
  /**
   *  Marshals the content of this object to the specified output stream.
   * 
   *  @param out the stream to be written
   */
  public void writeExternal(ObjectOutput out)
    throws IOException
  {
    super.writeExternal(out);
    sender.writeExternal(out);
    Flag.marshal(out, isObject);
    out.writeObject(ackListener);
    out.writeObject(payload);
    out.writeObject(chainId);
  }

} // END MulticastRequest

