/*
 * Copyright (c) 1998-2004 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.relacs.gm;

import java.rmi.RemoteException;
import java.rmi.server.ExportException;
import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.ThreadFactory;

import jgroup.core.ConfigManager;
import jgroup.core.JgroupException;
import jgroup.relacs.config.TransportConfig;
import jgroup.relacs.events.Event;
import jgroup.util.ThreadMonitor;
import net.jini.export.Exporter;
import net.jini.jeri.BasicILFactory;
import net.jini.jeri.BasicJeriExporter;
import net.jini.jeri.tcp.TcpServerEndpoint;

import org.apache.log4j.Logger;
import org.apache.log4j.MDC;

/**
 * Handle dispatching daemon events to other layers in the stack, e.g.,
 * the <code>MembershipLayer</code> and <code>MulticastLayer</code>.
 *
 * @author Hein Meling
 * @since Jgroup 0.1 (rewritten for Jgroup 3.0)
 */
public class NewDispatcherLayer
  implements DispatcherService, RemoteDispatcher, PingListener
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Logger
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Obtain logger for this class */
  private static final Logger log = Logger.getLogger(NewDispatcherLayer.class);


  //////////////////////////////////////////////////////////////////////////////////////////
  // Constants
  //////////////////////////////////////////////////////////////////////////////////////////

  /** The maximum number of event types that we allow */
  private static final int MAX_EVENT_TYPES = 20;


  //////////////////////////////////////////////////////////////////////////////////////////
  // Private fields
  //////////////////////////////////////////////////////////////////////////////////////////

  /** Last time when the dispatcher had interaction with the daemon */
  private volatile long lastPingReceived;

  /** Ping timeout value */
  private int daemonSuspectTimeout;

  /** The executor used to schedule events to the dispatcher listeners */
  private ExecutorService executorService;

  /** List of dispatcher listeners (indexed by their supported event types) */
  private final List<DispatcherListener> eventTypes;

  /** The exporter for this remote dispatcher object */
  private Exporter exporter;


  //////////////////////////////////////////////////////////////////////////////////////////
  // Constructors
  //////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Initializes a DispatcherLayer
   */
  private NewDispatcherLayer()
  {
    eventTypes = new ArrayList<DispatcherListener>(MAX_EVENT_TYPES);
    for (int i = 0; i < MAX_EVENT_TYPES; i++)
      eventTypes.add(null);

    /* Obtains configuration information */
    TransportConfig tconf =
      (TransportConfig) ConfigManager.getConfig(TransportConfig.class);
    daemonSuspectTimeout = tconf.getDaemonSuspectTimeout();
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Static factory
  ////////////////////////////////////////////////////////////////////////////////////////////

  public static NewDispatcherLayer getLayer()
  {
    return new NewDispatcherLayer();
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Layer methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /* (non-Javadoc)
   * @see jgroup.core.Layer#addListener(java.lang.Object)
   */
  public void addListener(Object listener)
  {
    if (listener instanceof DispatcherListener) {
      DispatcherListener dispListener = (DispatcherListener) listener;
      int[] events = dispListener.eventTypes();
      for (int i = 0; i < events.length; i++) {
        eventTypes.set(events[i], dispListener);
      }
    } else {
      throw new IllegalArgumentException("Specified listener is not a DispatcherListener");
    }
  }


  //////////////////////////////////////////////////////////////////////////////////////////
  // Methods from DispatcherService
  //////////////////////////////////////////////////////////////////////////////////////////

  /* (non-Javadoc)
   * @see jgroup.relacs.gm.DispatcherService#start(int)
   */
  public void start(final int gid)
  {
    ThreadFactory tFactory = new ThreadFactory() {
      public Thread newThread(Runnable r) {
        Thread dispatcher = new Thread(r, "Dispatcher-" + gid);
        dispatcher.setDaemon(true);
        ThreadMonitor.add(dispatcher);
        return dispatcher;
      }
    };
//    executorService = Executors.newSingleThreadExecutor(tFactory);
    executorService = Executors.newFixedThreadPool(3, tFactory);
  }

  /* (non-Javadoc)
   * @see jgroup.relacs.gm.DispatcherService#halt()
   */
  public void halt()
  {
    // Unexport the remote dispatcher
    exporter.unexport(true);
    executorService.shutdown();
  }

  /* (non-Javadoc)
   * @see jgroup.relacs.gm.DispatcherService#dispatch(java.lang.Object)
   */
  public void dispatch(Object obj)
  {
    if (log.isDebugEnabled())
      log.debug("Blocking on " + obj);
    synchronized (obj) {
      try {
        obj.wait(daemonSuspectTimeout);
      } catch (InterruptedException e) {
        if (log.isDebugEnabled())
          log.debug("Interrupted on " + obj);
        return;
      }
    }
    /* Check if the daemon is suspected to have failed. */
    long timePassed = System.currentTimeMillis() - lastPingReceived; 
    if (timePassed > daemonSuspectTimeout) {
      log.warn("Daemon suspected after: " + timePassed);
    } else {
      if (log.isDebugEnabled())
        log.debug("Unblocked on " + obj);
    }
  }

  /* (non-Javadoc)
   * @see jgroup.relacs.gm.DispatcherService#getRemoteDispatcher()
   */
  public RemoteDispatcher getRemoteDispatcher() throws JgroupException
  {
    exporter = new BasicJeriExporter(TcpServerEndpoint.getInstance(0),
        new BasicILFactory());
    try {
      return (RemoteDispatcher) exporter.export(this);
    } catch (ExportException e) {
      throw new JgroupException("Could not export the dispatcher", e);
    }
  }

  /* (non-Javadoc)
   * @see jgroup.relacs.gm.RemoteDispatcher#addEvent(jgroup.relacs.events.Event)
   */
  public void addEvent(final Event event) throws RemoteException
  {
    lastPingReceived = System.currentTimeMillis();
    /* Insert the event in the upcall queue */
    if (log.isDebugEnabled()) {
      MDC.put("group", "[Group: " + event.getGid() + "]");
      log.debug("Received daemon event: " + lastPingReceived + " " + event);
    }
    Runnable task = new Runnable() {
      public void run() {
        if (log.isDebugEnabled()) {
          MDC.put("group", "[Group: " + event.getGid() + "]");
          log.debug("Executing event: " + event);
        }
        DispatcherListener listener = eventTypes.get(event.getTag());
        listener.notifyEvent(event);
        if (log.isDebugEnabled())
          log.debug("Event execution completed");
      }
    };
    executorService.submit(task);
    if (log.isDebugEnabled())
      log.debug("Event submitted for execution");
  }

  /* (non-Javadoc)
   * @see jgroup.relacs.gm.PingListener#ping()
   */
  public void ping() throws RemoteException
  {
    lastPingReceived = System.currentTimeMillis();
    if (log.isDebugEnabled()) {
      log.debug("PingReceived: " + lastPingReceived);
    }
  }

  /* (non-Javadoc)
   * @see java.lang.Object#toString()
   */
  public String toString()
  {
    return " DISP ";
  }

} // END NewDispatcherLayer
