/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.relacs.gm;

import java.rmi.NotBoundException;
import java.rmi.RemoteException;

import jgroup.core.GroupManager;
import jgroup.core.IID;
import jgroup.core.InternalGMIService;
import jgroup.core.JgroupException;
import jgroup.core.MemberId;
import jgroup.core.MembershipListener;
import jgroup.core.MembershipService;
import jgroup.core.View;
import jgroup.core.registry.RegistryService;
import jgroup.util.log.Eventlogger;

import org.apache.log4j.Logger;


/**
 *  The <code>NotifyLayer</code> class keeps track of IIDs in 
 *  the current view, and compares this to the previous. If any
 *  differences are found, the <code>NotifyLayer</code> will 
 *  invoke an unbind on the dependable registry.
 *
 *  Note that for this layer should occur before the recovery
 *  layer in the layer stack.  This is since otherwise we may
 *  introduce a delay in the recovery layer view handling, and
 *  this may lead this layer to perform IGMI multicasts in such
 *  a manner that the membership service is unable to conclude
 *  a new view agreement protocol that has already been started,
 *  while processing the view in the recovery layer.
 *
 *  @author Henning Hommeland
 *  @author Jo Andreas Lind
 *  @since Jgroup 2.1
 */
public class NotifyLayer
  implements NotifyService, MembershipListener, InternalNotification
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Logger
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Obtain logger for this class */
  private static final Logger log = Logger.getLogger(NotifyLayer.class);


  //////////////////////////////////////////////////////////////////////////////////////////
  // Private fields
  //////////////////////////////////////////////////////////////////////////////////////////

  /** The IID for this layer */
  private IID iid;

  /** The RegistryService for this layer */
  private RegistryService rs;

  /** This table contains IIDs from the previous view */
  private Object[] iidTableOld = null;

  /** The GroupManager for this layer */
  private GroupManager gm;

  /**
   *  The internalNotifiaction will be used to invoke a internal 
   *  group method invocation to get the most recent IID of the other
   *  members so that the leader member can perform the unbind of
   *  crashed members.
   */
  private InternalNotification internalNotification;

  /** We need the local MemberId to determine its position in the current view */
  private MemberId id;


  //////////////////////////////////////////////////////////////////////////////////////////
  // Constructors
  //////////////////////////////////////////////////////////////////////////////////////////
	
  /**
   *  Initializes a NotifyLayer
   */
  public NotifyLayer(GroupManager gm, RegistryService rs, MembershipService ms)
  {
    this.gm = gm;
    this.rs = rs;
    this.id = ms.getMyIdentifier();
  }
	
	
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Static factory
  ////////////////////////////////////////////////////////////////////////////////////////////

  public static NotifyLayer getLayer(GroupManager gm, RegistryService rs, 
                                     MembershipService ms, InternalGMIService igmi)
  {
    return new NotifyLayer(gm, rs, ms);
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Layer methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  public void addListener(Object listener) {}


  ////////////////////////////////////////////////////////////////////////////////////////////
  // FinalizeLayer interface methods (invoked to complete the group manager construction)
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Complete the replica initialization, once all group manager components
   *  have been initialized.
   */
  public void complete(Object server)
    throws JgroupException
  {
    iid = rs.getIID();
    if (log.isDebugEnabled()) {
      log.debug("Binding identifier for this group manager: " + iid);
    }
    internalNotification = (InternalNotification) gm.getService(InternalNotification.class);
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from InternalNotification
  ////////////////////////////////////////////////////////////////////////////////////////////	

  /**
   *  This method returns the IID for this layer
   */
  public Object getIID()
    throws RemoteException
  {
    if (log.isDebugEnabled()) {
      log.debug("Returning binding identifier for this group manager: " + iid);
    }
    return iid;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from MembershipListener
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  The viewChange() method is invoked on the whole server
   *  group when a view change occur.
   */
  public void viewChange(View view)
  {
    try {
      Object[] iidTableNew = (Object[]) internalNotification.getIID();

      if (log.isDebugEnabled()) {
        for (int i = 0; i < iidTableNew.length; i++) {
          log.debug("new iids: " + iidTableNew[i]);
        }
      }

      /* The old table needs to be set in the first view change */
      if (iidTableOld == null)
        iidTableOld = iidTableNew;

      /* Compare old table with the new table and check unbind if necessary */
      if (view.memberHasPosition(-1, id) && iidTableOld.length >= iidTableNew.length) {

        boolean found = false;
        try {
          for (int i = 0; i < iidTableOld.length; i++) {
            found = false;
            for (int j = 0; j < iidTableNew.length && found == false; j++) {
              if (iidTableOld[i].equals(iidTableNew[j]))
                found = true;
            }
            if (!found) {
              if (Eventlogger.ENABLED)
                Eventlogger.logEventFlush("Unbinding: " + iidTableOld[i]);
              rs.unbind((IID) iidTableOld[i]);
              if (Eventlogger.ENABLED)
                Eventlogger.logEventFlush("Unbound");
            }
          }
        } catch (JgroupException je) {
          log.warn("Could not unbind", je);
        } catch (NotBoundException nbe) {
          log.warn("Could not unbind; not bound anymore", nbe);
        }

      }
      iidTableOld = iidTableNew;
    } catch (RemoteException re) {
      log.warn("Could not obtain iid or unbind", re);
    } 
  }

  public void hasLeft() {}

  public void prepareChange() {}

} // End NotifyLayer