/*
 * Copyright (c) 1998-2006 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.relacs.gm;

import java.rmi.RemoteException;
import java.util.HashMap;
import java.util.Map;

import jgroup.core.GroupManager;
import jgroup.core.InternalGMIService;
import jgroup.core.JgroupException;
import jgroup.core.MembershipListener;
import jgroup.core.MembershipService;
import jgroup.core.View;

import org.apache.log4j.Logger;

/**
 * The <code>TimestampLayer</code> creates an agreed timestamp
 * between all the members in a group.<p>
 * 
 * The timestamp can be retrieved as a counter (not unique between 
 * groups) or as a timestamp object (unique between groups).
 *
 * @author Rohnny Moland
 */
public class TimestampLayer 
  implements TimestampService, MembershipListener, InternalTimestamp
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Logger
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Obtain logger for this class */
  private static final Logger log = Logger.getLogger(TimestampLayer.class);


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** The GroupManager object */
  private GroupManager gm;

  /** The service proxies */
  private InternalTimestamp internalTimestamp;

  /** Holds the local timestamp value */  
  private long timestamp = 0;

  /** The timestamp objects */  
  private Map<Integer, Timestamp> timestamps = new HashMap<Integer, Timestamp>();

  /** The view size of the client group */
  private int vsize;

  /*
   * Initialize a timestamp layer. 
   */
  private TimestampLayer(GroupManager gm)
  {
    this.gm = gm;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Static factory
  ////////////////////////////////////////////////////////////////////////////////////////////

  public static TimestampLayer getLayer(GroupManager gm, MembershipService ms, InternalGMIService igmi) 
  {
    return new TimestampLayer(gm);
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from Layer
  ////////////////////////////////////////////////////////////////////////////////////////////

  /* (non-Javadoc)
   * @see jgroup.core.Layer#addListener(java.lang.Object)
   */
  public void addListener(Object listener) {}


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from FinalizeLayer
  ////////////////////////////////////////////////////////////////////////////////////////////

  /* (non-Javadoc)
   * @see jgroup.core.FinalizeLayer#complete(java.lang.Object)
   */
  public void complete(Object server)
    throws JgroupException
  {
    internalTimestamp = (InternalTimestamp) gm.getService(InternalTimestamp.class);
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from MembershipListener
  ////////////////////////////////////////////////////////////////////////////////////////////

  /* (non-Javadoc)
   * @see jgroup.core.MembershipListener#viewChange(jgroup.core.View)
   */
  public void viewChange(View view) 
  {
    vsize = view.size();
    for (Timestamp tsObj : timestamps.values()) {
      tsObj.viewChange(view);
    }
    //FIXME must also resync the timestamp field.
    try {
      Object[] tsMap = (Object[]) internalTimestamp.getInternalTimestamp();
      for (int i = 0; i < tsMap.length; i++) {
        for (Integer id : timestamps.keySet()) {
          // Get the local timestamp for id
          Timestamp ts = timestamps.get(id);
          // Get the remote timestamp for member host i for id
          Timestamp t = (Timestamp) ((Map) tsMap[i]).get(id);
          ts = ts.greatest(t);
          timestamps.put(id, ts);
        }
      }
    } catch (RemoteException e) {
      log.error("Failed to resync the timestamp object", e);
    }
  }


  /* (non-Javadoc)
   * @see jgroup.core.MembershipListener#prepareChange()
   */
  public void prepareChange() {}


  /* (non-Javadoc)
   * @see jgroup.core.MembershipListener#hasLeft()
   */
  public void hasLeft() {}


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from InternalTimestamp
  ////////////////////////////////////////////////////////////////////////////////////////////


  /* (non-Javadoc)
   * @see jgroup.relacs.gm.InternalTimestamp#getInternalTimestamp()
   */
  public Object getInternalTimestamp() 
    throws RemoteException 
  {
    return timestamps;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from TimestampService
  ////////////////////////////////////////////////////////////////////////////////////////////

  /* (non-Javadoc)
   * @see jgroup.relacs.gm.TimestampService#getTimestamp(int)
   */
  public synchronized Timestamp getTimestamp(int id) 
  {
    Timestamp ts = timestamps.get(id);
    if (ts == null) {
      ts = new Timestamp(gm.getGroupId(), id, vsize);
      timestamps.put(id, ts);
    }
    ts.updateTimestamp();
    if (log.isDebugEnabled())
      log.debug("TimestampLayer.getTimestamp(): " + ts);
    return ts;
  }


  /* (non-Javadoc)
   * @see jgroup.relacs.gm.TimestampService#getTimestampCounter()
   */
  public long getTimestampCounter()
  {
    if (log.isDebugEnabled())
      log.debug("TimestampLayer.getTimestampCounter() ts=" + timestamp);
    return timestamp++;
  }

} // END TimestampLayer
