/*
 * Copyright (c) 1998-2004 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */
package jgroup.relacs.gm;

import java.io.Serializable;

import jgroup.core.Layer;
import jgroup.core.View;
import jgroup.core.Layer.FinalizeLayer;

/**
 * Service interface for the timestamp layer.
 * 
 * @author Rohnny Moland
 */
public interface TimestampService
  extends Layer, FinalizeLayer
{
  /**
   * Returns a <code>Timestamp</code> object, and updates
   * the timestamp counter.
   */
  public Timestamp getTimestamp(int id);

  /**
   * Returns the timestamp counter value, and updates the
   * timestamp counter.  This is useful for applications that
   * do not need distinguish between groups and different views.
   * That is, if the application is such that the counter value
   * will only be used within the same group, and that view
   * changes are handled otherwise.
   */
  public long getTimestampCounter();


  /**
   * The timestamp object is used to detect duplicate invocations
   * from several members of the same group, and yet distinguish
   * the invocations amongst multiple groups.<p>
   */
  public class Timestamp
    implements Serializable
  {
    private static final long serialVersionUID = 1868153781418135225L;

    private long clientGroupViewId;
    private long timestamp;
    public int viewSize;
    private int gid;
    private int id;

    Timestamp(int gid, int id, int vsize)
    {
      this.gid = gid;
      this.id = id;
      this.viewSize = vsize;
    }

    /**
     * Invoked by the <code>TimestampLayer</code> of the client group
     * in response to a view change in the client group.
     */
    void viewChange(View view)
    {
      clientGroupViewId = view.getVid();
      viewSize = view.size();
    }

    void updateTimestamp()
    {
      timestamp++;
    }

    /**
     * Returns the timestamp object that is the greatest among the given
     * timestamp object and this timestamp object.
     */
    Timestamp greatest(Timestamp ts)
    {
      if (timestamp >= ts.timestamp)
        return this;
      else
        return ts;
    }

    public boolean equals(Object obj) 
    {
      if (obj instanceof Timestamp) {
        Timestamp t = (Timestamp) obj;
        if (timestamp == t.timestamp && gid == t.gid && id == t.id)
          return true;
        return false;
      }
      return false;
    }

    public int hashCode() 
    {
      return gid ^ viewSize ^ id
        ^ (int)(timestamp ^ (timestamp >>> 32))
        ^ (int)(clientGroupViewId ^ (clientGroupViewId >>> 32));
    }

    public String toString()
    {
      return "[ts=" + gid + ", " + timestamp + ", " + id + ", " + hashCode() + "]";
    }
  } // END Timestamp

} // End TimestampService
