/*
 * Copyright (c) 1998-2004 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.relacs.gmi;

import java.rmi.RemoteException;
import java.rmi.server.ExportException;

import jgroup.core.MemberId;
import jgroup.core.View;
import jgroup.core.multicast.AckListener;
import net.jini.export.Exporter;
import net.jini.jeri.BasicILFactory;
import net.jini.jeri.BasicJeriExporter;
import net.jini.jeri.tcp.TcpServerEndpoint;

import org.apache.log4j.Logger;
import org.apache.log4j.MDC;

/**
 * Handler for multicast return messages from all members of the current view.
 *
 * @author Hein Meling
 * @author Tor Arve Stangeland
 */
public class GroupAckListenerOld
  implements AckListener
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Logger
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Obtain logger for this class */
  private static final Logger log = Logger.getLogger(GroupAckListenerOld.class);


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constants
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** States */
  private static final int STARTING = 1;

  private static final int WAITING = 2;

  private static final int COMPLETED = 3;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Current state */
  private volatile int state = STARTING;

  /** Number of unacknowledged messages */
  private volatile int missing;

  /** Results from servers */
  protected Object[] results;

  /** Flag completed servers */
  protected boolean[] completed;

  /** The BasicJeriExporter used by this AckListener */
  private Exporter exporter;

  /** The object on which to synchronize */
  private final Object lockObj;

  /** The group identifier associated with this AckListener; initialized in notifyView() */
  private int groupId;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructor
  ////////////////////////////////////////////////////////////////////////////////////////////

  public GroupAckListenerOld(Object lockObj)
  {
    this.lockObj = lockObj;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Get a reference for the remote <code>AckListener</code>.
   * 
   * @return A remote proxy for the <code>AckListener</code>.
   * @throws ExportException
   *   Raised if the <code>AckListener</code> could not be exported.
   */
  public AckListener getRemoteAckListener()
    throws ExportException
  {
    /*
     * Create a new exporter on any available port for this AckListener, and
     * enable the distributed garbage collection (DGC), and enable the keep
     * alive attribute.
     */
    exporter = new BasicJeriExporter(TcpServerEndpoint.getInstance(0),
        new BasicILFactory(), true, true);
    return (AckListener) exporter.export(this);
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from AckListener
  ////////////////////////////////////////////////////////////////////////////////////////////

  /* (non-Javadoc)
   * @see jgroup.core.multicast.AckListener#ack(jgroup.core.MemberId, int, java.lang.Object)
   */
  public void ack(MemberId id, int pos, Object obj)
    throws RemoteException
  {
    if (log.isDebugEnabled()) {
      MDC.put("group", "[Group: " + groupId + "]");
      log.debug("Got ACK from " + id + ", pos=" + pos + ", reply=" + obj);
    }
    completed[pos] = true;
    results[pos] = obj;
    missing--;
    updateState();
  }

  /* (non-Javadoc)
   * @see jgroup.core.multicast.AckListener#notifyView(jgroup.core.View)
   */
  public void notifyView(View view)
    throws RemoteException
  {
    groupId = view.getGid();
    if (log.isDebugEnabled())
      MDC.put("group", "[Group: " + groupId + "]");
    missing = view.size();
    completed = new boolean[missing];
    results = new Object[missing];
    state = WAITING;
    if (log.isDebugEnabled())
      log.debug(this);
  }

  /* (non-Javadoc)
   * @see jgroup.core.multicast.AckListener#viewChange()
   */
  public void viewChange()
    throws RemoteException
  {
    if (log.isDebugEnabled()) {
      MDC.put("group", "[Group: " + groupId + "]");
      log.debug(this);
    }
    if (missing > 0) {
      for (int i = 0; i < completed.length; i++) {
        if (!completed[i]) {
          completed[i] = true;
          results[i] = new RemoteException("Member not reachable");
          if (log.isDebugEnabled())
            log.debug("Unreachable member at position " + i);
        }
      }
    }
    missing = 0;
    updateState();
    // Unexport this remote AckListener; no more remote calls should occur
    exporter.unexport(false);
  }

  /**
   * Update the internal state.
   */
  private void updateState()
  {
    synchronized (lockObj) {
      if (missing == 0) {
        state = COMPLETED;
        if (log.isDebugEnabled())
          log.debug(this);
        lockObj.notifyAll();
      } else {
        if (log.isDebugEnabled())
          log.debug(this);
      }
    }
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods for accessing the results array
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Retrieves the invocation result of any completed member.
   */
  public Object getResult()
  {
    pendingCompletion();
    // Select first completed result that is not a RemoteException
    for (int i = 0; i < results.length; ++i)
      if (completed[i] && !(results[i] instanceof RemoteException))
        return results[i];

    /*
     * If no actual results were returned, check if there is a RemoteException
     * caused by a viewChange() and if so wrap it in a InvocationResult object
     * for delivery to the client.
     */
    for (int i = 0; i < results.length; ++i)
      if (completed[i] && results[i] instanceof RemoteException)
        return new InvocationResult((RemoteException) results[i]);

    log.warn("No results have been received!");
    return null;
  }

  /**
   * Retrieves the invocation result obtained from the leader.
   */
  public Object getLeaderResult()
  {
    pendingCompletion();
    if (completed[0]) {
      if (results[0] instanceof RemoteException)
        return new InvocationResult((RemoteException) results[0]);
      else
        return results[0];
    }

    log.warn("No results from leader found!");
    return null;
  }

  /**
   * Retrieves the whole array of invocation results (from all members).
   * The results array <i>may</i> contain one or more <code>Exception</code>s.
   */
  public Object[] getResults()
  {
    pendingCompletion();
    return results;
  }

  /**
   * Blocks until the state of the listener is <code>COMPLETED</code>.
   */
  protected void pendingCompletion()
  {
    if (log.isDebugEnabled())
      log.debug(this);
    synchronized (lockObj) {
      while (state != COMPLETED) {
        try {
          lockObj.wait();
        } catch (InterruptedException e) {
        }
        if (log.isDebugEnabled())
          log.debug(this);
      }
    }
    /*
     * Unexport this remote AckListener; note that calls to the
     * AckListener.viewChange() method may occur after the remote AckListener
     * has been unexported. This should be dealt with at the client-side, as
     * this will cause a NoSuchObjectException
     */
    exporter.unexport(false);
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from Object
  ////////////////////////////////////////////////////////////////////////////////////////////

  public String toString()
  {
    StringBuilder b = new StringBuilder("GroupAckListener: missing=");
    b.append(missing);
    switch (state) {
      case STARTING:
        b.append(", STARTING");
        break;

      case WAITING:
        b.append(", WAITING");
        break;

      case COMPLETED:
        b.append(", COMPLETED");
        break;
    }
    b.append(", hash=");
    b.append(hashCode());
    return b.toString();
  }

} // END GroupAckListener
