/*
 * Copyright (c) 1998-2003 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */
package jgroup.relacs.gmi;

import java.io.Externalizable;
import java.io.IOException;
import java.io.ObjectInput;
import java.io.ObjectInputStream;
import java.io.ObjectOutput;
import java.io.ObjectOutputStream;
import java.io.OutputStream;
import java.lang.reflect.Method;
import java.lang.reflect.Proxy;
import java.rmi.RemoteException;

import jgroup.core.GroupManager;
import jgroup.core.IID;
import jgroup.core.JgroupException;
import jgroup.core.ExternalGMIService.BootstrapInvocationHandler;
import jgroup.relacs.gm.TimestampService;
import jgroup.relacs.gm.TimestampService.Timestamp;
import jgroup.relacs.rmi.IIDFactory;
import jgroup.util.log.Eventlogger;
import net.jini.jeri.ObjectEndpoint;
import net.jini.jeri.OutboundRequest;
import net.jini.jeri.OutboundRequestIterator;

import org.apache.log4j.Logger;

import com.sun.jini.jeri.internal.runtime.Util;


/**
 * Invocation handler (the proxy) at client side.
 * Marshalls call and passes it down to lower layers.
 *
 * @author Tor Arve Stangeland
 * @author Hein Meling
 */
public class GroupInvocationHandler
  implements BootstrapInvocationHandler, Externalizable
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Logger
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Obtain logger for this class */
  private static final Logger log = Logger.getLogger(GroupInvocationHandler.class);


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Static section
  ////////////////////////////////////////////////////////////////////////////////////////////

  private static final long serialVersionUID = 6090074369501693886L;

  /** Invocation handler type */
  public static final String TYPE = "GroupInvocationHandler";


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** GroupEndPoint associated with this invocation handler */
  private GroupEndPoint gep;

  /** Caching of method hash */
  private MethodTable methodTable;

  /** IID generator factory */
  private IIDFactory generator;

  /** The view identifier of the server group known to the client */
  private long clientViewId = -1;

  /** The timestamp service obtained from the client group */
  private TimestampService timestampService;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructors
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Constructor for externalization
   */
  public GroupInvocationHandler() {}

  /**
   * Construct a new GroupInvocationHandler for the given group ObjectEndpoint
   * @param oe the group object endpoint
   * @param methodTable the table of remote methods accessible through this
   *   invocation handler. 
   */
  public GroupInvocationHandler(ObjectEndpoint oe, MethodTable methodTable)
  {
    this.gep = (GroupEndPoint) oe;
    this.methodTable = methodTable;
    generator = new IIDFactory();
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from InvocationHandler
  ////////////////////////////////////////////////////////////////////////////////////////////

  /* (non-javadoc)
   * @see java.lang.reflect.InvocationHandler#invoke(Object, Method, Object[])
   */
  public Object invoke(Object proxy, Method method, Object[] args)
    throws Exception
  {
    if (method.getDeclaringClass() == Object.class) {
      return invokeObjectMethod(proxy, method, args);
    } else 
      return invokeRemoteMethod(proxy, method, args);    
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Instance methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Handles java.lang.Object methods.
   */
  private Object invokeObjectMethod(Object proxy, Method method, Object[] args) 
  {
    String name = method.getName();

    if (name.equals("hashCode")) {
      return new Integer(hashCode());      
    } else if (name.equals("equals")) {
      Object obj = args[0];
      boolean b = (proxy == obj) ||
        ((obj != null) && Util.sameProxyClass(proxy, obj) &&
          equals(Proxy.getInvocationHandler(obj)));
      return Boolean.valueOf(b);
    } else if (name.equals("toString")) {
      return proxyToString(proxy);
    } else {
      throw new IllegalArgumentException("Unexpected object method: " + method);
    }
  }


  /**
   * Returns a hash code value for this invocation handler.
   */
  public int hashCode() 
  {
    return gep.hashCode();  
  }


  /**
   * Returns a string representation for a proxy that uses this invocation
   * handler.
   **/
  private String proxyToString(Object proxy) 
  {
    Class[] interfaces = proxy.getClass().getInterfaces();

    if (interfaces.length == 0) {
      return "Proxy[" + this + "]";
    }
    String iface = interfaces[0].getName();
    int dot = iface.lastIndexOf('.');

    if (dot >= 0) {
      iface = iface.substring(dot + 1);
    }
    return "Proxy[" + iface + "," + gep + "]";
  }


  /**
   * Handles remote methods.
   */
  private Object invokeRemoteMethod(Object proxy, Method method, Object[] args)
    throws Exception
  {
    IID iid = generator.getId();
    MethodDetails info = methodTable.get(method);
    boolean invoked = false;
    InvocationResult result = null;

    do {
      // Get request iterator
      OutboundRequestIterator iter = gep.newCall(info.getInvocationSemantics());
      while (result == null && iter.hasNext()) {
        try {
          result = invokeRequest(iid, info, args, iter.next());
        } catch (IOException ioex) {
          /*
           * We just log it; the GroupEndPoint is responsible for
           * checking if group members are reachable, and updating
           * them if they are not.
           */
          if (Eventlogger.ENABLED)
            Eventlogger.logEvent("FailureDetected");
          if (log.isDebugEnabled())
            log.debug("IOException caught", ioex);
        }
      }
      
      if (result != null)
        invoked = true;
    } while (!invoked); // Keep retrying until exception is thrown or we succeed
    gep.updateClientView(result.getServerView());
    clientViewId = result.getServerViewId();
//    gep.notifyView(clientViewId);
    generator.addReply(iid);
    return result.getResult();    
  }


  /**
   * Try to invoke method using given request
   * @param iid Invocation Identifier for invocation
   * @param minfo Method to be invoked
   * @param args Arguments to pass to method
   * @param req OutboundRequest to use
   */
  protected InvocationResult invokeRequest(IID iid, MethodDetails minfo, Object[] args, OutboundRequest req)
    throws IOException, JgroupException
  {
    if (log.isDebugEnabled())
      log.debug("Invoking: " + minfo + " with iid=" + iid);

    OutputStream os = req.getRequestOutputStream();
    /*
     * Only the invocation semantics is written to the output stream,
     * since this is required to be unwrapped at the receiver.
     */
    os.write(minfo.getInvocationSemantics().ordinal());
    /*
     * The following is written to an object output stream that is
     * not unwrapped immediately at the receiver (in case it needs to
     * be multicast to the other group members.)  That is, if the method
     * needs to be multicasted, the following object stream is passed on
     * to the other members without unmarshalling and remarshalling.
     */
    ObjectOutputStream out = new ObjectOutputStream(os);
    // Write IID for this request
    out.writeObject(iid);

    // Write the timestamp object
    if (timestampService != null) {
      synchronized (timestampService) {
        // Write the timestamp object
        Timestamp ts = timestampService.getTimestamp(gep.getGroupId());
        if (log.isDebugEnabled())
          log.debug("GroupInvocationHandler.invoke ts = " + ts + " on " + minfo);
        out.writeObject(ts);
      }
    } else {
      out.writeObject(null);
    }
    // Write the client-side view
    out.writeLong(clientViewId);
    // Write method hash
    out.writeLong(minfo.getHash());    
    // Write args
    for (int i = 0; args != null && i < args.length; ++i) 
      out.writeObject(args[i]);
    out.flush();
    out.close();

    if (log.isDebugEnabled())
      log.debug("Data sent for " + minfo);

    // Execute call and await reply now that we have closed OutboundRequest.OutputStream.
    RemoteException remex = gep.executeCall(req);
    if (remex != null)
      throw remex;

    // Unmarshall results
    ObjectInputStream in = new ObjectInputStream(req.getResponseInputStream());
    if (log.isDebugEnabled())
      log.debug("Result received for " + minfo);
    try {
      return (InvocationResult) in.readObject();
    } catch (ClassNotFoundException cnfex) {
      throw new JgroupException("Returned class could not be deserialized!", cnfex);
    }
  }


  public GroupEndPoint getGroupEndPoint()
  {
    return gep;
  }


  /**
   * Initialize the timestamp service.  This is required when the proxy
   * is being used by a replicated client group.
   * 
   * @param gm
   *   The group manager of the client group.
   */
  public void setGroupManager(GroupManager gm)
  {
    timestampService = (TimestampService) gm.getService(TimestampService.class);
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from Externalization
  ////////////////////////////////////////////////////////////////////////////////////////////

  /* (non-Javadoc)
   * @see java.io.Externalizable#readExternal(java.io.ObjectInput)
   */
  public void readExternal(ObjectInput in)
    throws IOException, ClassNotFoundException
  {
    gep = new GroupEndPoint();
    gep.readExternal(in);
    methodTable = (MethodTable) in.readObject();
    generator = new IIDFactory();
  }

  /* (non-Javadoc)
   * @see java.io.Externalizable#writeExternal(java.io.ObjectOutput)
   */
  public void writeExternal(ObjectOutput out)
    throws IOException
  {
    gep.writeExternal(out);
    out.writeObject(methodTable);
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from BootstrapInvocationHandler
  ////////////////////////////////////////////////////////////////////////////////////////////

  public String getType()
  {
    return TYPE;
  }

  public void merge(BootstrapInvocationHandler handler)
    throws JgroupException, RemoteException
  {
    if (!handler.getType().equals(TYPE))
      throw new JgroupException("Incompatible type: " + handler.getType());
    GroupInvocationHandler h = (GroupInvocationHandler) handler;
    gep.addMembers(h.gep.getMembers());
  }

} // END GroupInvocationHandler
