/*
 * Copyright (c) 1998-2003 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */
package jgroup.relacs.gmi;

import java.io.IOException;
import java.io.InputStream;
import java.io.ObjectOutputStream;
import java.io.OutputStream;
import java.util.EnumMap;
import java.util.HashMap;
import java.util.Map;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.locks.Condition;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReentrantLock;

import jgroup.core.MemberId;
import jgroup.relacs.gmi.protocols.ProtocolDispatcher;
import net.jini.jeri.Endpoint;
import net.jini.jeri.InboundRequest;
import net.jini.jeri.RequestDispatcher;
import net.jini.jeri.ServerEndpoint;
import net.jini.jeri.ServerEndpoint.ListenCookie;
import net.jini.jeri.ServerEndpoint.ListenEndpoint;
import net.jini.jeri.tcp.TcpServerEndpoint;

import org.apache.log4j.Logger;
import org.apache.log4j.MDC;
import org.apache.log4j.NDC;

/**
 * Handles external group invocation requests.
 *
 * @author Tor Arve Stangeland
 * @author Hein Meling
 */
public class GroupRequestHandler
  implements RequestDispatcher
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Logger
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Obtain logger for this class */
  private static final Logger log = Logger.getLogger(GroupRequestHandler.class);


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Invocation dispatcher  */
  private final Map<ListenEndpoint,Record> handles = new HashMap<ListenEndpoint,Record>();

  /** Mapping from a method semantics to the corresponding protocol dispatcher. */
  private final EnumMap<MethodSemantics,ProtocolDispatcher> protocols =
    new EnumMap<MethodSemantics,ProtocolDispatcher>(MethodSemantics.class);

  /** My member identifier */
  private final MemberId me;

  /** Server group view identifier */
  private final int groupId;

  /** Server endpoint */
  private TcpServerEndpoint serverEndpoint;

  /**
   * True if invocation requests should be blocked awaiting a new view.
   * Initially, we do not accept any invocations until we have installed a view.
   */
  private volatile boolean blockInvocation = true;

  /** Lock and condition used to enable/disable method dispatching */
  private final Lock lock = new ReentrantLock();
  private final Condition notDispatching = lock.newCondition();


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructors
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Constructs a new request handler for the group.
   */
  public GroupRequestHandler(int gid, MemberId me)
  {
    this.groupId = gid;
    this.me = me;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Public methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Returns true if the given method semantics has registered its
   * protocol instance.
   */
  public boolean hasProtocol(MethodSemantics semantics)
  {
    return protocols.containsKey(semantics);
  }

  /**
   * Associate the given protocol with the given method invocation semantics
   * as stored in the internal table of protocols.  The method will override
   * any existing mapping for the given invocation semantics.
   */
  public void addProtocol(MethodSemantics semantics, ProtocolDispatcher protocol)
  {
    protocols.put(semantics, protocol);
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from net.jini.jeri.RequestDispatcher
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Handles inbound requests to this local endpoint (a member of a group).
   * 
   * @see net.jini.jeri.RequestDispatcher#dispatch(InboundRequest)
   */
  public void dispatch(InboundRequest request)
  {
    if (log.isDebugEnabled())
      MDC.put("group", "[Group: " + groupId + "]");
    doDispatch(request);
  }


  /**
   * Handles inbound requests to this local endpoint (a member of a group).
   * 
   * @see net.jini.jeri.RequestDispatcher#dispatch(InboundRequest)
   */
  public void BLOCKING_dispatch(InboundRequest request)
  {
    if (log.isDebugEnabled())
      MDC.put("group", "[Group: " + groupId + "]");

    /*
     * Check if the invocation should be blocked to await a new view installation.
     * Allowing layers (e.g. the merging layer) to complete the updating of the
     * server state before performing new invocations which may itself update the
     * server state.
     * 
     * Note that multiple threads (inbound invocation requests) may be blocked
     * here, pending execution on some local server.  We have no control over
     * which of the pending requests are executed first after the invocation
     * dispatching is again unblocked.
     */
    lock.lock();
    try {
      while (blockInvocation) {
        if (log.isDebugEnabled())
          log.debug("Blocking invocation " + request);
        // wait for signal to start dispatching again
//        notDispatching.awaitUninterruptibly();

        // wait for signal or timeout; and start dispatching again
        try {
          boolean elapsed = !notDispatching.await(5, TimeUnit.SECONDS);
          if (elapsed) {
            // Unblock since we haved waited for 5 seconds.
            blockInvocation = false;
            log.warn("UNBLOCKED DUE TO TIMEOUT");
          }
        } catch (InterruptedException e) { }

        if (!blockInvocation && log.isDebugEnabled())
          log.debug("Unblocking invocation " + request);
      }
    } finally {
      lock.unlock();
    }
    doDispatch(request);
  }


  /**
   * Handles inbound requests to this local endpoint (a member of a group).
   * 
   * @see net.jini.jeri.RequestDispatcher#dispatch(InboundRequest)
   */
  public void CLIENT_BLOCKING_dispatch(InboundRequest request)
  {
    if (log.isDebugEnabled())
      MDC.put("group", "[Group: " + groupId + "]");

    /*
     * Check if the invocation should be blocked to await a new view installation.
     * Allowing layers (e.g. the merging layer) to complete the updating of the
     * server state before performing new invocations which may itself update the
     * server state.
     * 
     * Note that multiple threads (inbound invocation requests) may be blocked
     * here, pending execution on some local server.  We have no control over
     * which of the pending requests are executed first after the invocation
     * dispatching is again unblocked.
     */
    if (blockInvocation) {
      if (log.isDebugEnabled())
        log.debug("Blocked invocation " + request);
      OutputStream out = request.getResponseOutputStream();
      try {
        out.write(InvocationResult.INVOCATION_BLOCKED);
        // Flush and close the return object stream
        out.flush();
        out.close();
      } catch (IOException ioex) {
        // If an exception is thrown, something is wrong with the network (?)
        log.warn("Failed to perform method invocation; aborting...", ioex);
        request.abort();
      }
    } else {
      doDispatch(request);
    }
  }

  /**
   * Perform the method dispatching according to the protocol type incorporated
   * within the inbound request object.  Note that methods may be dispatched in
   * such a way that they interleave with other invocations using the same or a
   * different protocol at the local endpoint.
   * 
   * Methods that cannot tolerate the above concurrency behavior must implement
   * its own synchronization policy (locking synchronous access to its variables)
   * at the server level, and may in addition use the Atomic protocol.
   */
  private void doDispatch(InboundRequest request)
  {
    // Get in/out streams from the inbound request object
    InputStream in = request.getRequestInputStream();
    OutputStream out = request.getResponseOutputStream();
    try {
      boolean invoked = false;
      InvocationResult res = null;
      int ordinal = in.read();
      try {
        MethodSemantics semantics = MethodSemantics.valueOf(ordinal);
        // no need to synchronize access to 'protocols' since it is not modified after init
        ProtocolDispatcher protocol = protocols.get(semantics);
        if (protocol == null) {
          log.warn("No protocol registered for: " + semantics);
          out.write(InvocationResult.UNKNOWN_INVOCATION_SEMANTICS);
        } else {
          if (log.isDebugEnabled())
            NDC.push(semantics.toString());
          /*
           * Note that we do not synchronize here to block invocations if other
           * requests are being handled; this has to be handled by the server
           * implementations themselves.
           */
          res = protocol.dispatch(in);
          if (log.isDebugEnabled())
            NDC.pop();
          invoked = true;
        }
      } catch (ArrayIndexOutOfBoundsException e) {
        log.warn("Unknown invocation semantics in call: " + ordinal);
        out.write(InvocationResult.UNKNOWN_INVOCATION_SEMANTICS);
      }

      if (invoked) {
        if (res != null) {
          // Write result to stream if successful
          out.write(InvocationResult.INVOCATION_COMPLETED);
          ObjectOutputStream oout = new ObjectOutputStream(out);
          oout.writeObject(res);
        } else {
          log.warn("Invocation failed!");
          out.write(InvocationResult.INVOCATION_FAILED);
        }
      }

      // Flush and close the return object stream
      out.flush();
      out.close();
    } catch (IOException ioex) {
      // If an exception is thrown, something is wrong with the network (?)
      log.warn("Failed to perform method invocation; aborting...", ioex);
      request.abort();
    }
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Public methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Update local view, and unblock invocation dispatching.
   */
  void enableDispatching()
  {
    if (log.isDebugEnabled())
      log.debug("Enable invocation dispatching");
    lock.lock();
    try {
      // Unblock all pending invocations
      blockInvocation = false;
      notDispatching.signalAll();
    } finally {
      lock.unlock();
    }
  }

  /**
   * To block invocation dispatching during view change periods.
   */
  void disableDispatching()
  {
    if (log.isDebugEnabled())
      log.debug("Disable invocation dispatching");
    lock.lock();
    try {
      // Block new invocations from now on and until enableDispatching is called
      blockInvocation = true;
    } finally {
      lock.unlock();
    }
  }

  Endpoint initServerEndpoint()
    throws IOException
  {
    if (serverEndpoint == null) {
      int port = me.getServerPort();
      // Create a ServerEndpoint on the given port.
      serverEndpoint = TcpServerEndpoint.getInstance(port);
      if (log.isDebugEnabled())
        log.debug("Created server endpoint on port: " + port);
    }
    // Enumerate endpoints
    return serverEndpoint.enumerateListenEndpoints(new GroupRequestHandler.ListenContext());
  }

  ServerEndpoint getServerEndpoint()
  {
    return serverEndpoint;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Nested class for holding the listen handle and cookie
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Describe this class */
  private class Record
  {
    public ServerEndpoint.ListenHandle handle;
    public ServerEndpoint.ListenCookie cookie;
  }
  

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Nested class implementing the ServerEndpoint.ListenContext interface
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Internal listen context
   */
  private final class ListenContext
    implements ServerEndpoint.ListenContext
  {

    /**
     * @see net.jini.jeri.ServerEndpoint.ListenContext#addListenEndpoint(ListenEndpoint)
     */
    public ListenCookie addListenEndpoint(ListenEndpoint listenEndpoint)
      throws IOException 
    {
      Record r = handles.get(listenEndpoint);
      if (r == null) {
        // Start listening
        r = new Record();
        r.handle = listenEndpoint.listen(GroupRequestHandler.this);
        r.cookie = r.handle.getCookie();
        handles.put(listenEndpoint, r);
      }
      return r.cookie;
    }

  } // END ListenContext

} // END GroupRequestHandler
