/*
 * Copyright (c) 1998-2004 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.relacs.gmi;

import java.io.Serializable;
import java.lang.reflect.Method;
import java.rmi.Remote;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import org.apache.log4j.Logger;

/**
 * Table mapping a method (represented by a hash of the method or by
 * a <code>Method</code> object) to its <code>MethodDetails</code> object.
 * 
 * The mapping from <code>Method</code> objects will only be populated
 * on the client-side since those are only used there.
 *
 * @author Hein Meling
 * @since Jgroup 3.0
 */
public class MethodTable
  implements Serializable, Iterable<MethodDetails>
{

  private static final long serialVersionUID = -7706887009118848315L;

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Logger
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Obtain logger for this class */
  private static final Logger log = Logger.getLogger(MethodTable.class);


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Table of methods (mapping from hash value to method details object) */
  private final Map<Long,MethodDetails> table = new HashMap<Long,MethodDetails>();

  /** Table mapping from method to method details object (only client-side) */
  private transient Map<Method,MethodDetails> methTable;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructor
  ////////////////////////////////////////////////////////////////////////////////////////////

  public MethodTable()
  {
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Add remote methods implemented by the given server to this
   * method table.  All methods defined in a <code>Remote</code>
   * interface will be added, and their associated invocation
   * semantics details will be computed.
   */
  public void addMethods(Object server)
  {
    /*
     * Find all the remote interfaces implemented by the given
     * server object, and compute relevant method information.
     */
    Class serverClass = server.getClass();
    Class[] interfaces = serverClass.getInterfaces();
    for (int i = 0; i < interfaces.length; i++) {
      if (Remote.class.isAssignableFrom(interfaces[i])) {
        Method methods[] = interfaces[i].getMethods();
        for (int j = 0; j < methods.length; j++) {
          MethodDetails meth = new MethodDetails(methods[j], server);
          table.put(meth.getHash(), meth);
          if (log.isDebugEnabled())
            log.debug("Adding: " + meth);
        }
      }
    }
  }

  /**
   * Get the method details associated with the given hash value.
   */
  public MethodDetails get(long hash)
  {
    return table.get(hash);
  }

  /**
   * Return the method details object associated with the given method.
   */
  public MethodDetails get(Method m)
  {
    if (methTable == null) {
      methTable = new HashMap<Method,MethodDetails>();
    }
    MethodDetails method = methTable.get(m);
    if (method == null) {
      long hash = MethodDetails.computeHash(m);
      method = get(hash);
      methTable.put(m, method);
    }
    return method;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from Iterable<MethodDetails>
  ////////////////////////////////////////////////////////////////////////////////////////////

  public Iterator<MethodDetails> iterator()
  {
    return table.values().iterator();
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Object methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  public String toString()
  {
    StringBuilder buf = new StringBuilder("MethodTable[");
    for (Map.Entry<Long,MethodDetails> entry : table.entrySet()) {
      MethodDetails meth = entry.getValue();
      buf.append("\n");
      buf.append(entry.getKey());
      buf.append(" -> ");
      buf.append(meth);
      Object server = meth.getServer();
      if (server != null) {
        // the server object is only available on the server-side
        buf.append("; server=");
        buf.append(server.getClass().getSimpleName());
      }
    }
    buf.append("]");
    return buf.toString();
  }

} // END MethodTable
