/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.relacs.mss;

import java.io.IOException;
import java.net.DatagramPacket;
import java.util.Iterator;
import java.util.SortedSet;
import java.util.TreeSet;

import jgroup.relacs.events.Event;
import jgroup.util.Queue;
import jgroup.util.ThreadMonitor;

import org.apache.log4j.Logger;


/**
 *  The <code>Ehandler</code> class handles all events generated
 *  internally by the timer and externally by the local members and by
 *  the network interface threads.
 *
 *  @author Alberto Montresor
 *  @author Hein Meling
 *  @since Jgroup 0.1
 */
final class Ehandler
  extends Thread
  implements NIListener, MssConstants, MssTag
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Logger
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Obtain logger for this class */
  private static final Logger log = Logger.getLogger(Ehandler.class);


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Upper layer */
  private final EhandlerUser ehuser;
  
  /** Sorted list of scheduled events */
  private final SortedSet<ScheduledEvent> events = new TreeSet<ScheduledEvent>();
  
  /** Local message queue */
  private final Queue lmessages = new Queue();
  
  /** Remote message queue*/
  private final Queue rmessages = new Queue();

  /** Next remote message */
  private DatagramPacket nextRemoteMessage;

  /** Next local message */
  private Event nextLocalMessage;

  /** Next nextScheduledEvent */
  private ScheduledEvent nextScheduledEvent;

  /** Set to true when the Ehandler thread has to be stopped */
  private boolean stop;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructor
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Constructs and starts the nextScheduledEvent handler.
   *  
   * @param name        
   *   the name of the nextScheduledEvent handler thread
   * @param ehuser
   *   reference to the nextScheduledEvent handler user (used to generate upcalls).
   */
  Ehandler(String name, EhandlerUser ehuser)
    throws IOException
  {
    /* Thread initialization */
    super(name);

    /* Data structure initialization */
    this.ehuser = ehuser;

    /* Thread management */
    stop = false;
    this.setDaemon(true);
    this.setPriority(EHANDLER_PRIORITY);
    this.start();
    ThreadMonitor.add(this);
  }

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** 
   *  Stops the Ehandler thread by setting stop to true.
   */
  void kill()
  {
    ThreadMonitor.remove(this);
    stop = true;
  }

  /**
   *  Inserts a new timeout event.  The event will be handled
   *  appropriately at the given time.
   *
   *  @param timeout
   *    Timeout length in milliseconds till the event should be handled.
   *  @param event
   *    The event description to associate this timeout value with.
   */
  synchronized ScheduledEvent setTimeout(long timeout, ScheduledEvent event)
  {
    long time = System.currentTimeMillis() + timeout;
    event.setTimeout(time);
    /*
     * Check if the event set contains several events at the exact same
     * time (millisecond granularity); if so, we avoid this problem by
     * incrementing the time value until we find an open time value for
     * the event to occur on.
     */
    while (events.contains(event)) {
      time++;
      event.setTimeout(time);
    }
    events.add(event);
    notifyAll();
    return event;
  }


  /**
   *  Abort the timeout associated to the specified <CODE>event</CODE>.
   *  If <CODE>event</CODE> is different from <CODE>null</CODE>, it
   *  is removed from the event list.
   */
  synchronized void abortTimeout(ScheduledEvent event)
  {
    if (event != null)
      events.remove(event);
  }

  
  /**
   *  Sends messages for local members by storing them in the <code>
   *  lmessages </code> queue.
   *
   *  @param msg
   *    Local message to send
   */
  synchronized void lnotify(Object msg)
  {
    lmessages.insert(msg);
    notifyAll();
  }


  /**
   *  Sends messages for remote members by storing them in the <code>
   *  rmessages </code> queue.
   *
   *  @param msg
   *    Local message to send
   */
  public synchronized void rnotify(DatagramPacket packet)
  {
    rmessages.insert(packet);
    notifyAll();
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Run method of the thread
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Run method of the thread.  It remains blocked until (i) a Datagram
   *  packet is received through the net; (ii) a message from a local
   *  member is received; (iii) a scheduled timeout expires.
   */
  public void run()
  {
    while (!stop) {
      await();
      if (log.isDebugEnabled())
        logEventQueue();
      if (nextScheduledEvent != null) {
        if (log.isDebugEnabled())
          log.debug("Scheduling event: " + nextScheduledEvent);
        ehuser.treceive(nextScheduledEvent);
        yield();
      }
      if (nextRemoteMessage != null) {
        if (log.isDebugEnabled())
          log.debug("Scheduling remote mesg: " + nextRemoteMessage);
        ehuser.rreceive(nextRemoteMessage);
        nextRemoteMessage = null;
        yield();
      }
      if (nextLocalMessage != null) {
        if (log.isDebugEnabled())
          log.debug("Scheduling local mesg: " + nextLocalMessage);
        ehuser.lreceive(nextLocalMessage);
        nextLocalMessage = null;
        yield();
      }
    }
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Private methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Select next nextScheduledEvent and next messages 
   */
  private synchronized void await()
  {
    /* Select next events and messages */
    if (!events.isEmpty()) {
      nextScheduledEvent = events.first();
    }
    nextRemoteMessage = (DatagramPacket) rmessages.removeFirst();
    nextLocalMessage  = (Event) lmessages.removeFirst();

    if (nextRemoteMessage == null && nextLocalMessage == null) {
      if (nextScheduledEvent == null) {
        try { wait(); } 
        catch (InterruptedException e) {
          log.warn("Ehandler:await:withoutTimeout", e);
        }
      } else {
        long timeout = nextScheduledEvent.getTimeout() - System.currentTimeMillis();
        if (timeout > 0) {
          try { wait(timeout); }
          catch (InterruptedException e) {
            log.warn("Ehandler:await:withTimeout", e);
          }
        }
      }
    }

    /*
     *  Check if the first timeout is expired
     */
    if (nextScheduledEvent != null) {
      long timeout = nextScheduledEvent.getTimeout() - System.currentTimeMillis();
      if (timeout <= 0) {
        events.remove(nextScheduledEvent);
        if (log.isDebugEnabled())
          log.debug("Timeout expired: " + (-timeout) + " msec ago");
      } else {
        nextScheduledEvent = null;
      }
    }
  }


  /**
   * Returns true if the local message queue contains undelivered
   * messages for the specified group; otherwise, false is returned.
   */
  public boolean hasUndeliveredLocalMsgs(int gid)
  {
    for (Iterator iter = lmessages.iterator(); iter.hasNext();) {
      Event event = (Event) iter.next();
      if (event.getGid() == gid) {
        return true;
      }
    }
    return false;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from Object
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Returns a string representation of this object
   */
  public String toString()
  {
    StringBuilder buf = new StringBuilder();
    buf.append("[Ehandler: LocalMsgs(");
    buf.append(lmessages.size());
    buf.append("), RemoteDatagrams(");
    buf.append(rmessages.size());
    buf.append("), ScheduledEvents(");
    buf.append(events.size());
    buf.append("), Events={ ");
    for (ScheduledEvent scheduledEvent : events) {
      buf.append(scheduledEvent);
      buf.append(", ");
    }
    buf.append(" }]");
    return buf.toString();
  }

  //mos
  /** last time show() was invoked or 0 */
  private long lastShowTime = 0;

  /** interval between show times */
  private final long interval = 2000;

  /**
   *  Shows the contents of the event queues.  This method is called
   *  from the run() loop of Ehandler.
   */
  private void logEventQueue()
  {
    long now = System.currentTimeMillis();
    if (now-lastShowTime > interval) {
      lastShowTime = now;
      log.debug(this);
    }
  }

} // END Ehandler
