/*
 * Copyright (c) 1998-2005 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */
package jgroup.relacs.mss;

import java.net.DatagramPacket;
import java.util.concurrent.Executors;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.ScheduledFuture;
import java.util.concurrent.ThreadFactory;
import java.util.concurrent.TimeUnit;

import jgroup.relacs.events.Event;
import jgroup.util.ThreadMonitor;

import org.apache.log4j.Logger;

/**
 *  The <code>EventHandler</code> class handles all events, packets and
 *  timeout events generated by the various mss layer components and by
 *  the network interface threads.
 *
 *  @author Hein Meling
 *  @since Jgroup 3.0
 */
public class EventHandler
  implements NIListener
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Logger
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Obtain logger for this class */
  private static final Logger log = Logger.getLogger(EventHandler.class);


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Upper layer */
  private final EhandlerUser ehuser;

  /** The executor used to schedule events, timeouts and packets */
  private final ScheduledExecutorService executorService;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructor
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Constructs and starts the event handler.
   *  
   * @param ehuser
   *   reference to the receiver of upcall events.
   */
  EventHandler(EhandlerUser ehuser)
  {
    /* Initialize the upper layer reference */
    this.ehuser = ehuser;
    /* Thread management */
    ThreadFactory tFactory = new ThreadFactory() {
      public Thread newThread(Runnable r) {
        Thread ehandler = new Thread(r, "MssEventHandler");
        ehandler.setDaemon(true);
        ThreadMonitor.add(ehandler);
        return ehandler;
      }
    };
    executorService = Executors.newSingleThreadScheduledExecutor(tFactory);
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** 
   * Stops the EventHandler task execution.
   */
  void kill()
  {
    executorService.shutdown();
  }

  /**
   * Schedule remote packet for delivery up to the mss layer.
   * 
   * @see jgroup.relacs.mss.NIListener#rnotify(java.net.DatagramPacket)
   */
  public void rnotify(final DatagramPacket msg)
  {
    Runnable task = new Runnable() {
      /* Insert the packet in the upcall queue */
      public void run() {
        if (log.isDebugEnabled())
          log.debug("remote msg: " + msg);
        ehuser.rreceive(msg);
        if (log.isDebugEnabled())
          log.debug("remote msg execution completed");
      }
    };
    executorService.submit(task);
    if (log.isDebugEnabled())
      log.debug("remote msg submitted for execution");
  }

  /**
   * Schedule event for delivery down to the mss layer.
   * 
   * @param event the event to be passed to the mss layer.
   */
  void lnotify(final Event event)
  {
    Runnable task = new Runnable() {
      /* Insert the event in the downcall queue */
      public void run() {
        if (log.isDebugEnabled())
          log.debug("local msg: " + event);
        ehuser.lreceive(event);
        if (log.isDebugEnabled())
          log.debug("local msg execution completed");
      }
    };
    executorService.submit(task);
    if (log.isDebugEnabled())
      log.debug("local msg submitted for execution");
  }

  /**
   * Schedule a new timeout event.  The event will be handled
   * appropriately at the given time.
   *
   * @param timeout
   *   Timeout length in milliseconds till the event should be handled.
   * @param event
   *   The event description to associate this timeout value with.
   */
  ScheduledEvent setTimeout(long timeout, final ScheduledEvent event)
  {
    Runnable task = new Runnable() {
      /* Insert the event in the downcall queue */
      public void run() {
        if (log.isDebugEnabled())
          log.debug("Delivering timeout event: " + event);
        ehuser.treceive(event);
        if (log.isDebugEnabled())
          log.debug("Timeout execution completed");
      }
    };
    ScheduledFuture handle = executorService.schedule(task, timeout, TimeUnit.MILLISECONDS);
    event.setTimeout(timeout, handle);
    return event;
  }


  /**
   * Abort the timeout associated to the specified event.
   */
  void abortTimeout(ScheduledEvent event)
  {
    if (event != null)
      event.abortTimeout();
  }
  
} // END EventHandler
