/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.relacs.mss;

import java.util.HashMap;
import java.util.Iterator;

import jgroup.core.ConfigurationException;
import jgroup.core.EndPoint;

/**
 * The <code> HostTable</code> class
 *
 * @author  Jgroup Team
 * @since Jgroup 0.5
 */
public final class HostTable
{
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constants
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Initial capacity of the table */
  private final static int CAPACITY = 32;
 
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Hash map containing the hosts */
  private HashMap map;
  
  /** Table of MssHost */
  private MssHost[] table;

  /** Actual size of the table */
  private int size;
  
  /** This host */
  private static MssHost me;

  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructors
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Creates an empty table.
   */
  public HostTable()
  {
    map = new HashMap();
    table = new MssHost[CAPACITY];
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Public methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Inserts host <code>host</code> in the table, unless an host with the same key
   *  is already in the table.
   *
   *  @param     host     the host to be inserted.
   *  @exception ConfigurationException if the element is already in the table
   */
  public void insert(MssHost host)
    throws ConfigurationException
  {
    if (host == null)
      throw new NullPointerException();
    if (map.containsKey(host.getEndPoint()))
      throw new ConfigurationException(host + " already present in hosttable.");
    if (host.isLocal()) {
      if (me == null) {
        me = host;
      }
    }
    if (size >= table.length) {
      doubleCapacity();
    }
    table[size++] = host;
    map.put(host.getEndPoint(), host);
    host.setIndex(size-1);
  }


  /**
   * Returns the host reference to which this address is mapped.
   *
   * @param   endpoint  host endpoint
   * @return the host reference to which this address is mapped.
   */
  public MssHost lookup(EndPoint endpoint)
  {
    return (MssHost) map.get(endpoint);
  }

  
  /**
   * Returns the host reference having the specified index
   *
   * @param   index     host index
   * @return            the host reference having the specified index.
   */
  public MssHost get(int index)
  {
    if (index >= size)
      throw new IndexOutOfBoundsException();
    return table[index];
  }

  
  /**
   *  Returns the <code>MssHost</code> descriptor of the local host.
   *
   *  @throws IllegalStateException
   *    Raised if the local host has not yet been initialized.
   */
  public static MssHost getLocalHost()
  {
    if (me == null)
      throw new IllegalStateException("Local host not yet initialized in the HostTable.");
    return me;
  }


  /**
   *  Return the number of hosts contained in the table.
   */
  public int size()
  {
    return size;
  }


  /**
   *  Return an iterator over the values in this host table.
   */
  public Iterator iterator()
  {
    return map.values().iterator();
  }


  /**
   *  Reset the message flow of all hosts in this
   *  <code>HostTable</code>.
   */
  void resetMsgFlow()
  {
    for (Iterator iter = map.values().iterator(); iter.hasNext(); ) {
      MssHost host = (MssHost) iter.next();
      host.resetMsgFlow();
    }
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from Object
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Returns a string representation of this object
   */
  public String toString()
  {
    StringBuilder buffer = new StringBuilder();
    buffer.append("[HostTable: {");
    for (int i=0; i < size; i++) {
      if (i!=0)
        buffer.append(",");
      buffer.append(table[i]);
    }
    buffer.append("}]");
    return buffer.toString();
  }

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Private methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Double the capacity of the table
   *  
   */
  private void doubleCapacity()
  {
    MssHost[] newtable = new MssHost[table.length*2];
    System.arraycopy(table, 0, newtable, 0, table.length);
    table = newtable;
  }
  

} // END HostTable
