/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.relacs.mss;

import java.util.Iterator;
import java.util.SortedMap;
import java.util.TreeMap;

import jgroup.relacs.config.TransportConfig;

import org.apache.log4j.Logger;

/**
 *  The <code>MsgFlowSndrSide</code> class holds the data structures
 *  needed to maintain reliable message delivery.
 *
 *  @author Salvatore Cammarata
 *  @author Hein Meling
 *  @since Jgroup 1.2
 */
final class MsgFlowSndrSide
  implements MssConstants
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Logger
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Obtain logger for this class */
  private static final Logger log = Logger.getLogger(MsgFlowSndrSide.class);


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Static section
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Token for blocking the sender if the sent window is closed */
  private static Token token = new Token();

  /** */
  private static Token sleepToken = new Token();  


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constants
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  The NACK threshold; the number of NACKs required to mark the local
   *  cluster as congested.
   */
  private static final int NACK_THRESHOLD = 1;

  /** The slow start threshold */
  private static final int SLOW_THRESHOLD = 10;

  /** The delay for each iteration of the slow start */
  private static final int SLOW_DELAY  = 10;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Sliding window parameters
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** The width of the sending window (when starting) */
  private int startingWin;

  /** Maximum width of the sending window */
  private int maxWin;

  /** Mimimum width of the sending window */
  private int minWin;

  /** Width of the congestion window */
  private int congestionWindow;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Associated cluster */
  private Cluster cluster;

  /** Right edge of the sliding window */
  private int lastMsgSent;

  /** Left edge of the sliding window */
  private int lastMsgAcked; 

  /** Number of NACKs received */
  private int nackCount = 0;

  /** Slow start activated */
  private boolean slowStart = true; 

  /** Number of iterations so far for the slow start algorithm */
  private int slowCount = 0;

  /** Queue of the message sent and not yet acked */
  private SortedMap<Integer, FragmentIterator> sent = new TreeMap<Integer, FragmentIterator>();

  /** Queue of the message sent and not yet acked */
  //FIXED HEIN: was not used; consider if it will be used later?!
  //  TreeMap frozen = new TreeMap();


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructors
  ////////////////////////////////////////////////////////////////////////////////////////////

  MsgFlowSndrSide(Cluster cluster, TransportConfig config) 
  {
    this.cluster = cluster;
    this.startingWin = config.getStartingWindow();
    this.maxWin  = config.getMaxWindow();
    this.minWin  = config.getMinWindow();
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * 
   */
  void reset(int mid) 
  {
    token.go();
    if (log.isDebugEnabled()) {
      log.debug("Before: " + this);
    }

    // right edge of the sliding window
    lastMsgSent = mid;
    // left edge of the sliding window
    lastMsgAcked = mid;
    // reset the congestion window size
    congestionWindow = startingWin;

    /* Clear the sent queue */
    sent.clear();
    //     frozen.clear();

    if (log.isDebugEnabled()) {
      log.debug(" After: " + this);
    }
  }


  /**
   *  Returns true if the sender side message flow is initialized and
   *  can be used; otherwise false is returned.
   */
  boolean isInitialized()
  {
    return (lastMsgSent != UNDEF);
  }


  /**
   *  Insert the given message fragment (that will be sent next) in the
   *  ordered <i>sent</i> queue.
   *
   *  @param msgFragment
   *    The message fragment to store in the <i>sent</i> queue.
   *  @return
   *    The fragment identifier for this message.
   */
  int insertSentMsgFrag(FragmentIterator fragIter)
  {
    /* Increments the number of messages sent. */
    int fid = ++lastMsgSent;
    sent.put(fid, fragIter);
    if (log.isDebugEnabled()) {
      log.debug("Inserted msg fragment (" + fid + ") in sent queue of cluster "+ cluster);
      log.debug(this);
    }

    /* Slow start */
    slowDelay();
    return fid;
  }


  /**
   *  Get a previously sent message fragment, based on the NACK message
   *  identifier.  This will be used to resend the message fragment.
   *
   *  @param nack
   *    The <code>MsgNACK</code> object containing the message
   *    identifier that should be resent.
   *  @return 
   *    The fragment iterator associated with the identifier of the NACK
   *    message.
   */
  FragmentIterator getSentMsgFrag(MsgNACK nack)
  {
    if (log.isDebugEnabled()) {
      log.debug(this);
    }
    return sent.get(nack.mid);
  }


  /**
   *  Flush the sent queue.  Remove all messages already acknowledged.
   */
  void flush(int lastMsgAcked)
  {
    setLastMsgAcked(lastMsgAcked);

    /* Remove received messages from the sent queue in sorted order. */
    for (Iterator<Integer> iter = sent.keySet().iterator(); iter.hasNext(); ) {
      int mid = iter.next();
      if (mid <= lastMsgAcked) {
        if (log.isDebugEnabled())
          log.debug("flush: Removing msg (" + mid + ") from sent queue of cluster: "
            + cluster.getEndPoint());
        iter.remove();
      } else if (log.isDebugEnabled()) {
        log.debug("flush: Keeping msg (" + mid + ") in sent queue of cluster: "
          + cluster.getEndPoint());
      }
    }
    if (log.isDebugEnabled()) {
      log.debug("flush: LastMsgAcked (" + lastMsgAcked + ") for cluster: "
        + cluster.getEndPoint() + "; remaining messages in sent queue: " + sent.size());
    }
  }


  /**
   *  Slow start
   */
  private void slowDelay()
  {
    if (slowStart && slowCount < SLOW_THRESHOLD) {
      slowCount++;
      sleepToken.sleep(SLOW_DELAY);
    } else {
      slowStart = false;
      slowCount = 0;
    }
  }

  /**
   *  Returns the number of messages sent.
   */
  int getLastMsgSent() 
  {
    return lastMsgSent;
  }

  /**
   * 
   */
  boolean isCongested()
  {
    return nackCount > NACK_THRESHOLD;
  }

  void clearNackCount()
  {
    nackCount = 0;
  }

  void incNackCount()
  {
    nackCount++;
  }

  synchronized int getActualWindow()
  {
    if (log.isDebugEnabled()) {
      log.debug(cluster.getEndPoint() + ": * actualWindow=" 
        + (congestionWindow < maxWin ? congestionWindow : maxWin));
    }
    return (congestionWindow < maxWin ? congestionWindow : maxWin);
  }

  synchronized int decCongWin()
  {
    int newWin = congestionWindow / 2;
    congestionWindow = (minWin < newWin ? newWin : minWin);
    if (log.isDebugEnabled()) {
      log.debug(cluster.getEndPoint() + ": - congestionWindow=" + congestionWindow);
    }
    return congestionWindow; 
  }
  
  synchronized int incCongWin(int npack)
  {
    if (npack > 0) {
      int inc = (npack > SLOW_THRESHOLD) ? SLOW_THRESHOLD : npack;
      congestionWindow = Math.min(congestionWindow + inc, maxWin);
      if (log.isDebugEnabled()) 
        log.debug(cluster.getEndPoint() + ": + congestionWindow=" + congestionWindow);
      return congestionWindow;
    }
    return congestionWindow;
  }

  public void waitOpen()
  {
    token.stop();
    slowStart = true;
    slowCount = 0;
  }


  public int getLastMsgAcked()
  {
    return lastMsgAcked;
  }


  private void setLastMsgAcked(int mid)
  {
    if (mid != UNDEF && mid > lastMsgAcked) {
      int llma = lastMsgAcked;
      lastMsgAcked = mid;

      // congestion control
      incCongWin(mid - llma);
      token.go(); 
    }
  }

  boolean isWindowOpen()
  {
    boolean openWindow = (lastMsgSent < (lastMsgAcked + getActualWindow()));
    if (log.isDebugEnabled()) {
      log.debug("isWindowOpen: " + openWindow);
    }
    return openWindow;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Object methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  public String toString()
  {
    StringBuilder buf = new StringBuilder("[MsgFlowSndrSide: ");
    buf.append("lastMsgSent=");
    buf.append(lastMsgSent);
    buf.append(", lastMsgAcked=");
    buf.append(lastMsgAcked);
    buf.append(", actualWindow=");
    buf.append(getActualWindow());
    buf.append(", congestionWindow=");
    buf.append(congestionWindow);
    buf.append(", sent={");
    for (Iterator iter = sent.keySet().iterator(); iter.hasNext(); ) {
      Integer fragid = (Integer) iter.next();
      FragmentIterator frag = sent.get(fragid);
      buf.append("\n(");
      buf.append(fragid);
      buf.append(") --> ");
      buf.append(frag);
    }
    buf.append(" }]\n");
    return buf.toString();
  }

} // END MsgFlowSndrSide
