/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.relacs.mss;

import java.io.IOException;
import java.io.ObjectInput;
import java.util.Iterator;
import java.util.List;

import jgroup.core.EndPoint;
import jgroup.relacs.daemon.DaemonMsg;
import jgroup.relacs.types.EndPointImpl;
import jgroup.relacs.types.GroupIndex;
import jgroup.relacs.types.MessageId;
import jgroup.util.InMessage;
import jgroup.util.OutMessage;

import org.apache.log4j.Logger;

/**
 *  The <code>MsgJG</code> class is the main <i>mss</i> level message,
 *  used for passing Jgroup related multicast messages to group members.
 *
 *  @author Hein Meling
 *  @since Jgroup 1.2
 */
public final class MsgJG
  implements Msg, MssConstants, MssTag
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Logger
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Obtain logger for this class */
  private static Logger log = Logger.getLogger(MsgJG.class);


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Static section
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Constant null array used when no receivers are set; meaning that
   *  all should receive the message.  Object reuse pattern; see Item 27
   *  in Effective Java.
   */
  private final static EndPoint[] NULL_ENDPOINT_ARRAY = new EndPoint[0];


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Message Identifier Generator
  ////////////////////////////////////////////////////////////////////////////////////////////

  private static transient int fragIdGen = 0;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Message Fields (static part)
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Fragment identifier */
  private int fragId;

  /** Tag assigned in upper level (@see jgroup.relacs.daemon.Tag) */
  private byte JGtag;

  /** The cluster to which the message will be sent (only endpoint is transmitted) */
  private transient Cluster cluster;

  /**
   *  Pointer to the message trailer, which contains the dynamic part of
   *  the message (see below)
   */
  private int trailerStartPos;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Message Fields (dynamic part)
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Receivers array (zero length if ALL should receive this message).
   *  This array contains all receivers, when accessed from the sending
   *  side, while on the receiver side it contains only those whom was
   *  marshalled into the message destined for a particular cluster.
   */
  private EndPoint[] receivers;

  /** Distinguish between local and external flow control information */
  private byte FCtype;

  /** Array of flow control entries */
  private FCEntry[] fc;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Transient fields (recomputed during unmarshalling)
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** The MssDS object */
  private transient MssDS mssds;

  /** Message to be sent */
  private transient OutMessage outmsg;

  /** Message received from the Mss */
  private transient InMessage inmsg;

  /** True if this host is in the destination set of this message */
  private transient boolean isForMe;

  /** The index of the flow control entry associated with the local host */
  private transient int localFCIndex = UNDEF; 

  /** The local host (where this message resides; sender or receiver) */
  private transient MssHost localHost;

  /** The sender host */
  private transient MssHost sender;

  /** The start position of the cluster entry in the outmsg stream */
  private transient int clusterStartPosition;

  /** The start position of the receiver set in the outmsg stream */
  private transient int receiversStartPosition;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructors and marshalling methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Marshal a message with receiver set all.
   */
  static MsgJG marshal(DaemonMsg dmsg, byte JGtag, MssDS mssds)
    throws IOException
  {
    return new MsgJG(dmsg, JGtag, NULL_ENDPOINT_ARRAY, mssds);
  }


  static MsgJG marshal(DaemonMsg dmsg, byte JGtag, EndPoint[] receivers, MssDS mssds)
    throws IOException
  {
    return new MsgJG(dmsg, JGtag, receivers, mssds);
  }


  /**
   *  Marshalling constructor: Creates a new <code>MsgJG</code>
   *  instance, based on an upper level message originated by the Jgroup
   *  daemon.
   *
   *  @param dmsg
   *    A <code>DaemonMsg</code> from the upper level.  The set
   *    of possible messages are in the @see jgroup.relacs.daemon
   *    package.
   *  @param JGtag
   *    A one byte tag indicating an upper level message type that is
   *    encapsulated within the given <code>outmsg</code>.  For the set
   *    of possible values @see jgroup.relacs.daemon.Tag.
   *  @param dest  a <code>Cluster</code> value
   *  @param receivers  a <code>MssHost[]</code> value
   *  @param mssds  a <code>MssDS</code> value
   *
   *  @exception IOException if an error occurs
   */
  private MsgJG(DaemonMsg dmsg, byte JGtag, EndPoint[] receivers, MssDS mssds)
    throws IOException
  {
    localHost = HostTable.getLocalHost();
    sender = localHost;
    this.receivers = receivers;
    this.mssds = mssds;
    this.outmsg = dmsg.getOutMessage();

    this.fragId = fragIdGen++;
    this.JGtag = JGtag;

    /*
     * The static <i>mss</i> data associated with a <code>MsgJG</code>
     * is kept in the beginning of the stream (in the header), while the
     * dynamic <i>mss</i> level data is kept at the end of the
     * <code>OutMessage</code> stream (in the trailer).  In addition to
     * the static data, we need to keep the trailer index in the header
     * to allow unmarshalling, starting from the beginning of the
     * <code>OutMessage</code> stream.
     *
     * Note that the <code>dmsg.size()</code> returns the number of
     * bytes that originates from the daemon layer, and since the first
     * part of every Mss level message is <code>MSS_HEADER_SIZE</code>
     * bytes, we have to add this to get the correct trailer position.
     */
    trailerStartPos = dmsg.size() + MSS_HEADER_SIZE;
    log.assertLog(dmsg.size() != 0, "Illegal position for provided DaemonMsg");
    outmsg.seek(0);

    /*
     * Store the trailer index in the static section of the <i>mss</i>
     * level data.  The trailer is written using the methods
     * <code>setCluster()</code> and <code>setReceivers()</code>.
     */
    outmsg.writeInt(trailerStartPos);
    MessageId.marshal(outmsg, fragId);
    outmsg.writeByte(JGtag);
    clusterStartPosition = outmsg.getPosition();
  }


  /**
   *  Returns a <code>MsgJG</code> object that contains a partially
   *  decoded m-received input stream, containing only the static part
   *  of the message structure.  To decode the whole stream, including
   *  the dynamic part of the <i>mss</i> level message, all message
   *  fragments must have been added into the message object using the
   *  <code>addFragment()<code> method.
   *
   *  @param inmsg
   *    The message input stream to decode.
   */
  static MsgJG unmarshal(ObjectInput inmsg, FragmentHeader header, MssDS mssds)
    throws IOException, ClassNotFoundException
  {
    return new MsgJG((InMessage) inmsg, header, mssds);
  }


  /**
   *  Unmarshalling constructor: Constructs a new <code>MsgJG</code>
   *  data structure from an m-received message.
   *
   *  @param inmsg
   *    The m-received <code>InMessage</code> object
   *
   *  @exception IOException
   *    Raised if there was problem unmarshalling the message.
   */
  private MsgJG(InMessage inmsg, FragmentHeader header, MssDS mssds)
    throws IOException, ClassNotFoundException
  {
    this.inmsg = inmsg;
    this.mssds = mssds;

    /* Get the local host of the receiver */
    localHost = HostTable.getLocalHost();
    sender = header.getSender();

    /*
     * Unmarshalling static header part.
     */
    /*
     * Retrive pointer to the trailer start position in which the
     * <i>mss</i> level message data is collected.
     */
    trailerStartPos = inmsg.readInt();
    fragId = MessageId.unmarshal(inmsg);
    JGtag = inmsg.readByte();
    EndPointImpl endpoint = new EndPointImpl();
    endpoint.readExternal(inmsg);
    cluster = mssds.getClusterTable().lookup(endpoint);
  }


  /**
   *  Unmarshalling method to complete the unmarshalling of a
   *  <code>MsgJG</code> object.  It assumes that all fragments has been
   *  added to the stream using the <code>addFragment()</code> method.
   */
  InMessage complete()
    throws IOException, ClassNotFoundException
  {
    // Seek to the start of the dynamic data; the trailer
    inmsg.seek(trailerStartPos);

    // Flow control data
    FCtype = inmsg.readByte();
    int len = GroupIndex.unmarshal(inmsg);
    fc = new FCEntry[len];
    for (int i = 0; i < len; i++) {
      fc[i] = new FCEntry();
      fc[i].readExternal(inmsg);
      if (localFCIndex == UNDEF && fc[i].key.isLocal())
        localFCIndex = i;
    }

    // Set of receivers
    int nreceivers = GroupIndex.unmarshal(inmsg);
    if (nreceivers == ALL) {
      isForMe = true;
      receivers = NULL_ENDPOINT_ARRAY;
    } else {
      isForMe = false;
      receivers = new EndPoint[nreceivers];
      for (int i = 0; i < nreceivers; i++) {
        receivers[i] = new EndPointImpl();
        receivers[i].readExternal(inmsg);
        if (receivers[i].isLocal())
          isForMe = true;
      }
    }

    /*
     * We are now ready to update the flow control information, attached
     * with this message.
     */
    updateFC();

    /*
     * Seek to position <code>MSS_HEADER_SIZE</code> and mark that
     * position so that any reset will return to that position
     * instead; this is to avoid that external message users see the
     * static header part of this message. <p>
     *
     * PS; Note that the parameter to the mark() method is not used.
     */
    inmsg.seek(MSS_HEADER_SIZE);
    inmsg.mark(0);
    return inmsg;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fragmentation handling (from Msg interface)
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Returns a <code>FragmentIterator</code> for this
   *  <code>MsgJG</code> object.  This iterator allows to send the
   *  entire message as multiple fragments of specified size (payload).
   *  At the same time, it marks each fragment with a tag and message
   *  identifier provided through the <code>next()</code> method of the
   *  iterator.
   *
   *  Note that <code>MsgJG</code> messages cannot reuse the same
   *  iterator for sending to multiple clusters.  This is because, at
   *  each iteration reuse, the fragment identifier is changed, causing
   *  problems when doing lookup in the sent queue to resend a message.
   */
  public FragmentIterator iterator(MsgCntrl msgCntrl)
  {
    return new MsgFragmentIterator(this, msgCntrl);
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Msg interface methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Returns the tag associated with this message.
   */
  public byte getTag()
  {
    return JG;
  }


  /**
   *  Returns the message identifier for this message.
   */
  public int getMid()
  {
    return fragId;
  }


  /**
   *  Returns the sender of this message.
   */
  public MssHost getSender()
  {
    return sender;
  }


  /**
   *  Returns true if this message has to be routed to a different
   *  cluster.
   */
  public boolean hasToBeRouted()
  {
    return !cluster.isLocal();
  }


  /**
   *  Returns the message flow controller for the sender side.
   */
  public MsgFlowSndrSide getMsgFlow()
  {
    return cluster.getMsgFlow();
  }


  /**
   *  Returns the <code>OutMessage</code> associated with this message.
   */
  public OutMessage getOutMessage()
  {
    return outmsg;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // MsgJG specified methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Marshal receivers into the stream of this message.  If there are
   *  no receivers specified (the <code>receivers</code> list is empty),
   *  this means that all members should receive this message. <p>
   *
   *  This method is invoked for each cluster, to marshal the receivers
   *  that belong to that cluster into the stream.  That means, not all
   *  actual receivers may be marshalled into the stream.
   *
   *  @param receivers
   *    List of receivers for this message; if the list is empty, all
   *    should receive the message.
   */
  public void setReceivers(List receivers)
  {
    outmsg.seek(receiversStartPosition);
    try {
      GroupIndex.marshal(outmsg, receivers.size());
      for (Iterator iter = receivers.iterator(); iter.hasNext(); ) {
        EndPoint recv = (EndPoint) iter.next();
        recv.writeExternal(outmsg);
        if (log.isDebugEnabled()) {
          log.debug("setReceivers: " + recv);
        }
      }
    } catch (IOException e) {
      log.warn("setReceivers: Could not update receiver information in message stream");
    }
  }


  public void setCluster(Cluster cluster)
  {
    if (log.isDebugEnabled()) {
      log.debug("setCluster: " + cluster);
    }

    this.cluster = cluster;
    if (cluster.isLocal()) {
      FCtype = LOCALFC;
      fc = mssds.getAllFCEntry();
    } else {
      FCtype = EXTERNFC;
      fc = mssds.getClusterFCEntry(cluster.getEndPoint());
    }

    /* Find the location for writting the cluster. */
    outmsg.seek(clusterStartPosition);
    try {
      cluster.getEndPoint().writeExternal(outmsg);
    } catch (IOException e) {
      log.warn("setCluster: Could not update cluster in message stream");
    }

    /*
     * Search for the trailer location, and write out the flow control
     * information.
     */
    outmsg.seek(trailerStartPos);
    try {
      outmsg.writeByte(FCtype);
      GroupIndex.marshal(outmsg, fc.length);
      for (int i = 0; i < fc.length; i++) {
        fc[i].writeExternal(outmsg);
      }
    } catch (IOException e) {
      log.warn("setCluster: Could not update flow control information in message stream");
    }

    /* Store the location for the receiver set. */
    receiversStartPosition = outmsg.getPosition();
  }


  /**
   *  Returns the destination cluster of this message.
   */
  public Cluster getCluster()
  {
    return cluster;
  }


  /**
   *  Returns the receiver set for this message.  On the sending side,
   *  this includes all receivers, while on the receiver side this may
   *  only include those that were marshalled into the message for a
   *  particular cluster.
   */
  public EndPoint[] getReceivers()
  {
    return receivers;
  }


  private void updateFC()
  {
    switch (FCtype) {

    case EXTERNFC:
      /* Check if there is a flow control entry for the local host */
      if (localFCIndex != UNDEF) {
        /* Update congestion information (sender side flow control) */
        sender.flush(fc[localFCIndex].lastMsgRcvd);
      }
      break;

    case LOCALFC:
      if (sender.isLocal()) {

        localHost.flush();

      } else {

        for (int i = 0; i < fc.length; i++) {
          MssHost host = mssds.hostLookup(fc[i].key);
          if (host == null) {
            log.warn("Unavailable host: " + fc[i].key);
            continue;
          }
          host.getMsgFlow().clusterWindow.set(sender.getClusterIndex(), fc[i].lastMsgRcvd);
        }

      }
      break;

    default:
      log.warn("Erroneous flow control type");
    }
  }


  /**
   *  Returns the upper level Jgroup tag associated with this message.
   *
   *  @see jgroup.relacs.daemon.Tag
   */
  public byte getJGTag()
  {
    return JGtag;
  }


  /**
   *  Returns true if this message is destined for the local host.
   */
  boolean isForMe()
  {
    return isForMe;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from Object
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Returns a string representation of this object
   */
  public String toString()
  {
    StringBuilder buf = new StringBuilder("MsgJG: jgtag=");
    buf.append(JGtag);
    buf.append(", sender=");
    buf.append(sender);
    buf.append(", cluster=");
    buf.append(cluster);
    buf.append(", nreceivers=");
    buf.append(receivers.length);
    for (int i = 0; i < receivers.length; i++) {
      buf.append(", ");
      buf.append(receivers[i]);
      if (receivers[i].isLocal())
        buf.append(" (LOCAL)");
    }
    if (outmsg != null) {
      buf.append(", ");
      buf.append(outmsg.toString());
    } else if (inmsg != null) {
      buf.append(", ");
      buf.append(inmsg.toString());
    }
    return buf.toString();
  }

}
