/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.relacs.mss;

import java.io.IOException;
import java.io.ObjectInput;

import jgroup.relacs.types.MessageId;
import jgroup.util.InMessage;
import jgroup.util.MsgFactory;
import jgroup.util.OutMessage;

/**
 *  The <code>MsgSYN</code> class is used to synchronize the message
 *  (fragment) identifiers of the different hosts within the distributed
 *  system.  This message type is never broadcast, and it will always
 *  fit within a fragment, given that the payload is set above 9.
 *
 *  Note that this message is currently queued in the MssHost, enabling
 *  us to check the correctness of a ASYN wrt. a previously sent QSYN
 *  message.  This prevents us from using a single message instance for
 *  all SYN messages.
 *
 *  @author Hein Meling
 *  @since Jgroup 1.2
 */
final class MsgSYN
  implements Msg, MssConstants, MssTag
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Size
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Message size (1+4+4) bytes */
  private static final int MSG_SIZE = 9;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Message Fields (static part)
  ////////////////////////////////////////////////////////////////////////////////////////////

  /* Type of synchronization message (QSYN, ASYN) */
  private byte synchType;

  /** Message identifier for this message */
  private int nonce;

  /** Last message sent by the sender of this synchronization message */
  private int lastMsgSent;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Transient fields (recomputed during unmarshalling)
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Message to be sent */
  private transient OutMessage outmsg;

  /** Receiver of this SYN message (not transmitted) */
  private transient MssHost receiver = null;

  /** The message sender (obtained from the fragment header) */
  private transient MssHost sender;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructors and marshalling methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Used for ASYN
   */
  static MsgSYN marshal(byte synchType, MsgSYN msg, int lastMsgSent)
    throws IOException
  {
    return new MsgSYN(synchType, msg.getSender(), msg.nonce, lastMsgSent);
  }


  /**
   *  Used for QSYN
   */
  static MsgSYN marshal(byte synchType, MssHost receiver)
    throws IOException
  {
    return new MsgSYN(synchType, receiver, RandomGenerator.newNonce(), UNDEF);
  }


  /**
   *  Marshalling constructor
   */
  private MsgSYN(byte synchType, MssHost receiver, int nonce, int lastMsgSent)
    throws IOException
  {
    sender = HostTable.getLocalHost();
    this.receiver = receiver;
    outmsg = MsgFactory.get(MSG_SIZE);

    this.synchType   = synchType;
    this.nonce       = nonce;
    this.lastMsgSent = lastMsgSent;

    outmsg.writeByte(synchType);
    MessageId.marshal(outmsg, nonce);
    MessageId.marshal(outmsg, lastMsgSent);
  }


  /**
   *
   */
  static MsgSYN unmarshal(ObjectInput inmsg, FragmentHeader header)
    throws IOException, ClassNotFoundException
  {
    return new MsgSYN((InMessage) inmsg, header);
  }


  /**
   *  Unmarshalling contructor
   */
  private MsgSYN(InMessage in, FragmentHeader header)
    throws IOException, ClassNotFoundException
  {
    sender      = header.getSender();
    synchType   = in.readByte();
    nonce       = MessageId.unmarshal(in);
    lastMsgSent = MessageId.unmarshal(in);
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  byte getSYNType()
  {
    return synchType;
  }


  int getLastMsgSent()
  {
    return lastMsgSent;
  }


  /**
   *  Returns the receiver of this message.  This method should only be
   *  invoked on the sending side, since it does not make sense to query
   *  the receiver at the receiving end (the receiver would be local
   *  host).  Therefore the receiver field is not transmitted either.
   *
   *  @exception IllegalStateException
   *    Thrown if invoked on the receiving side.
   */
  MssHost getReceiver()
  {
    if (receiver == null)
      throw new IllegalStateException("Cannot obtain receiver on receiving side (it is the localhost)");
    return receiver;
  }


  /**
   *  Returns true if the sender of this message is the local host.
   */
  public boolean isLocal()
  {
    return sender.isLocal();
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fragmentation handling (from Msg interface)
  ////////////////////////////////////////////////////////////////////////////////////////////

  /*
   * The fragmentation iterator instance for this message; it can be
   * reused for multiple iterations over this message.
   */
  private transient MsgFragmentIterator fragIter = null;


  /**
   *  Returns a <code>FragmentIterator</code> for this
   *  <code>MsgSYN</code> object.  This iterator allows to send the
   *  entire message as multiple fragments of specified size (payload).
   *  At the same time, it marks each fragment with a tag and message
   *  identifier provided through the <code>next()</code> method of the
   *  iterator.
   */
  public FragmentIterator iterator(MsgCntrl msgCntrl)
  {
    /* If there is already an iterator for this message, we reuse it. */
    if (fragIter == null) 
      fragIter = new MsgFragmentIterator(this, msgCntrl);
    else 
      fragIter.reset(outmsg);
    return fragIter;
  }



  ////////////////////////////////////////////////////////////////////////////////////////////
  // Msg interface methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Returns the tag associated with this message.
   */
  public byte getTag() 
  {
    return SYN;
  }


  /**
   *  Returns the message identifier for this message.
   */
  public int getMid()
  {
    return nonce;
  }


  /**
   *  Returns the sender of this message.
   */
  public MssHost getSender()
  {
    return sender;
  }


  /**
   *  Returns false always, since synchronization messages should never
   *  be routed.
   */
  public boolean hasToBeRouted()
  {
    return false;
  }


  /**
   *  Returns the message flow controller for the sender side.
   */
  public MsgFlowSndrSide getMsgFlow()
  {
    return sender.getCluster().getMsgFlow();
  }


  /**
   *  Returns the <code>OutMessage</code> associated with this message.
   */
  public OutMessage getOutMessage()
  {
    return outmsg;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from Object
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Returns a string representation of this object
   */
  public String toString()
  {
    StringBuilder buf = new StringBuilder();
    switch (synchType) {
    case QSYN:
      buf.append("QSYN");
      break;

    case ASYN:
      buf.append("ASYN");
      break;

    default:
      buf.append("Illegal synchronization type: " + synchType);
    }
    buf.append(", sender: ");
    buf.append(sender);
    if (receiver != null) {
      buf.append(", receiver: ");
      buf.append(receiver);
    }
    buf.append(", nonce=");
    buf.append(nonce);
    buf.append(", lastMsgSent=");
    buf.append(lastMsgSent);
    return buf.toString();
  }

} // END MsgSYN
