/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.relacs.mss;

/**
 *  The <code>MssConstants</code> interface declares a number of
 *  constants mainly relevant for the <i>mss</i> level.
 *
 *  @author Hein Meling
 *  @author Alberto Montresor
 *  @author Salvatore Cammarata
 *  @since Jgroup 1.2
 */
public interface MssConstants
{

  /**
   *  Enabling (setting it to <code>true</code>) NACK suppression sends
   *  only a single NACK for each cluster.  Otherwise
   *  <i>cluster.size</i> messages will be sent for each missing
   *  message, for each of the clusters that are missing the message.
   */
  public static final boolean NACKSUPPRESSION = false;


  /**
   *  Setting this to true enables routing of messages through
   *  intermediate clusters, if a node (or cluster) is not directly
   *  connected to the destination cluster.
   *
   *  FIXME NOTE: There is currently a problem with the routing
   *  mechanism, so don't turn it on.
   */
  public static final boolean ROUTING_ENABLED = false;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Thread priorities
  ////////////////////////////////////////////////////////////////////////////////////////////

  public static final int MSSUSER_PRIORITY    = Thread.NORM_PRIORITY;

  public static final int EHANDLER_PRIORITY   = MSSUSER_PRIORITY;

  public static final int NI_PRIORITY         = EHANDLER_PRIORITY;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Reachability and message constants
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  If the alive counter of a host is equal to this value, the host is
   *  considered to be unreachable.
   */
  public static final int UNREACHABLE         = 0;

  /**
   *  If the alive counter of a host is greater or equal to this value,
   *  the host is considered to be alive.
   */
  public static final int MINIMUM_ALIVE_VALUE = 1;

  /**
   *  Constant used to indicate that <i>all</i> members of a cluster
   *  should receive a message.
   */
  public static final int ALL                 = 0;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // General constants
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Used to indicate that an <code>int</code> value is undefined.
   */
  public static final int UNDEF               = -1;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Message and Packet constants
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Header size for <i>mss</i> level Jgroup (<code>MsgJG</code>)
   *  messages.  Mss messages contains a static header and a dynamic
   *  trailer of unspecified size, whose size is determined at runtime.
   *  The content of the header includes a message tag, message
   *  identifer, pointer to the trailer's position in the message stream
   *  etc.  The content of the trailer cannot be accessed until all
   *  message fragments (packets) have been received.
   */
  public static final int MSS_HEADER_SIZE = 15;

  /**
   *  Header size for <i>network</i> level packets (fragments).
   *  Messages sent from the <i>mss</i> level may be split into several
   *  smaller fragments (or packets).  Each of these fragments will be
   *  attached a header of the size given below.  This fragment header
   *  comes in addition to the payload limit specified in the
   *  <code>config.xml</code> file.  The header contains information
   *  about the message type, broadcast, fragment identifier (fid),
   *  sender and incarnation identifier.
   */
  public static final int HEADER_SIZE   = 16;

  /**
   *  There is currently no trailer added for each of the
   *  <code>mss</code> level packets.
   */
  public static final int TRAILER_SIZE  = 0;

  /**
   *  The overall overhead for each packet sent to the network.
   */
  public static final int OVERHEAD_SIZE = HEADER_SIZE + TRAILER_SIZE;

}
