/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.relacs.mss;

import java.io.Externalizable;
import java.io.IOException;
import java.io.ObjectInput;
import java.io.ObjectOutput;

import jgroup.core.EndPoint;
import jgroup.relacs.types.GroupIndex;

/**
 *  The <code>RoutingTable</code> class
 *
 *  @author Salvatore Cammarata
 *  @author Hein Meling
 *  @since Jgroup 1.2
 */
public final class RoutingTable 
  implements MssConstants, Externalizable
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Table with topology information */
  TopologyEntry[] table;

  /** The number of bytes required to marshal this routing table instance */
  private transient int marshalSize;

  /** The reachability threshold for the hosts to include in the routing table. */
  private transient int threshold;

  /** The clusters in this system */
  private transient ClusterTable clustertable;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructors
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** 
   *  Default constructor for externalization
   */
  public RoutingTable()
  {
  }

  /** 
   *  
   */
  public RoutingTable(ClusterTable ct, int threshold)
  {
    marshalSize = GroupIndex.SIZE;
    this.clustertable = ct;
    this.threshold = threshold;

    table = new TopologyEntry[ct.size()];
    for (int i=0; i<table.length; i++) {
      Cluster cluster = ct.get(i);
      table[i] = new TopologyEntry(cluster.getRoutingEntry(), cluster.getReachable(threshold));
      marshalSize += table[i].getMarshalSize();
    }
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Returns the number of bytes required to marshal this routing table
   *  instance.
   */
  public int getMarshalSize()
  {
    return marshalSize;
  }


  /**
   *  Returns the maximum number of bytes required to marshal a routing
   *  table instance, when the reachability set includes all hosts in
   *  the system.
   */
  public static int getMaxSize(int numOfHosts, int numOfClusters)
  {
    return GroupIndex.SIZE + (numOfClusters*TopologyEntry.getMaxSize(numOfHosts));
  }


  /**
   *  Updates the reachability information for all clusters in the
   *  clustertable.
   *
   *  @return 
   *    true if the reachability information in the routing table was
   *    changed as a consequence of this invocation; false otherwise.
   */
  public boolean updateReachability()
  {
    marshalSize = GroupIndex.SIZE;
    boolean changed = false;

    /*
     * This at least avoids creating a new array, given that there is no
     * new clusters in the system.  New clusters should not occur very
     * often, at least not until we implement dynamic system
     * configuration.
     */
    if (table.length != clustertable.size()) {
      table = new TopologyEntry[clustertable.size()];
      for (int i = 0; i < table.length; i++) {
        Cluster cluster = clustertable.get(i);
        table[i] = new TopologyEntry(cluster.getRoutingEntry(), cluster.getReachable(threshold));
        marshalSize += table[i].getMarshalSize();
      }
      changed = true;
    } else {
      for (int i = 0; i < table.length; i++) {
        Cluster cluster = clustertable.get(i);
        /*
         * Note the compound assignment operator below; used to ensure
         * that we reflect a change in any one of the clusters.
         */
        changed |= table[i].updateReachability(cluster.getReachable(threshold));
        marshalSize += table[i].getMarshalSize();
      }
    }
    return changed;
  }


  /** 
   *  Negate the cost as reverse poison 
   */
  public void splitHorizonOn(EndPoint key) 
  {
    for (int i=0; i < table.length; i++)
      if (table[i].routing.route.equals(key))
        table[i].routing.cost = -table[i].routing.cost;
  }


  /** 
   *  Reverts all negative routing costs to positive.
   */
  public void splitHorizonOff() 
  {
    for (int i=0; i < table.length; i++)
      if (table[i].routing.cost < 0)
        table[i].routing.cost = -table[i].routing.cost;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from Externalizable
  ////////////////////////////////////////////////////////////////////////////////////////////
  
  /**
   *  Restores the content of this object from the marshalled data contained
   *  in the specified input stream.
   * 
   *  @param in the stream to be read
   */
  public void readExternal(ObjectInput in)
    throws IOException, ClassNotFoundException
  {
    /* Read the number of entries in the stream */
    int entries = GroupIndex.unmarshal(in);
    table = new TopologyEntry[entries];
    for (int i=0; i < entries; i++) {
      table[i] = new TopologyEntry();
      table[i].readExternal(in);
    }
  }
  
  
  /**
   *  Marshals the content of this object to the specified output stream.
   * 
   *  @param out the stream to be written
   */
  public void writeExternal(ObjectOutput out)
    throws IOException
  {
    GroupIndex.marshal(out, table.length);
    for (int i=0; i<table.length; i++)
      table[i].writeExternal(out);
  }

  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from Object
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Returns a string representation of this object
   */
  public String toString() 
  {
    StringBuilder buf = new StringBuilder("[RoutingTable: length=");
    buf.append(table.length);
    buf.append(", {");
    for (int i = 0; i < table.length; i++) {
      buf.append((i != 0) ? ", " : "");
      buf.append(table[i].toString());
    }
    buf.append("}]");
    return buf.toString();
  }

} // END RoutingTable

