/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.relacs.mss;

import java.io.Externalizable;
import java.io.IOException;
import java.io.ObjectInput;
import java.io.ObjectOutput;

import jgroup.core.EndPoint;
import jgroup.relacs.types.EndPointImpl;


/**
 *  The <code>RoutingTableEntry</code> class.
 *
 *  @author Salvatore Cammarata
 *  @since Jgroup 1.2
 */
final class RoutingTableEntry 
  implements Externalizable
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constants
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Size of this object in bytes (those that are marshalled) */
  public static final int SIZE = EndPointImpl.SIZE + EndPointImpl.SIZE + 4;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Unique identifier for the cluster */
  EndPoint key;

  /** Route for the cluster */
  EndPoint route;

  /** Cost for this route */
  int cost;

  /** TTL */
  int TTL;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructors
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Default constructor for externalization. 
   */
  public RoutingTableEntry()
  {
  }

  /** 
   *  Builds a new routing table entry by setting the specified
   *  endpoint, the initial route to be used for this endpoint,
   *  the initial cost for this endpoint.
   */
  RoutingTableEntry(EndPoint id, EndPoint route, int cost)
  {
    this.key = id;
    this.route = route;
    this.cost = cost;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Marshaling / unmarshaling
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Restores the content of this object from the marshalled data contained
   *  in the specified input stream.
   * 
   *  @param in the stream to be read
   */
  public void readExternal(ObjectInput in)
    throws IOException, ClassNotFoundException
  {
    key = new EndPointImpl();
    key.readExternal(in);
    route = new EndPointImpl();
    route.readExternal(in);
    cost = in.readInt();
  }


  /**
   *  Marshals the content of this object to the specified output stream.
   * 
   *  @param out the stream to be written
   */
  public void writeExternal(ObjectOutput out)
    throws IOException
  {
    key.writeExternal(out);
    route.writeExternal(out);
    if (cost < 0) {
      out.writeInt(-cost);
    } else {
      out.writeInt(cost);
    }
  }

  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods
  ////////////////////////////////////////////////////////////////////////////////////////////


  /**
   *  Returns a string representation of this object
   */
  public String toString() 
  {
    return "[" + key + ", " + route + ", " + cost + "]";
  } 


  /**
   * Compares two objects for content equality.
   *
   * @param obj the object to compare with
   * @return  true if these objects are equal; false otherwise.
   */
  public boolean equals(Object obj)
  {
    if (this == obj)
      return true;
    else
      if (obj instanceof RoutingTableEntry)
        return key.equals(((RoutingTableEntry) obj).key);
      else 
        return false;
  }

} // END RoutingTableEntry
