/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.relacs.mss;

import java.util.concurrent.Future;


/**
 *  The <code>ScheduledEvent</code> class
 *
 *  @author Alberto Montresor
 *  @author Hein Meling
 *  @since  Jgroup 0.1
 */
final class ScheduledEvent 
  implements Comparable, MssTag
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Timeout value */
  private long timeout;
  
  /** Tag associated to this event */
  private byte tag;
  
  /** Object associated to this event */
  private Object obj;

  private Future<?> handle;

  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructors
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *
   */
  ScheduledEvent(byte tag)
  {
    this(tag, null);
  }

  /**
   *
   */
  ScheduledEvent(byte tag, Object obj)
  {
    this.tag = tag;
    this.obj = obj;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Set the timeout of this event. 
   */
  void setTimeout(long timeout) 
  { 
    this.timeout = timeout;
  }

  /**
   *  Set the timeout of this event and schedule the timer task.
   */
  public void setTimeout(long timeout, Future<?> handle)
  {
    this.timeout = timeout;
    this.handle = handle;
  }

  public void abortTimeout()
  {
    handle.cancel(false);
  }

  /**
   *  Returns the actual timeout associated to this event. 
   */
  long getTimeout() 
  { 
    return timeout;
  }

  /**
   *  Returns the event object associated to this event. 
   */
  Object getData() 
  { 
    return obj;
  }

  /**
   *  Returns the tag associated to this event 
   */
  byte getTag() 
  {
    return tag;
  }
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from Object and Comparable
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Compares this object with the specified object for order.
   */
  public int compareTo(Object obj)
  {
    if (this == obj) {
      return 0;
    } else {
      ScheduledEvent event = (ScheduledEvent) obj;
      if (timeout < event.timeout) {
        return -1;
      } else if (timeout == event.timeout) {
        return 0;
      } else {
        return 1;
      }
    }
  }
  
  /**
   *
   */
  public int hashCode()
  {
    return (int) timeout;
  }

  /**
   * Compares two objects for content equality.
   *
   * @param obj the object to compare with
   * @return  true if these objects are equal; false otherwise.
   */
  public boolean equals(Object obj)
  {
    if (this == obj) {
      return true;
    } else {
      if ( (obj != null) && (obj instanceof ScheduledEvent) ) {
        ScheduledEvent event = (ScheduledEvent) obj;
        return (event.timeout == timeout);
      } else {
        return false;
      }
    }
  }

  /**
   *  Returns a string representation of this object
   */
  public String toString()
  {
    StringBuilder buffer = new StringBuilder();
    buffer.append("[");
    if (tag >= TAGTYPE.length || tag < 0) {
      buffer.append("INVALID TAG");
    } else {
      buffer.append(TAGTYPE[tag]);
    }
    buffer.append(", timeout=");
    buffer.append(timeout);
//    buffer.append(timeout - System.currentTimeMillis());
    if (obj != null) {
      buffer.append(", obj=");
      buffer.append(obj);
    }
    buffer.append("]");
    return buffer.toString();
  }

} // END ScheduledEvent
