/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.relacs.registry;

import java.io.Externalizable;
import java.io.IOException;
import java.io.ObjectInput;
import java.io.ObjectOutput;

import jgroup.core.VMID;
import jgroup.relacs.types.VMIDImpl;
import jgroup.util.IntList;

/**
 *  This class is a compact representation of the set of bindings 
 *  (active and removed) currently maintained in a registry replica.
 *  Objects of this class are generated by the registry during merging
 *  operations, and contains all currently active bindings, together
 *  with the identifiers of the bindings that have been removed from
 *  the registry.
 *
 *  @author Alberto Montresor
 *  @since Jgroup 0.4
 */
public final class MergingData
  implements Externalizable
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  private static final long serialVersionUID = -3216043980643626103L;

  /** 
   *  Array of virtual machine identifiers for servers that are registered
   *  in the registry that generated this object.
   */
  VMID[] vmids;

  /**
   *  For each of the virtual machines in vmids, the numeric identifiers
   *  of each of the bindings that have been removed from the registry.
   */
  IntList[] removed;

  /**
   *  For each of the virtual machines in vmids, the array of bindings
   *  still active.
   */
  BindingList[][] bindings;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Object methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Returns a string representation of this object
   */
  public String toString()
  {
    StringBuilder buf = new StringBuilder("[MergingData: ");
    int initlen = buf.length();
    if (bindings.length > 0) {
      buf.append("; bindings: ");
      for (int i = 0; i < bindings.length; i++) {
        for (int j = 0; j < bindings[i].length; j++) {
          buf.append(bindings[i][j]);
          buf.append(", ");
        }
      }
    }
    if (removed.length > 0) {
      buf.append("; removed: ");
      for (int i = 0; i < removed.length; i++) {
        buf.append(removed[i]);
        buf.append(", ");
      }
    }
    if (vmids.length > 0) {
      buf.append("VMIDS: ");
      for (int i = 0; i < vmids.length; i++) {
        buf.append(vmids[i]);
        buf.append(", ");
      }
    }
    if (buf.length() == initlen)
      buf.append("<empty> ]");
    else
      buf.append("]");
    return buf.toString();
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Externalization
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Marshals the content of this object to the specified output stream.
   * 
   *  @param out the stream to be written
   */
  public void writeExternal(ObjectOutput out)
    throws IOException
  {
    // Write vmids
    int size = vmids.length;
    out.writeInt(size);
    for (int i = 0; i < size; i++)
      vmids[i].writeExternal(out);

    // Write removed
    size = removed.length;
    out.writeInt(size);
    for (int i = 0; i < size; i++)
      removed[i].writeExternal(out);

    // Write bindings
    size = bindings.length;
    out.writeInt(size);
    for (int i = 0; i < size; i++) {
      int size2 = bindings[i].length;
      out.writeInt(size2);
      for (int j = 0; j < size2; j++)
        bindings[i][j].writeExternal(out);
    }
  }


  /**
   *  Restores the content of this object from the marshalled data contained
   *  in the specified input stream.
   * 
   *  @param in the stream to be read
   */
  public void readExternal(ObjectInput in)
    throws IOException, ClassNotFoundException
  {
    // Read vmids
    int size = in.readInt();
    vmids = new VMID[size];
    for (int i = 0; i < size; i++) {
      vmids[i] = new VMIDImpl();
      vmids[i].readExternal(in);
    }

    // Read removed
    size = in.readInt();
    removed = new IntList[size];
    for (int i = 0; i < size; i++) {
      removed[i] = new IntList();
      removed[i].readExternal(in);
    }

    // Read bindings
    size = in.readInt();
    bindings = new BindingList[size][];
    for (int i = 0; i < size; i++) {
      int size2 = in.readInt();
      bindings[i] = new BindingList[size2];
      for (int j = 0; j < size2; j++) {
        bindings[i][j] = new BindingList();
        bindings[i][j].readExternal(in);
      }
    }
  }

} // END MergingData
