/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.relacs.registry;

import java.rmi.NotBoundException;
import java.rmi.RemoteException;

import jgroup.core.ExternalGMIListener;
import jgroup.core.ExternalGMIService;
import jgroup.core.IID;
import jgroup.core.JgroupException;
import jgroup.core.MembershipListener;
import jgroup.core.MembershipService;
import jgroup.core.View;
import jgroup.core.registry.DependableRegistry;
import jgroup.core.registry.RegistryFactory;
import jgroup.core.registry.RegistryService;
import jgroup.relacs.config.AppConfig;

import org.apache.log4j.Logger;

/**
 *  This layer implements the registry service interface, allowing
 *  the server associated with the local group manager, or layers within
 *  the group manager to gain access to the content of the dependable
 *  registry.
 *
 *  @author Alberto Montresor
 *  @author Hein Meling
 *  @since Jgroup 2.1
 */
public class RegistryLayer
  implements RegistryService, MembershipListener
{
	
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Logger
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Obtain logger for this class */
  private static final Logger log = Logger.getLogger(RegistryLayer.class);


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** The external GMI service reference */
  private ExternalGMIService egmis;

  /** IID representing this replica's binding in the dependable registry. */
  private IID iid = null;

  /** The server object using this layer. */
  private Object server = null;

  /** The application configuration information for the server. */
  private AppConfig thisApp;

  /** The lease time to be used by the dependable registry (initially disabled = -1) */
  private long leaseTime = -1;

  /** The shutdown hook thread used to unbind from the dependable registry */
  private Thread shutdownHook = null;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructor / Static factory
  ////////////////////////////////////////////////////////////////////////////////////////////

  private RegistryLayer(ExternalGMIService egmis)
  {
    this.egmis = egmis;
  }

  public static RegistryLayer getLayer(ExternalGMIService egmis, MembershipService pgms)
  {
    return new RegistryLayer(egmis);
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Layer interface methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  This method does not need to do any processing since the
   *  RegistryListener interface is an empty marker.
   */
  public void addListener(Object listener)
  {
    if (listener == null)
      throw new NullPointerException("No replica specified for the registry layer.");
    this.server = listener;
    /* Get the application object for this replica */
    thisApp = AppConfig.getApplication(server);
    if (thisApp.hasService("LeaseRefresh")) {
      leaseTime = thisApp.getIntParam("LeaseRefresh.leaseTime", DEFAULT_LEASE_TIME);
    }

    /*
     * If we are not using the lease layer or the notify layer, we install
     * the shutdown hook so as to perform a clean release of the replica,
     * once a CTRL-C is pressed.  We don't want to use this shutdown hook
     * when testing the lease and notify layers.
     */
    if (!thisApp.hasService("LeaseRefresh") && !thisApp.hasService("NotifyRefresh")) {
      /*
       * Create a shutdown hook thread that will be executed when the server
       * exits relatively cleanly (e.g., through CTRL-C), it will unbind the
       * replica from the dependable registry before exiting.
       */
      shutdownHook = new Thread("RegistryLayer-ShutdownHook") {
        public void run() {
          doUnbind();
        }};
    }
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // RegistryService interface methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  public IID bind()
    throws RemoteException, JgroupException
  {
    return bind(server);
  }

  public IID bind(Object server)
    throws RemoteException, JgroupException
  {
    String name = AppConfig.getApplication(server).getRegistryName();
    return bind(name, server);
  }

  public IID bind(String name, Object server)
    throws RemoteException, JgroupException
  {
    DependableRegistry registry = RegistryFactory.getRegistry();
    return bind(name, server, registry);
  }

  public IID bind(String name, Object server, DependableRegistry registry)
    throws RemoteException, JgroupException
  {
    if (server == null) 
      throw new NullPointerException("Server is null; cannot bind");
    Class serverClass = server.getClass();
    if (!ExternalGMIListener.class.isAssignableFrom(serverClass))
      throw new JgroupException("Cannot bind server since it does not implement an external GMI interface");
    return registry.bind(name, egmis.getRegistryEntry(name), serverClass, leaseTime);
  }

  public void unbind()
    throws RemoteException, NotBoundException, JgroupException
  {
    if (iid == null)
      throw new NullPointerException("Cannot unbind(null)");
    unbind(iid);
  }

  public void unbind(IID bid)
    throws RemoteException, NotBoundException, JgroupException
  {
    DependableRegistry registry = RegistryFactory.getRegistry();
    unbind(bid, registry);
  }

  public void unbind(IID bid, DependableRegistry registry)
    throws RemoteException, NotBoundException
  {
    if (shutdownHook != null) {
      /* If unbind was successful, remove the shutdown hook to avoid extra unbinds */
      Runtime.getRuntime().removeShutdownHook(shutdownHook);
    }
    registry.unbind(bid);
  }

  public void refreshLease()
    throws RemoteException, JgroupException
  {
    DependableRegistry registry = RegistryFactory.getRegistry();
    try {
      registry.refresh(iid);
    } catch (NotBoundException nbe) {
      log.warn(nbe.getMessage());
    }
  }

  public IID getIID()
    throws NullPointerException
  {
    if (iid == null)
      throw new NullPointerException("iid is null; Replica not yet bound to the dependable registry.");
    return iid;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // MembershipListener interface methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Bind the local server with the dependable registry if:
   * 
   * <ol>
   * <li> it has not already been bound,
   * <li> it is configured to auto bind and
   * <li> if the this view represents a fully replicated service view.
   * </ol>
   * 
   * @see jgroup.core.MembershipListener#viewChange(jgroup.core.View)
   */
  public void viewChange(View view)
  {
    if (iid == null &&
        thisApp.getBooleanParam("Registry.auto") &&
        view.size() == thisApp.getInitialRedundancy()) {
      /*
       * Perform the bind method in a separate thread to avoid delays in
       * the view install phase.
       */
      new Thread("BindThread-" + thisApp.getGroupId()) {
        public void run() {
          try {
            if (log.isDebugEnabled())
              log.debug("Group is fully replicated; auto binding with registry");
            iid = bind();
            if (log.isDebugEnabled())
              log.debug("Successfully bound: " + thisApp.getRegistryName());
            if (shutdownHook != null) {
              /*
               * Note that we don't start the shutdown hook thread here; instead
               * when the JVM exit, it will start all registered shutdown hooks.
               */
              Runtime.getRuntime().addShutdownHook(shutdownHook);
            }
          } catch (Exception e) {
            log.error("Unable to bind " + thisApp.getRegistryName() + " to the dependable registry", e);
          }
        }
      }.start();
    }
  }

  /* (non-Javadoc)
   * @see jgroup.core.MembershipListener#prepareChange()
   */
  public void prepareChange() {}

  /* (non-Javadoc)
   * @see jgroup.core.MembershipListener#hasLeft()
   */
  public void hasLeft()
  {
    doUnbind();
  }

  /**
   * Do unbind from the dependable registry, but only if it was previously bound.
   */
  private void doUnbind()
  {
    // It may happen that we never bind with the registry; no need to unbind then.
    if (iid == null)
      return;
    try {
      /* Unbind the replica from the dependable registry proxy. */
      unbind(iid);
      if (log.isDebugEnabled())
        log.debug("Replica was unbound from registry: " + iid);
    } catch (Exception e) {
      log.warn("Failed to unbind replica from registry", e);
    }
  }

} // END RegistryLayer
