/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.relacs.rmi;

import java.lang.reflect.InvocationHandler;
import java.lang.reflect.Proxy;
import java.rmi.Remote;
import java.rmi.RemoteException;
import java.util.ArrayList;

import jgroup.core.ExternalGMIListener;

/**
 *  
 */
public class Exporter
{

  
  /**
   *  Returns a proxy object for the specified server class, using the 
   *  specified invocation handler.
   *
   *  If <CODE>serverClass</CODE> does not implement any external interface
   *  (i.e., an interface that extends <tt>jgroup.core.ExternalGMIListener</tt>), a
   *  <CODE>RemoteException</CODE> is thrown.
   *
   *  @exception RemoteException Thrown if for some reason, it is impossible to 
   *    instantiate a proxy class.
   */
  public static Remote getProxy(Class serverClass, InvocationHandler handler)
    throws RemoteException
  {
    Class[] interfaces = serverClass.getInterfaces();
    ArrayList array = new ArrayList();
    for (int i=0; i < interfaces.length; i++)
      if (ExternalGMIListener.class.isAssignableFrom(interfaces[i]))
        array.add(interfaces[i]);
    interfaces = (Class[]) array.toArray(new Class[array.size()]);
    if (interfaces.length == 0)
      throw new RemoteException("Unable to instantiate a dynamic proxy for class " + 
        serverClass.getName() + ": does not implement any external GMI interface");

    try {
      return (Remote) Proxy.newProxyInstance(serverClass.getClassLoader(), 
        interfaces, handler);
    } catch (Exception e) {
      throw new RemoteException("Unable to instantiate a dynamic proxy " + 
        "for class " + serverClass.getName(), e);
    }
  }

  /**
   *  Returns a proxy object for set of specified interfaces, using
   *  the specified invocation handler.
   *
   *  @exception RemoteException Thrown if for some reason, it is impossible to 
   *    instantiate a proxy class.
   */
  public static Remote getProxy(Class[] interfaces, InvocationHandler handler)
    throws RemoteException
  {
    /* Warning message */
    if (interfaces == null || interfaces.length == 0)
      throw new RemoteException(
        "Unable to instantiate a dynamic proxy: no interfaces provided");
    try {
      return (Remote) Proxy.newProxyInstance(interfaces[0].getClassLoader(), 
        interfaces, handler);
    } catch (Exception e) {
      StringBuilder buffer = new StringBuilder();
      buffer.append("Unable to instantiate a dynamic proxy for interfaces ");
      for (int i=0; i<interfaces.length; i++) { 
        buffer.append(interfaces[i].getName());
        if (i < interfaces.length - 1)
          buffer.append(", ");
      }
      throw new RemoteException(buffer.toString(), e);
    }
  }
  
} // END Exporter
