/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.relacs.rmi;

import jgroup.core.IID;
import jgroup.core.VMID;
import jgroup.relacs.types.IIDImpl;
import jgroup.relacs.types.VMIDImpl;
import jgroup.util.IntList;


/*
 *  FIXME: Not sure that having acknowledgement info piggybacked int
 *  IID is a good idea. Probably we should study some alternative
 *  design.
 */

/**
 *  This class is a factory for invocation identifiers.  It uses a
 *  sequence number to generate unique identifiers, and keeps track of
 *  invocation identifiers for which a response has been received.  This
 *  information is used to generate acknowledgement information sent to
 *  servers in order that they can remove replies from their tables.
 *
 *  @author Alberto Montresor
 *  @since Jgroup 0.8
 */
public class IIDFactory
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Unique id generated for this factory */
  private VMID vmid; 

  /** Counter used to create unique identifiers */
  private int counter;

  /**
   *  Set containing the identifiers of the invocations for which a
   *  reply has been received.  Because all identifiers are created
   *  through this class, we may safely maintain only the sequence
   *  number of those identifiers, instead of complete <code>IID</code>
   *  objects.
   */
  private IntList replies;

  /** 
   *  Contains the final value of the first sequence of consecutive
   *  identifiers for which a reply has been received.
   */
  private int ack;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructor
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Initializes a new <code>IIDFactory</code> object by setting the
   *  sequence number to 0 and creating an empty replies list.  Ack
   *  information is also set to 0.
   */
  public IIDFactory()
  {
    vmid = VMIDImpl.getLocalVMID();
    counter = 0;
    replies = new IntList();
    ack = 0;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Creates a new invocation identifier.  It increases the sequence
   *  number and constructs a new <code>IIDImpl</code> object with the
   *  new sequence number; acknowledgement information is piggybacked in
   *  the identifier.
   */
  public synchronized IID getId()
  {
    return new IIDImpl(vmid, ++counter, ack);
  }

  /**
   *  Adds the given invocation identifier to the set of identifiers for
   *  which a reply has been received.
   */
  public synchronized void addReply(IID id)
  {
    replies.insert(id.getCounter());
    ack = replies.getMaxConsecutive();
    if (ack < 0)
      ack = 0;
  }

} // END IIDFactory
