/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.relacs.rmi;

import java.io.ByteArrayOutputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.lang.reflect.Method;
import java.security.AccessController;
import java.security.DigestOutputStream;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.security.PrivilegedAction;
import java.util.HashMap;
import java.util.Iterator;

import jgroup.core.InternalGMIListener;

/**
 *  Helper class usueful to associate a "method hash" to each of the
 *  methods contained in a set of interfaces. The method hash is a long
 *  containing the first 64 bits of the SHA digest from the UTF encoded
 *  string of the method name and descriptor. The table maintains two
 *  hash table, one for methods and another for hash numbers.
 *
 *  (Partly copied from Java 2 SDK 1.3)
 *
 *  @author Alberto Montresor
 *  @author Hein Meling
 *  @since Jgroup 0.9
 */
public final class ServerMethodTable
{
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  private HashMap   table;
  private HashMap   invertedTable;
  private HashMap   serverTable;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructors
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Creates a new <code>ServerMethodTable</code> instance, with empty
   *  tables.
   */
  public ServerMethodTable()
  {
    table = new HashMap();
    invertedTable = new HashMap();
    serverTable = new HashMap();
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Access methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Add methods from the given GMI interface implemented by the
   *  provided server.  For each remote method in
   *  the given GMI interfaces, a mapping is created between a hash of the
   *  method and the java.lang.reflect.Method object (and the reverse
   *  mapping).  In addition, the hash of the method is also associated
   *  with the server object that is actually implementing these
   *  methods.  The latter is used to enable multiple listeners (servers/layers)
   *  to receive method invocations, through the same EGMI/IGMI layer.
   */
  public void addMethods(Object server, Class gmiListenerClass)
  {
    /*
     * Determine the GMI listener interfaces implemented by the
     * server object, and compute relevant method information.
     */
    Class[] interfaces = server.getClass().getInterfaces();
    for (int i = 0; i < interfaces.length; i++) {
      if (gmiListenerClass.isAssignableFrom(interfaces[i])) {
        Method methods[] = interfaces[i].getMethods();
        for (int j = 0; j < methods.length; j++) {
          /*
           * Special handling for InternalGMI interfaces; they are not
           * allowed to have other return types than Object or void.
           * This is since the return type from the IGMI proxy is an
           * array of values.  That is, the Object class can also
           * encapsulate an Object[], and thus IGMI is allowed to return
           * multiple values.
           */
          if (gmiListenerClass.equals(InternalGMIListener.class)) {
            Class retType = methods[j].getReturnType();
            if (!(retType.toString().equals("void") || retType.equals(Object.class))) {
              throw new IllegalArgumentException(
                "Return type for methods in the internal interfaces must be Object or void:\n"
                + methods[j]);
            }
          }
          String name = getMethodNameAndDescriptor(methods[j]);
          Long hashLong = new Long(computeInfo(methods[j]));
          table.put(hashLong, methods[j]);
          serverTable.put(hashLong, server);
          invertedTable.put(name, hashLong);
        }
      }
    }
  }


  /**
   *  Get the method associated with the given hash value.
   */
  public Method get(long hash)
  {
    Long hashLong = new Long(hash);
    return (Method) table.get(hashLong);
  }


  /**
   *  Get the hash value associated with the given method.
   */
  public long get(Method m)
  {
    Long hashLong = (Long) invertedTable.get(getMethodNameAndDescriptor(m));
    return hashLong.longValue();
  }


  /**
   *  Get the server associated with the given hash value representing a
   *  method.
   */
  public Object getServer(long hash)
  {
    return serverTable.get(new Long(hash));
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Object methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  public String toString()
  {
    StringBuilder buf = new StringBuilder("ServerMethodTable[");
    for (Iterator iter = table.keySet().iterator(); iter.hasNext();) {
      Long hash = (Long) iter.next();
      buf.append("\n");
      buf.append(hash);
      buf.append(" -> ");
      buf.append(((Method) table.get(hash)).getName());
      buf.append("; server=");
      buf.append(serverTable.get(hash).getClass().getName());
    }
    buf.append("]");
    return buf.toString();
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Static methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Computes and returns a hash of the given method. Used when 
   * serializing remote invocations.
   */
  static long computeInfo(final Method method)
  {
    /*
     * Set this Method object to override language
     * access checks so that the dispatcher can invoke
     * methods from non-public remote interfaces.
     */
    AccessController.doPrivileged(new PrivilegedAction() {
      public Object run() {
        method.setAccessible(true);
        return null;
      }
    });

    return computeHash(method);
  }

  /**
   * Compute the "method hash" of a remote method.  The method hash
   * is a long containing the first 64 bits of the SHA digest from
   * the UTF encoded string of the method name and descriptor.
   */
  static long computeHash(Method m)
  {
    long hash = 0;
    ByteArrayOutputStream sink = new ByteArrayOutputStream(127);
    try {
      MessageDigest md = MessageDigest.getInstance("SHA");
      DataOutputStream out = new DataOutputStream(
        new DigestOutputStream(sink, md));

      String s = getMethodNameAndDescriptor(m);

      out.writeUTF(s);

      // use only the first 64 bits of the digest for the hash
      out.flush();
      byte hasharray[] = md.digest();
      for (int i = 0; i < Math.min(8, hasharray.length); i++) {
        hash += ((long) (hasharray[i] & 0xFF)) << (i * 8);
      }
    } catch (IOException ignore) {
      /* can't happen, but be deterministic anyway. */
      hash = -1;
    } catch (NoSuchAlgorithmException complain) {
      throw new SecurityException(complain.getMessage());
    }
    return hash;
  }


  /**
   * Return a string consisting of the given method's name followed by
   * its "method descriptor", as appropriate for use in the computation
   * of the "method hash".
   *
   * See section 4.3.3 of The Java Virtual Machine Specification for
   * the definition of a "method descriptor".
   */
  static String getMethodNameAndDescriptor(Method m)
  {
    StringBuilder desc = new StringBuilder(m.getName());
    desc.append("(");
    Class[] paramTypes = m.getParameterTypes();
    for (int i = 0; i < paramTypes.length; i++) {
      desc.append(getTypeDescriptor(paramTypes[i]));
    }
    desc.append(")");
    Class returnType = m.getReturnType();
    if (returnType == void.class) { // optimization: handle void here
      desc.append("V");
    } else {
      desc.append(getTypeDescriptor(returnType));
    }
    return desc.toString();
  }

  /**
   * Get the descriptor of a particular type, as appropriate for either
   * a parameter or return type in a method descriptor.
   */
  static String getTypeDescriptor(Class type)
  {
    if (type.isPrimitive()) {
      if (type == int.class) {
        return "I";
      } else if (type == boolean.class) {
        return "Z";
      } else if (type == byte.class) {
        return "B";
      } else if (type == char.class) {
        return "C";
      } else if (type == short.class) {
        return "S";
      } else if (type == long.class) {
        return "J";
      } else if (type == float.class) {
        return "F";
      } else if (type == double.class) {
        return "D";
      } else if (type == void.class) {
        return "V";
      } else {
        throw new Error("unrecognized primitive type: " + type);
      }
    } else if (type.isArray()) {
      /*
       * According to JLS 20.3.2, the getName() method on Class does
       * return the VM type descriptor format for array classes (only);
       * using that should be quicker than the otherwise obvious code:
       *
       *     return "[" + getTypeDescriptor(type.getComponentType());
       */
      return type.getName().replace('.', '/');
    } else {
      return "L" + type.getName().replace('.', '/') + ";";
    }
  }

} // END ServerMethodTable
