/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.relacs.types;

import java.io.IOException;
import java.io.ObjectInput;
import java.io.ObjectOutput;

/**
 *  The <code>Checksum</code> class wraps a value of the relacs daemon type 
 *  "checksum" in an object. A checksum is used to quickly verify if two
 *  proposals are equal or not. If the checksum is different, the proposal
 *  are different. Otherwise, we have to check their content.
 *
 *  @author Alberto Montresor
 *  @since  Jgroup 1.2
 */

public final class Checksum
{
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constants
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Size of this object in bytes */
  public static final int SIZE = 1;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////
  
  /** Wrapped value */
  private int value;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructor
  ////////////////////////////////////////////////////////////////////////////////////////////
  
  /**
   *  Creates a <CODE>Checksum</CODE> object containing the specified
   *  value.
   */
  public Checksum(int value)
  {
    this.value = value;
  }
  
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Accessor method
  ////////////////////////////////////////////////////////////////////////////////////////////
  
  /**
   *  Returned the wrapped value.
   */
  public int getValue()
  {
    return value;
  }
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Marshaling / unmarshaling methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Marshals the specified checksum to an outgoing message.
   */
  public static void marshal(ObjectOutput msg, int value)
    throws IOException
  {
    msg.write(value);
  }
  
  /**
   *  Unmarshals a checksum from an incoming message.
   */
  public static int unmarshal(ObjectInput msg)
    throws IOException
  {
    return msg.read();
  }

  /**
   * Returns the checksum of a sequence of bytes starting at position <code> start</code>
   * and ending at position <code> stop</code> of array <code> data</code>.
   * index and an ending index.
   *
   * @param data  the byte array
   * @param start       the starting index.
   * @param stop  the ending index.
   */
  static int computeChecksum(byte[] data, int start, int stop)
  {
    byte  check = 0;          // Return value
    
    for (int i=start; i<stop; i++)
      check ^= data[i];
    return check;
  }
  
} // END Checksum
