/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.relacs.types;

import java.io.IOException;
import java.io.ObjectInput;
import java.io.ObjectOutput;
import java.net.InetAddress;
import java.net.UnknownHostException;

import jgroup.core.EndPoint;
import jgroup.util.Abort;
import jgroup.util.Network;

/**
 *
 *
 *  @author Alberto Montresor
 *  @author Hein Meling
 *  @since Jgroup 1.2
 */
public class EndPointImpl
  implements EndPoint
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constants
  ////////////////////////////////////////////////////////////////////////////////////////////

  private static final long serialVersionUID = -8548927439519782852L;


  /** Size of this object (in bytes) when marshalled. */
  public static final int SIZE = 6;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////
  
  /** Internet address */
  protected InetAddress address;

  /** Internet address represented as an integer */
  private int packedAddress;
  
  /** Port number */
  protected int port;

  /** True if this endpoint is the local endpoint; false otherwise */
  protected boolean local;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructors
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Default constructor for externalization
   */
  public EndPointImpl()
  {
  }

  /**
   *  Constructs a new endpoint for the specified internet
   *  address and port number. Computes the packed address
   *  and checks if the endpoint is local.
   */
  public EndPointImpl(InetAddress address, int port)
  {
    this.address = address;
    this.port = port;
    packedAddress = Network.translate(address);
    local = Network.isLocal(packedAddress);
  }

  /**
   *  Constructs a new endpoint for the internet address 
   *  specified as integer, and the specified port number.
   *  The InetAddress version of the address is computed
   *  only when needed, to avoid useless DNS lookups.
   */
  public EndPointImpl(int packedAddress, int port)
  {
    this.packedAddress = packedAddress;
    this.port = port;
    this.local = Network.isLocal(packedAddress);
    address = null;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Returns the <CODE>InetAddress</CODE> associated to this endpoint.
   *  If needed, performs the translation between the integer version
   *  of the address and the <CODE>InetAddress</CODE> version.
   */
  public InetAddress getAddress()
  {
    checkAddress();
    return address;
  }

  /**
   *  Return the port number associated to this end point
   */
  public int getPort()
  {
    return port;
  }

  /**
   *
   */
  public int getIntAddress()
  {
    return packedAddress;
  }

  /**
   *
   */
  public boolean isMulticastEndPoint()
  {
    checkAddress();
    return address.isMulticastAddress();
  }

  /**
   *
   */
  public boolean isLocal()
  {
    return local;
  }


  /**
   * If <code>address</code> is <code>null</code>, tries to convert
   * packed address to <code>InetAddress</code>.
   */
  private void checkAddress()
  {
    if (address == null) {
      try {
        address = Network.translate(packedAddress);
      } catch (UnknownHostException e) {
        Abort.exit("Integer could not be converted into a valid IP-address.", e);
      }
    }
  }
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from Object and Comparable
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Compares this object with the specified object for order.
   */
  public int compareTo(Object obj)
  {
    if (this == obj) {
      return 0;
    } else {
      EndPointImpl endpoint = (EndPointImpl) obj;
      if (packedAddress < endpoint.packedAddress) {
        return -1;
      } else if (packedAddress == endpoint.packedAddress) {
        if (port < endpoint.port)
          return -1;
        else if (port == endpoint.port)
          return 0;
        else
          return 1;
      } else {
        return 1;
      }
    }
  }

  /**
   *
   */
  public int hashCode()
  {
    return packedAddress ^ port;
  }


  /**
   *  Compares two <code>EndPointImpl</code> objects for content
   *  equality.
   *
   *  @param obj
   *    The object to compare this object with.
   *  @return 
   *    True if the objects are equal; false otherwise.
   */
  public boolean equals(Object obj)
  {
    if (!(obj instanceof EndPointImpl)) {
      return false;
    }
    EndPointImpl endpoint = (EndPointImpl) obj;
    return (endpoint.packedAddress == packedAddress && endpoint.port == port);
  }


  /**
   *  Returns a string representation of this object
   */
  public String toString()
  {
    StringBuilder buf = new StringBuilder();
    buf.append("[");
    buf.append(Network.translateToString(packedAddress));
//FIXME enable port # printing again later
//    buf.append(":");
//    buf.append(port);
    buf.append("]");
    return buf.toString();
  }

  
  ////////////////////////////////////////////////////////////////////////////////////////////
  //  Marshaling/unmarshaling methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Restores the content of this object from the marshalled data contained
   *  in the specified input stream.
   * 
   *  @param in the stream to be read
   */
  public void readExternal(ObjectInput in)
    throws ClassNotFoundException, IOException
  {
    packedAddress = in.readInt();
    port = in.readUnsignedShort();
    local = Network.isLocal(packedAddress);
    address = null;
  }

  /**
   *  Marshals the content of this object to the specified output stream.
   * 
   *  @param out the stream to be written
   */
  public void writeExternal(ObjectOutput out)
    throws IOException
  {
    out.writeInt(packedAddress);
    out.writeShort(port);
  }

} // END EndPointImpl
