/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.relacs.types;

import java.io.IOException;
import java.io.ObjectInput;
import java.io.ObjectOutput;

import jgroup.core.IID;
import jgroup.core.VMID;

/**
 *  Instances of this class uniquely identify invocations issued by a
 *  client.  Each invocation identifier is composed of a virtual machine
 *  identifier, identifying the client, and a progressive counter used
 *  to distinguish among invocations performed by the same client.
 *
 *  @author Alberto Montresor
 *  @since Jgroup 0.9
 */
public final class IIDImpl
  implements IID
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  private static final long serialVersionUID = 3101263354396256254L;

  /* Identifier for the virtual machine that generated this invocation. */
  private VMID vmid;

  /* Sequence number for this invocation. */
  private int counter;

  /* Acknowledge counter for this invocation. */
  private int ack;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructors
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Default public constructor for externalization.
   */
  public IIDImpl()
  {
  }


  /**
   *  Constructs a new invocation identifier.
   */
  public IIDImpl(VMID vmid, int counter, int ack)
  {
    this.ack = ack;
    this.counter = counter;
    this.vmid = vmid;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // IID interface methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Returns the <code>VMID</code> of the (client) virtual machine that
   *  generated this invocation.
   */
  public VMID getVmid()
  {
    return vmid;
  }


  /**
   *  Returns the sequential counter identifying this particular
   *  invocation.
   */
  public int getCounter()
  {
    return counter;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Package methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Returns ack information for this invocation identifier.
   */
  public int getAck()
  {
    return ack;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Object methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Returns a string representation of this object
   */
  public String toString()
  {
    StringBuilder buf = new StringBuilder();
    buf.append("[IID: ");
    buf.append(vmid.toString());
    buf.append(", ");
    buf.append(counter);
    buf.append(", ");
    buf.append(ack);
    buf.append("]");
    return buf.toString();
  }


  /**
   *  Returns a hashcode for the <code>IID</code>.  Two
   *  <code>IID</code>s will have the same hashcode if they are equal
   *  with respect to their content.
   */
  public int hashCode()
  {
    return vmid.hashCode() + counter;
  }


  /**
   *  Compares two objects for content equality.
   *
   *  @param obj 
   *    the object to compare with
   *  @return 
   *    true if these objects are equal; false otherwise.
   */
  public boolean equals(Object obj)
  {
    if (!(obj instanceof IIDImpl))
      return false;
    IIDImpl iid = (IIDImpl) obj;
    return (vmid.equals(iid.vmid) && counter == iid.counter);
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Externalizable methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Write the content of this IID on stream <tt>out</tt>.  Note that
   *  method writeExternal of vmid is called, to improve the efficiency
   *  of this method.
   * 
   *  @param out the stream to be written
   */
  public void writeExternal(ObjectOutput out)
    throws IOException
  {
    out.writeInt(ack);
    out.writeInt(counter);
    vmid.writeExternal(out);
  }


  /**
   *  Restores the content of this object from the marshalled data contained
   *  in the specified input stream.
   * 
   *  @param in the stream to be read
   */
  public void readExternal(ObjectInput in)
    throws IOException, ClassNotFoundException
  {
    ack = in.readInt();
    counter = in.readInt();
    vmid = new VMIDImpl();
    vmid.readExternal(in);
  }

} // END IIDImpl
