/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.relacs.types;

import java.io.Externalizable;
import java.io.IOException;
import java.io.ObjectInput;
import java.io.ObjectOutput;

import jgroup.core.EndPoint;

/**
 *  A <code>LocalId</code> object is used to distinguish multiple local
 *  members associated with a single Jgroup daemon.  It is comprised of
 *  the member number and its <code>EndPoint</code>.
 *
 *  @author Hein Meling
 *  @since Jgroup 1.2
 */
public final class LocalId
  implements Externalizable
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constants
  ////////////////////////////////////////////////////////////////////////////////////////////

  private static final long serialVersionUID = -340854940914476631L;

  /** Size of this object in bytes */
  public static final int SIZE = 2 + EndPointImpl.SIZE;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////
  
  /** Member number */
  private int member;

  /** The local member endpoint address */
  private EndPoint endpoint;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructor
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Default constructor for externalization */
  public LocalId() {}


  /**
   *  Creates a <CODE>LocalId</CODE> object containing the specified
   *  member number, and member <code>EndPoint</code> address.
   */
  public LocalId(int member, EndPoint membEndPoint)
  {
    this.member = member;
    this.endpoint = membEndPoint;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Access methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  public EndPoint getEndpoint()
  {
    return endpoint;
  }

  public int getMemberNo()
  {
    return member;
  }

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Marshaling / unmarshaling methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Restores the content of this object from the marshalled data contained
   *  in the specified input stream.
   * 
   *  @param in the stream to be read
   */
  public void readExternal(ObjectInput in)
    throws IOException, ClassNotFoundException
  {
    member = in.readUnsignedShort();
    endpoint = new EndPointImpl();
    endpoint.readExternal(in);
  }


  /**
   *  Marshals the content of this object to the specified output stream.
   * 
   *  @param out the stream to be written
   */
  public void writeExternal(ObjectOutput out)
    throws IOException
  {
    out.writeShort(member);
    endpoint.writeExternal(out);
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Object methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Returns <code>true</code> if and only if the given object is a non-null
   *  <code>LocalId</code> object, whose member number and endpoint address
   *  is equal to this object.  Otherwise, <code>false</code> is returned.
   */
  public boolean equals(Object obj) 
  {
    if (!(obj instanceof LocalId)) {
      return false;
    } else {
      LocalId localId = (LocalId) obj;
      return (member == localId.member && endpoint.equals(localId.endpoint));
    }
  }


  /**
   *  Returns a hashcode for this <code>LocalId</code> object.
   */
  public int hashCode() 
  {
    return member ^ endpoint.hashCode();
  }

  /**
   *  Returns a string representation of this LocalId.
   */
  public String toString()
  {
    StringBuilder buf = new StringBuilder();
    buf.append(endpoint);
    buf.append(".");
    buf.append(member);
    return buf.toString();
  }
  
} // END LocalId
