/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.util;

import java.io.OutputStream;
import java.io.PrintWriter;
import java.io.StringWriter;
import java.io.Writer;

/**
 *  Utility class for dealing with serious failures.
 *
 *  @author <a href="mailto:meling@item.ntnu.no">Hein Meling</a>
 *  @since Jgroup 1.1
 */
public class Abort
  extends Error
{

  /////////////////////////////////////////////////////////////////////////////
  // Private Fields
  /////////////////////////////////////////////////////////////////////////////

  private String message;

  private Throwable exception;

  /////////////////////////////////////////////////////////////////////////////
  // Constructors
  /////////////////////////////////////////////////////////////////////////////

  public Abort()
  {
    this(null, null);
  }

  public Abort(String message)
  {
    this(message, null);
  }

  public Abort(Throwable exception)
  {
    this(null, exception);
  }

  public Abort(String message, Throwable exception)
  {
    this.message = message;
    this.exception = exception;
  }

  /////////////////////////////////////////////////////////////////////////////
  // Static Methods
  /////////////////////////////////////////////////////////////////////////////

  public static Abort exit()
  {
    return exit(null, null, 1);
  }

  public static Abort exit(int code)
  {
    return exit(null, null, code);
  }

  public static Abort exit(String error)
  {
    return exit(error, null, 1);
  }

  public static Abort exit(String error, int code)
  {
    return exit(error, null, code);
  }

  public static Abort exit(Throwable exception)
  {
    return exit(null, exception, 1);
  }

  public static Abort exit(Throwable exception, int code)
  {
    return exit(null, exception, code);
  }

  public static Abort exit(String error, Throwable exception)
  {
    return exit(error, exception, 1);
  }

  public static Abort exit(String error, Throwable exception, int code)
  {
    if (error != null)
      System.err.println(error);
    if (exception != null)
      exception.printStackTrace();
    System.exit(code);
    return new Abort("System.exit(" + code + ")");
  }

  public static String stackTrace(Throwable exception)
  {
    StringWriter writer = new StringWriter();
    printStackTrace(exception, writer);
    return writer.toString();
  }

  public static void printStackTrace(Throwable exception, StringBuffer buf)
  {
    buf.append(stackTrace(exception));
  }

  public static void printStackTrace(Throwable exception, OutputStream stream)
  {
    PrintWriter printer = new PrintWriter(stream);
    exception.printStackTrace(printer);
  }

  public static void printStackTrace(Throwable exception, Writer writer)
  {
    PrintWriter printer = new PrintWriter(writer);
    exception.printStackTrace(printer);
  }

  /**
   *  Prints the given usage message on stderr and exits with zero error code.
   *
   *  @param message
   *    The usage message to print.
   */
  public static void usage(String message)
  {
    usage(message, 0);
  }

  /**
   *  Prints the given usage message on stderr and exits with the given error code.
   *
   *  @param message
   *    The usage message to print.
   *  @param errorcode
   *    The error code for exit.
   */
  public static void usage(String message, int errorcode)
  {
    System.err.println(message);
    System.exit(errorcode);
  }

  /////////////////////////////////////////////////////////////////////////////
  // Object Methods
  /////////////////////////////////////////////////////////////////////////////

  /**
   *  Returns a string representation of this object
   */
  public String toString() 
  {
    StringWriter writer = new StringWriter();
    if (exception != null)
      printStackTrace(exception, writer);
    StringBuffer buf = writer.getBuffer();
    buf.append(getClass().getName());
    if (message != null) {
      buf.append(": ");
      buf.append(message);
    }
    return buf.toString();
  }

} // END Abort class
