/*
 * Copyright (c) 1998-2005 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */
package jgroup.util;

import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import org.apache.log4j.Logger;

/**
 * Implementation of a map with limited access times to a key/value pair. When
 * an entry is inserted in the map, you must specify the number of times the
 * entry can be retrieved before it is inaccessible/removed.<p>
 * 
 * This map can be useful when caching objects, and you need a way to remove old
 * items from the cache.
 * 
 * @author Rohnny Moland
 */
public class CacheMap<K, V>
  extends HashMap<K, V>
{
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Logger
  ////////////////////////////////////////////////////////////////////////////////////////////

  private static final Logger log = Logger.getLogger(CacheMap.class);

  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** The serial version uid */
  private static final long serialVersionUID = 8045886013359387147L;

  /** Map holding access times for the cache */
  private Map<Object, Integer> accessTimes = Collections
      .synchronizedMap(new HashMap<Object, Integer>());

  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructor and instance methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  public CacheMap()
  {
    super();
  }


  /**
   * Returns the value of the given key, if key is accessible. If not, 
   * return null.
   */  
  @Override
  public V get(Object key)
  {
    if (key == null)
      return null;
    Integer a = (Integer) accessTimes.get(key);
    if (a == null)
      return null;

    int access = a.intValue() - 1;
    if (access > 1) {
      if (log.isDebugEnabled())
        log.debug("Access value=" + access);
      accessTimes.put(key, new Integer(access));
      return super.get(key);
    } else {
      accessTimes.remove(key);
      return super.remove(key);
    }
  }


  /**
   * Insert a key/value in a <code>HashMap</code>, and set the number of
   * times it can be accessed.
   */
  public void put(K key, V value, int maxAccesses)
  {
      if (maxAccesses > 1) {
        accessTimes.put(key, new Integer(maxAccesses));
        super.put(key, value);
    }      
  }


  @Override
  public V put(K key, V value)
  {
    throw new UnsupportedOperationException("put(K,V) is not supported by CacheMap");
  }


  @Override
  public void putAll(Map<? extends K, ? extends V> m)
  {
    throw new UnsupportedOperationException("putAll(..) is not supported by CacheMap");
  }

  
  /**
   * Clear the map. 
   */
  @Override
  public void clear()
  {
    super.clear();
    accessTimes.clear();
  }


  /**
   *  Returns the size of cache map 
   */
  @Override
  public int size()
  {
    return super.size();
  }


  @Override
  public String toString()
  {
    StringBuffer buf = new StringBuffer();
    buf.append("CacheMap.accessTimes");
    buf.append("\n");
    synchronized (this) {
      for (Iterator iter = accessTimes.entrySet().iterator(); iter.hasNext();) {
        Map.Entry entry = (Map.Entry) iter.next();
        buf.append("[");
        buf.append(entry.getKey());
        buf.append("]: ");
        buf.append(entry.getValue());
        buf.append("\n");
      }
      return buf.toString();
    }
  }
  
} // END CacheMap
