/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.util;

import jgroup.core.ConfigurationException;
import jgroup.relacs.mss.FCEntry;
import jgroup.relacs.mss.MsgRouting;
import jgroup.relacs.mss.MssConstants;
import jgroup.relacs.types.EndPointImpl;
import jgroup.relacs.types.GroupIndex;
import jgroup.relacs.types.MessageTag;

/**
 *  The <code>MsgFactory</code> is used to create messages valid for the
 *  Jgroup Mss layer.
 *
 *  @author Alberto Montresor
 *  @author Hein Meling
 *  @since Jgroup 0.8
 */
public final class MsgFactory
  implements MssConstants
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Constants for the message factory.
   */
  private static int payload;
  private static int header;
  private static int trailer;

  /* Static part of the message size */
  private static int msgJGsize = MessageTag.SIZE + GroupIndex.SIZE*2;

  /**
   *  Static instance of the message factory.
   */
  private static MsgFactory msgFactory;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Static initialization method
  ////////////////////////////////////////////////////////////////////////////////////////////

  public static void initMsgFactory(int payload, int numOfHosts, int numOfClusters)
    throws ConfigurationException
  {
    /*
     * We simply return if multiple initializations were issued.  That
     * is we assume all GroupManagers in the same JVM use the same
     * TransportConfig object.
     */
    if (msgFactory != null)
      return;
    int minPayload = computeMinPayload(numOfHosts, numOfClusters);
    if (payload < minPayload) {
      throw new ConfigurationException("The payload is too small for this distributed system: "
        + minPayload + " is the smallest payload for the current configuration");
    }
    msgFactory = new MsgFactory(payload, HEADER_SIZE, TRAILER_SIZE, numOfHosts);
  }

  /**
   *  Compute the minimal payload, since we don't want to fragment
   *  certain low level messages.  Note that <code>MsgRouting</code>
   *  messages will always be larger than <code>MsgIamAlive</code>
   *  messages, thus we only need to return the largest possible payload
   *  that a routing message may introduce.
   */
  private static int computeMinPayload(int numOfHosts, int numOfClusters)
  {
    return MsgRouting.getMaxSize(numOfHosts, numOfClusters);
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Private constructor 
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  
   */
  private MsgFactory(int payl, int headr, int trailr, int numOfHosts)
  {
    payload = payl;
    header  = headr;
    trailer = trailr;
    /* Compute partially the dynamic size of the message */
    msgJGsize   += FCEntry.SIZE*numOfHosts;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Access methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *
   */
  public static OutMessage get()
  {
    return new OutMessage(payload, header, trailer);
  }


  /**
   *
   */
  public static OutMessage get(int size)
  {
    return new OutMessage(payload, header, trailer, size);
  }


  /**
   *
   */
  public static OutMessage get(int size, int nreceivers)
  {
    /* Finalize computing the dynamic size of the message */
    size += msgJGsize + EndPointImpl.SIZE*nreceivers;
    return new OutMessage(payload, header, trailer, size);
  }

} // END MsgFactory
