/*
 * Copyright (c) 1998-2004 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.util;

import java.util.Collections;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

import org.apache.log4j.Logger;

/**
 * Monitor threads to detect thread failures, and perform corrective action
 * in case some thread has crashed. <p>
 *
 * This class is helpful with respect to debugging deadlock problems, especially
 * when <code>DUMP_ALL_THREADS</code> is set to true.  This will, at regular
 * intervals, dump the stack trace of all threads allowing the developer to
 * identify threads that might have entered into a deadlock situation awaiting
 * a lock being held by another thread. <p>
 *
 * This class requires JDK 5.0, but is not mandatory for general use of Jgroup/ARM,
 * and as such use of this class may be disabled.
 *
 * @author Hein Meling
 * @since Jgroup 2.1
 */
public class ThreadMonitor
  extends Thread
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Logger
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Obtain logger for this class */
  private static final Logger log = Logger.getLogger(ThreadMonitor.class);


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /* The default delay to wait between checking thread liveness */
  private static final int THREAD_DUMP_DELAY = 5000;

  /*
   * Option to dump the stack trace of all threads (true),
   * rather than just the trace for the registered threads (false).
   */
  private static boolean DUMP_ALL_THREADS = true;

  /* Singleton instance of the thread monitor object */
  private static ThreadMonitor threadMonitor = null;

  /* The list of threads to watch for failures */
  private final Set<Thread> threadSet = Collections.synchronizedSet(new HashSet<Thread>(5));


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructors
  ////////////////////////////////////////////////////////////////////////////////////////////

  private ThreadMonitor(String name)
  {
    super(name);
    DUMP_ALL_THREADS = Boolean.getBoolean("jgroup.log.dump.threads");
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Add the given thread to the monitor.
   */
  public static void add(Thread thread)
  {
    if (threadMonitor == null) {
      threadMonitor = new ThreadMonitor("ThreadMonitor");
      threadMonitor.setDaemon(true);
      threadMonitor.start();
    }
    threadMonitor.threadSet.add(thread);
  }


  /**
   * Remove the given thread from being monitored.
   */
  public static void remove(Thread thread)
  {
    if (threadMonitor == null) {
      throw new IllegalStateException("Cannot remove entry from monitor; not initialized");
    }
    threadMonitor.threadSet.remove(thread);
  }

  public static void dumpAllThreads()
  {
    Set traceSet = Thread.getAllStackTraces().entrySet();
    synchronized (traceSet) {
      for (Iterator iter = traceSet.iterator(); iter.hasNext();) {
        Map.Entry element = (Map.Entry) iter.next();
        StackTraceElement[] arr = (StackTraceElement[]) element.getValue();
        Thread thread = (Thread) element.getKey();
        log.debug("------------- " + thread.getName()
            + "." + thread.getId()
            + ", state=" + thread.getState());
        for (int i = 0; i < arr.length; i++) {
          log.debug("  " + arr[i]);
        }
      }
      log.debug("*************************");
    }
  }


  /* (non-Javadoc)
   * @see java.lang.Runnable#run()
   */
  public void run()
  {
    while (true) {
      try {
        sleep(THREAD_DUMP_DELAY);
      } catch (InterruptedException e) { }

      if (DUMP_ALL_THREADS) {
        if (log.isDebugEnabled()) {
          dumpAllThreads();
        }
      } else {
        boolean noFailedThreads = true;
        synchronized (threadSet) {
          for (Iterator iter = threadSet.iterator(); noFailedThreads && iter.hasNext();) {
            Thread thread = (Thread) iter.next();
            if (log.isDebugEnabled()) {
              StackTraceElement[] trace = thread.getStackTrace();
              log.debug("--------------------- " + thread.getName());
              for (int i = 0; i < trace.length; i++) {
                log.debug("  " + trace[i]);
              }
            }
            noFailedThreads &= thread.isAlive();
          }
          if (log.isDebugEnabled())
            log.debug("*************************");
        }
        if (noFailedThreads == false) {
          /* One or more threads have failed */
          if (log.isDebugEnabled())
            log.debug("One or more threads have failed...");
        }
      }
    }
  }

} // END of ThreadMonitor
