/*
 * Copyright (c) 1998-2005 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */
package jgroup.util.log;

import static jgroup.util.log.ConnectionPatternEvent.ConnectionPatternType.ConnectionPattern;

import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.Serializable;
import java.net.InetAddress;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.NoSuchElementException;
import java.util.Set;

import jgroup.core.ConfigManager;
import jgroup.relacs.config.Domain;
import jgroup.relacs.config.DomainSet;

/**
 * A connection pattern event to be logged.
 * 
 * @author Hein Meling
 */
public class ConnectionPatternEvent
  extends Event
  implements Iterable<Domain>
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constants
  ////////////////////////////////////////////////////////////////////////////////////////////

  private static final long serialVersionUID = 227099539664502309L;

  /**
   * Supported connection event types.
   */
  public enum Type implements EventType { Partition, Merge }

  public enum ConnectionPatternType implements EventType { ConnectionPattern }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** The list of connection events comprising this event pattern */
  private List<ConnectionEvent> events = new ArrayList<ConnectionEvent>(3);

  /** The connection pattern string */
  private String pattern;

  /** Set of partitions computed from the events list */
  private transient Set<DomainSet> partitions;

  /** Mapping between InetAddress and domain set (used to lookup the partition of a host) */
  private transient Map<InetAddress, DomainSet> hostPartitions;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructor
  ////////////////////////////////////////////////////////////////////////////////////////////

  public ConnectionPatternEvent()
  {
    super();
  }

  public ConnectionPatternEvent(String description, String pattern, List<ConnectionEvent> events)
  {
    super(ConnectionPattern, description);
    this.pattern = pattern;
    this.events = events;
  }

  public void addEvent(Type type, Domain fromDomain, Domain toDomain)
  {
    events.add(new ConnectionEvent(type, fromDomain, toDomain));
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Inner class for connection event
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Connection event describing the connection between to domains.
   * 
   * @author Hein Meling
   */
  private class ConnectionEvent
    implements Serializable, Comparable<ConnectionEvent>
  {

    private static final long serialVersionUID = -5851464130156734196L;

    /**
     * The connection type for this event
     */
    private final Type type;

    /**
     * The two domains involved in this connection event.
     */
    private Domain fromDomain, toDomain;

    /**
     * The connection pattern for this ConnectionEvent
     */
    private String connectPattern;

    public ConnectionEvent(Type type, Domain fromDomain, Domain toDomain)
    {
      this.type = type;
      this.fromDomain = fromDomain;
      this.toDomain = toDomain;
      switch (type) {
        case Partition:
          this.connectPattern = fromDomain.getName() + " | " + toDomain.getName();
          break;

        case Merge:
          this.connectPattern = fromDomain.getName() + " - " + toDomain.getName();
          break;
      }
    }

    public String toString()
    {
      return connectPattern;
    }

    /**
     * Comparing two events for the purpose of sorting so that merge events are
     * first in the list.
     * 
     * @param o
     * @return
     */
    public int compareTo(ConnectionEvent o)
    {
      if (o.type == type) {
        return 0;
      }
      if (o.type == Type.Partition && type == Type.Merge) {
        return -1;
      } else {
        return 1;
      }
    }

    ////////////////////////////////////////////////////////////////////////////////////////////
    // Serialization Methods
    ////////////////////////////////////////////////////////////////////////////////////////////

    /**
     * Override readObject to update the domain objects to contain all hosts;
     * this is required since the Domain class is not serialize; only the EndPointImpl
     * class is serialized, which contains the InetAddress.
     */
    private void readObject(ObjectInputStream in)
      throws IOException, ClassNotFoundException
    {
      in.defaultReadObject();
      DomainSet domains = ConfigManager.getDistributedSystem().getDomainSet();
      // Replace the unmarshaled domains with real domains 
      fromDomain = domains.getDomain(fromDomain);
      toDomain = domains.getDomain(toDomain);
    }

  } // END ConnectionEvent


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Access Methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * @return Returns the list of connection events.
   */
  public List<ConnectionEvent> getConnectionEvents()
  {
    return events;
  }

  /**
   * @return True if at least one of the events is a partition event.
   */
  public boolean containsPartition()
  {
    return partitionCount() > 1;
  }

  /**
   * @return The number of partitions that will result from this connection pattern.
   */
  public int partitionCount()
  {
    int pCount = 0;
    for (ConnectionEvent ev : events) {
      if (ev.type == Type.Partition)
        pCount++;
    }
    if (pCount == 0)
      pCount = 1;
    return pCount;
  }

  /**
   * Returns the domain set (the partition) of the given host.
   * 
   * @param inetAdr the host for which to obtain the domain set
   * @return Returns the domain set (the partition) of the given host.
   * @throws NoSuchElementException Raised if the given host is not
   *   in any of the partitions.
   */
  public DomainSet getPartition(InetAddress inetAdr)
  {
    if (hostPartitions == null)
      hostPartitions = new HashMap<InetAddress, DomainSet>();
    DomainSet domainSet = hostPartitions.get(inetAdr);
    if (domainSet == null) {
      domainSet = searchForHost(inetAdr);
      hostPartitions.put(inetAdr, domainSet);
    }
    return domainSet;
  }

  private DomainSet searchForHost(InetAddress inetAdr)
  {
    Set<DomainSet> parts = getPartitions();
    for (DomainSet domainSet : parts) {
      if (domainSet.getHostDomain(inetAdr) != null)
        return domainSet;
    }
    throw new NoSuchElementException(inetAdr.toString());
  }

  /**
   * Returns a set of domain sets representing the partitions of this connection
   * pattern event.
   * 
   * Implementation note: this method requires that merge events are stored first
   * in the events list.  Currently this is accomplished by implementing the
   * <code>Comparable</code> interface and sorting this list before constructing
   * the set of domain sets.
   * 
   * @return Returns a set of domain sets representing the partitions of this
   *   connection pattern event.
   */
  public Set<DomainSet> getPartitions()
  {
    if (partitions == null)
      partitions = new HashSet<DomainSet>();
    if (!partitions.isEmpty())
      return partitions;

    Collections.sort(events);
    for (ConnectionEvent ev : events) {
      DomainSet xPartition = findPartition(ev.fromDomain);
      DomainSet yPartition = findPartition(ev.toDomain);

      switch (ev.type) {
        case Partition:
          if (xPartition == null) {
            xPartition = new DomainSet(DomainSet.domainComparator);
            partitions.add(xPartition);
          }
          if (yPartition == null) {
            yPartition = new DomainSet(DomainSet.domainComparator);
            partitions.add(yPartition);
          }
          xPartition.addDomain(ev.fromDomain);
          yPartition.addDomain(ev.toDomain);
          break;

        case Merge:
          if (xPartition == null && yPartition == null) {
            xPartition = new DomainSet(DomainSet.domainComparator);
            partitions.add(xPartition);
          } else if (xPartition == null && yPartition != null) {
            xPartition = yPartition;
          } else {
            // Should not happen
            assert false;
          }
          xPartition.addDomain(ev.fromDomain);
          xPartition.addDomain(ev.toDomain);
          break;
      }
    }
    return partitions;
  }

  private DomainSet findPartition(Domain domain)
  {
    for (DomainSet domains : partitions) {
      if (domains.containsDomain(domain)) {
        return domains;
      }
    }
    return null;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from Object
  ////////////////////////////////////////////////////////////////////////////////////////////

  public String toString()
  {
    StringBuilder buf = commonToString(true);
    buf.append(" [");
    buf.append(pattern);
    buf.append("]");
    return buf.toString();
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from Iterable<Domain>
  ////////////////////////////////////////////////////////////////////////////////////////////

  public Iterator<Domain> iterator()
  {
    Set<Domain> domains = new HashSet<Domain>(5);
    for (ConnectionEvent ev : events) {
      domains.add(ev.fromDomain);
      domains.add(ev.toDomain);
    }
    return domains.iterator();
  }

} // END ConnectionPatternEvent
