/*
 * Copyright (c) 1998-2005 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */
package jgroup.util.log;

import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.Serializable;
import java.net.InetAddress;

import jgroup.util.Network;

/**
 * A general event that can be logged.  This class is designed to be
 * subclassed, and it should be noted that subclassing requires that
 * the subclass be serializable to be written to disk.  Since this class
 * does implement <code>Serializable</code> the subclasses do not need
 * to do that.
 * 
 * @author Rune Vestvik
 * @author Hein Meling
 * @since Jgroup 3.0
 */
public class Event
  implements Comparable<Event>, Serializable
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constants
  ////////////////////////////////////////////////////////////////////////////////////////////

  private static final long serialVersionUID = 4121134753270542643L;

  /**
   * The time of the event to which other events should be relative to.
   * If not set, the output generated for the events will be the actual
   * time values of the logged events.
   */
  private static long timeOfRelativeEvent = 0;

  /** The length of the longest hostname ever recorded */
  private static int longestHostname = 0;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Event fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  protected long time;

  protected EventType type;

  protected String description;

  private String hostname;

  protected InetAddress eventHost;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructors
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Super constructor for sub events with late initialization */
  public Event()
  {
  }

  public Event(EventType type, String description)
  {
    this(System.currentTimeMillis(), type, description);
  }

  public Event(long time, EventType type, String description)
  {
    this.time = time;
    this.type = type;
    this.description = description;
    this.hostname = Network.getLocalMachineName();
    this.eventHost = Network.getLocalHost();
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Access methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * @return Returns the occurrance time of this event
   */
  public long getTime()
  {
    return time;
  }

  /**
   * Set the time value of this event; used to adjust the time for events that
   * was logged on machines with incorrect clock.
   * 
   * @param time the occurrance time of this event
   */
  public void setTime(long time)
  {
    this.time = time;
  }

  /**
   * @return Returns the type of this event
   */
  public EventType getType()
  {
    return type;
  }

  /**
   * @return Returns the description of this event
   */
  public String getDescription()
  {
    return description;
  }

  /**
   * @return Returns the hostname.
   */
  public String getHostname()
  {
    return hostname;
  }

  /**
   * @return Returns the host on which this event was generated.
   */
  public InetAddress getHost()
  {
    return eventHost;
  }

  /**
   * Returns true if the given type is the same as the type of this instance.
   */
  public boolean isType(EventType theType)
  {
    return theType == type;
  }

  /**
   * This method must be overriden by the event type which will be used to indicate
   * that it is the last event.
   * 
   * @return Always returns false.
   */
  public boolean isLastEvent()
  {
    return false;
  }

  /**
   * Set the event whose time will be the used to compute relative time
   * of consecutive events.  Typically, the first event should be the
   * relative event.
   * 
   * @param relativeEvent
   */
  public static void setRelativeTo(Event relativeEvent)
  {
    timeOfRelativeEvent = relativeEvent.time;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from Object
  ////////////////////////////////////////////////////////////////////////////////////////////

  public String toString()
  {
    StringBuilder buf = commonToString(true);
    return buf.toString();
  }

  protected StringBuilder commonToString(boolean includeDescription)
  {
    StringBuilder buf = new StringBuilder();
    buf.append(String.format("%6d", time - timeOfRelativeEvent));
    buf.append(" (");
    buf.append(hostname);
    int pad = longestHostname - hostname.length();
    for (int i = 0; i < pad; i++) {
      buf.append(' ');
    }
    buf.append("): ");
    if (type != null) {
      buf.append(type);
    }
    if (includeDescription && description.length() > 0) {
      buf.append(' ');
      buf.append(description);
    }
    return buf;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from Comparable
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Allows the system to compare two events to determine which are first
   * in some ordering; hence making it simple to sort the events.
   */
  public int compareTo(Event event)
  {
    if (time > event.time)
      return 1;
    else if (time < event.time)
      return -1;
    return 0;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Serialization Methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Override readObject to allow us to compute the length of longest hostname so that
   * we can do pretty printing. 
   */
  private void readObject(ObjectInputStream in)
    throws IOException, ClassNotFoundException
  {
    in.defaultReadObject();
    if (hostname != null) {
      int hostnameLen = hostname.length();
      if (longestHostname < hostnameLen)
        longestHostname = hostnameLen;
    }
  }

} // END Event
