/*
 * Copyright (c) 1998-2005 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */
package jgroup.util.log;

import static jgroup.util.log.ViewEvent.RecoveryState.FULLY_REPLICATED;
import static jgroup.util.log.ViewEvent.RecoveryState.NEED_RECOVERY;
import static jgroup.util.log.ViewEvent.RecoveryState.NEW_MEMBER;
import static jgroup.util.log.ViewEvent.RecoveryState.NORMAL;
import static jgroup.util.log.ViewEvent.RecoveryState.TOO_MANY_REPLICAS;
import jgroup.core.MemberId;
import jgroup.core.View;
import jgroup.relacs.config.AppConfig;
import jgroup.relacs.types.ViewId;
import jgroup.util.Network;

/**
 * A view event to be logged.
 * 
 * @author Rune Vestvik
 * @author Hein Meling
 */
public final class ViewEvent
  extends Event
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constants
  ////////////////////////////////////////////////////////////////////////////////////////////

  private static final long serialVersionUID = 4049640096401864757L;

  /**
   * Supported view change event types.
   */
  public enum Type implements EventType { ARM, Merge, Server, Client }

  enum RecoveryState { NEW_MEMBER, NORMAL, NEED_RECOVERY, FULLY_REPLICATED, TOO_MANY_REPLICAS }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * The view event to be logged.
   */
  private final View view;

  private int minimalRedundancy;

  private int initialRedundancy;

  private RecoveryState state;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructor
  ////////////////////////////////////////////////////////////////////////////////////////////

  public ViewEvent(EventType type, View view)
  {
    super(type, "ViewChange");
    this.view = (View) view.clone();
    AppConfig thisApp = AppConfig.getApplication(this.view.getGid());
    try {
      this.minimalRedundancy = thisApp.getMinimalRedundancy();
      this.initialRedundancy = thisApp.getInitialRedundancy();
    } catch (IllegalStateException e) {
      /*
       * Ignored, since we may be running an application without a specified
       * redundancy, e.g. helloserver.
       */
    }
    computeState();
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Private Methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  private void computeState()
  {
    int vsize = view.size();
    if (vsize == initialRedundancy) {
      state = FULLY_REPLICATED;
    } else if (vsize < minimalRedundancy) {
      if (isFirstView())
        state = NEW_MEMBER;
      else
        state = NEED_RECOVERY;
    } else if (vsize > initialRedundancy) {
      state = TOO_MANY_REPLICAS;
    } else if (vsize >= minimalRedundancy && vsize < initialRedundancy) {
      state = NORMAL;
    }
    assert state != null;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Access Methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  public int getViewSize()
  {
    return view.size();
  }

  public long getViewId()
  {
    return view.getVid();
  }

  public View getView()
  {
    return view;
  }

  public int getGroupId()
  {
    return view.getGid();
  }

  /**
   * @return True if this view is the first view installed by the member of this view.
   *   False is returned otherwise.  Meaning that if there is more than one member or
   *   if the single member view is not the first that the member has installed, then
   *   false will be returned.
   */
  public boolean isFirstView()
  {
    if (view.size() == 1) {
      MemberId memb = view.getMembers()[0];
      // Compute the initial vid for this single member
      long theInitialVid = ViewId.create(memb.getEndPoint(), 0);
      return theInitialVid == view.getVid();
    } else {
      return false;
    }
  }

  public boolean isFullyReplicated()
  {
    return state == FULLY_REPLICATED;
  }

  public boolean isOK()
  {
    return state == NORMAL || state == FULLY_REPLICATED;
  }

  public boolean needRecovery()
  {
    return state == NEED_RECOVERY;
  }

  public boolean hasTooManyReplicas()
  {
    return state == TOO_MANY_REPLICAS;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from Object
  ////////////////////////////////////////////////////////////////////////////////////////////

  public String toString()
  {
    StringBuilder buf = commonToString(true);
    commonToString(buf);
    return buf.toString();
  }

  /**
   *  Print the content of this view event without the host and time.
   */
  public String shortToString()
  {
    StringBuilder buf = new StringBuilder("ViewEvent");
    commonToString(buf);
    return buf.toString();
  }

  protected void commonToString(StringBuilder buf)
  {
    buf.append(" gid=");
    buf.append(view.getGid());
    buf.append(" view.size=");
    buf.append(view.size());
    buf.append(" (");
    buf.append(state);
    buf.append(") vid=");
    buf.append(view.getVid());
    buf.append(" members: ");
    MemberId[] members = view.getMembers();
    for (int i = 0; i < members.length; i++) {
      buf.append(Network.getMachineName(members[i].getCanonicalHostName()));
      buf.append(" ");
    }
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Serialization Methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Override readObject to update the recovery state, given a particular 
   * redundancy configuration.
   */
//  private void readObject(ObjectInputStream in)
//    throws IOException, ClassNotFoundException
//  {
//    in.defaultReadObject();
//    computeState();
//  }

} // END ViewEvent
